/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#ifndef __iscsipdu_h__
#define __iscsipdu_h__

/*++

    Filename:

    iscsipdu.h - iSCSI Protocol Data Unit Definitions

    Purpose:

        The structures for the iSCSI Protocol Data Unit(PDU) and the prototypes
        of the iSCSI layers intrenal functions.
        
    Maintainers:

        01/05/04 : Parul Samdarshi  : File created.

    Copyright:

        Copyright 2004 Adaptec, Inc. All rights reserved.

        This software contains the valuable trade secrets of Adaptec or its
        licensors. The software is protected under international copyright
        laws and treaties. This software may only be used in accordance with
        the terms of its accompanying license agreement.

--*/

#ifndef _WINISCSI_INBOX
#include "setypes.h"
#endif
                                /*--------**
                                ** Macros **
                                **--------*/
//
// Given a pointer to the PDU header, return the data_segment_len
//
#define PDU_DATA_SEG_LENGTH(_pMsg_) ((u32)(((PPDU_BASE)_pMsg_)->data_len_lo) |      \
                                    ((u32)(((PPDU_BASE)_pMsg_)->data_len_hi)<<16))  \

//
// Give a pointer to a Network Byte Ordered PDU header, return the data_segment_len
//
#define PDU_NETWORK_DATA_SEG_LENGTH(_pMsg_) (ACIT_GET_BE16(&((PPDU_BASE)_pMsg_)->data_len_lo) |      \
                                          ((u32)(((PPDU_BASE)_pMsg_)->data_len_hi)<<16))  \

// Returns number of padding bytes in pdu.
#define PDU_PADDING(_pMsg_)                                                                          \
    (PDU_NETWORK_DATA_SEG_LENGTH (_pMsg_) & 3 ? 4 - (PDU_NETWORK_DATA_SEG_LENGTH (_pMsg_) & 3) : 0)  \

// Return the header digest in network byte order.
#define PDU_NETWORK_HEADER_DIGEST(_pMsg_) (*((u32*)((Pu8)_pMsg_ + ISCSI_PDU_HEADER_SIZE)))

// Return the data digest.  PDU Length is the header + both digests + data + padding length.
#define PDU_NETWORK_DATA_DIGEST(_pMsg_, _pduLength_)                    \
    (*((u32*)((Pu8)_pMsg_ + _pduLength_ - ISCSI_CRC32C_LENGTH)))  \

#define PDUH_SZ_DATA_SEG_LEN    3
#define PDUH_SZ_SAM2LUN 8
#define PDUH_SZ_CDB 16
#define PDUH_SZ_BOOKMARK 8
#define PDUH_ASYNC_MSG_MAX_PARAMS   3

#define PDUH_M_OPCODE   0x3F

#define PDU_LEN_PAD_MASK    3
#define PDU_TAG_RESERVED    (~ 0)

/*SNA = Serial Number Arithmetic*/
#define SNA_SERIAL_BITS 32
#define SNA_SMALLEST_INVALID_DELTA  (ACIT_UEXACT32) (1 << (SNA_SERIAL_BITS - 1))

//
// iSCSI version
//
#define ISCSI_MAX_VERSION               0   // draft 20 is current
#define ISCSI_MIN_VERSION               0
                                
/*
 * prabal - fixing a compilation warning. As long as SNA_SERIAL_BITS is
 * 32, this should be ok
 */
/*
#define SNA_ADD(Base, Delta)    (((Base) + (Delta)) & ((1 << SNA_SERIAL_BITS) - 1))
*/
#define SNA_ADD(Base, Delta)    (((Base) + (Delta)) & (0 - 1))
    /*Not using the modulo operator because (1 << SNA_SERIAL_BITS) could be 0*/
#define SNA_SUBTRACT(Base, Delta)   (((Base) - (Delta)) & ((1 << SNA_SERIAL_BITS) - 1))
    /*SNA subtraction is not really defined, but it works.*/

#define SNA_FORMER_IS_LESSER(Former, Latter)    (   \
    ((Former) < (Latter) && (Latter) - (Former) < SNA_SMALLEST_INVALID_DELTA) ||    \
    ((Former) > (Latter) && (Former) - (Latter) > SNA_SMALLEST_INVALID_DELTA)   \
)

#define SNA_FORMER_IS_GREATER(Former, Latter)   (   \
    ((Former) > (Latter) && (Former) - (Latter) < SNA_SMALLEST_INVALID_DELTA) ||    \
    ((Former) < (Latter) && (Latter) - (Former) > SNA_SMALLEST_INVALID_DELTA)   \
)

#define SNA_ARE_EQUAL(Former, Latter)   ((Former) == (Latter))


//  S3_Lun must be 8 byte char array
#define S2_TO_S3_LUN( S3_Lun,  S2_Lun )          \
        OSM_OSM_MEMSET (S3_Lun, 0, SCSI3_LUN_SIZE);  \
        S3_Lun[1] = (S2_Lun);                    \


#define IS_PDU_STATUS_OVERRUN(status)      ((status & 0x04) == 0x04)
#define IS_PDU_STATUS_UNDERRUN(status)     ((status & 0x02) == 0x02)

typedef enum _iSCSI_ASYNC_EVENT_CODES
{
    ASYNC_SCSI_EVENT = 0,
    ASYNC_LOGOUT_REQUEST,            // Target request logout
    ASYNC_DROP_CONNECTION,           // Drop a specific connection in a session
    ASYNC_DROP_ALL_CONNECTIONS,      // Drop all connections in the session
    ASYNC_PARAM_NEGOTIATION_REQUEST, // Target requesting parameters negotiation
    ASYNC_VENDOR_SPECIFIC_ISCSI_EVENT = 0xff
} iSCSI_ASYNC_EVENT_CODES;

#define ISCSI_PDU_HEADER_SIZE           48
/* 
 * iSCSI Initiator OpCodes (IIOC) 
 */
#define IIOC_NOP_OUT                    0x00
#define IIOC_SCSI_CMD                   0x01
#define IIOC_SCSI_TASK_MGMT_CMD         0x02
#define IIOC_LOGIN_CMD                  0x03
#define IIOC_TEXT_CMD                   0x04
#define IIOC_SCSI_DATA                  0x05  /* Write Operation */
#define IIOC_LOGOUT_CMD                 0x06
#define IIOC_SNACK_REQ                  0x10


/* 
 * iSCSI Target OpCodes  (ITOC) 
 */
#define ITOC_NOP_IN                     0x20
#define ITOC_SCSI_RESPONSE              0x21
#define ITOC_SCSI_TASK_MGMT_RESPONSE    0x22
#define ITOC_LOGIN_RESPONSE             0x23
#define ITOC_TEXT_RESPONSE              0x24
#define ITOC_SCSI_DATA                  0x25  /* Read Operation */
#define ITOC_LOGOUT_RESPONSE            0x26
#define ITOC_R2T                        0x31
#define ITOC_ASYNC_EVENT                0x32
#define ITOC_REJECT                     0x3F

/* 
 * Bit field definition 
 */
#define BIT_0   0x01
#define BIT_1   0x02
#define BIT_2   0x04
#define BIT_3   0x08
#define BIT_4   0x10
#define BIT_5   0x20
#define BIT_6   0x40
#define BIT_7   0x80


/*
 * Logout Reason Code of IIOC_LOGOUT_CMD
 */
enum LRC_ENUM {
    LRC_SESSION_IS_CLOSING = 0,
    LRC_FOR_RECOVERY,
    LRC_TARGET_REQUEST
};

/*
 * Logout Response Status of ITOC_LOGOUT_RESPONSE
 */
enum LORS_ENUM {
    LORS_SUCCESS=0,
    LORS_CLEANUP_FAILED
};


//
// Logout response code
//
enum ISCSI_PDU_LOGOUT_RSP_CODE {
    ISCSI_PDU_LOGOUT_RESPONSE_SUCCESS       = 0x00,
    ISCSI_PDU_LOGOUT_RESPONSE_BAD_CID       = 0x01,  // CID not found
    ISCSI_PDU_LOGOUT_RESPONSE_NO_RECOVERY   = 0x02,  // connection recovery not support
    ISCSI_PDU_LOGOUT_RESPONSE_FAIL          = 0x03   // various reason cleanup failure
};

/*
 * Reject Reason Code for 
 */
enum RRC_ENUM {
    RRC_RESERVED = 1,
    RRC_DATA_DIGEST_ERROR,
    RRC_SNACK_REJECT,
    RRC_PROTOCOL_ERROR,
    RRC_COMMAND_UNSUPPORTED,
    RRC_REJECT_IMMD_COMMAND,
    RRC_TASK_IN_PROGRESS,
    RRC_INVALID_DATA_ACK,
    RRC_INVALID_PDU_FIELD,
    RRC_LONG_OPERATION_REJECT,
    RRC_NEGOTIATION_RESET,
    RRC_WAITING_FOR_LOGOUT
};

#if ACIT_CFG_DEBUG_BUILD
static const ACIT_UEXACT8 rrc_string[RRC_WAITING_FOR_LOGOUT+1][16] = {
    "not_used",
    "reserved",
    "data_digest_err",
    "snack_reject",
    "protocol_err",
    "cmd_unsupported",
    "reject_immd_cmd",
    "in_progress",
    "invld_data_ack",
    "invld_pdu",
    "long_op_reject",
    "nego_reset",
    "wait_4_logout"
};
#endif



/*iSCSI Task Management Functions*/
enum    TMF_ENUM    {
    TMF_INVALID,
    TMF_ABORT_TASK, 
    TMF_ABORT_TASK_SET, 
    TMF_CLEAR_ACA, 
    TMF_CLEAR_TASK_SET,
    TMF_LOGICAL_UNIT_RESET, 
    TMF_TARGET_WARM_RESET, 
    TMF_TARGET_COLD_RESET, 
    TMF_TASK_REASSIGN
};
#define PDUH_M_TASK_MGMT_FUNCTION   0x7F

/*iSCSI Task Management Responses*/
enum    TMR_ENUM    {
    TMR_FUNCTION_COMPLETE, 
    TMR_TASK_NOT_IN_SET, 
    TMR_LUN_INEXISTENT,
    TMR_TASK_STILL_ALLEGIANT, 
    TMR_TASK_FAILOVER_UNSUPPORTED,
    TMR_TASK_MGMT_FUNCTION_UNSUPPORTED, 
    TMR_FUNCTION_AUTH_FAILED,
    TMR_FUNCTION_REJECTED = 255
};

/*iSCSI Task Management Response Qualifiers*/
enum    TMRQ_ENUM   {
    TMRQ_FUNCTION_EXECUTED, TMRQ_NOT_AUTHORIZED
};


//
// SCSI CMDBITS field type definitions
// Draft 20, Section 10.3.1 Flags and Task Attibutes 1 byte
//
typedef enum {
    ISCSI_TASK_TAG_UNTAGGED = 0,
    ISCSI_TASK_TAG_SIMPLE = 1,
    ISCSI_TASK_TAG_ORDERED = 2,
    ISCSI_TASK_TAG_HEAD_OF_QUEUE = 3,
    ISCSI_TASK_TAG_ACA = 4
} ISCSI_PDU_TASK_ATTR;


                        /*------------------------------**
                        ** Structures, unions and types **
                        **------------------------------*/

/*!
@brief
    ISID format type enum definition
*/
enum ISCSI_PDU_ISID_TYPE {
    ISCSI_PDU_ISID_TYPE_OUI,
    ISCSI_PDU_ISID_TYPE_EN,
    ISCSI_PDU_ISID_TYPE_RANDOM,
    ISCSI_PDU_ISID_TYPE_RESERVED
};

/*!
@brief
    Basic ISID format
*/
typedef struct _ISCSI_ISID_BASIC {
    u8   isid_A : 6;         //! A field
    u8   isid_T : 2;         //! T
    u8   isid_B_H;           //! B-1
    u8   isid_B_L;           //! B-2
    u8   isid_C;             //! C
    u16  isid_D;             //! D
} ISCSI_ISID_BASIC, *PISCSI_ISID_BASIC;

/*!
@brief
    OUI ISID format
*/
typedef struct _ISCSI_OUI_ISID {
    u8   ouiHigh : 6;        //! A field
    u8   type : 2;           //! T; value = 00b
    u16  ouiLow;             //! B field  (A+B = 22 bit field)
    u8   qualifierHigh;      //! C  (C+D = 24 bit qualifier)
    u16  qualifierLow;       //! D
} ISCSI_OUI_ISID, *PISCSI_OUI_ISID;

/*!
@brief
    EN ISID format
*/
typedef struct _ISCSI_EN_ISID {
    u8   reserved : 6;       //! A field
    u8   type : 2;           //! T; Value = 01b
    u16  enHigh;             //! B  (B+C = IANA Enterprise Num)
    u16  enLow;              //! C  
    u16  qualifier;          //! D
} ISCSI_EN_ISID, *PISCSI_EN_ISID;

/*!                                                                      
@brief
    Random ISID format
*/
typedef struct _ISCSI_RANDOM_ISID {
    u8   reserved : 6;       //! A field
    u8   type : 2;           //! T; Value = 10b
    u8   randHi;             //! B-1
    u16  randLo;             //! B-2 and C
    u16  qual;               //! D, qualifier
} ISCSI_RANDOM_ISID, *PISCSI_RANDOM_ISID;

/*!
@brief
    ISID definition
*/
typedef struct _ISCSI_ISID {
    union
    {                        
        ISCSI_ISID_BASIC    basic;          //! Basic ISID format
        ISCSI_OUI_ISID      oui;            //! Organization Unique ID format
        ISCSI_EN_ISID       en;             //! IANA Enterprise Number format 
        ISCSI_RANDOM_ISID   random;         //! Random ISID format
    } s;
} ISCSI_ISID, *PISCSI_ISID;

// -----------------------------------------------------------------------------
// Common iSCSI PDU Fields
// -----------------------------------------------------------------------------
typedef struct _PDU_BASE {       
    u8  opcode : 6;
    u8  i_bit : 1;               // immediate bit
    u8  x_bit : 1;               // reserved, always 0
    u8  reserved1[3];           // opcode-specific fields
    u8  ahs_length;              // length units is 4 byte words
    u8  data_len_hi;
    u16 data_len_lo;              // DatasegmentLength
    u8  lun[8];                 // lun or opcode-specific fields
    u32 itt;                    // initiator task tag
    u8  reserved4[28];
} PDU_BASE, * PPDU_BASE;

// -----------------------------------------------------------------------------
// SCSI Command Request
// -----------------------------------------------------------------------------
typedef struct _PDU_COMMAND {
    u8  opcode : 6;             // opcode 0x01
    u8  i_bit : 1;               // Immediate Bit
    u8  x_bit : 1;               // reserved 0
    u8  attribute : 3;          // ATTR
    u8  flag1resv : 2;          // reserved
    u8  w_bit : 1;               // W bit
    u8  r_bit : 1;               // R bit
    u8  f_bit : 1;               // F bit
    u16 reserved1;              // reserved;
    u8  ahs_length;
    u8  data_len_hi;
    u16 data_len_lo;              // DataSegmentLength
    u8  lun[8];
    u32 task_tag;                // ITT; initiator task tag
    u32 data_trans_len;           // expected data transfer length
    u32 cmd_sn;                  // cmd sequence number
    u32 exp_stat_sn;              // expected status/data sequence number
    u8  cdb[PDUH_SZ_CDB];
    u32 header_digest;
} PDU_COMMAND, *PPDU_COMMAND;


STATIC INLINE void
pdu_cmd_swap_bytes (PPDU_COMMAND pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->task_tag   = ACIT_GET_BE32( &pdu->task_tag );
    pdu->data_trans_len = ACIT_GET_BE32( &pdu->data_trans_len );
    pdu->cmd_sn     = ACIT_GET_BE32( &pdu->cmd_sn );
    pdu->exp_stat_sn = ACIT_GET_BE32( &pdu->exp_stat_sn );
}

STATIC INLINE void 
pdu_build_command (PPDU_COMMAND pdu, 
                   u8 attribute, 
                   u32 data_transfer_length, 
                   u8* cdb, 
                   void* s3_lun, 
                   u8 read_bit, 
                   u8 write_bit)
{
    //OSM_ASSERT ((data_segment_length >> 24) == 0);

    // BYTE 0
    pdu->opcode       = ISCSI_OPCODE_SCSI_CMD;
    pdu->i_bit        = 0;
    pdu->x_bit        = 0;

    // BYTE 1
    pdu->attribute    = attribute;
    pdu->flag1resv    = 0;
    pdu->f_bit         = ACIT_TRUE; // 1 always
    pdu->r_bit         = read_bit;
    pdu->w_bit         = write_bit;

    // BYTE 2 and 3
    pdu->reserved1     = 0;

    // BYTE 4
    pdu->ahs_length    = 0;

    // BYTE 5
    // pdu->data_len_hi    = (ACIT_UEXACT8)((data_segment_length>>16) & 0x00ff);
    pdu->data_len_hi = 0;

    // BYTE 6 and 7
    // ACIT_PUT_BE16(pdu->data_len_lo, (ACIT_UEXACT16)(data_segment_length & 0x0000ffff));
    pdu->data_len_lo = 0;

    // BYTES 8 to 15
    OSM_MEMCPY (&pdu->lun, s3_lun, PDUH_SZ_SAM2LUN);  

    // BYTES 16 to 19
    // ACIT_PUT_BE32(pdu->task_tag, itt);

    // BYTES 20 to 23
    ACIT_PUT_BE32(pdu->data_trans_len, data_transfer_length);

    // BYTES 24 to 27
    // ACIT_PUT_BE32(pdu->cmd_sn, cmd_sn);    

    // BYTES 28 to 31 - patched by EP f/w for ExpStatSN
	// EP firmware expects driver to set this field to 0
	pdu->exp_stat_sn = 0;

    // BYTES 32 to 47
    OSM_MEMCPY(&pdu->cdb, cdb, sizeof(pdu->cdb));
}


// -----------------------------------------------------------------------------
// Data-In
// -----------------------------------------------------------------------------
typedef struct _PDU_DATA_IN {
    u8  opcode : 6;          // opcode 0x25
    u8  rsvd0 : 2;           // should be 0    
    u8  status_bit : 1;       // S bit if set piggy-back status 
    u8  underflow_bit : 1;    // U bit
    u8  overflow_bit : 1;     // O bit
    u8  rsvd1 : 3;           // Reserved
    u8  a_bit: 1;             // A Bit
    u8  f_bit : 1;            // F bit  
    u8  rsvd2;
    u8  status;              // SCSI status field 
    u8  ahs_length;           // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;           // DataSegmentLength
    u32 lun[2];
    u32 itt;                 // initiator task tag
    u32 ttt;                 // valid or 0xffffffff
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u32 data_sn;
    u32 buffer_offset;
    u32 residual_count;
} PDU_DATA_IN, *PPDU_DATA_IN;


STATIC INLINE
void
pdu_data_in_swap_bytes (PPDU_DATA_IN pdu) {

    pdu->data_len_lo     = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->itt           = ACIT_GET_BE32( &pdu->itt );
    pdu->ttt           = ACIT_GET_BE32( &pdu->ttt );
    pdu->stat_sn        = ACIT_GET_BE32( &pdu->stat_sn );
    pdu->exp_cmd_sn      = ACIT_GET_BE32( &pdu->exp_cmd_sn );
    pdu->max_cmd_sn      = ACIT_GET_BE32( &pdu->max_cmd_sn );
    pdu->data_sn        = ACIT_GET_BE32( &pdu->data_sn );
    pdu->buffer_offset  = ACIT_GET_BE32( &pdu->buffer_offset );
    pdu->residual_count = ACIT_GET_BE32( &pdu->residual_count );
}


// -----------------------------------------------------------------------------
// Data-Out
// -----------------------------------------------------------------------------
typedef struct _PDU_DATA_OUT {
    u8  opcode : 6; // opcode 0x05
    u8  rsvd0 : 2;  // should be 0    
    u8  rsvd1 : 7;
    u8  final_bit : 1;   // F bit 
    u8  rsvd2[2];
    u8  ahs_length;  // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;  // DataSegmentLength
    u32 lun[2]; 
    u32 itt;        // ITT; initiator task tag
    u32 ttt;        // TTT; target transfer tag; valid for R2T or 0xffffffff
    u8  rsvd3[4];
    u32 exp_stat_sn;
    u8  rsvd4[4];
    u32 data_sn;
    u32 buffer_offset;
    u8  rsvd5[4];
} PDU_DATA_OUT, *PPDU_DATA_OUT;

STATIC INLINE void
pdu_data_out_swap_bytes (PPDU_DATA_OUT pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->itt = ACIT_GET_BE32( &pdu->itt );
    pdu->ttt = ACIT_GET_BE32( &pdu->ttt);
    pdu->exp_stat_sn = ACIT_GET_BE32( &pdu->exp_stat_sn );
    pdu->data_sn = ACIT_GET_BE32( &pdu->data_sn);
    pdu->buffer_offset = ACIT_GET_BE32( &pdu->buffer_offset );
}

STATIC INLINE void
pdu_build_dataout (PPDU_DATA_OUT pdu,
                   void* s3_lun
                   )
{
    OSM_ASSERT (pdu);
    OSM_MEMSET(pdu, 0, ISCSI_PDU_HEADER_SIZE);

    pdu->opcode       = ISCSI_OPCODE_SCSI_DATA_OUT;
    pdu->final_bit    = ACIT_TRUE;
    
    //ACIT_PUT_BE32(pdu->itt, itt);

    //OSM_ASSERT ((data_segment_len >> 24) == 0);
    //pdu->data_len_hi    = (u8)((data_segment_len>>16) & 0x00ff);
    //ACIT_PUT_BE16(pdu->data_len_lo, (u16)(data_segment_len & 0x0000ffff));  

    OSM_MEMCPY (&pdu->lun, s3_lun, PDUH_SZ_SAM2LUN);  
}

// -----------------------------------------------------------------------------
// SCSI Command Response
// -----------------------------------------------------------------------------
typedef struct _PDU_COMMAND_RESPONSE {
    u8  opcode : 6;             // opcode 0x21
    u8  i_bit : 1;               // reserved; should be 0
    u8  x_bit : 1;               // reserved; should be 0
    u8  flag1_filler : 1;        // should be 0
    u8  underflow : 1;          // U bit; residual underflow
    u8  overflow : 1;           // O bit; residual overflow       
    u8  bidir_read_under : 1;     // u bit; bidi read residual underflow
    u8  bidir_read_over : 1;      // o bit; bidi read residual overflow
    u8  flag1_filler1 : 2;       // reserved   
    u8  f_bit : 1;               // reserved; should be 0
    u8  response;               // iSCSI service response
    u8  status;                 // SCSI status field 
    u8  ahs_length;              // no AHS; TotalAHSlength=0
    u8  data_len_hi;
    u16 data_len_lo;              // DataSegmentLength
    u8  reserved3[8];           // reserved
    u32 itt;                    // initiator task tag
    u32 reserved4;                      
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u32 exp_data_sn;
    u32 bidi_read_residual_count;
    u32 residual_count;
} PDU_COMMAND_RESPONSE, *PPDU_COMMAND_RESPONSE;

STATIC INLINE 
void
pdu_cmd_response_swap_bytes (PPDU_COMMAND_RESPONSE pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->itt = ACIT_GET_BE32( &pdu->itt );
    pdu->residual_count = ACIT_GET_BE32( &pdu->residual_count );
    pdu->stat_sn = ACIT_GET_BE32( &pdu->stat_sn );
    pdu->exp_cmd_sn = ACIT_GET_BE32( &pdu->exp_cmd_sn );
    pdu->max_cmd_sn = ACIT_GET_BE32( &pdu->max_cmd_sn );
    pdu->exp_data_sn = ACIT_GET_BE32( &pdu->exp_data_sn );
    pdu->bidi_read_residual_count = ACIT_GET_BE32( &pdu->bidi_read_residual_count );
}

// -----------------------------------------------------------------------------
// Text Request
// -----------------------------------------------------------------------------
typedef struct _PDU_TEXT {
    u8  opcode : 6;             // opcode 0x04
    u8  i_bit : 1;               // Immediate Bit
    u8  x_bit : 1;               // should be 0
    u8  fp_bit_filler1 : 6;
    u8  c_bit : 1;               // C Bit
    u8  f_bit : 1;               // F Bit
    u8  reserved1[2];
    u8  ahs_length;              // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;              // DataSegmentLength
    u32 lun[2];
    u32 task_tag;                // ITT; initiator task tag
    u32 transfer_tag;            // TTT; target transfer tag; valid (used for bookmark) or 0xffffffff
    u32 cmd_sn;
    u32 exp_stat_sn;  
    u8  reserved5[16]; 
    u32 header_digest;
} PDU_TEXT, * PPDU_TEXT;

// -----------------------------------------------------------------------------
// Text Response
// -----------------------------------------------------------------------------
typedef struct _PDU_TEXT_RESPONSE {
    u8  opcode : 6;             // opcode 0x24
    u8  i_bit : 1;               // should be 0
    u8  x_bit : 1;               // should be 0
    u8  fp_bit_filler1 : 6;
    u8  c_bit : 1;               // C bit
    u8  f_bit : 1;               // F bit
    u8  reserved1[2];
    u8  ahs_length;              // no AHS 
    u8  data_len_hi;     
    u16 data_len_lo;              // DatSegmentLength
    u32 lun[2];
    u32 itt;                // ITT; initiator task tag
    u32 ttt;            // TTT; target transfer tag; valid (used for bookmark) or 0xffffffff
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  reserved6[12];
    u32 header_digest;
} PDU_TEXT_RESPONSE, * PPDU_TEXT_RESPONSE;

// -----------------------------------------------------------------------------
// Login
// -----------------------------------------------------------------------------
typedef struct _PDU_LOGIN {
    u8  opcode : 6;             // opcode 0x03  
    u8  i_bit : 1;               // should be 1; login always immediate
    u8  x_bit : 1;               // should be 0;
    u8  nsg : 2;                // next stage
    u8  csg : 2;                // current stage
    u8  flag1_filler : 2;
    u8  c_bit : 1;               // C (Continue) Bit
    u8  t_bit : 1;               // T (transit) bit, set to 1 indicate ready to transit to next stage
    u8  versionMax;             // maximum version number support
    u8  versionMin;             // minimum version number support, 0x0 for draft 12 
    u8  ahs_length;              // no AHS
    u8  data_len_hi;
    u16 data_len_lo;              // DataSegmentLength
    u8  isid_A : 6;             // isid A field
    u8  isid_T : 2;             // isid Type
    u8  isid_B_Hi;              // isid B high byte
    u8  isid_B_Lo;              // isid B low byte
    u8  isid_C;                 // isid C field
    u16 isid_D;                 // isid D field, qualifier
    u16 tsid;                   // TSID: target session id 
    u32 task_tag;                // ITT; initiator task tag
    u16 cid;                    // connection iD, unique for this connection within the session
    u8  reserved4[2];
    u32 cmd_sn;                  // initial command sequence number of a session
    u32 exp_stat_sn;              // exp_stat_sn for old connection for connection reinstatement
    u8  reserved5[16];
} PDU_LOGIN, *PPDU_LOGIN;

STATIC INLINE void
pdu_login_swap_bytes( PPDU_LOGIN pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->isid_D    = ACIT_GET_BE16( &pdu->isid_D );
    pdu->tsid      = ACIT_GET_BE16( &pdu->tsid );
    pdu->task_tag   = ACIT_GET_BE32( &pdu->task_tag );
    pdu->cid       = ACIT_GET_BE16( &pdu->cid );
    pdu->cmd_sn     = ACIT_GET_BE32( &pdu->cmd_sn );
    pdu->exp_stat_sn = ACIT_GET_BE32( &pdu->exp_stat_sn );
}

STATIC INLINE void
pdu_login_build (PPDU_LOGIN    pdu, 
               u8             csg, 
               u8             nsg, 
               u32            data_segment_len, 
               const ISCSI_ISID* isid, 
               u16            cid, 
               u16            tsid, 
               u32            itt, 
               u32            cmd_sn, 
               u32            exp_stat_sn, 
               u8           continue_bit, 
               u8           transition_bit)
{
    OSM_MEMSET (pdu, 0, ISCSI_PDU_HEADER_SIZE);

    pdu->i_bit         = 1;
    pdu->opcode        = ISCSI_OPCODE_LOGIN; 
    pdu->c_bit         = continue_bit;
    pdu->t_bit         = transition_bit;
    pdu->nsg           = nsg;
    pdu->csg           = csg;   
    pdu->versionMax    = ISCSI_MAX_VERSION;
    pdu->versionMin    = ISCSI_MIN_VERSION;       
    if (data_segment_len)
    {
        OSM_ASSERT ((data_segment_len >> 24) == 0);
        pdu->data_len_hi = (u8)((data_segment_len >> 16) & 0x00ff);
        pdu->data_len_lo = (u16)(data_segment_len & 0x0000ffff);
    }
    pdu->cid           = cid;
    pdu->isid_A        = isid->s.basic.isid_A;
    pdu->isid_T        = isid->s.basic.isid_T;
    pdu->isid_B_Hi     = isid->s.basic.isid_B_H;
    pdu->isid_B_Lo     = isid->s.basic.isid_B_L;
    pdu->isid_C        = isid->s.basic.isid_C;
    pdu->isid_D        = isid->s.basic.isid_D;
    pdu->tsid          = tsid;    
    pdu->task_tag      = itt;
    pdu->cmd_sn        = cmd_sn;         
    pdu->exp_stat_sn   = exp_stat_sn;      

    pdu_login_swap_bytes (pdu);
}

// -----------------------------------------------------------------------------
// Login Response
// -----------------------------------------------------------------------------
typedef struct _PDU_LOGIN_RESPONSE {
    u8  opcode : 6;             // opcode 0x23
    u8  i_bit : 1;               // should be 0
    u8  x_bit : 1;               // should be 0
    u8  nsg : 2;
    u8  csg : 2;
    u8  rsvd : 2;
    u8  c_bit : 1;               // C Bit
    u8  t_bit : 1;               // T(transit) Bit
    u8  versionMax;
    u8  versionMin;             // active version number or lowest version support by target
    u8  ahs_length;              // no AHS
    u8  data_len_hi;
    u16 data_len_lo;              // DataSegmentLength
    u8  isid[6];                // initiator session id
    u16 tsid;                   // tag set by traget for initiator identifying
    u32 task_tag;                // ITT; initiator task tag
    u8  reserved4[4];
    u32 stat_sn;                 // starting status sequence number
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  statusClass;
    u8  statusDetail;
    u8  reserved6[2];
    u8  reserved7[8];
} PDU_LOGIN_RESPONSE, * PPDU_LOGIN_RESPONSE;

STATIC INLINE void 
pdu_login_response_swap_bytes ( PPDU_LOGIN_RESPONSE pdu )
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->tsid      = ACIT_GET_BE16( &pdu->tsid );
    pdu->task_tag   = ACIT_GET_BE32( &pdu->task_tag );
    pdu->stat_sn    = ACIT_GET_BE32( &pdu->stat_sn );
    pdu->exp_cmd_sn  = ACIT_GET_BE32( &pdu->exp_cmd_sn );
    pdu->max_cmd_sn  = ACIT_GET_BE32( &pdu->max_cmd_sn );
}


// -----------------------------------------------------------------------------
// NOP-out
// -----------------------------------------------------------------------------
typedef struct _PDU_NOP_OUT {
    u8  opcode : 6;             // opcode 0x00
    u8  i_bit : 1;               // I Bit
    u8  x_bit : 1;               // reserved; should be 0
    u8  fp_bit_filler1 : 7;
    u8  f_bit : 1;               // always 1
    u8  reserved1[2];
    u8  ahs_length;              // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;              // DataSegmentLength
    u32 lun[2];                 
    u32 itt;            // initiator task tag; initiator assigned id for ping or 0xffffffff
    u32 ttt;            // target transfer tag; target assigned id for ping or 0xffffffff
    u32 cmd_sn;
    u32 exp_stat_sn;
    u8  reserved5[16];
} PDU_NOP_OUT, *PPDU_NOP_OUT;

STATIC INLINE void 
pdu_nop_out_swap_bytes (PPDU_NOP_OUT pdu )
{
    pdu->data_len_lo   = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->ttt           = ACIT_GET_BE32( &pdu->ttt );
    pdu->itt           = ACIT_GET_BE32( &pdu->itt );
    pdu->cmd_sn        = ACIT_GET_BE32( &pdu->cmd_sn );
    pdu->exp_stat_sn   = ACIT_GET_BE32( &pdu->exp_stat_sn );
}

STATIC INLINE void
pdu_build_nop_out (PPDU_NOP_OUT pdu, 
                   u8 i_bit, 
                   ACIT_LUN* s3_lun, 
                   u32 itt, 
                   u32 ttt, 
                   u32 cmd_sn)
{
    OSM_MEMSET (pdu, 0, ISCSI_PDU_HEADER_SIZE);

    pdu->opcode = ISCSI_OPCODE_NOP_OUT;
    pdu->i_bit = i_bit;
    
    // pdu->x_bit = 0;
    // pdu->fp_bit_filler1 = 0;
    
    pdu->f_bit = 1;

    // pdu->ahs_length = 0;
    //pdu->data_len_hi = 0;    // Don't support
    //pdu->data_len_lo = 0;    

    if (s3_lun) {
        // optional
        OSM_MEMCPY (&pdu->lun, s3_lun, PDUH_SZ_SAM2LUN);  
    }

    ACIT_PUT_BE32(pdu->itt, itt);   // initiator task tag; valid value for ping or 0xffffffff
    ACIT_PUT_BE32(pdu->ttt, ttt);   // target transfer tag; valid value for ping or 0xffffffff
    ACIT_PUT_BE32(pdu->cmd_sn, cmd_sn);
}


// -----------------------------------------------------------------------------
// NOP-IN
// -----------------------------------------------------------------------------
typedef struct _PDU_NOP_IN {
    u8  opcode : 6;             // opcode 0x20
    u8  i_bit : 1;               // should be 0
    u8  x_bit : 1;               // should be 0
    u8  fp_bit_filler1 : 7;       
    u8  f_bit : 1;               // always 1
    u8  reserved1[2];
    u8  ahs_length;              // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;              // DatSegmentLength
    u32 lun[2];
    u32 itt;                    // ITT; initiator task tag; valid value for ping or 0xffffffff
    u32 ttt;                    // TTT; target transfer tag; valid value for ping or 0xffffffff
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  reserved6[12];
} PDU_NOP_IN, *PPDU_NOP_IN;

STATIC INLINE void
pdu_nop_in_swap_bytes (PPDU_NOP_IN pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->itt = ACIT_GET_BE32( &pdu->itt );
    pdu->ttt = ACIT_GET_BE32( &pdu->ttt );
    pdu->stat_sn = ACIT_GET_BE32( &pdu->stat_sn );
    pdu->exp_cmd_sn = ACIT_GET_BE32( &pdu->exp_cmd_sn );
    pdu->max_cmd_sn = ACIT_GET_BE32( &pdu->max_cmd_sn );
}


// -----------------------------------------------------------------------------
// Logout
// -----------------------------------------------------------------------------
typedef struct _PDU_LOGOUT {
    u8  opcode : 6;             // opcode 0x06
    u8  i_bit : 1;               // Immediate Bit
    u8  x_bit : 1;               // should be 0
    u8  reasonCode : 7;
    u8  f_bit : 1;               // F bit, shoule be 1
    u8  reserved1[2];
    u8  ahs_length;              // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;              // no data_segment_len
    u8  reserved3[8];           // no lun
    u32 itt;                    // initiator task tag;
    u16 cid;                    // only valid if reason code not close session
    u8  reserved4[2];   
    u32 cmd_sn;
    u32 exp_stat_sn;              // last exp_stat_sn value for the connection to be closed
    u8  reserved5[16];
} PDU_LOGOUT,* PPDU_LOGOUT;

STATIC INLINE 
void 
pdu_logout_swap_bytes (PPDU_LOGOUT pdu) {
    pdu->data_len_lo = ACIT_GET_BE16(&pdu->data_len_lo);
    pdu->itt         = ACIT_GET_BE32(&pdu->itt);
    pdu->cid         = ACIT_GET_BE16(&pdu->cid);
    pdu->cmd_sn      = ACIT_GET_BE32(&pdu->cmd_sn);
    pdu->exp_stat_sn = ACIT_GET_BE32(&pdu->exp_stat_sn);    
}

STATIC INLINE
void
pdu_logout_build (PPDU_LOGOUT pdu, 
                  u8 immediate, 
                  u8 reason_code, 
                  u16 cid, 
                  u32 itt, 
                  u32 cmd_sn)
{
    OSM_MEMSET (pdu, 0, sizeof (PDU_LOGOUT));

    pdu->i_bit = (immediate ? 1 : 0);
    pdu->opcode = ISCSI_OPCODE_LOGOUT;
    pdu->f_bit = 1;
    pdu->reasonCode = reason_code;
    ACIT_PUT_BE16(pdu->cid, cid);
    ACIT_PUT_BE32(pdu->itt, itt);
    ACIT_PUT_BE32(pdu->cmd_sn, cmd_sn);
}



// -----------------------------------------------------------------------------
// Logout Response
// -----------------------------------------------------------------------------
typedef struct _PDU_LOGOUT_RESPONSE {
    u8  opcode : 6;             // opcode 0x26
    u8  i_bit : 1;               // should be 0
    u8  x_bit : 1;               // should be 0
    u8  fp_bit_filler1 : 7;
    u8  f_bit : 1;               // F bit, shoule be 1
    u8  response;
    u8  reserved1;
    u8  ahs_length;              // no AHS; 
    u8  data_len_hi;              // no Data Segment
    u16 data_len_lo;    
    u8  reserved3[8];           // no lun
    u32 itt;                    // initiator task tag;
    u8  reserved4[4];
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  reserved6[4];
    u16 time_wait;              // min time in seconds to wait before login for reinstating
    u16 time_retain;            // max time to wait for login to new connection
    u8  reserved8[4];
} PDU_LOGOUT_RESPONSE,* PPDU_LOGOUT_RESPONSE;

STATIC INLINE void
pdu_logout_response_swap_bytes (PPDU_LOGOUT_RESPONSE pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16(&pdu->data_len_lo);
    pdu->itt = ACIT_GET_BE32(&pdu->itt);
    pdu->stat_sn = ACIT_GET_BE32(&pdu->stat_sn);
    pdu->exp_cmd_sn = ACIT_GET_BE32(&pdu->exp_cmd_sn);
    pdu->max_cmd_sn = ACIT_GET_BE32(&pdu->max_cmd_sn);
    pdu->time_wait = ACIT_GET_BE16(&pdu->time_wait);
    pdu->time_retain = ACIT_GET_BE16(&pdu->time_retain);
}


// -----------------------------------------------------------------------------
// Async Message
// -----------------------------------------------------------------------------
typedef struct _PDU_ASYNC_MSG {
    u8  opcode : 6;             // opcode 0x32
    u8  i_bit : 1;               // should be 0
    u8  x_bit : 1;               // should be 0
    u8  fp_bit_filler1 : 7;
    u8  f_bit : 1;               // F bit, shoule be 1
    u8  reserved1[2];
    u8  ahs_length;              // no AHS
    u8  data_len_hi;     
    u16 data_len_lo;    
    u32 lun[2];
    u32 task_tag;                // should be 0xffffffff only
    u8  reserved4[4];           
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  async_event;             // asynchronous event code
    u8  async_vcode;             // vendor specific detail code
    u16 parameter1;
    u16 parameter2;
    u16 parameter3;
    u8  reserved8[4];
} PDU_ASYNC_MSG,* PPDU_ASYNC_MSG;

STATIC INLINE
void pdu_async_msg_swap_bytes (PPDU_ASYNC_MSG pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16( &pdu->data_len_lo );
    pdu->stat_sn = ACIT_GET_BE32( &pdu->stat_sn );
    pdu->exp_cmd_sn = ACIT_GET_BE32( &pdu->exp_cmd_sn );
    pdu->max_cmd_sn = ACIT_GET_BE32( &pdu->max_cmd_sn );
    pdu->parameter1 = ACIT_GET_BE16( &pdu->parameter1 );
    pdu->parameter2 = ACIT_GET_BE16( &pdu->parameter2 );
    pdu->parameter3 = ACIT_GET_BE16( &pdu->parameter3 );
}

// -----------------------------------------------------------------------------
// Task Management Request
// -----------------------------------------------------------------------------

typedef struct _PDU_TASK_MGMT {
    u8  opcode : 6;             // opcode 0x02
    u8  i : 1;               // I-bit
    u8  x : 1;               // reserved; should be 0
    u8  function : 7;
    u8  f : 1;               // should be 1
    u8  reserved1[2];                
    u8  ahs_length;              // no AHS; TotalAHSlength=0
    u8  data_len_hi;
    u16 data_len_lo;              // DataSegmentLength
    u32 lun[2];
    u32 itt;                    // ITT; initiator task tag
    u32 ref_task_tag;      // valid or 0xffffffff 
    u32 cmd_sn;
    u32 exp_stat_sn;
    u32 ref_cmd_sn;               // RefCmdSN
    u32 exp_data_sn;              // ExpDataSN 
    u8  reserved7[8];
} PDU_TASK_MGMT, *PPDU_TASK_MGMT;


STATIC INLINE
void pdu_tmf_swap_bytes (PPDU_TASK_MGMT pdu)
{
    pdu->data_len_lo  = ACIT_GET_BE16(&pdu->data_len_lo);
    pdu->itt          = ACIT_GET_BE32(&pdu->itt);
    pdu->ref_task_tag = ACIT_GET_BE32(&pdu->ref_task_tag);
    pdu->cmd_sn       = ACIT_GET_BE32(&pdu->cmd_sn);
    pdu->exp_stat_sn  = ACIT_GET_BE32(&pdu->exp_stat_sn);
    pdu->ref_cmd_sn   = ACIT_GET_BE32(&pdu->ref_cmd_sn);
    pdu->exp_data_sn  = ACIT_GET_BE32(&pdu->exp_data_sn);
}

STATIC INLINE
void
pdu_build_tmf (PPDU_TASK_MGMT pdu,
               u8 function,
               u8 immediate,
               ACIT_LUN* s3_lun,
               u32 itt,
               u32 ref_task_tag,
               u32 cmd_sn,
               u32 ref_cmd_sn)
{
    OSM_MEMSET (pdu, 0, sizeof (PDU_TASK_MGMT));

    pdu->i = (immediate ? 1 : 0);
    pdu->f = 1;
    pdu->opcode = ISCSI_OPCODE_SCSI_TASK_MGMT;
    pdu->function = function;

    OSM_MEMCPY (&pdu->lun, s3_lun, PDUH_SZ_SAM2LUN);  
    
    ACIT_PUT_BE32(pdu->itt, itt);
    ACIT_PUT_BE32(pdu->ref_task_tag, ref_task_tag);
    ACIT_PUT_BE32(pdu->cmd_sn, cmd_sn);
    ACIT_PUT_BE32(pdu->ref_cmd_sn, ref_cmd_sn);
}
               


// -----------------------------------------------------------------------------
// Task Management Response
// -----------------------------------------------------------------------------

typedef struct _PDU_TASK_MGMT_RESPONSE {
    u8  opcode : 6;             // opcode 0x22
    u8  i : 1;               // reserved; should be 0
    u8  x : 1;               // reserved; should be 0
    u8  rsvd0 : 7;
    u8  f : 1;               // should be 1
    u8  response;               // target response
    u8  reserved1;              // reserved;
    u8  ahs_length;              // no AHS; TotalAHSlength=0
    u8  data_len_hi;     
    u16 data_len_lo;              // DataSegmentLength
    u8  rsvd2[8];           // no lun
    u32 itt;                // ITT; initiator task tag
    u32 reserved3;              // no refence task tag; should be 0
    u32 stat_sn;
    u32 exp_cmd_sn;
    u32 max_cmd_sn;
    u8  reserved6[12];        
} PDU_TASK_MGMT_RESPONSE, * PPDU_TASK_MGMT_RESPONSE;

STATIC INLINE
void pdu_tmf_response_swap_bytes (PPDU_TASK_MGMT_RESPONSE pdu)
{
    pdu->data_len_lo = ACIT_GET_BE16(&pdu->data_len_lo);
    pdu->itt         = ACIT_GET_BE32(&pdu->itt);
    pdu->stat_sn     = ACIT_GET_BE32(&pdu->stat_sn);
    pdu->exp_cmd_sn  = ACIT_GET_BE32(&pdu->exp_cmd_sn);
    pdu->max_cmd_sn  = ACIT_GET_BE32(&pdu->max_cmd_sn);
}

typedef struct _PDUH_REJECT
{
    ACIT_UEXACT8     xiopcode;
    ACIT_UEXACT8     flags;
    ACIT_UEXACT8     reason;
    ACIT_UEXACT8     reservedbyte03 [2];
    ACIT_UEXACT8     data_segment_len [PDUH_SZ_DATA_SEG_LEN];
    ACIT_UEXACT8     reservedbyte08 [16];
    ACIT_UEXACT32    statsn;
    ACIT_UEXACT32    expcmdsn;
    ACIT_UEXACT32    maxcmdsn;
    ACIT_UEXACT32    datasn;
    ACIT_UEXACT8     reservedbyte40 [8];
}PDUH_REJECT, *PPDUH_REJECT;

#endif 
