/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
  @file
  acitproc.c

  @brief
    This file will be used for the the ACIT Linux/ESX debug features. All the 
    debug routines of Linux/ESX using the ACIT Layer should be added in this 
    file. 

    This file will not be included by Windows for compilation, only Linux/ESX will
    add this file for compilation.
 */
#include "acitpch.h"
#ifndef __VMKLNX__
#include <linux/seq_file.h>
#endif

#define OPTION_ALL  0xF /* All the CID */
#define OPTION_VAL  0xA /* Valid CID   */
#define OPTION_CUR  0xC /* Current CID */

#define PIOH_SGL_MEM    0x1 /* Check the SGL Handle */
#define PIOH_CRI_MEM    0x2 /* Check the CRI Member */

#define DMP_ALL_ICD (-1)
#define DMP_ALL_CID (-1)

#define GET_DATA(_bstruct,_struct,_fieldname) \
    (((_bstruct *)_struct)->_fieldname)

typedef enum _ACIT_HLIST_TYPE {
    ACIT_IO_HLIST,
    ACIT_TMF_HLIST,
    ACIT_LOGOUT_HLIST,
    ACIT_TGT_HLIST
}ACIT_HLIST_TYPE;


/***** Common Start ******/

static int 
check_cid_valid(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       cid,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32 cid_start = 0;

    /* Check if the CID is Valid */
    cid_start = pws_ctrl->fw_config.iscsi_cid_start;

    if(cid_start != (cid & 0x01)){
        seq_printf(pseq,"CID : %d is not valid\n",cid);
        return -EINVAL;
    }
    else if ( cid > (pws_ctrl->fw_config.iscsi_cid_count +
                pws_ctrl->fw_config.iscsi_cid_start) ) {
        seq_printf(pseq,"CID : %d is not valid, Max CID possible is : %d\n",
                cid,
                (pws_ctrl->fw_config.iscsi_cid_count +
                 pws_ctrl->fw_config.iscsi_cid_start) );

        return -EINVAL;
    }

    return 0;
}


static PACIT_TARGET_WS
ptgt_from_cid(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       cid,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       cri;
    PISCSI_CONNECTION   pcxn;
    PACIT_TARGET_WS     ptarget = ACIT_NULL;
    PISCSI_TARGET_WS    piscsi_tgt;

    cri = CRI_FROM_CID(pws_ctrl,cid);
    pcxn = GET_ISCSI_CXN(pws_ctrl,cri);

    if(pcxn){
        if(pcxn->cid != cid){
            seq_printf(pseq,"ERROR!!! CID mismatch actual: %d pcxn : %d cri : %d\n",
                    cid,pcxn->cid,cri);

            return ACIT_NULL;
        }

        piscsi_tgt = pcxn->piscsi_target_ws;

        if(piscsi_tgt)
            ptarget = piscsi_tgt->pacit_ws;

    }

    return ptarget;
}

static void 
target_dump(
        PACIT_CONTROLLER_WS     pws_ctrl,
        ACIT_UEXACT32           cid,
        struct seq_file         *pseq)
{
    PISCSI_CONNECTION   pcxn;
    PISCSI_TARGET_WS    piscsi_tgt;
    PACIT_TARGET_WS     ptarget;

    ptarget = ptgt_from_cid(pws_ctrl,cid,pseq);
    if(!ptarget){
        seq_printf(pseq,"ptarget is NULL\n");
        return ;
    }

    pcxn = GET_ISCSI_CXN(pws_ctrl,CRI_FROM_CID(pws_ctrl,cid));
    piscsi_tgt = pcxn->piscsi_target_ws; 

    if(!piscsi_tgt){
        seq_printf(pseq,"pscsi_tgt is NULL\n");
        return ;
    }

    seq_printf(pseq,"\n\n\t\tDumping Target WS \n");

    seq_printf(pseq,"Session ID : %08d Target Table Index  : %08d\n",
            piscsi_tgt->session_handle,ptarget->target_table_index);

    seq_printf(pseq,"CmdSn      : 0x%08x          ExpCmdSn : 0x%08x MaxCmdSn : 0x%08x\n",
            piscsi_tgt->cmd_sn,piscsi_tgt->exp_cmd_sn,piscsi_tgt->max_cmd_sn);

    seq_printf(pseq,"Busy       : 0x%08x          SubState : 0x%08x     Flag : 0x%08x\n",
            ptarget->busy,ptarget->sub_state,ptarget->flags);

    seq_printf(pseq,"EventFlag  : 0x%08x           Cleanup : 0x%08x\n",
            ptarget->event_flag,ptarget->cleanup_flag);

//    seq_printf(pseq," Last IO Completed at tick 0x%x (Current Tick is 0x%x)\n",
//            ptarget->stats.last_completed_tick,
//            pws_ctrl->counters.current_tick_count);

    return ;
}


void dump_pwrb(
        PISCSI_WRB      pwrb,
        struct seq_file *pseq
        )
{
    seq_printf(pseq,"\n\t\t Dumping WRB : %p\n",pwrb);

    seq_printf(pseq,"WRB Index : 0x%08x     CmdSn : 0x%08x      ICD : 0x%08x\n",
            pwrb->wrb_idx,
            pwrb->cmdsn_itt,
            pwrb->sgl_icd_idx);

    seq_printf(pseq,"WRB Type  : 0x%08x   BHS Low : 0x%08x BHS High : 0x%08x\n",
            pwrb->type,
            pwrb->iscsi_bhs_addr_lo,
            pwrb->iscsi_bhs_addr_hi);

}

void dump_ptmf_handle(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_TMF_HANDLE    ptmf_handle,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   cid;

    cid = CID_FROM_CRI(pws_ctrl,ptmf_handle->cri);

    seq_printf(pseq,"\n\t\t TMF : ACIT_TMF_HANDLE\n");  
    seq_printf(pseq,"CmdSn   : 0x%08x             ITT : 0x%08x\n",
            ptmf_handle->cmd_sn,ptmf_handle->itt);
    seq_printf(pseq,"CID     : 0x%08x    Driver Flags : 0x%08x \n",
            cid,ptmf_handle->flags);
#if 0   
    seq_printf(pseq,"Target     : %p\n",ptmf_handle->ptarget);
    seq_printf(pseq,"LUN        : %p\n",ptmf_handle->plun);
    seq_printf(pseq,"SGLHandle  : %p\n",ptmf_handle->psgl_handle);
    seq_printf(pseq,"WRBHandle  : %p\n",ptmf_handle->pwrb_handle);
    
    if(ptmf_handle->pwrb_handle)
        seq_printf(pseq,"WRB        : %p\n",ptmf_handle->pwrb_handle->pwrb);
    
#endif  
    seq_printf(pseq,"Starting Tick : 0x%x (Current Tick : 0x%x)\n",
            ptmf_handle->starting_tick_count,
            pws_ctrl->counters.current_tick_count);


    if(ptmf_handle->pwrb_handle)
        dump_pwrb(ptmf_handle->pwrb_handle->pwrb,pseq);

    return;
}

void dump_plogout_handle(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_LOGOUT_HANDLE plo_handle,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   cid;

    cid = CID_FROM_CRI(pws_ctrl,plo_handle->cri);

    seq_printf(pseq,"\n\t\t LOGOUT : ACIT_LOGOUT_HANDLE\n");    
    seq_printf(pseq,"CmdSn   : 0x%08x             ITT : 0x%08x\n",
            plo_handle->cmd_sn,plo_handle->itt);
    seq_printf(pseq,"CID     : 0x%08x    Driver Flags : 0x%08x Reason Code : 0x%08x\n",
            cid,plo_handle->flags,plo_handle->reason_code);
#if 0   
    seq_printf(pseq,"Target     : %p\n",plo_handle->ptarget);
    seq_printf(pseq,"SGLHandle  : %p\n",plo_handle->psgl_handle);
    seq_printf(pseq,"WRBHandle  : %p\n",plo_handle->pwrb_handle);

    if(plo_handle->pwrb_handle)
        seq_printf(pseq,"WRB        : %p\n",plo_handle->pwrb_handle->pwrb);
#endif
    seq_printf(pseq,"Starting Tick : 0x%x (Current Tick : 0x%x)\n",
            plo_handle->starting_tick_count,
            pws_ctrl->counters.current_tick_count);

    if(plo_handle->pwrb_handle)
        dump_pwrb(plo_handle->pwrb_handle->pwrb,pseq);

    return;
}


void dump_pio_handle(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_IO_HANDLE     pio_handle,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   cid;    

    cid = CID_FROM_CRI(pws_ctrl,pio_handle->cri);

    seq_printf(pseq,"\n\t\t IOH : ACIT_IO_HANDLE\n");   
    seq_printf(pseq,"CmdSn   : 0x%08x             ITT : 0x%08x\n",
            pio_handle->cmd_sn,pio_handle->itt);
    seq_printf(pseq,"CID     : 0x%08x    Driver Flags : 0x%08x\n",
            cid,pio_handle->flags);
#if 0   
    seq_printf(pseq,"Target     : %p\n",pio_handle->ptarget);
    seq_printf(pseq,"LUN        : %p\n",pio_handle->plun);
    seq_printf(pseq,"PIO        : %p\n",pio_handle->pio);
    seq_printf(pseq,"SGLHandle  : %p\n",pio_handle->psgl_handle);
    seq_printf(pseq,"WRBHandle  : %p\n",pio_handle->pwrb_handle);
    
    if(pio_handle->pwrb_handle)
        seq_printf(pseq,"WRB        : %p\n",pio_handle->pwrb_handle->pwrb);
#endif
    seq_printf(pseq,"Starting Tick : 0x%x (Current Tick : 0x%x)\n",
            pio_handle->starting_tick_count,
            pws_ctrl->counters.current_tick_count);

    if(pio_handle->pwrb_handle)
        dump_pwrb(pio_handle->pwrb_handle->pwrb,pseq);
    
    return;
}

static ACIT_UEXACT32
dump_list(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_DC_LIST       pioh_list,
        ACIT_UEXACT32       start_idx,
        ACIT_UEXACT32       num_entry,
        ACIT_HLIST_TYPE     hlist_type,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32       count = 0,prnt = 0;
    ACIT_DC_LIST        *pioh_next,*pioh_element;
    PACIT_IO_HANDLE     pio_handle = ACIT_NULL;

    pioh_element = pioh_list;
    pioh_next = pioh_element->next;

    while(pioh_next != pioh_list){
        
        if(hlist_type == ACIT_IO_HLIST){
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                controller_link,pioh_element);
        }
        else if(hlist_type == ACIT_TGT_HLIST){
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                    target_link,pioh_element);
        }
        else if(hlist_type == ACIT_TMF_HLIST){
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                    tmf_link,pioh_element);
        }

        if(count >= start_idx){
            dump_pio_handle(pws_ctrl,pio_handle,pseq);

            if(++prnt > num_entry)
                break;
        }

        count++;
        pioh_element = pioh_next;
        pioh_next = pioh_next->next;
        
    }

    return prnt;
}

/***** Comon End ******/


/***** IOS QD Start  ******/

static void
dump_io_request(
        PACIT_IO_REQUEST    pio_req,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   i = 0;

#if 0   
    seq_printf(pseq,"         TargetKey    : 0x%p\n", pio_req->ptarget_key);
    seq_printf(pseq,"         LunKey       : 0x%p \n",pio_req->plun_key);
    seq_printf(pseq,"         IOKey        : 0x%p \n",pio_req->pio_key);
#endif  
    seq_printf(pseq,"         Flags        : 0x%x \n",*((ACIT_UEXACT32 *)&pio_req->flags));
    seq_printf(pseq,"         TimeOut      : %d\n", pio_req->timeout);
    seq_printf(pseq,"         DataXferLen  : %d (0x%x) Bytes \n", 
            pio_req->data_transfer_length, 
            pio_req->data_transfer_length);
    seq_printf(pseq,"         CDBLen       : %d \n", 
            pio_req->cdb_length);
    seq_printf(pseq,"         CDB          : ");

    for(i=0; i<pio_req->cdb_length; i++)
    {
        seq_printf(pseq,"0x%x ", pio_req->cdb[i]);
    }

    seq_printf(pseq,"\n");

    seq_printf(pseq,"         IOStatus     : 0x%x \n", pio_req->io_status);
    seq_printf(pseq,"         SCSIStatus   : 0x%x \n", pio_req->scsi_status);
}



static void
dump_pio_ws(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_IO_WS         pio_ws,
        struct seq_file     *pseq)
{
#if 0
    seq_printf(pseq,"       IOWS         : 0x%p (_ACIT_IO_WS)\n",pio_ws); 
    seq_printf(pseq,"       IOKey        : 0x%p (_ACIT_IO_KEY)\n",pio_ws->pio_key);
    seq_printf(pseq,"       IORequest    : 0x%p (_ACIT_IO_REQUEST)\n",pio_ws->pio_request);
#endif
    if(pio_ws->pio_request)
        dump_io_request(pio_ws->pio_request,pseq);

    seq_printf(pseq,"CurrentTick:0x%x\n", 
            pws_ctrl->counters.current_tick_count);

    seq_printf(pseq,"       OriginalTicks: 0x%x \n", 
            pio_ws->original_ticks);
    
    seq_printf(pseq,"       AvailableTicks:0x%x \n", 
            pio_ws->available_ticks);
    
    seq_printf(pseq,"       InternalAvailableTicks: 0x%x \n",
            pio_ws->internal_available_ticks);
    
    seq_printf(pseq,"       Queued       : %d \n", pio_ws->queued);
    
    seq_printf(pseq,"\n");



}

static ACIT_UEXACT32
dump_ioq_list (
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_DC_LIST       iq_list,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32   count = 0;
    ACIT_DC_LIST    *element,*next;
    PACIT_IO_WS     pio_ws;

    element = iq_list;
    next = element->next;

    while(next != iq_list){

        pio_ws = STRUCT_BASE(ACIT_IO_WS,
                target_link,element);

        dump_pio_ws(pws_ctrl,pio_ws,pseq);

        element = next;
        next = next->next;

        count++;

        if(count > 1024){
            seq_printf(pseq,"List might be corrupted, breaking off the loop\n");
            break;
        }
    }

    return count;
    
}

static void
dump_target_queued_ios(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_TARGET_WS     ptarget_ws,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   prntd = 0;
    ACIT_DC_LIST    *ioq_list;

    if(ptarget_ws)
        ioq_list = ptarget_ws->io_queued_list.next;
    else
        ioq_list = pws_ctrl->io_queued_list.next;

    if(!ioq_list)
        goto EXIT_FUNC;

    prntd = dump_ioq_list(pws_ctrl,ioq_list,pseq);
    
EXIT_FUNC:
        seq_printf(pseq,"Completed dump of %d queued IOs\n", prntd);
        return;
}

static void
dump_queued_ios(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_SEXACT32       cid,
        struct seq_file     *pseq)
{

    PACIT_TARGET_WS ptarget_ws;
    
    if(cid != DMP_ALL_CID){

        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);

        if(!ptarget_ws){
            seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
            seq_printf(pseq,"Use cid command to get the list of current cids on this controller\n");

            return;
        }

        dump_target_queued_ios(pws_ctrl,ptarget_ws,pseq);
    }
    else {
        seq_printf(pseq,"Checking all queued IOs \n");
        dump_target_queued_ios(pws_ctrl,NULL,pseq);
    }
}


void
acit_process_proc_iosqd(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_SEXACT32           cid,
        struct seq_file         *pseq)
{

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    if(cid < 0 &&
            cid != DMP_ALL_CID){
        seq_printf(pseq,"cid : %d is not valid\n",cid);
        return;
    }
    
    if(cid != DMP_ALL_CID)
        if(check_cid_valid(pws_ctrl,cid,pseq))
            return ;

    dump_queued_ios(pws_ctrl,cid,pseq);

}

/***** IOS QD End ******/

/***** INVALID Start  ******/

void dump_invalidation_list(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32           io_count = 0,inv_count = 0,prntd = 0;
    ACIT_DC_LIST            *inv_list,*element,*next,*ioh_list;
    PACIT_INVALIDATE_HANDLE pinv_handle = ACIT_NULL;

    inv_list = pws_ctrl->inv_handle_busy_list.next;

    if(!inv_list)
        return;

    element = inv_list;
    next = element->next;

    while(next != inv_list){
        
        pinv_handle = STRUCT_BASE(ACIT_INVALIDATE_HANDLE,
                controller_link,element);

        io_count += pinv_handle->num_invs;

        seq_printf(pseq," Dumping %d IOHandles in invalidate handle 0x%p (_ACIT_INVALIDATE_HANDLE)\n",
                                pinv_handle->num_invs, pinv_handle);

        ioh_list = pinv_handle->list_header.next;
        prntd =  dump_list(pws_ctrl,
                ioh_list,
                0,
                pinv_handle->num_invs,
                ACIT_IO_HLIST,  
                pseq);

        if(prntd != pinv_handle->num_invs)
            seq_printf(pseq,"Error!! No of IOH in the list %d, count is %d\n",
                    pinv_handle->num_invs, prntd);

        element = next;
        next = next->next;
        
        inv_count++;

        if(inv_count > 1024){
            seq_printf(pseq,"List might be corrupted, breaking off the loop\n");
            break;
        }
    }
    seq_printf(pseq,"\nCompleted dump of %d invalidation requests for a total of %d IO invalidations\n",
                    inv_count, io_count);
}


void acit_process_proc_invl(
        PACIT_CONTROLLER_KEY    pcontroller,
        struct seq_file         *pseq)
{

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    dump_invalidation_list(pws_ctrl,pseq);
}


/***** INVALIDEnd ******/

/***** TMF Start  ******/

ACIT_UEXACT32
dump_lun_qd_tmf(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_LUN_WS    plun,
        struct seq_file *pseq)
{
    
    ACIT_UEXACT32   prntd = 0;
    ACIT_DC_LIST    *abort_list;

    abort_list = plun->tmf_abort_list.next;

    if(!abort_list)
        return prntd;

    prntd =  dump_list(pws_ctrl,
            abort_list,
            0,
            1024,
            ACIT_TMF_HLIST, 
            pseq);
    
    seq_printf(pseq,"Found %d Abort TMFs queued for LUN\n", prntd);

    return prntd;
}

ACIT_UEXACT32
dump_tgt_queued_tmf(
    PACIT_CONTROLLER_WS pws_ctrl,
    PACIT_TARGET_WS     ptarget_ws,
    struct seq_file     *pseq)
{
    ACIT_UEXACT32   prntd = 0;
    
    ACIT_DC_LIST    *lun_list,*element,*next;
    PACIT_LUN_WS    plun;   

    seq_printf(pseq,"Checking queued TMFs on target 0x%p\n", ptarget_ws);

    lun_list = ptarget_ws->tmf_lun_list.next;
    
    if(!lun_list)
        return prntd;

    element = lun_list;
    next = element->next;

    while(next != lun_list){
        plun = STRUCT_BASE(ACIT_LUN_WS,
                tmf_link,element);

        prntd = dump_lun_qd_tmf(pws_ctrl,plun,pseq);

        element = next;
        next = next->next;
    }

    return prntd;
}

ACIT_UEXACT32   
dump_queued_tmf (
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_SEXACT32       cid,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   prntd = 0;
    ACIT_DC_LIST    *tgt_list,*element,*next;
    PACIT_TARGET_WS ptarget_ws;

    if(cid != DMP_ALL_CID){
        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);

        if(!ptarget_ws) {
            seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
            seq_printf(pseq,"Use cid command to get the list of current cids on this controller\n");
            return prntd;
        }

        prntd = dump_tgt_queued_tmf(pws_ctrl,ptarget_ws,pseq);
    }
    else{
        seq_printf(pseq,"Checking all queued TMFs \n");
    
        tgt_list = pws_ctrl->io_queued_list.next;

        if(!tgt_list)
            return prntd;

        element = tgt_list;
        next = element->next;

        while(next != tgt_list){

            ptarget_ws = STRUCT_BASE(ACIT_TARGET_WS,
                    tmf_link,element);

            prntd = dump_tgt_queued_tmf(pws_ctrl,ptarget_ws,pseq);

            element = next;
            next = next->next;
            
        }
    }

    return prntd;
}

ACIT_UEXACT32
dump_tmf_ws(
        PACIT_CONTROLLER_WS pws_ctrl,
        PACIT_TMF_WS        ptmf_ws,
        PACIT_TARGET_WS     ptarget_ws,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       valid = 1,prntd = 0;
    PACIT_TMF_REQUEST ptmf_req;

    ptmf_req = ptmf_ws->ptmf_request;

    if(ptarget_ws){

        if(ptmf_req->ptarget_key->acit_key != ptarget_ws)
            valid = 0;
    }

    if(valid){
#if 0
        seq_printf(pseq," Printing TMF Workspace : 0x%p (_ACIT_TMF_WS)\n", ptmf_ws);
        seq_printf(pseq,"TMF Request    : 0x%p (_ACIT_TMF_REQUEST)\n", ptmf_req);
#endif
        
        seq_printf(pseq,"TMF Function   : %8d  TMF Status      : %8d\n",
                GET_DATA(ACIT_TMF_REQUEST,ptmf_req,tmf_function),
                GET_DATA(ACIT_TMF_WS,ptmf_ws,tmf_status));

        seq_printf(pseq,"Original Ticks : %8d  Available Ticks : %8d Timeout %d\n", 
                GET_DATA(ACIT_TMF_WS,ptmf_ws,original_ticks), 
                GET_DATA(ACIT_TMF_WS,ptmf_ws,available_ticks),
                GET_DATA(ACIT_TMF_REQUEST,ptmf_req,timeout));
#if 0
        seq_printf(pseq,"OSM Reference  : 0x%p\n", 
                GET_DATA(ACIT_TMF_REQUEST,ptmf_req,posm_reference));
#endif      
        seq_printf(pseq,"\n");
    }
    return ++prntd;
}

ACIT_UEXACT32
dump_osm_issued_tmf(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_SEXACT32       cid,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32   prntd = 0;
    ACIT_DC_LIST    *tmf_list,*element,*next;
    PACIT_TARGET_WS ptarget_ws = ACIT_NULL;
    PACIT_TMF_WS    ptmf_ws;


    if(cid != DMP_ALL_CID){
        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);

        if(!ptarget_ws) {
            seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
            seq_printf(pseq,"Use cid command to get the list of current cids on this controller\n");
            return prntd;
        }
    }

    tmf_list = pws_ctrl->tmf_queued_list.next;

    if(!tmf_list)
        return prntd;

    element = tmf_list;
    next = element->next;

    while(next != tmf_list){

        ptmf_ws = STRUCT_BASE(ACIT_TMF_WS,
                controller_link,element);

        prntd += dump_tmf_ws(pws_ctrl,ptmf_ws,ptarget_ws,pseq);

        element = next;
        next = next->next;
    }

    return prntd;
}

void acit_process_proc_tmf(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_SEXACT32           cid,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   prntd = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    if(cid < 0 &&
            cid != DMP_ALL_CID){
        seq_printf(pseq,"cid : %d is not valid\n",cid);
        return;
    }

    if(cid != DMP_ALL_CID)
        if(check_cid_valid(pws_ctrl,cid,pseq))
            return ;

    prntd = dump_queued_tmf(pws_ctrl,cid,pseq);

    seq_printf(pseq,"Found a total of %d ACIT internal Aborts waiting to be submitted to hardware\n\n", prntd);

    seq_printf(pseq,"Checking for TMFs issued by OSM...\n");
    prntd = dump_osm_issued_tmf(pws_ctrl,cid,pseq);

    seq_printf(pseq,"Found a total of %d OSM issued TMFs waiting to be submitted to hardware\n", prntd);

}


/***** TMF End ******/

/***** Interrupt Start  ******/

static void
dump_intr(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32   qidx;

    PACIT_CONTROLLER_INT_STATS      pintr_stats;
    PACIT_CONTROLLER_DIC_COUNTERS   pdic_stats;


    for(qidx = 0 ; qidx < pws_ctrl->params.num_cq ; qidx++){
        
        pintr_stats = &pws_ctrl->partition[qidx].interrupt_stats;
        pdic_stats  = &pws_ctrl->partition[qidx].dic;

        seq_printf(pseq,"ValidInterrupts      : %8d EQMajorInterrrupt    : %8d ISCSICQInterrupts  :%8d\n",
                GET_DATA(ACIT_CONTROLLER_DIC_COUNTERS,pdic_stats,interrupts),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,eq_maj_ints),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,cq_ints));

        seq_printf(pseq,"DefaultPDUInterrups  : %8d DrvrMsgInterrupts    : %8d\n",
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,dft_pdu_ints),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,drvr_msg_ints));

        seq_printf(pseq,"ISolCqeInterrupts    : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,iscsi_cmd_ints));

        seq_printf(pseq,"\n");

        seq_printf(pseq,"UnknownInterrupts    : %8d DelayedEQDmas        : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,unknown_ints),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,delayed_eq_dma));

        seq_printf(pseq,"UnknownEQEntries     : %8d UnknownCQEntries     : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,unknown_eqes),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,unknown_cqes));

        seq_printf(pseq,"UnknownISolCQE       : %8d UnknownCQErrors      : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,unknown_cmds),
                GET_DATA(ACIT_CONTROLLER_INT_STATS,pintr_stats,unknown_cmd_errors));

        seq_printf(pseq,"\n");

        seq_printf(pseq,"DicEnabled           : %8d DicInterruptGoal     : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_DIC_COUNTERS,pdic_stats,enabled),
                GET_DATA(ACIT_CONTROLLER_DIC_COUNTERS,pdic_stats,interrupt_rate));

        seq_printf(pseq,"MaxInterruptsPerSec  : %8d EQTimerValue         : %8d\n", 
                GET_DATA(ACIT_CONTROLLER_DIC_COUNTERS,pdic_stats,max_interrupts_per_sec),
                GET_DATA(ACIT_CONTROLLER_DIC_COUNTERS,pdic_stats,eq_timer));

        seq_printf(pseq,"\n");


    }
}


void
acit_process_proc_intr(
        PACIT_CONTROLLER_KEY    pcontroller,
        struct seq_file         *pseq)
{

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    dump_intr(pws_ctrl,pseq);

    return; 
}
/***** Interrupt End ******/


/***** ETO Find  Start  ******/

static void
target_dump_err_stats(
        PACIT_TARGET_WS ptarget_ws,
        struct seq_file     *pseq)
{

    seq_printf(pseq,"\n   Dumping EH_Stats:\n");

    seq_printf(pseq,"DigestErr  : 0x%8x TimeoutErr : 0x%8x FormatErr  : 0x%8x\n", 
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.ssn_digest_errors),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.ssn_cxn_timeout_errors),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.ssn_format_errors));

    seq_printf(pseq,"TMFTgtRest : 0x%8x Recoveries : 0x%8x CmdWndTO   : 0x%8x\n", 
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.tmf_target_resets),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.recoveries),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.ping_wait_timeouts));

    seq_printf(pseq,"TMFTO      : 0x%8x TMFIntTO   : 0x%8x TMFErrs    : 0x%8x\n",
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.tmf_timeouts),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.tmf_internal_timeouts),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.tmf_errors));

    seq_printf(pseq,"IORetryXce : 0x%8x RejectPDU  : 0x%8x AsyncLogout: 0x%8x\n", 
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.ioretry_exceeded),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.rejects),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.async_logouts));


    seq_printf(pseq,"AsyncReopen: 0x%8x AsyncRenego: 0x%8x IOStsErrs  : 0x%8x\n", 
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.async_reopens),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.async_param_renego),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.io_status_errors));

    seq_printf(pseq,"ReopenIOCTL: 0x%8x ReopenSucc : 0x%8x ReopenFail : 0x%8x\n", 
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.reopen_issued),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.reopen_success),
            GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.reopen_error));

    seq_printf(pseq,"ARMFWAsyncReopen : 0x%8x Offloads   : 0x%8x UnknownDPDUs : 0x%8x\n",
        GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.armfw_async_reopens),
        GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.offloads), 
        GET_DATA(ACIT_TARGET_WS,ptarget_ws,eh_stats.unknown_defpdu));

}

static ACIT_UEXACT32 
dump_etofind(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   cri,cid;
    ACIT_UEXACT32   eto_count = 0;

    PACIT_TARGET_WS ptarget_ws;

    seq_printf(pseq,"\n");
    seq_printf(pseq,"Dumping controller error statistics:\n");

    seq_printf(pseq,"ETO Expiry           : %8d TargetRecoveries     : %8d\n",
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.eto_expiry),
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.target_recoveries));

    seq_printf(pseq,"TMFTimeouts          : %8d TMFInternalTimeouts  : %8d\n",
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.tmf_timeouts), 
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.tmf_internal_timeouts));

    seq_printf(pseq,"LogoutsNormal        : %8d LogoutOthers         : %8d\n",
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.ssn_logout_normals),
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.ssn_logout_others));

    seq_printf(pseq,"DigestErrors         : %8d TimeoutErrors        : %8d\n", 
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.ssn_digest_errors),
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.ssn_cxn_timeout_errors));

    seq_printf(pseq,"OffloadFailures      : %8d CmdWindowTimeouts    : %8d\n", 
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.offload_failures),
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.ping_wait_timeouts));

    seq_printf(pseq,"Cxn Invalidations    : %8d Cxn Uploads          : %8d\n",
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.cxn_invalidations),
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.cxn_uploads));

    seq_printf(pseq,"CommandTimeouts      : %8d \n", 
            GET_DATA(ACIT_CONTROLLER_WS,pws_ctrl,eh_stats.io_timeouts));


    for(cri = 0;cri < BE_MAX_CXNS;cri++){
        
        cid = CID_FROM_CRI(pws_ctrl,cri);
    
        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);

        if(!ptarget_ws)
            continue;

        if(ptarget_ws->busy & TARGET_BUSY_ETO_EXPIRED){
            
            seq_printf(pseq,"\n");
            seq_printf(pseq,"Target 0x%p has expired ETO\n", ptarget_ws);


            target_dump(pws_ctrl,cid,pseq);
            target_dump_err_stats(ptarget_ws,pseq);

            eto_count++;
        }
    }

    seq_printf(pseq,"\n");
    seq_printf(pseq,"Completed dumping of %d targets that currently have expired ETO\n", eto_count);

    return eto_count;
}

void
acit_process_proc_etofind(
        PACIT_CONTROLLER_KEY    pcontroller,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   r_val = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    r_val = dump_etofind(pws_ctrl,pseq);

    return; 
}

/***** ETO FInd End ******/

/***** CQ Start  ******/

static ACIT_UEXACT32 
dump_cqids(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       num_cq = 0,qidx;
    ACIT_UEXACT32       cq_dumped = 0;

    PSA_RING            pcq_ring;

    PHWI_CONTROLLER_WS  phwi_ctrl;

    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    if(!phwi_ctrl)
        return cq_dumped;

    num_cq = pws_ctrl->params.num_cq;

    for(qidx = 0 ; qidx < pws_ctrl->params.num_cq ; qidx++){

        pcq_ring = &phwi_ctrl->partition[qidx].cq;

        seq_printf(pseq,"ID : %d ConsumerID : %d(0x%x) NumPages : %d NumEnteries : %d VA:0x%p - 0x%p\n",
                pcq_ring->id,
                pcq_ring->cidx,
                pcq_ring->cidx,
                pcq_ring->pages,
                pcq_ring->num, 
                pcq_ring->va,
                (pcq_ring->va + (pcq_ring->pages * PAGE_SIZE) - 1));

        cq_dumped++;
    }

    return cq_dumped;
}

static ACIT_UEXACT32
dump_cq(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       q_id,
        ACIT_SEXACT32       count,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       prntd = 0,i;
    ACIT_UEXACT32       index = 0;
    ACIT_UEXACT32       num_cq = 0,qidx;
    ACIT_UEXACT32       cq_dumped = 0;

    PSA_RING            pcq_ring;
    PISCSI_CQE          pcqe;
        
    PHWI_CONTROLLER_WS  phwi_ctrl;
    
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    if(!phwi_ctrl)
        return prntd;

    num_cq = pws_ctrl->params.num_cq;

    for(qidx = 0 ; qidx < pws_ctrl->params.num_cq ; qidx++){
    
        pcq_ring = &phwi_ctrl->partition[qidx].cq;

        if(q_id != -1){
            if(q_id != pcq_ring->id)
                continue;
        }

        seq_printf(pseq,"ID : %d ConsumerID : %d(0x%x) NumPages : %d NumEnteries : %d VA:0x%p - 0x%p\n",
                pcq_ring->id,
                pcq_ring->cidx,
                pcq_ring->cidx,
                pcq_ring->pages,
                pcq_ring->num, 
                pcq_ring->va,
                (pcq_ring->va + (pcq_ring->pages * PAGE_SIZE) - 1));

        cq_dumped++;

        if(count >= 0){
            index = pcq_ring->cidx;
            seq_printf(pseq,"Dumping next %d entries from cidx %d (0x%x)\n", count, index, index);
        }
        else {

            index = pcq_ring->cidx - abs(count);
            
            /* Check if the Enteries to dump goes backwards 
             * i.e beyond 1024. If cidx is at 3 and count to
             * dump is -10
             */
            if((index > pcq_ring->num)){
                index = pcq_ring->num - abs(count);

                if(index == pcq_ring->num)
                    index = 0;
            }
            seq_printf(pseq,"Dumping last %ld entries from cidx %d (0x%x)\n", abs(count), index, index);
        }

        for(i = 0 ; i < abs(count) ; i++){

            pcqe = ( PISCSI_CQE)((ACIT_UEXACT8 *)pcq_ring->va + (index * pcq_ring->itemSize));
            
            seq_printf(pseq,"( %d) %p : DW[0]: 0x%x DW[1]: 0x%x DW[2]: 0x%x DW[3]: 0x%x Code: %d CID : %d Valid : %d \n",
                    index,
                    pcqe,
                    pcqe->dw[0],
                    pcqe->dw[1],
                    pcqe->dw[2],
                    pcqe->dw[3],
                    pcqe->code,
                    pcqe->cid,
                    pcqe->valid);

            if(++index == pcq_ring->num)
                index = 0;

            prntd++;
        }
    }

    if(!cq_dumped)
        seq_printf(pseq,"No CQ was found with the matching ID, run cqfind to list all CQ IDs\n");

    return prntd;
}

void
acit_process_proc_cqfind(
        PACIT_CONTROLLER_KEY    pcontroller,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   r_val = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    r_val = dump_cqids(pws_ctrl,pseq);
    seq_printf(pseq,"Completed dump of %d CQ IDS\n", r_val);
    
    return; 
}

void 
acit_process_proc_cq(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_UEXACT32           q_id,
        ACIT_SEXACT32           count,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   r_val = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    if(!count) count = 8;

    r_val = dump_cq(pws_ctrl,q_id,count,pseq);
    seq_printf(pseq,"Completed dump of %d CQEs\n", r_val);
    
    return; 
}



/***** CQ End ******/

/***** EQ Start ******/
static ACIT_UEXACT32
dump_eq(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       eq_id,
        ACIT_SEXACT32       count,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       prntd = 0,i;
    ACIT_UEXACT32       index = 0;
    ACIT_UEXACT32       num_eq = 0,qidx;
    ACIT_UEXACT32       eq_dumped = 0;

    PSA_RING            peq_ring;
    EQ_ENTRY            *peqe;
        
    PHWI_CONTROLLER_WS  phwi_ctrl;
    
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    if(!phwi_ctrl)
        return prntd;

    num_eq = pws_ctrl->params.num_cq;

    for(qidx = 0 ; qidx < pws_ctrl->params.num_cq ; qidx++){
    
        peq_ring = &phwi_ctrl->partition[qidx].eq;

        if(eq_id != -1){
            if(eq_id != peq_ring->id)
                continue;
        }

        seq_printf(pseq,"ID : %d ConsumerID : %d(0x%x) NumPages : %d NumEnteries : %d VA:0x%p - 0x%p\n",
                peq_ring->id,
                peq_ring->cidx,
                peq_ring->cidx,
                peq_ring->pages,
                peq_ring->num, 
                peq_ring->va,
                (peq_ring->va + (peq_ring->pages * PAGE_SIZE) - 1));

        eq_dumped++;

        if(count >= 0){
            index = peq_ring->cidx;
            seq_printf(pseq,"Dumping next %d entries from cidx %d (0x%x)\n", count, index, index);
        }
        else {

            index = peq_ring->cidx - abs(count);
            
            /* Check if the Enteries to dump goes backwards 
             * i.e beyond 1024. If cidx is at 3 and count to
             * dump is -10
             */
            if((index > peq_ring->num)){
                index = peq_ring->num - abs(count);

                if(index == peq_ring->num)
                    index = 0;
            }
            /* seq_printf(pseq,"Dumping last %ld entries from cidx %d (0x%x)\n", abs(count), index, index);*/
             seq_printf(pseq,"Dumping last %ld entries from cidx %d (0x%x)\n", abs(count),peq_ring->num,peq_ring->num);
        }

        for(i = 0 ; i < abs(count) ; i++){

            peqe = (EQ_ENTRY *)((ACIT_UEXACT8 *)peq_ring->va + (index * peq_ring->itemSize));
            
            seq_printf(pseq,"( %d) %p : MajorCode : 0x%x MinorCode : 0x%x ResourceID : 0x%x Valid : %d DWORD : 0x%08x\n",
                    index,
                    peqe,
                    peqe->MajorCode,
                    peqe->MinorCode,
                    peqe->ResourceID,
                    peqe->Valid,
                    peqe->dw);

            if(++index == peq_ring->num)
                index = 0;

            prntd++;
        }
    }

    if(!eq_dumped)
        seq_printf(pseq,"No EQ was found with the matching ID, run eqfind to list all EQ IDs\n");

    return prntd;
}

static ACIT_UEXACT32 
dump_eqids(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       num_eq = 0,qidx;
    ACIT_UEXACT32       eq_dumped = 0;

    PSA_RING            peq_ring;
        
    PHWI_CONTROLLER_WS  phwi_ctrl;
    
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    if(!phwi_ctrl)
        return eq_dumped;

    num_eq = pws_ctrl->params.num_cq;

    for(qidx = 0 ; qidx < pws_ctrl->params.num_cq ; qidx++){
    
        peq_ring = &phwi_ctrl->partition[qidx].eq;

        seq_printf(pseq,"ID : %d ConsumerID : %d(0x%x) NumPages : %d NumEnteries : %d VA:0x%p - 0x%p\n",
                peq_ring->id,
                peq_ring->cidx,
                peq_ring->cidx,
                peq_ring->pages,
                peq_ring->num, 
                peq_ring->va,
                (peq_ring->va + (peq_ring->pages * PAGE_SIZE) - 1));

        eq_dumped++;
    }

    return eq_dumped;
}

void
acit_process_proc_eqfind(
        PACIT_CONTROLLER_KEY    pcontroller,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   r_val = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    r_val = dump_eqids(pws_ctrl,pseq);
    seq_printf(pseq,"Completed dump of %d EQ IDS\n", r_val);
    
    return; 
}
    
void 
acit_process_proc_eq(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_UEXACT32           eq_id,
        ACIT_SEXACT32           count,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   r_val = 0;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    if(!count) count = 8;

    r_val = dump_eq(pws_ctrl,eq_id,count,pseq);
    seq_printf(pseq,"Completed dump of %d EQEs\n", r_val);
    
    return; 
}


/***** EQ End ******/


/***** WRB Start ******/

static void 
get_wrb_type(int type,char *type_str)
{
    switch(type) {

        case HWH_TYPE_IO:
            strcpy(type_str, "HWH_TYPE_IO");
            break;

        case HWH_TYPE_LOGOUT:
            strcpy(type_str, "HWH_TYPE_LOGOUT");
            break;

        case HWH_TYPE_TMF:
            strcpy(type_str, "HWH_TYPE_TMF");
            break;

        case HWH_TYPE_NOP:
            strcpy(type_str, "HWH_TYPE_NOP");
            break;
        case HWH_TYPE_INVALID:
            strcpy(type_str, "HWH_TYPE_INVALID");
            break;
    }
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("type : %d get_wrb_type : %s\n",
            type,type_str));
}

static void 
dump_wrb_info(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       wrb_idx,
        ACIT_UEXACT32       cid,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       cri;
    ACIT_SEXACT8        type_str[20] = {0};

    PHWI_CONTROLLER_WS  phwi_ctrl;
    PHWI_WRB_CONTEXT    pwrb_context;
    PWRB_HANDLE         pwrb_handle_base;
    PWRB_HANDLE         pwrb_handle;

    seq_printf(pseq,"Dumping info about wrb %d on cid %d \n", wrb_idx,cid);

    cri = CRI_FROM_CID(pws_ctrl,cid);
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    pwrb_context = &phwi_ctrl->wrb_context[cri];
    pwrb_handle_base = pwrb_context->pwrb_handle_base;
    pwrb_handle = (PWRB_HANDLE)pwrb_handle_base + wrb_idx;

    get_wrb_type(pwrb_handle->type,&type_str[0]);

    seq_printf(pseq,"WrbIndex:%d Type:%s WrbHandle:0x%p (_WRB_HANDLE)\n",
            pwrb_handle->wrb_index, 
            type_str,
            pwrb_handle);

    if(pwrb_handle->pio_handle){

        seq_printf(pseq,"RequestHandle:0x%p Wrb:0x%p (ISCSI_WRB)\n",
                pwrb_handle->pio_handle,
                pwrb_handle->pwrb);

        switch(pwrb_handle->type){
            case HWH_TYPE_IO :
            case HWH_TYPE_TMF:
                dump_pio_handle(pws_ctrl,
                        pwrb_handle->pio_handle,pseq);
                break;
        case HWH_TYPE_LOGOUT:
        case HWH_TYPE_NOP:
            break;
        default:
            seq_printf(pseq,"Error!! Invalid type %d\n", 
                    pwrb_handle->type);
        }
    }
    else{
        seq_printf(pseq,"WrbIndex %d is not currently active\n", wrb_idx);
        seq_printf(pseq,"Dumping Wrb 0x%p (_ISCSI_WRB)\n", pwrb_handle->pwrb);
        dump_pwrb(pwrb_handle->pwrb,pseq);
    }

    return;
}

void 
acit_process_proc_wrb(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_UEXACT32           wrb_idx,
        ACIT_UEXACT32           cid,
        struct seq_file         *pseq)
{
    PACIT_CONTROLLER_WS pws_ctrl;
    PACIT_TARGET_WS ptarget_ws;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    if(check_cid_valid(pws_ctrl,cid,pseq))
        return;
    
    /* Check if the CID is present or not */
    ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);

    if(!ptarget_ws){
        seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
        seq_printf(pseq,"Use (cid 0 0xa) to get the list of current cids on this controller\n");
        return;
    }

    /* Check if the wrb is valid*/
    if(wrb_idx > pws_ctrl->params.num_cmd_per_cxn){
        seq_printf(pseq,"WrbIndex %d is not present for CID %d\n",wrb_idx, cid);
        return;
    }

    dump_wrb_info(pws_ctrl,wrb_idx,cid,pseq);

}
/***** WRBEnd ******/

/***** ICD Start ******/
#define MAX_ISCSI_FRAGS 32




ACIT_UEXACT32
check_list_rev(
        PACIT_CONTROLLER_WS pws_ctrl,
        void *              phndl_mem,
        PACIT_DC_LIST       plist,
        ACIT_HLIST_TYPE     hlist_type,
        ACIT_UEXACT32       pioh_mem_check,
        ACIT_UEXACT32       num_entry,
        struct seq_file     *pseq
        )
{
    ACIT_UEXACT32       found = 0,count = 0;
    ACIT_SEXACT32       cri = 0;

    ACIT_DC_LIST        *phndl_head,*phndl_prev;
    PACIT_IO_HANDLE     pio_handle;
    PACIT_LOGOUT_HANDLE plogout_handle;
    PACIT_TMF_HANDLE    ptmf_handle;
    PSGL_HANDLE         psgl_hndl = ACIT_NULL;


    phndl_head = plist->prev;
    phndl_prev = phndl_head->prev;


    if(pioh_mem_check == PIOH_CRI_MEM)
        cri = *((ACIT_UEXACT32 *)phndl_mem);
    else
        psgl_hndl = (PSGL_HANDLE)phndl_mem;

    while(phndl_prev != phndl_head){

        if(hlist_type == ACIT_IO_HLIST){
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                    controller_link,phndl_prev);

            if(pio_handle){

                /* Check if SGL Handle Matches */
                if(pioh_mem_check == PIOH_SGL_MEM){
                    if(pio_handle->psgl_handle == psgl_hndl){
                        dump_pio_handle(pws_ctrl,pio_handle,pseq);
                        found++;
                        break;
                    }
                }
                else if(pioh_mem_check == PIOH_CRI_MEM){
                    if(pio_handle->cri == cri || cri == -1){
                        dump_pio_handle(pws_ctrl,pio_handle,pseq);
                        found++;
                        goto CONTINUE;
                    }
                }
            }
        }/* End of if ACIT_IO_HLIST */

        if(hlist_type == ACIT_LOGOUT_HLIST){
            plogout_handle = STRUCT_BASE(ACIT_LOGOUT_HANDLE,
                    controller_link,phndl_prev);

            if(plogout_handle){

                /* Check if SGL Handle Matches */
                if(plogout_handle->psgl_handle == psgl_hndl){
                    dump_plogout_handle(pws_ctrl,plogout_handle,pseq);
                    found++;
                    break;
                }
            }
        }/* End of ACIT_LOGOUT_HLIST*/

        if(hlist_type == ACIT_TMF_HLIST){
            ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                    controller_link,phndl_prev);

            if(ptmf_handle){

                /* Check if SGL Handle Matches */
                if(ptmf_handle->psgl_handle == psgl_hndl){
                    dump_ptmf_handle(pws_ctrl,ptmf_handle,pseq);
                    found++;
                    break;
                }
            }
        } /* End of ACIT_TMF_HLIST */
CONTINUE:
        if(found >= num_entry)
            break;

        if(++count > 1024){
            seq_printf(pseq,"List might be corrupted, breaking off the loop\n");
            break;
        }
        phndl_prev = phndl_prev->prev;

    }

    return found;
    
}

void find_icd_list(
        PACIT_CONTROLLER_WS pws_ctrl,
        PSGL_HANDLE         psgl_hndl,
        struct seq_file     *pseq)
{
    PACIT_DC_LIST   icd_list;
    ACIT_UEXACT32   icd_prsnt = 0;

    /* Check the IOH Busy List */
    icd_list = pws_ctrl->io_handle_busy_list.next;

    if(icd_list){
        icd_prsnt = check_list_rev(pws_ctrl,
                        (void *)psgl_hndl,
                        icd_list,
                        ACIT_IO_HLIST,
                        PIOH_SGL_MEM,
                        1024,
                        pseq);

        if(icd_prsnt){
            
            seq_printf(pseq,"%p psgl_handle is present in the IOH Busy List\n",psgl_hndl);
            seq_printf(pseq,"%p therefore is currently active with an IO Request\n",psgl_hndl);

            return;
        }
    }

    /* Check the IOH Free List */
    icd_list = pws_ctrl->io_handle_free_list.next;
    if(icd_list){
        icd_prsnt = check_list_rev(pws_ctrl,
                        (void *)psgl_hndl,
                        icd_list,
                        ACIT_IO_HLIST,
                        PIOH_SGL_MEM,
                        1024,
                        pseq);

        if(icd_prsnt){
            
            seq_printf(pseq,"%p psgl_handle is present in the IOH Free List\n",psgl_hndl);
            seq_printf(pseq,"%p therefore is currently NOT active\n",psgl_hndl);

            return;
        }
    }

    /* Check the TMF Busy List */
    icd_list = pws_ctrl->tmf_handle_busy_list.next;
    if(icd_list){
        icd_prsnt = check_list_rev(pws_ctrl,
                        (void *)psgl_hndl,
                        icd_list,
                        ACIT_TMF_HLIST,
                        0,
                        1024,
                        pseq);

        if(icd_prsnt){
            
            seq_printf(pseq,"%p psgl_handle is present in the TMF Busy List\n",psgl_hndl);
            seq_printf(pseq,"%p therefore is currently active with TMF Request\n",psgl_hndl);

            return;
        }
    }

    /* Check the Logout Busy List */
    icd_list = pws_ctrl->logout_handle_busy_list.next;
    if(icd_list){
        icd_prsnt = check_list_rev(pws_ctrl,
                        (void *)psgl_hndl,
                        icd_list,
                        ACIT_LOGOUT_HLIST,
                        0,
                        1024,
                        pseq);

        if(icd_prsnt){
            
            seq_printf(pseq,"%p psgl_handle is present in the Logout Busy List\n",psgl_hndl);
            seq_printf(pseq,"%p therefore is currently active with Logout Request\n",psgl_hndl);

            return;
        }
    }
}


void dump_icd_ctx(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       icd_idx,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32       frag_idx;
    ACIT_UEXACT32       icd_normal;
    
    PISCSI_SGE          pfrag;
    PHWI_CONTROLLER_WS  phwi_ctrl;
    PSGL_HANDLE         psgl_base = ACIT_NULL;
    PSGL_HANDLE         psgl_hndl = ACIT_NULL;
    
    
    icd_normal = (icd_idx - pws_ctrl->fw_config.iscsi_icd_start);
    
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);
    psgl_base = phwi_ctrl->psgl_handle_base;
    psgl_hndl = (psgl_base + icd_normal);

    if(psgl_hndl->sgl_index != icd_idx){
        seq_printf(pseq,"ERROR !!! ICD %d did not match with expected %d\n",
                psgl_hndl->sgl_index,icd_idx);
        return ;
    }

    pfrag = psgl_hndl->pfrag;

    for(frag_idx = 0 ; frag_idx < MAX_ISCSI_FRAGS ; frag_idx++){
        seq_printf(pseq,"PAHi : 0x%08x  PALo : 0x%08x   Offset : 0x%08x  Len : 0x%08x   Last SGE : %d\n",
                pfrag->addr_hi,pfrag->addr_lo,pfrag->sge_offset,pfrag->len,pfrag->last_sge);

        if(pfrag->last_sge)
            break;
        else
            pfrag++;
    }

    if(!pfrag->last_sge)
        seq_printf(pseq,"NOTE !!! Last SGE Not Found\n");
    
    /* Find the list to which ICD is assigned */
    find_icd_list(pws_ctrl,psgl_hndl,pseq);
}



static void dump_icd_history(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_SEXACT32       cid,
        ACIT_SEXACT32       count,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   ioh_hndl_avail = 0;
    ACIT_SEXACT32   prnt = 0,cri = -1;
    PACIT_TARGET_WS ptarget_ws;

    PACIT_DC_LIST   ioh_list;

    ioh_hndl_avail = pws_ctrl->counters.io_handles_available;

    if(cid == DMP_ALL_CID){
        if(count == DMP_ALL_ICD)
            seq_printf(pseq,"Checking all completed ICDs\n");
        else
            seq_printf(pseq,"Dumping last %ld completed ICDs\n",abs(count));
    }
    else {
        /* Check if the CID is present or not */
        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);
        
        if(!ptarget_ws){
            seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
            seq_printf(pseq,"Use (cid 0 0xa) to get the list of current cids on this controller\n");
            return;
        }

        if(count == DMP_ALL_ICD)
            seq_printf(pseq,"Checking all completed ICDs on CID %d \n",cid);
        else
            seq_printf(pseq,"Dumping last %ld completed ICDs on CID %d \n",abs(count), cid);


        cri = CRI_FROM_CID(pws_ctrl,cid);
    }
    /* If user has not entered the number of entries to 
     * dump. Dump all the entries  
     */
    if(count == DMP_ALL_ICD)
        count = ioh_hndl_avail;
    else /* if user has entered the value as -ive */
        count = abs(count);
    
    ioh_list = pws_ctrl->io_handle_free_list.next;

    if(!ioh_list) {
        seq_printf(pseq," IO Free Handle List is NULL\n");
        return;
    }


    prnt = check_list_rev(pws_ctrl,
            (void *)&cri,
            ioh_list,
            ACIT_IO_HLIST,
            PIOH_CRI_MEM,
            count,
            pseq);
    
    seq_printf(pseq,"Completed dump of %d ICDs\n",prnt);
    return;
}

static void dump_icd_active(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_SEXACT32       cid,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32   list_type = 0;
    ACIT_UEXACT32   count = 0,start_count = 0;
    PACIT_DC_LIST   ioh_list;
    PACIT_TARGET_WS ptarget_ws;

    if(cid != DMP_ALL_CID){
        ptarget_ws = ptgt_from_cid(pws_ctrl,cid,pseq);
        
        if(!ptarget_ws){
            seq_printf(pseq,"Error: Target could not be determined for the specified cid 0x%x\n", cid);
            seq_printf(pseq,"Use (cid 0 0xa) to get the list of current cids on this controller\n");
            return;
        }

        /* Get the target busy list */
        ioh_list = ptarget_ws->io_handle_busy_list.next;
        seq_printf(pseq,"Dumping pending ICDs on CID %d\n",cid);
        seq_printf(pseq,"ptarget_pending IOS : %d\n",
                ptarget_ws->counters.pending_ios);

        list_type = ACIT_TGT_HLIST;
    }
    else {
        seq_printf(pseq,"Dumping pending ICDs\n");

        /* Get the busy list of the controller */
        ioh_list = pws_ctrl->io_handle_busy_list.next;

        list_type = ACIT_IO_HLIST;
    }

    if(ioh_list){
        count = dump_list(pws_ctrl,
                ioh_list,
                start_count,
                1024,
                list_type,  
                pseq);  
    }
    seq_printf(pseq,"Number of ICDs pending is %d\n",count);
}

void 
acit_process_proc_icdfind(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_SEXACT32           cid,
        ACIT_UEXACT32           options,
        struct seq_file         *pseq)
{
    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);
    
    if(cid < 0 &&
            cid != DMP_ALL_CID){
        seq_printf(pseq,"cid : %d is not valid\n",cid);
        return;
    }
    
    if(cid != DMP_ALL_CID)
        if(check_cid_valid(pws_ctrl,cid,pseq))
            return ;

    if(options)
        dump_icd_history(pws_ctrl,cid,options,pseq);
    else
        dump_icd_active(pws_ctrl,cid,pseq);
    return; 
}

void 
acit_process_proc_icd(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_UEXACT32           icd_idx,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32   icd_start,icd_count;

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    /* Check if the ICD is valid */
    icd_start = pws_ctrl->fw_config.iscsi_icd_start;
    icd_count = pws_ctrl->fw_config.iscsi_icd_count;

    if(icd_idx < icd_start || 
            icd_idx >= (icd_start + icd_count)){

        seq_printf(pseq,"ICD : %d Not Valid\n",icd_idx);
        return;
    }

    dump_icd_ctx(pws_ctrl,icd_idx,pseq);
    return; 
}

/***** ICD End ******/



static void 
cxn_dump(
        PACIT_CONTROLLER_WS     pws_ctrl,
        ACIT_UEXACT32           cri,
        struct seq_file         *pseq)
{
    ACIT_UEXACT32       cid;
    PISCSI_CONNECTION   pcxn;

    pcxn = GET_ISCSI_CXN(pws_ctrl,cri);

    if(!pcxn){
        seq_printf(pseq,"\n\n\t\tThe CXN is NULL\n");
        return; 
    }

    
    cid = CID_FROM_CRI(pws_ctrl,cri);
    if(pcxn->cid != cid){
        seq_printf(pseq,"\n\nERROR!!! CID mismatch actual: %d pcxn : %d cri : %d\n",
                cid,pcxn->cid,cri);
        return;
    }


    seq_printf(pseq,"\n\n\t\tDumping Connection Details  \n");
    seq_printf(pseq,"     CmdPdus : %08d   ResponsePdus : %08d\n",
            pcxn->iscsi_cmd_pdus,pcxn->iscsi_rsp_pdus);
    
    seq_printf(pseq,"Pending Cmds : %08d Pending Writes : %08d\n",
            pcxn->num_of_active_cmds,pcxn->num_of_active_writes);

    seq_printf(pseq,"  Digest Err : %08d    Timeout Err : %08d Format Err: %08d\n",
            pcxn->digest_errors,pcxn->timeout_errors,pcxn->format_errors);

    return ;
}

static void 
wrb_ctx_dump(
        PHWI_CONTROLLER_WS  phwi_ctrl,
        ACIT_UEXACT32       cri,
        struct seq_file     *pseq)
{
    PISCSI_WRB plast_wrb = NULL;
    PHWI_WRB_CONTEXT pwrb_ctxt;

    pwrb_ctxt = &phwi_ctrl->wrb_context[cri];
    plast_wrb = pwrb_ctxt->plast_wrb;

    seq_printf(pseq,"\n\t\t Dumping WRB Context  %p\n",plast_wrb);
    seq_printf(pseq,"WRB Handles Available : %d\n",
            pwrb_ctxt->wrb_handles_available);
    
    if(!plast_wrb){
            seq_printf(pseq,"No plast wrb for this CID\n");
        return ;
    }

    dump_pwrb(plast_wrb,pseq);
    return ;
}

static void 
dump_cid_info(
        PACIT_CONTROLLER_WS pws_ctrl,
        ACIT_UEXACT32       cid,
        struct seq_file     *pseq)
{
    ACIT_UEXACT32 cri;
    PHWI_CONTROLLER_WS  phwi_ctrl;
    
    phwi_ctrl = GET_HWI_CONTROLLER_WS(pws_ctrl);

    /* Check for the CID validity */
    if (check_cid_valid(pws_ctrl,cid,pseq))
        return;

    /* Get the CRI for the CID */
    cri = CRI_FROM_CID(pws_ctrl,cid);

    seq_printf(pseq,"\n\t\tDumping CID: %d   CRI: %d\n",
            cid,cri);

    /* Dump the WRB Context */
    wrb_ctx_dump(phwi_ctrl,cri,pseq);

    /* Dump the CXN Info */
    cxn_dump(pws_ctrl,cri,pseq);

    /* Dump the target details */
    target_dump(pws_ctrl,cid,pseq);
    
    seq_printf(pseq,"***************************************\n");
    
    return;

}

static void
dump_cid_valid(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       cid = 0 ,cri,count = 0;

    for(cri = 0;cri < BE_MAX_CXNS;cri++){
        
        cid = CID_FROM_CRI(pws_ctrl,cri);
        
        if(ptgt_from_cid(pws_ctrl,cid,pseq)){
            dump_cid_info(pws_ctrl,cid,pseq);
            count++;
        }
    }

    seq_printf(pseq,"\n\t\tDumped %d Current CID\n",count);
}


static void
dump_cid_all(
        PACIT_CONTROLLER_WS pws_ctrl,
        struct seq_file     *pseq)
{

    ACIT_UEXACT32       cid = 0, cri;

    for(cri = 0;cri < BE_MAX_CXNS;cri++){
        
        cid = CID_FROM_CRI(pws_ctrl,cri);
        dump_cid_info(pws_ctrl,cid,pseq);
    }
    seq_printf(pseq,"\n\t\tDumped %d CID\n",cid);
}



void 
acit_process_proc_cid(
        PACIT_CONTROLLER_KEY    pcontroller,
        ACIT_UEXACT32           cid,
        ACIT_UEXACT32           flag,
        struct seq_file         *pseq)
{

    PACIT_CONTROLLER_WS pws_ctrl;

    pws_ctrl = GET_CONTROLLER_WS(pcontroller);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("In the CID : cid : %d flag : %d\n",
                cid,flag));
    

    switch(flag){
        
        case OPTION_ALL:
            dump_cid_all(pws_ctrl,pseq);
            break;
            
        case OPTION_VAL:
            dump_cid_valid(pws_ctrl,pseq);
            break;
            
        case OPTION_CUR:
            dump_cid_info(pws_ctrl,cid,pseq);
            break;
    }

    return;
}
