/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_rxf_h__
#define __be2_gen_rxf_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"
#include "be2_gen_gpio.h"
#include "be2_gen_mdio.h"
#include "be2_gen_misc.h"

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_Rxf_Table
//
//   Accesses an RXF table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// table_type         - The RXF table type to access. 
// start_offset       - The starting offset in the RXF table to access. 
// dword_count        - The number of dwords (4-bytes) to access. 
// p_data             - Pointer to destination or source data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_Rxf_Table (
   IN void*    function_context,
   IN u8       table_type,
   IN u16      start_offset,
   IN u32      dword_count,
   IN OUT u32* p_data,
   IN boolean  write_flag )
{
    u32 curr_dword;
    RXF_RTPR_CSR rxf_rtpr;
   
    // Set the table type and the starting index in the RXF table programming CSR
    rxf_rtpr.dw = 0;
    rxf_rtpr.indx = table_type;
    rxf_rtpr.offset = start_offset;
    BE_GEN_CSR_WRITE( function_context, rxf.rtpr, rxf_rtpr.dw );

    switch(write_flag)
    {
    case FALSE: // read
        for (curr_dword = 0; curr_dword < dword_count; curr_dword++) {
            *(p_data + curr_dword) = BE_GEN_CSR_READ( function_context, rxf.rtdr );
        }
        break;

    default:    // write
        for (curr_dword = 0; curr_dword < dword_count; curr_dword++) {
            BE_GEN_CSR_WRITE( function_context, rxf.rtdr, *(p_data + curr_dword) );
        }
        break;
    }
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_MAC_Address_Table
//
//   Accesses the MAC address table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - The index of the MAC address table entry to access. 
// p_data             - Pointer to source or destination MAC address table entry data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_MAC_Address_Table (
   IN void*   function_context,
   IN u16     index,
   IN u32*    p_data,
   IN boolean write_flag )
{
    BE2_Gen_Access_Rxf_Table(function_context, MAC_ADDRESS_TABLE_INDEX, index*sizeof(MAC_ADDRESS_TABLE_ENTRY)/4, sizeof(MAC_ADDRESS_TABLE_ENTRY)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_Multicast_Address_Table
//
//   Accesses the Multicast address table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - The index of the multicast hash address table entry to access. 
// p_data             - Pointer to source or destination Multicast address table entry
//                      data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_Multicast_Address_Table (
   IN void*   function_context,
   IN u16     index,
   IN u32*    p_data,
   IN boolean write_flag )
{
    BE2_Gen_Access_Rxf_Table(function_context, MULTICAST_TABLE_INDEX, index*sizeof(MULTICAST_TABLE_ENTRY)/4, sizeof(MULTICAST_TABLE_ENTRY)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_VLAN_Table
//
//   Accesses the VLAN table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - The index of the VLAN table entry to access. 
// p_data             - Pointer to source or destination VLAN table entry data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_VLAN_Table (
   IN void*   function_context,
   IN u16     index,
   IN u32*    p_data,
   IN boolean write_flag )
{
    BE2_Gen_Access_Rxf_Table(function_context, VLAN_TABLE_INDEX, index*sizeof(VLAN_TABLE_ENTRY)/4, sizeof(VLAN_TABLE_ENTRY)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_FCOE_Table
//
//   Accesses the Fibre Channel Over Etherent table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - The index of the FCOE table entry to access. 
// p_data             - Pointer to source or destination FCOE table entry data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_FCOE_Table (
   IN void*   function_context,
   IN u16     index,
   IN u32*    p_data,
   IN boolean write_flag )
{
    BE2_Gen_Access_Rxf_Table(function_context, FCOE_TABLE_INDEX, index*sizeof(FCOE_TABLE_ENTRY)/4, sizeof(FCOE_TABLE_ENTRY)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_DA_Table
//
//   Accesses the Destination Address table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - The index of the DA table entry to access. 
// p_data             - Pointer to source or destination DA table entry data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_DA_Table (
   IN void*   function_context,
   IN u16     index,
   IN u32*    p_data,
   IN boolean write_flag )
{
    BE2_Gen_Access_Rxf_Table(function_context, DA_TABLE_INDEX, index*sizeof(DA_TABLE_ENTRY)/4, sizeof(DA_TABLE_ENTRY)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_ACPI_Filter_Table
//
//   Accesses the ACPI Filter table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// offset             - The dword offset of the ACPI Filter table entry to access. 
// p_data             - Pointer to source or destination ACPI Filter table entry data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_ACPI_Filter_Table (
   IN void*   function_context,
   IN u16     offset,
   IN u32*    p_data,
   IN boolean write_flag )
{
    // FIXME TODO @@@ ACPI filter programming API probably needs to have some usage 
    // mode defined before this API is useful. For now, it will just assume reading/writing
    // a single dword in the table.
    BE2_Gen_Access_Rxf_Table(function_context, ACPI_TABLE_INDEX, offset, sizeof(u32)/4, p_data, write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_XGMAC_Table
//
//   Accesses the XGMAC table for read or write operation. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number indicating which register space to access. 
// offset             - The dword offset of the XGMAC register to access. 
// p_data             - Pointer to destination or source data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_XGMAC_Table (
   IN void*   function_context,
   IN u8      port,
   IN u16     offset,
   IN u32*    p_data,
   IN boolean write_flag )
{
    
    BE_GEN_ASSERT (port == 0 || port == 1);

    BE2_Gen_Access_Rxf_Table(function_context, 
                             (port == 1 ? MAC1_REGISTER_TABLE_INDEX : MAC0_REGISTER_TABLE_INDEX),
                             offset,
                             1,
                             p_data,
                             write_flag);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Post_MAC_Command
//
//   Issues an AMIIM command via the XGMAC register space. Useful to access the internal
//   PCS (in XAUI mode) or external PHY (in RGMII mode). 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number indicating which register space to access. 
// amiim_cmd          - The AMIIM command register value. Reference the A-XGMAC
//                      specification for details. 
// amiim_field        - The AMIIM field register value. Reference the A-XGMAC
//                      specification for details. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Post_MAC_Command (
   IN void* function_context,
   IN u8    port,
   IN u32   amiim_cmd,
   IN u32   amiim_field )
{
    AMIIM_INDICATOR_REG indicator;
    u32 xgmac_data;

    const u32 amiim_cmd_loc = BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_cmd) / 4;
    const u32 amiim_field_loc = BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_field) / 4;
    const u32 amiim_indicator_loc = BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_indicator) / 4;
    
    // Just to prove the offsets are working...
    BE_GEN_C_ASSERT (BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_cmd) / 4 == 16);

    xgmac_data = 0;
    BE2_Gen_Access_XGMAC_Table (function_context, port, amiim_cmd_loc, &xgmac_data, TRUE);
    BE2_Gen_Access_XGMAC_Table (function_context, port, amiim_field_loc, &amiim_field, TRUE);
    BE2_Gen_Access_XGMAC_Table (function_context, port, amiim_cmd_loc, &amiim_cmd, TRUE);

    // Wait for operation to be done
    indicator.miimbusy = 1;
    while (indicator.miimbusy != 0) {
         BE2_Gen_Access_XGMAC_Table (function_context, port, amiim_indicator_loc, &indicator.dw, FALSE);
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_Internal_PHY_Reg
//
//   Accesses a PHY register via the XGMAC. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number indicating which register space to access. 
// dev_addr           - The MDI0 device address value. 
// addr               - The MMD specific address. 
// p_data             - Pointer to the destination or source data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_Internal_PHY_Reg (
   IN void*   function_context,
   IN u8      port,
   IN u8      dev_addr,
   IN u16     addr,
   IN u16*    p_data,
   IN boolean write_flag )
{
    AMIIM_CMD_REG cmd;
    AMIIM_FIELD_REG field;

    // note: hstphyadx is the MDIO field PRTAD and is fixed at 0x01 for the internal PCS
    // note: hstregadx is the MDIO field DEVAD and is only valid for values 0x01 (PMA/PMD) or 0x03 (PCS) for the internal PCS

    // Post the address to access
    field.dw = 0; // clear reg
    field.hststfield  = 0;
    field.hstopfield  = 0;
    field.hstphyadx  = 1;
    field.hstregadx = dev_addr;
    field.hsttafield = 2;
    field.hstmiimwrdat = addr;
    
    cmd.dw = 0;
    cmd.hstldcmd = 1;
    cmd.hstmiimcmd = 5;

    BE2_Gen_Post_MAC_Command (function_context, port, cmd.dw, field.dw);

    if (!write_flag) {
        // now post the read command
        field.dw = 0; // clear reg
        field.hststfield = 0;
        field.hstopfield = 2;
        field.hstphyadx = 1;
        field.hstregadx = dev_addr;
        field.hsttafield = 2;
        
        cmd.dw = 0;
        cmd.hstldcmd = 1;
        cmd.hstmiimcmd = 5;

        BE2_Gen_Post_MAC_Command (function_context, port, cmd.dw, field.dw);

        // Read the result
        BE2_Gen_Access_XGMAC_Table (function_context, port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_field)/4, &field.dw, FALSE);

        // Assign to destination data pointer
        *p_data = field.hstmiimwrdat;
    } else {
        // now post the write command
        field.dw = 0; // clear reg
        field.hststfield = 0;
        field.hstopfield = 1;
        field.hstphyadx = 1;
        field.hstregadx = dev_addr;
        field.hsttafield = 2;
        field.hstmiimwrdat = *p_data;
        
        cmd.dw = 0;
        cmd.hstldcmd = 1;
        cmd.hstmiimcmd = 5;

        BE2_Gen_Post_MAC_Command (function_context, port, cmd.dw, field.dw);

        // Pop the result
        BE2_Gen_Access_XGMAC_Table (function_context, port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_field) / 4, &field.dw, FALSE); 
    }

    return;
}


//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_Internal_Gigabit_PHY_Reg
//
//   Accesses a PHY register via the XGMAC. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number indicating which register space to access. 
// addr               - The MMD specific address. 
// p_data             - Pointer to the destination or source data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_Internal_Gigabit_PHY_Reg (
   IN void*   function_context,
   IN u8      port,
   IN u16     addr,
   IN u16*    p_data,
   IN boolean write_flag )
{
    AMIIM_CMD_REG cmd;
    AMIIM_FIELD_REG field;
    
    cmd.dw = 0;
    field.dw = 0; // clear reg
    // Write to the MDIO field register amii_field
    // Host ST field (ignored during access to legacy PHY (10/100/1000)
    field.hststfield = 0;
    // Host OP field (ignored during access to legacy PHY (10/100/1000)
    field.hstopfield = 0;
    // Host PHY device address field (Clause 22 PHY address)
    // Port-0: 0x10, Port-1: 0x11 
    field.hstphyadx = (port ? 0x11 : 0x10);
    // Host PHY register address field (Clause 22 Register address)
    field.hstregadx = addr;
    // Host Turn Around bit field (ignored during access to legacy PHY (10/100/1000)
    field.hsttafield = 2;
    if (!write_flag) 
    {
        // Lower 15 bits ignored during legacy PHY read

        // This bit triggers the MDIO operation
        cmd.hstldcmd = 1;
        // Legacy Read operation : 0x2
        cmd.hstmiimcmd = 2;
        // This operation starts the MDIO Read
        BE2_Gen_Post_MAC_Command (function_context, port, cmd.dw, field.dw);

        // Read data from lower 16bits of AMIIM Field Register
        BE2_Gen_Access_XGMAC_Table (function_context, port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, amiim_field)/4, &field.dw, FALSE);
        *p_data = field.hstmiimwrdat;
    }
    else
    {
        field.hstmiimwrdat = *p_data;

        // This bit triggers the MDIO operation
        cmd.hstldcmd = 1;
        // Legacy Write operation : 0x1
        cmd.hstmiimcmd = 1;
        // This operation starts the MDIO Write
        BE2_Gen_Post_MAC_Command (function_context, port, cmd.dw, field.dw);

    }
}



//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Is_10Gb_Link_Up
//
//   Queries if the link on a given port is up or not. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number. 
// if_module          - Designates the interface module type. 
// phy_type           - Designates the PHY type. 
// return return_type - Returns TRUE if the link is up, FALSE if the link is down. 
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
BE2_Gen_Is_10Gb_Link_Up (
   IN void* function_context,
   IN u8    port,
   IN u8    if_module,
   IN u8    phy_type )
{
    AXGMAC_CFG2_REG config2;

    // Read the (internal) XGMAC config2 register
    BE2_Gen_Access_XGMAC_Table(function_context, port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, config2) / 4, &config2.dw, FALSE);
    if (config2.rflnkflt) {
        // A local or remote link fault is present, return bad link.
        return FALSE;
    }
#if 0
    if (if_module) {
        //
        // For non-CX4 modules we use GPIO to determine if a module is plugged-in to the
        // slot or not.
        // 
        // Port0 uses GPIO14. Port1 uses GPIO15.
        //
        u8 gpi_bit = (port) ? 15 : 14;
        BE2_Gen_GPIO_Set_Mode(function_context, gpi_bit, BE_GEN_GPIO_IN_MODE);

        // Read the GPI bit of interest. 
        if (BE2_Gen_GPI_Get_Data(function_context, gpi_bit)) {
            // No module is present, return bad link.
            return FALSE;
        }

        //
        // A module is present, and there are no XGMII link faults.
        // Drop through to the common code to check PCS status.
        // We expect the internal PCS status to be good most of the time.
        // However, the PCS controls the link up/down LED, and in cases
        // where we transition link status, we need to reset the PCS faults
        // in order for the LED to lite.
        //

    }
#endif 
    //else 
    {   
        u16 value;
    
        // Read PCS status register (3.1)
        BE2_Gen_Access_Internal_PHY_Reg(function_context, port, 3, 1, &value, FALSE);
        if (value & 4) {
            // link status is up
            return TRUE;
        }

        //
        // PCS status does not indicate good link. Reset any latched faults and 
        // re-check.
        //

        // Clear current link status, link fault, and sync (30.2)
        value = 0xC080;
        BE2_Gen_Access_Internal_PHY_Reg(function_context, port, 30, 2, &value, TRUE);

        // Reset register
        value = 0;
        BE2_Gen_Access_Internal_PHY_Reg(function_context, port, 30, 2, &value, TRUE);

        // Read PCS status register (3.1)
        BE2_Gen_Access_Internal_PHY_Reg(function_context, port, 3, 1, &value, FALSE);
        if (value & 4) {
            // link status is up
            return TRUE;
        } else {
            return FALSE;
        }
    }
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Is_Link_Up
//
//   Queries if the link on a given port is up or not. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// port               - The port number. 
// if_module          - Designates the interface module type. 
// phy_type           - Designates the PHY type. 
// return return_type - Returns TRUE if the link is up, FALSE if the link is down. 
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
BE2_Gen_Is_Link_Up (
   IN void* function_context,
   IN u8    port,
   IN u8    if_module,
   IN u8    phy_type )
{
    RXF_PHY_CONFIG_CSR phy_config;
    boolean link_up;

    phy_config.dw = BE_GEN_CSR_READ(function_context, rxf.phy_config);
    if (phy_config.eth_speed) {
        u8 phy_addr;
#ifdef BE_ARM_FIRMWARE
    extern u32 g_be_phy_swapped;
    phy_addr = (( TRUE == g_be_phy_swapped) ? !port : port);
#else
        phy_addr = port;
#endif
        // Operating in 1Gb/s mode.
        if ((BE2_GEN_PHY_LSI_BRCM1 == phy_type) || (BE2_GEN_PHY_LSI_BRCM2 == phy_type)) {
            //
            // Using a Broadcom PHY. 
            // The register map for this PHY is not made public to ServerEngines. The offsets and
            // values passed via MDIO are dictated by Broadcom/LSI. Read the register twice.
            // On LSI board the PHY is connected in reverse order, thus the inversion of the port.
            //
            u32 reg_val;
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 0x8304, !port, BE_GEN_MDIO_OP_10G);
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 0x8304, !port, BE_GEN_MDIO_OP_10G);

            // bit 1 indicates link up
            if (!(reg_val & 0x02)) {
                // Link is down
                return FALSE;
            }
        }
        else if ((BE2_GEN_PHY_QT2025 == phy_type) || (BE2_GEN_PHY_QT2225 == phy_type)) {
            u32 reg_val;
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 3, 0xD7FD, phy_addr, BE_GEN_MDIO_OP_10G);
            // 0x70  indicates link up
            if (reg_val != 0x70) {
                // Link is down
                return FALSE;
            }

        }

        else if (BE2_GEN_PHY_BRCM_5931 == phy_type) {
            u32 reg_val;
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 0x8304, phy_addr, BE_GEN_MDIO_OP_10G);
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 0x8304, phy_addr, BE_GEN_MDIO_OP_10G);
            // bit 1 indicates linkup
            if (!(reg_val & ( 1 << 1))) {
                // Link is down
                return FALSE;
            }
        }

        else if (BE2_GEN_PHY_SOLARFLARE == phy_type) {
            u32 reg_val;
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 1, phy_addr, BE_GEN_MDIO_OP_10G);
            reg_val = BE2_Gen_MDIO_Read_Reg(function_context, 7, 1, phy_addr, BE_GEN_MDIO_OP_10G);
            // bit 2 indicates linkup
            if (!(reg_val & ( 1 << 2))) {
                // Link is down
                return FALSE;
            }
        }

        // Internal PHY status
        link_up = (port ? phy_config.port1_link_up : phy_config.port0_link_up);
    } else {
        // Operating in 10Gb/s mode.
        link_up = BE2_Gen_Is_10Gb_Link_Up(function_context, port, if_module, phy_type);
    }
    return (link_up);
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Phys_Port_Local_Fault
//
//   This routine is used to Enable/Disable a forced Fault. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// phys_port          - BE Port. 
// mode               - 1-Enable 0- Disable. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Phys_Port_Local_Fault (
   IN void* function_context,
   IN u8    phys_port,
   IN u8    mode
)
{
    AXGMAC_CFG2_REG config2;
    BE2_Gen_Access_XGMAC_Table(NULL, phys_port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, config2) / 4, &config2.dw, FALSE);
    // Setting 27th bit will enable a local fault.
    if(mode)
        config2.dw |= ( 1 << 27);
    else 
        config2.dw &= ~( 1 << 27);
    BE2_Gen_Access_XGMAC_Table(NULL, phys_port, BE_GEN_FIELD_OFFSET (MAC_REGISTER_TABLE, config2) / 4, &config2.dw, TRUE);

}

#endif  // __be2_gen_rxf_h__

