/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!

@file
    chipobj.cpp

@brief
    Chip object implementation.  A chip object represents the BladeEngine as a
    whole.  Each chip is capable of of representing up to two function objects
    (representing true or logical PCI functions for storage and networking).

@author jflores

@note
    jflores   8/03/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/chipobj_ll.c,v 1.9 2006/02/21 03:20:07 menz Exp $
    
    ChipObjects locking strategy uses a single per-chip-object lock to serialize
    access to resources managed by the chip object.  The once exception to this
    rule is during ChipObjectFindOrCreate(), where the global lock will be
    used instead.

*/

#include "pch.h"


BESTATUS
be_chip_create (    
    OUT PBE_CHIP_OBJECT  chip)
{
    BESTATUS status = BE_SUCCESS;
    SA_TRACE_ENTRY ();

    ASSERT (chip);
    SA_ZERO_MEM (chip);

    TRACE (DL_INFO, "Create chip object. object:0x%p", chip);

    chip->magic     = BE_CHIP_MAGIC;
    chip->ref_count = 0;    

    sa_initialize_list_head (&chip->function_list_head);

    be_lock_init (&chip->lock);
    
//<no_inbox>
    // Initialize the MCC emulation layer.
    be_mccemul_init (&chip->mcc_emul);
//</no_inbox>

    return status;
}

void
be_chip_destroy (PBE_CHIP_OBJECT chip)
{
    SA_TRACE_ENTRY ();
    CHIP_ASSERT  (chip);    
    ASSERT (chip->ref_count == 0);
   
    SA_ZERO_MEM (chip);
    TRACE (DL_INFO, "destroying chip %p", chip);
}

BESTATUS    
be_chip_insert_function_object (
    IN      PBE_CHIP_OBJECT         chip_object, 
    IN      PBE_FUNCTION_OBJECT     function_object)
{
    SA_TRACE_ENTRY ();

    CHIP_ASSERT (chip_object);
    FUNCTION_ASSERT (function_object);

    TRACE (DL_INFO, "Insert function into chip object.");

    ASSERT (function_object->parent_chip == NULL ||
            function_object->parent_chip == chip_object);
    function_object->parent_chip = chip_object;

    be_chip_lock (chip_object);

    chip_object->ref_count++;
    sa_insert_tail_list (&chip_object->function_list_head, &function_object->function_list);

    be_chip_unlock (chip_object);   

    #if defined(BE_EMULATION) 
    {
        PVOID pd_bar_va;
        u32 func = be_function_get_function_number (function_object);

        pd_bar_va = sa_dev_get_bar_offset (
                        function_object->sa_dev,
                        sa_get_pd_bar (function_object->sa_dev));
        ASSERT (pd_bar_va);

        if (chip_object->pd0_bar_va[func] == NULL) {
            // For emulation, assume the first chip created is pd 0.    
            chip_object->pd0_bar_va[func] = pd_bar_va;
        }

        // Set the PD ID in the function object for allow emulation to function.
        function_object->pd_id = 
            (u32)(
            (SA_PTR_TO_U64 (pd_bar_va) - SA_PTR_TO_U64 (chip_object->pd0_bar_va[func])) /
            sizeof (PROTECTION_DOMAIN_DBMAP));

        TRACE (DL_NOTE, "Assigned PD id:%d func:%d", 
               function_object->pd_id,
               func
               );

        ASSERT (function_object->pd_id <
                SA_NUMBER_OF_FIELD (BLADE_ENGINE_PCI0_DBMAP, pci_func0_protection_domain_dbmap));

        // initialize the MPU during emulation
        if (!chip_object->mpu_initialized) {
            _be_mpuemul_init (function_object);
            chip_object->mpu_initialized = TRUE;
        }
    }
    #endif

    return BE_SUCCESS;
}


void
be_chip_remove_function_object (
    IN      PBE_CHIP_OBJECT         chip_object, 
    IN      PBE_FUNCTION_OBJECT     function_object)
{
    SA_TRACE_ENTRY ();

    CHIP_ASSERT (chip_object);
    FUNCTION_ASSERT (function_object);

    function_object->parent_chip = NULL;

    be_chip_lock (chip_object);

    ASSERT (chip_object->ref_count > 0);
    ASSERT (!sa_is_list_empty (&chip_object->function_list_head));

    chip_object->ref_count--;
    sa_remove_entry_list (&function_object->function_list);    

    be_chip_unlock (chip_object);
}

/*!

@brief
    This routine serializes access to resources maintained through a chip object.

@param    
    ChipObject      - The chip object to acquire the lock for.

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
be_chip_lock (IN PBE_CHIP_OBJECT chip_object)
{
    be_lock_acquire (&chip_object->lock);
}

/*!
@brief
    This routine removes serialization done by ChipObjectLock. 
@param    
    ChipObject      - The chip object to drop the lock for.
@return   
@note
    IRQL < DISPATCH_LEVEL    
*/
void
be_chip_unlock (
    IN      PBE_CHIP_OBJECT chip_object)
{
    be_lock_release (&chip_object->lock);
}



