/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "pch.h"

// LANCER TODO: Remove once LANCER specific defs find their right place
#include "lancer_temp.h"

void be_eth_sq_create_v1_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_ETHSQ_OBJECT eth_sq = (PBE_ETHSQ_OBJECT)context;

    ASSERT (eth_sq);

    if (status != BE_SUCCESS) {
        
        TRACE (DL_ERR, "Create eth sq failed in internal callback.\n");

    } else {

        IOCTL_ETH_TX_CREATE_V1 *ioctl = NULL;

        ASSERT (wrb); // non embedded by definition

        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (eth_sq->parent_function, wrb, ETH_TX_CREATE_V1);

        // save the butler ID
        eth_sq->bid = ioctl->params.response.id;
        eth_sq->rid = ioctl->params.response.rid;

        //save doorbell offset and class_queue_id
        eth_sq->doorbell_offset = ioctl->params.response.doorbell_offset;
        eth_sq->class_queue_id = ioctl->params.response.class_queue_id;

        //TRACE(DL_ERR,"doorbell offset: %d", ioctl->params.response.doorbell_offset);
        //TRACE(DL_ERR,"response class_queue_id: %d", ioctl->params.response.class_queue_id);

        // add a reference to the corresponding CQ
        be_cq_object_reference (eth_sq->cq_object);

        // add this object to the function object
        _be_function_add_eth_sq (eth_sq->parent_function, eth_sq);

        TRACE (DL_INFO, "eth sq created. function:%d pd:%d bid:%d cq_id:%d",
               be_function_get_function_number (eth_sq->parent_function),
               be_function_get_pd_number (eth_sq->parent_function),
               eth_sq->bid,
               be_cq_get_id (eth_sq->cq_object));
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create_ex
//
//   Creates an ethernet send ring - extended version with additional parameters. 
//
// function_object - 
// sgl             - no virtual address required
// length_in_bytes - 
// type            - The type of ring to create.
// ulp             - The requested ULP number for the ring. This should be zero based,
//                   i.e. 0,1,2. This must be valid NIC ULP based on the firmware
//                   configuration. Drivers may query the firmware configuration with
//                   be_function_query_firmware_config. All doorbells for this ring must
//                   be sent to this ULP. 
// cq_object       - cq object for completions
// ex_parameters   - Additional parameters (that may increase in future revisions). These
//                   parameters are only used for certain ring types -- see
//                   BE_ETH_SQ_PARAMETERS for details. 
// eth_sq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create_v1 (
   IN PBE_FUNCTION_OBJECT   function_object,
   IN PSA_SGL               sgl,
   IN u32                   length,
   IN u32                   type,   
   PBE_RXIF_OBJECT          rxif_object,   
   IN PBE_CQ_OBJECT         cq_object, 
   OUT PBE_ETHSQ_OBJECT     eth_sq,
   IN MCC_WRB_CQE_CALLBACK  callback,
   IN PVOID                 callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context
)
{
    IOCTL_ETH_TX_CREATE_V1 *ioctl = NULL;
    IOCTL_ETH_TX_CREATE_V1_LANCER *v1_ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;    
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};

    ASSERT (sgl);
    ASSERT (eth_sq);

    FUNCTION_ASSERT (function_object);
    be_lock_wrb_post (function_object);

    sa_zero_mem (eth_sq, sizeof(*eth_sq));

    eth_sq->magic           = BE_ETHSQ_MAGIC;
    eth_sq->parent_function = function_object;
    eth_sq->bid             = 0xFFFFFFFF;
    eth_sq->cq_object       = cq_object;

    // Translate beclib interface to arm interface.
    switch (type) {
        case BE_ETH_TX_RING_TYPE_FORWARDING: 
            type = ETX_QUEUE_TYPE_FORWARDING; 
            break;
        case BE_ETH_TX_RING_TYPE_STANDARD:   
            type = ETX_QUEUE_TYPE_STANDARD;   
            break;
        case BE_ETH_TX_RING_TYPE_PRIORITY:
            type = ETX_QUEUE_TYPE_PRIORITY;
            break;
        default:
            // TRACE (DL_ERR, "Invalid eth tx ring type:%d", type);
            status = BE_NOT_OK;
            goto Error;
            break;
    }

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_TX_CREATE_V1);

    ioctl->header.request.version = 1;

    ioctl->params.request.num_pages = sa_ceiling (length, SA_PAGE_SIZE);
    ASSERT (sa_sgl_get_page_count (sgl) >= ioctl->params.request.num_pages);
    ioctl->params.request.type = type;

    if (function_object->adapter_family != ADAPTER_FAMILY_LANCER) {
        ioctl->params.request.iface_id = rxif_object->handle;
        ioctl->params.request.queue_size = be_ring_length_to_encoding (length, sizeof (BE2_ETH_WRB));
        ioctl->params.request.cq_id = cq_object->cq_id;
    } else {
        // Program V1 specific fields
        v1_ioctl = (PVOID)ioctl;
        v1_ioctl->params.request.context.ctx_valid    = TRUE;
        v1_ioctl->params.request.context.cq_id_send   = cq_object->cq_id;
        v1_ioctl->params.request.context.iface_id     = rxif_object->handle;
        v1_ioctl->params.request.context.tx_ring_size = be_ring_length_to_encoding (length, sizeof (BE2_ETH_WRB));
    }

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the ioctl
    status = be_function_post_mcc_wrb_complete (
                function_object,
                wrb,
                generic_context,   // queue context
                callback,
                callback_context,
                be_eth_sq_create_v1_internal_callback,
                eth_sq, 
                ioctl,
                response_copy);

    if (status < 0) {
        TRACE (DL_ERR, "SQ Create ioctl failed.");
        sa_zero_mem (eth_sq, sizeof(*eth_sq));
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

void be_eth_sq_create_ex_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_ETHSQ_OBJECT eth_sq = (PBE_ETHSQ_OBJECT)context;

    ASSERT (eth_sq);

    if (status != BE_SUCCESS) {
        
        TRACE (DL_ERR, "Create eth sq failed in internal callback.\n");

    } else {

        IOCTL_ETH_TX_CREATE *ioctl = NULL;

        ASSERT (wrb); // non embedded by definition

        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (eth_sq->parent_function, wrb, ETH_TX_CREATE);

        // save the butler ID
        eth_sq->bid = ioctl->params.response.cid;
        eth_sq->rid = ioctl->params.response.rid;

        // add a reference to the corresponding CQ
        be_cq_object_reference (eth_sq->cq_object);

        // add this object to the function object
        _be_function_add_eth_sq (eth_sq->parent_function, eth_sq);

        TRACE (DL_INFO, "eth sq created. function:%d pd:%d bid:%d cq_id:%d",
               be_function_get_function_number (eth_sq->parent_function),
               be_function_get_pd_number (eth_sq->parent_function),
               eth_sq->bid,
               be_cq_get_id (eth_sq->cq_object));
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create_ex
//
//   Creates an ethernet send ring - extended version with additional parameters. 
//
// function_object - 
// sgl             - no virtual address required
// length_in_bytes - 
// type            - The type of ring to create.
// ulp             - The requested ULP number for the ring. This should be zero based,
//                   i.e. 0,1,2. This must be valid NIC ULP based on the firmware
//                   configuration. Drivers may query the firmware configuration with
//                   be_function_query_firmware_config. All doorbells for this ring must
//                   be sent to this ULP. 
// cq_object       - cq object for completions
// ex_parameters   - Additional parameters (that may increase in future revisions). These
//                   parameters are only used for certain ring types -- see
//                   BE_ETH_SQ_PARAMETERS for details. 
// eth_sq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create_ex (
   IN PBE_FUNCTION_OBJECT   function_object,
   IN PSA_SGL               sgl,
   IN u32                   length,
   IN u32                   type,
   IN u32                   ulp,
   IN u32                   bound_port,
   IN PBE_CQ_OBJECT         cq_object,
   IN BE_ETH_SQ_PARAMETERS* ex_parameters,
   OUT PBE_ETHSQ_OBJECT     eth_sq,
   IN MCC_WRB_CQE_CALLBACK  callback,
   IN PVOID                 callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context
)
{
    IOCTL_ETH_TX_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    u32 cofe = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};

    // Not supported by LANCER family of adapters. Use version 1 instead
    if (function_object->adapter_family == ADAPTER_FAMILY_LANCER) {
        return BE_STATUS_NOT_SUPPORTED;
    }

    ASSERT (sgl);
    ASSERT (eth_sq);    
    ASSERT (ex_parameters);

    FUNCTION_ASSERT (function_object);
    be_lock_wrb_post (function_object);
    
    sa_zero_mem (eth_sq, sizeof (*eth_sq));

    eth_sq->magic           = BE_ETHSQ_MAGIC;
    eth_sq->parent_function = function_object;
    eth_sq->bid             = 0xFFFFFFFF;
    eth_sq->cq_object       = cq_object;        

    // Translate beclib interface to arm interface.
    switch (type) {
        case BE_ETH_TX_RING_TYPE_FORWARDING: 
            type = ETH_TX_RING_TYPE_FORWARDING; 
            break;
        case BE_ETH_TX_RING_TYPE_STANDARD:   
            type = ETH_TX_RING_TYPE_STANDARD;   
            break;
        case BE_ETH_TX_RING_TYPE_LOW_LATENCY:
            type = ETH_TX_RING_TYPE_LOW_LATENCY;
            break;
        case BE_ETH_TX_RING_TYPE_BOUND:
            type = ETH_TX_RING_TYPE_BOUND;
            break;
        case BE_ETH_TX_RING_TYPE_STANDARD_NO_QOS:
            type = ETH_TX_RING_TYPE_STANDARD;   
            cofe = 1;
            break;
        case BE_ETH_TX_RING_TYPE_PRIORITY:
            type = ETH_TX_RING_TYPE_STANDARD;
            break;
        default:
            // TRACE (DL_ERR, "Invalid eth tx ring type:%d", type);
            status = BE_NOT_OK;
            goto Error;
            break;
    }
    
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // NIC must be supported by the current config.
    ASSERT (ulp < SA_NUMBER_OF (function_object->fw_cfg.ulp));
    ASSERT (function_object->fw_cfg.ulp[ulp].ulp_mode & NIC_MODE);
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_TX_CREATE);

    ioctl->params.request.context.pd_id        = be_function_get_pd_number (function_object);
    ioctl->params.request.context.ctx_valid    = TRUE;
    ioctl->params.request.context.tx_ring_size = be_ring_length_to_encoding (length, sizeof (BE2_ETH_WRB));
    ioctl->params.request.context.cq_id_send   = cq_object->cq_id;
    ioctl->params.request.context.pci_func_id  = be_function_get_function_number(function_object); 
    ioctl->params.request.context.local_flow_id = be_function_get_pd_number (function_object);
    ioctl->params.request.context.no_remain_allowed = 1;
    ioctl->params.request.context.disable_qos = cofe;

    ioctl->params.request.type                 = type;
    ioctl->params.request.ulp_num              = ulp;
    ioctl->params.request.bound_port           = bound_port;

    ASSERT (ulp < SA_NUMBER_OF (function_object->fw_cfg.ulp));
    ASSERT (function_object->fw_cfg.ulp[ulp].ulp_mode & NIC_MODE);

    ioctl->params.request.num_pages = sa_ceiling (length, SA_PAGE_SIZE);   
    ASSERT (sa_sgl_get_page_count (sgl) >= ioctl->params.request.num_pages);

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the ioctl
    status = be_function_post_mcc_wrb_complete (
                function_object,
                wrb,
                generic_context,   // queue context
                callback,
                callback_context,
                be_eth_sq_create_ex_internal_callback,
                eth_sq, 
                ioctl,
                response_copy);

    if (status < 0) {
        TRACE (DL_ERR, "SQ Create ioctl failed.");
        sa_zero_mem (eth_sq, sizeof(*eth_sq));
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create
//
//   Creates an ethernet send ring. 
//
// function_object - 
// sgl             - no virtual address required
// length_in_bytes - 
// type            - The type of ring to create.
// ulp             - The requested ULP number for the ring. This should be zero based,
//                   i.e. 0,1,2. This must be valid NIC ULP based on the firmware
//                   configuration. Drivers may query the firmware configuration with
//                   be_function_query_firmware_config. All doorbells for this ring must
//                   be sent to this ULP. 
// cq_object       - cq object for completions
// eth_sq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length_in_bytes,
   IN u32                 type,
   IN u32                 ulp,
   IN u32                 port,
   IN PBE_CQ_OBJECT       cq_object,
   OUT PBE_ETHSQ_OBJECT   eth_sq )
{
    BE_ETH_SQ_PARAMETERS ex_parameters;
    SA_ZERO_MEM(&ex_parameters);
    return be_eth_sq_create_ex (function_object,
                                sgl,
                                length_in_bytes,
                                type,
                                ulp,
                                port,
                                cq_object,
                                &ex_parameters,
                                eth_sq,
                                NULL,
                                NULL,
                                NULL);
}


/*!

@brief
    This routine returns the RID for an Etx SQ

@param    
    EthSq - EthSq Handle returned from EthSqCreate

@return   
    The RID for an Etx SQ

@note    

*/
u32
be_eth_sq_get_id (
    IN      PBE_ETHSQ_OBJECT        eth_sq)
{
    ETHSQ_ASSERT (eth_sq);
    return  eth_sq->bid;
}

u32
be_eth_sq_get_completion_id (
    IN      PBE_ETHSQ_OBJECT        eth_sq)
{
    ETHSQ_ASSERT (eth_sq);
    return  eth_sq->rid;
}

/*!

@brief
    This routine destroys an ethernet send queue     

@param    
    EthSq - EthSq Handle returned from EthSqCreate

@return
    This function always return BE_SUCCESS.

@note
    This function frees memory allocated by EthSqCreate for the EthSq Object.

*/

void
be_eth_sq_destroy_callback (
    PVOID Context,
    BESTATUS Status,
    MCC_WRB* OptionalWrb)
{
    PBE_ETHSQ_OBJECT eth_sq = (PBE_ETHSQ_OBJECT)Context;

    // Derefence any associated CQs.
    be_cq_object_dereference (eth_sq->cq_object);

    // Remove from function
    _be_function_remove_eth_sq (eth_sq->parent_function, eth_sq);

    // Clear tracking object
    sa_zero_mem (eth_sq, sizeof (*eth_sq));
}


BESTATUS 
be_eth_sq_destroy_async (
    IN PBE_ETHSQ_OBJECT eth_sq,
    MCC_WRB_CQE_CALLBACK callback,
    PVOID                callback_context,
    BE_GENERIC_QUEUE_CONTEXT *queue_context)
{
    BESTATUS status = 0;

    ETHSQ_ASSERT (eth_sq);

    // Send ioctl to destroy the queue.
    status = be_function_ring_destroy_async (eth_sq->parent_function, 
                                             eth_sq->bid, 
                                             IOCTL_RING_TYPE_ETH_TX,
                                             queue_context,
                                             callback,
                                             callback_context,
                                             be_eth_sq_destroy_callback,
                                             eth_sq);
    return status;
}

BESTATUS 
be_eth_sq_destroy (
    IN PBE_ETHSQ_OBJECT eth_sq)
{
    return be_eth_sq_destroy_async (eth_sq, NULL, NULL, NULL);
}

/*!

@brief
    This routine attempts to set the transmit flow control parameters.  
    
@param
    FunctionObject      - Handle to a function object                           
                          
@param
    txfc_enable         - transmit flow control enable - true for enable, false for disable

@param
    rxfc_enable         - receive flow control enable - true for enable, false for disable

@return
    BE_SUCCESS if successfull, otherwise a useful BESTATUS error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    
    This function always fails in non-privileged machine context.  
*/    
BESTATUS
be_eth_set_flow_control (
   IN PBE_FUNCTION_OBJECT function_object,
   IN boolean             txfc_enable,
   IN boolean             rxfc_enable)
{
    IOCTL_COMMON_SET_FLOW_CONTROL *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_FLOW_CONTROL);

    ioctl->params.request.rx_flow_control = rxfc_enable;
    ioctl->params.request.tx_flow_control = txfc_enable;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "set flow control ioctl failed.");
        goto error;
    }

error:
    be_unlock_wrb_post (function_object);

    return status;
}


BESTATUS
be_eth_set_flow_control_ex (
   IN PBE_FUNCTION_OBJECT function_object,
   IN boolean             txfc_enable,
   IN boolean             rxfc_enable,
   IN MCC_WRB_CQE_CALLBACK             callback,
   IN PVOID                            callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context)
{
    IOCTL_COMMON_SET_FLOW_CONTROL *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto error;
        }
    }
       
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_FLOW_CONTROL);

    ioctl->params.request.rx_flow_control = rxfc_enable;
    ioctl->params.request.tx_flow_control = txfc_enable;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 callback, // callback
                 callback_context, // context 
                 ioctl);
    
error:
    be_unlock_wrb_post (function_object);

    return status;
}

/*!

@brief
    This routine attempts to get the transmit flow control parameters.  
    
@param
    FunctionObject      - Handle to a function object                           
                          
@param
    txfc_enable         - transmit flow control enable - true for enable, false for disable

@param
    rxfc_enable         - receive flow control enable - true for enable, false for disable

@return
    BE_SUCCESS if successfull, otherwise a useful BESTATUS error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    
    This function always fails in non-privileged machine context.  
*/    
BESTATUS
be_eth_get_flow_control (
   IN PBE_FUNCTION_OBJECT function_object,
   IN boolean*            txfc_enable,
   IN boolean*            rxfc_enable)
{
    IOCTL_COMMON_GET_FLOW_CONTROL *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_FLOW_CONTROL);

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "get flow control ioctl failed.");
        goto error;
    }

    *txfc_enable = ioctl->params.response.tx_flow_control;
    *rxfc_enable = ioctl->params.response.rx_flow_control;

error:
    be_unlock_wrb_post (function_object);

    return status;
}


//---------------------------------------------------------------------------------------
// Function: be_eth_set_qos
//
//   This function sets the ethernet transmit Quality of Service (QoS) characteristics of
//   BladeEngine for the domain. All ethernet transmit rings of the domain will evenly
//   share the bandwidth. The exeception to sharing is the host primary (super) ethernet
//   transmit ring as well as the host ethernet forwarding ring for missed offload data. 
//
// function_object - 
// max_bps         - the maximum bits per second in units of 10 Mbps (valid 0-100)
// max_pps         - the maximum packets per second in units of 1 Kpps (0 indicates no
//                   limit)
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_set_qos (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 max_bps,
   IN u32                 max_pps)
{
    IOCTL_COMMON_SET_QOS *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_QOS);

    // Set fields in ioctl
//    ioctl->params.request.fabric_domain_ID = be_function_get_pd_number (function_object);
    ioctl->params.request.max_bits_per_second_NIC = max_bps;
    ioctl->params.request.max_packets_per_second_NIC = 0; // max_pps;
    ioctl->params.request.valid_flags = QOS_BITS_NIC; //  | QOS_PKTS_NIC;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "network set qos ioctl failed.");
    }

error:
    be_unlock_wrb_post (function_object);

    return status;
}

//TODO_BE2: Review
//---------------------------------------------------------------------------------------
// Function: be_eth_set_qos_ex
//
//   This function sets the ethernet transmit Quality of Service (QoS) characteristics of
//   BladeEngine for the domain. All ethernet transmit rings of the domain will evenly
//   share the bandwidth. The exeception to sharing is the host primary (super) ethernet
//   transmit ring as well as the host ethernet forwarding ring for missed offload data. 
//
// function_object    - 
// parameters         - 
// callback           - Callback function on completion
// callback_context   - Callback context
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_set_qos_ex (
   IN PBE_FUNCTION_OBJECT              function_object,
   IN BE_QOS_PARAMETERS*               parameters,
   IN MCC_WRB_CQE_CALLBACK             callback,
   IN PVOID                            callback_context,
   IN PBE_COMMON_SET_QOS_QUEUE_CONTEXT queue_context
)
{
    IOCTL_COMMON_SET_QOS *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_QOS);

    // Set fields in ioctl
    ASSERT (parameters);
    ioctl->header.request.domain = parameters->pdid;
    ioctl->params.request.max_bits_per_second_NIC = parameters->max_bps;
    ioctl->params.request.max_packets_per_second_NIC = 0; // parameters->max_pps;
    ioctl->params.request.valid_flags = QOS_BITS_NIC;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "network set qos ioctl failed.");
    }

error:
    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_eth_get_qos
//
//   This function retrieves the ethernet transmit Quality of Service (QoS)
//   characteristics for the domain. 
//
// function_object - 
// max_bps         - the maximum bits per second in units of 10 Mbps (valid 0-100)
// max_pps         - the maximum packets per second in units of 1 Kpps (0 indicates no
//                   limit)
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_get_qos (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32*                max_bps,
   IN u32*                max_pps)
{
    IOCTL_COMMON_GET_QOS *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_QOS);

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "network get qos ioctl failed.");
        goto error;
    }

    *max_bps = ioctl->params.response.max_bits_per_second_NIC;
    *max_pps = 0; // ioctl->params.response.max_packets_per_second_NIC;

error:
    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_eth_set_frame_size
//
//   This function sets the ethernet maximum frame size. The previous values are returned.
//
// function_object - 
// tx_frame_size   - maximum transmit frame size in bytes
// rx_frame_size   - maximum receive frame size in bytes
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_eth_set_frame_size (
   IN PBE_FUNCTION_OBJECT function_object,
   IN OUT u32*            tx_frame_size,
   IN OUT u32*            rx_frame_size )
{
    IOCTL_COMMON_SET_FRAME_SIZE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_FRAME_SIZE);
    ioctl->params.request.max_tx_frame_size = *tx_frame_size;
    ioctl->params.request.max_rx_frame_size = *rx_frame_size;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "network set frame size ioctl failed.");
        goto error;
    }

    *tx_frame_size = ioctl->params.response.chip_max_tx_frame_size;
    *rx_frame_size = ioctl->params.response.chip_max_rx_frame_size;

error:
    be_unlock_wrb_post (function_object);

    return status;
}
