/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!

@file
    flash.cpp

@brief
    Flash managment routines.

@author jflores

@note
    jflores   8/03/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/flash_ll.c,v 1.4 2006/02/05 00:27:19 menz Exp $

*/

#include "pch.h"

//============================================================================
//                  F O R W A R D   D E C L A R A T I O N S
//============================================================================


//============================================================================
//                      P R I V A T E  R O U T I N E S
//============================================================================

/*!

@brief
    This routine reads or writes data to the specified address in the controller's flash.

@param        
    FunctionObject  - Function object handle.
    
@param        
    FlashAddress    - Address in the flash to read or write.

@param        
    Buffer          - Buffer to read or write to flash
    
@param        
    BufferLength    - Length of the data to be read or written

@param        
    IsRead          - Set to TRUE if this is for a flash read.
            
@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL

*/                     
STATIC
BESTATUS 
be_flash_read_write (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 flash_address,
    IN OUT PVOID           buffer,
    IN u32                 buffer_length,
    IN boolean             is_read)
{
    BESTATUS                status = BE_SUCCESS;

    FUNCTION_ASSERT( function_object );    
    ASSERT (buffer);
    ASSERT (flash_address % 4 == 0);
    ASSERT (buffer_length);
    ASSERT (buffer_length < 8*1024);
    
    if (is_read) {
//        status = be_function_post_mcc_wrb (function_object, &wrb, 0, NULL, buffer_length, buffer, NULL, NULL);
    } else {
//        status = be_function_post_mcc_wrb (function_object, &wrb, buffer_length, buffer, 0, NULL, NULL, NULL);
    }

    return  status;
    
}


//============================================================================
//                       P U B L I C  R O U T I N E S
//============================================================================

/*!

@brief
    This routine reads data to the specified address in the controller's flash.

@param        
    FunctionObject  - Function object handle.
    
@param        
    FlashAddress    - Address in the flash to read.

@param        
    Buffer          - Buffer to read to flash
    
@param        
    BufferLength    - Length of the data to be read

@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL

*/ 
BESTATUS 
be_flash_read (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 flash_address,
    IN PVOID               buffer,
    IN u32                 buffer_length)
{
    return  be_flash_read_write (function_object, 
                                 flash_address, 
                                 buffer, 
                                 buffer_length, 
                                 TRUE);
}

/*!

@brief
    This routine writes data to the specified address in the controller's flash.

@param        
    FunctionObject  - Function object handle.
    
@param        
    FlashAddress    - Address in the flash to write.

@param        
    Buffer          - Buffer to write to flash
    
@param        
    BufferLength    - Length of the data to be written

@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL

*/ 
BESTATUS 
be_flash_write (
    IN PBE_FUNCTION_OBJECT          function_object,
    IN IOCTL_COMMON_WRITE_FLASHROM *va_for_ioctl,
    IN u64                          pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK         callback,
    IN PVOID                        callback_context,
    IN PBE_FLASHACCESS_QUEUE_CONTEXT  queue_context)
{
    MCC_WRB *wrb=NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BESTATUS status=0;
    PIOCTL_REQUEST_HEADER header = NULL;

    ASSERT(va_for_ioctl);
    ASSERT(pa_for_ioctl);

    header = (PIOCTL_REQUEST_HEADER)va_for_ioctl;

    be_lock_wrb_post(function_object);

    wrb = be_function_peek_mcc_wrb(function_object);

    if (!wrb) {
        if (queue_context && callback) {
            wrb = (MCC_WRB*)&queue_context->wrb_header;
            generic_context=(PBE_GENERIC_QUEUE_CONTEXT)queue_context;
            generic_context->context.bytes = sizeof(*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }
    }

    wrb->embedded = 0;
    wrb->payload_length = sizeof(IOCTL_COMMON_FLASHROM)+va_for_ioctl->params.data_buffer_size-4;

    // Assume one fragment. The caller may override the SGL by rewriting the 0th length
    // and adding more entries.  They will also need to update the sge_count.
    wrb->sge_count = 1;
    wrb->payload.sgl[0].length = wrb->payload_length;
    wrb->payload.sgl[0].pa_lo = sa_lo(pa_for_ioctl);
    wrb->payload.sgl[0].pa_hi = sa_hi(pa_for_ioctl);

    header->timeout         = 0; // todo -- be_function_get_default_ioctl_timeout (function_object);
    header->domain          = be_function_get_pd_number(function_object);
    header->request_length  = wrb->payload_length;
    header->opcode          = BE_OPCODE_NAME(COMMON_WRITE_FLASHROM);
    header->subsystem       = IOCTL_SUBSYSTEM_COMMON;

    status = be_function_post_mcc_wrb_with_queue_context(
        function_object,
        wrb,
        generic_context,
        callback,
        callback_context,
        va_for_ioctl);

Error:
    be_unlock_wrb_post(function_object);

    return status;
}


BESTATUS 
be_mgmt_ioctl_passthru (
    IN PBE_FUNCTION_OBJECT          function_object,
    IN PVOID                        va_for_ioctl,
    IN u64                          pa_for_ioctl,
    IN int                          ioctlSize,
    IN MCC_WRB_CQE_CALLBACK         callback,
    IN PVOID                        callback_context,
    IN PBE_MGMTIOCTLS_QUEUE_CONTEXT  queue_context)
{
    MCC_WRB *wrb=NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BESTATUS status=0;
    PIOCTL_REQUEST_HEADER header = NULL;

    ASSERT(va_for_ioctl);
    ASSERT(pa_for_ioctl);

    header = (PIOCTL_REQUEST_HEADER)va_for_ioctl;

    be_lock_wrb_post(function_object);

    wrb = be_function_peek_mcc_wrb(function_object);

    if (!wrb) {
        if (queue_context && callback) {
            wrb = (MCC_WRB*)&queue_context->wrb_header;
            generic_context=(PBE_GENERIC_QUEUE_CONTEXT)queue_context;
            generic_context->context.bytes = sizeof(*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }
    }

    wrb->embedded = 0;
    wrb->payload_length = ioctlSize;

    // Assume one fragment. The caller may override the SGL by rewriting the 0th length
    // and adding more entries.  They will also need to update the sge_count.
    wrb->sge_count = 1;
    wrb->payload.sgl[0].length = wrb->payload_length;
    wrb->payload.sgl[0].pa_lo = sa_lo(pa_for_ioctl);
    wrb->payload.sgl[0].pa_hi = sa_hi(pa_for_ioctl);

    // header->timeout         = 0; // todo -- be_function_get_default_ioctl_timeout (function_object);
    header->domain          = be_function_get_pd_number(function_object);
    header->request_length  = ioctlSize;

    status = be_function_post_mcc_wrb_with_queue_context(
        function_object,
        wrb,
        generic_context,
        callback,
        callback_context,
        va_for_ioctl);

Error:
    be_unlock_wrb_post(function_object);

    return status;
}

