/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwiasync.c

    Purpose:

        This file implements functions needed for handing async PDUs (also.
        called default PDUs. hwi.c file invokes the main function 
        hwi_process_default_pdu_ring which takes care of processing the ring,
        forwarding the async message to iSCSI and then posting the buffers
        back to hardware
        
--*/

#include "acitpch.h"

ACIT_VOID
hwi_init_async_pdu_ctx (
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi;
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    PASYNC_PDU_HANDLE       pasync_header_h, pasync_data_h;
    ACIT_UEXACT32           index;

    // Initializes the ASYNC_PDU_CONTEXT structure. This will setup the
    // context to enable posting of buffers later.

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    phwi->phwic->pasync_ctx = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_PDU_CONTEXT);
    pasync_ctx = phwi->phwic->pasync_ctx;

    OSM_MEMSET(pasync_ctx, 0, sizeof(HWI_ASYNC_PDU_CONTEXT));

    pasync_ctx->buffer_size = p->default_pdu_buf_sz;
    pasync_ctx->num_entries = p->num_async_pdus;

    pasync_ctx->async_header.pa_base.u.a64.address = GET_MEM_DESC_PA64(pc, HWI_MEM_ASYNC_HEADER_BUF);
    pasync_ctx->async_header.va_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_HEADER_BUF);
    pasync_ctx->async_header.ring_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_HEADER_RING);
    pasync_ctx->async_header.handle_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_HEADER_HANDLE);
    pasync_ctx->async_header.writables = 0;
    LIST_INIT_HEADER(&pasync_ctx->async_header.free_list);

    pasync_ctx->async_data.pa_base.u.a64.address = GET_MEM_DESC_PA64(pc, HWI_MEM_ASYNC_DATA_BUF);
    pasync_ctx->async_data.va_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_DATA_BUF);
    pasync_ctx->async_data.ring_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_DATA_RING);
    pasync_ctx->async_data.handle_base = GET_MEM_DESC_VA(pc, HWI_MEM_ASYNC_DATA_HANDLE);
    pasync_ctx->async_data.writables = 0;
    LIST_INIT_HEADER(&pasync_ctx->async_data.free_list);

    pasync_header_h = (PASYNC_PDU_HANDLE)pasync_ctx->async_header.handle_base;
    pasync_data_h = (PASYNC_PDU_HANDLE)pasync_ctx->async_data.handle_base;

    for(index = 0; index < p->num_async_pdus; index++)
    {
        //
        // initialize the common part
        //

        // initialize wait queue
        LIST_INIT_HEADER(&pasync_ctx->async_entry[index].wait_queue.list);

        //
        // initialize the header part
        //

        // setup the free_list
        pasync_header_h->cri = ACIT_U16_MINUS_ONE;
        pasync_header_h->index = (ACIT_UEXACT8)index;
        LIST_INIT_ELEMENT(&pasync_header_h->link);
        pasync_header_h->pbuffer = \
            (PACIT_VOID)((ACIT_PTR)(pasync_ctx->async_header.va_base)+ 
                            (p->default_pdu_buf_sz * index));

        pasync_header_h->pa.u.a64.address = \
            pasync_ctx->async_header.pa_base.u.a64.address + 
                            (p->default_pdu_buf_sz * index);

        LIST_ADDTO_TAIL(&pasync_header_h->link, 
                        &pasync_ctx->async_header.free_list);
        pasync_header_h++;
        pasync_ctx->async_header.free_entries++;

        // initialize the busy list
        LIST_INIT_HEADER(&pasync_ctx->async_entry[index].header_busy_list);
        pasync_ctx->async_header.writables++;

        //
        // initialize the data part
        //

        // setup the free_list
        pasync_data_h->cri = ACIT_U16_MINUS_ONE;
        pasync_data_h->index = (ACIT_UEXACT8)index;
        LIST_INIT_ELEMENT(&pasync_data_h->link);
        pasync_data_h->pbuffer = \
            (PACIT_VOID)((ACIT_PTR)(pasync_ctx->async_data.va_base)+ 
                            (p->default_pdu_buf_sz * index));

        pasync_data_h->pa.u.a64.address = \
            pasync_ctx->async_data.pa_base.u.a64.address + 
                            (p->default_pdu_buf_sz * index);

        LIST_ADDTO_TAIL(&pasync_data_h->link, 
                        &pasync_ctx->async_data.free_list);
        pasync_data_h++;
        pasync_ctx->async_data.free_entries++;

        // initialize the busy list
        LIST_INIT_HEADER(&pasync_ctx->async_entry[index].data_busy_list);
        pasync_ctx->async_data.writables++;
    }

    pasync_ctx->async_header.host_write_ptr = 0;
    pasync_ctx->async_header.ep_read_ptr = ACIT_U32_MINUS_ONE;
    pasync_ctx->async_data.host_write_ptr = 0;
    pasync_ctx->async_data.ep_read_ptr = ACIT_U32_MINUS_ONE;
}

PACIT_DC_LIST
hwi_get_async_busy_list (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT32           is_header,
    ACIT_UEXACT32           host_write_ptr)
{
    if(is_header)
    {
        return &pasync_ctx->async_entry[host_write_ptr].header_busy_list;
    }
    else
    {
        return &pasync_ctx->async_entry[host_write_ptr].data_busy_list;
    }
}

PPHYS_ADDR
hwi_get_ring_address (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT32           is_header,
    ACIT_UEXACT32           host_write_ptr)
{
    PPHYS_ADDR              pasync_sge = ACIT_NULL;

    if(is_header)
    {
        pasync_sge = pasync_ctx->async_header.ring_base;
    }
    else
    {
        pasync_sge = pasync_ctx->async_data.ring_base;
    }

    return (pasync_sge + host_write_ptr);
}

PASYNC_PDU_HANDLE
hwi_get_async_handle (
    PACIT_CONTROLLER_WS     pcontroller,
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    PI_T_DPDU_CQE           pdpdu_cqe,
    PACIT_UEXACT32          pcq_index)
{
    ACIT_BUS_ADDRESS        phys_addr;
    PACIT_DC_LIST           pbusy_list = ACIT_NULL, plink;
    PASYNC_PDU_HANDLE       pasync_handle = ACIT_NULL;
    ACIT_UEXACT32           buffer_len=0;	
    ACIT_UEXACT8            buffer_index = ACIT_U8_MINUS_ONE;
    ACIT_UEXACT8            is_header = 0;

    // This function is invoked to get the right async_handle structure from
    // a given default PDU CQ entry.
    // The CQ entry contains the address offset where the DMA last ended. So subtract
    // the dataplacementlength to get to the base

    phys_addr.u.a32.address_lo = pdpdu_cqe->db_addr_lo - pdpdu_cqe->DataPlacementLength;
    phys_addr.u.a32.address_hi = pdpdu_cqe->db_addr_hi;

    phys_addr.u.a64.address = ACIT_GET_LE64(&phys_addr.u.a64.address);

    switch(pdpdu_cqe->code)
    {
    case UNSOL_HDR_NOTIFY:

        is_header = 1;

        // the index in the CQ entry gives the vertical index..
        pbusy_list = hwi_get_async_busy_list(pasync_ctx, 1, pdpdu_cqe->index);

        // and the buf_index from the PA will help walk the list horizontally
        buffer_len = (ACIT_UEXACT32)((phys_addr.u.a64.address  - 
                                        pasync_ctx->async_header.pa_base.u.a64.address));

        buffer_index = (ACIT_UEXACT8) ( buffer_len / pasync_ctx->buffer_size);

        break;

    case UNSOL_DATA_NOTIFY:

        // the index in the CQ entry gives the vertical index..
        pbusy_list = hwi_get_async_busy_list(pasync_ctx, 0, pdpdu_cqe->index);

        // and the buf_index from the PA will help walk the list horizontally
        buffer_len = (ACIT_UEXACT32)((phys_addr.u.a64.address - 
                                        pasync_ctx->async_data.pa_base.u.a64.address));
        buffer_index = (ACIT_UEXACT8) (buffer_len/pasync_ctx->buffer_size);

        break;
    }

    OSM_ASSERT(buffer_index <= pasync_ctx->num_entries);
    OSM_ASSERT(IS_LIST_NOT_EMPTY(pbusy_list));

    // walk the list and locate the default_handle with index buf_index
    plink = pbusy_list->next;

    while(plink != pbusy_list)
    {
        pasync_handle = STRUCT_BASE(ASYNC_PDU_HANDLE, 
                                      link,
                                      plink);

        OSM_ASSERT(pasync_handle);

        if(pasync_handle->index == buffer_index)
        {
            OSM_ASSERT(!pasync_handle->completed);
            pasync_handle->completed = 1;
            // found a match, exit out of the loop
            break;
        }

        plink = plink->next;
        pasync_handle = ACIT_NULL;
    }

    OSM_ASSERT(pasync_handle);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("CQ Index 0x%x Async Handle Index 0x%x Is_Hdr %d DataLen %d\n",
                                        pdpdu_cqe->index,
                                        pasync_handle->index,
                                        is_header,
                                        pdpdu_cqe->DataPlacementLength));

    pasync_handle->cri = CRI_FROM_CID(pcontroller, pdpdu_cqe->cid);
    pasync_handle->is_header = is_header;
    pasync_handle->buffer_len = pdpdu_cqe->DataPlacementLength;

    *pcq_index = pdpdu_cqe->index;
    return pasync_handle;
}

ACIT_UEXACT32
hwi_gather_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PASYNC_PDU_HANDLE       pasync_handle)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    ACIT_UEXACT32           bytes_needed = 0, status = 0;
    ACIT_UEXACT16           cri = pasync_handle->cri;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    // remove the element from the busy list and insert into the wait list
    LIST_REMOVE_ELEMENT(&pasync_handle->link);

    if(pasync_handle->is_header)
    {
        pasync_ctx->async_header.busy_entries--;

        // if there is an already stored header, then evict that header and store
        // this one. Only 1 header can be queued currently. It should be ok since
        // the only non-erroroneous case where we could see 2 headers is when an
        // offload is in progress.
        if(pasync_ctx->async_entry[cri].wait_queue.hdr_received)
        {
            OSM_ASSERT(0);
            hwi_free_async_msg(pcontroller, cri);
        }

        pasync_ctx->async_entry[cri].wait_queue.bytes_received = 0;
        pasync_ctx->async_entry[cri].wait_queue.hdr_received = 1;
        pasync_ctx->async_entry[cri].wait_queue.hdr_len = \
                        (ACIT_UEXACT8)pasync_handle->buffer_len;
        LIST_ADDTO_TAIL(&pasync_handle->link, 
                        &pasync_ctx->async_entry[cri].wait_queue.list);

        status = iscsi_check_async_pdu(pcontroller,
                                       cri,
                                       pasync_handle->pbuffer,
                                       &bytes_needed);

        if(status == ACIT_STATUS_SUCCESS)
        {
            pasync_ctx->async_entry[cri].wait_queue.bytes_needed = bytes_needed;

            if(bytes_needed == 0)
            {
                // all data has been received, forward it to iscsi
                status = hwi_fwd_async_msg(pcontroller, pasync_ctx, cri);
            }
        }
    }
    else
    {
        pasync_ctx->async_data.busy_entries--;

        if(pasync_ctx->async_entry[cri].wait_queue.hdr_received)
        {
            LIST_ADDTO_TAIL(&pasync_handle->link, 
                            &pasync_ctx->async_entry[cri].wait_queue.list);
            // Compilation note!
            // Microsoft C4244 - do not use += on non-integer types
            //
            pasync_ctx->async_entry[cri].wait_queue.bytes_received = 
                            pasync_ctx->async_entry[cri].wait_queue.bytes_received +
                            (ACIT_UEXACT16)pasync_handle->buffer_len;

            if(pasync_ctx->async_entry[cri].wait_queue.bytes_received >= 
                            pasync_ctx->async_entry[cri].wait_queue.bytes_needed)
            {
                // all data has been received, forward it to iscsi
                status = hwi_fwd_async_msg(pcontroller, pasync_ctx, cri);
            }
        }
    }

    return status;
}

ACIT_UEXACT32
hwi_fwd_async_msg (
    PACIT_CONTROLLER_WS     pcontroller,
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT16           cri)
{
    PACIT_DC_LIST           plink;
    PASYNC_PDU_HANDLE       pasync_handle;
    PACIT_VOID              phdr = ACIT_NULL;
    ACIT_UEXACT32           hdr_len = 0, buf_len = 0;
    ACIT_UEXACT32           status, index = 0, offset = 0;

    //ACIT_UEXACT8            buffer[48] = {0};

    OSM_MEMSET(pasync_ctx->buffer, 0, sizeof(pasync_ctx->buffer));

    plink = pasync_ctx->async_entry[cri].wait_queue.list.next;

    while((plink != &pasync_ctx->async_entry[cri].wait_queue.list))
    {
        pasync_handle = STRUCT_BASE(ASYNC_PDU_HANDLE,
                                      link,
                                      plink);

        OSM_ASSERT(pasync_handle);

        if(index == 0)
        {
            OSM_ASSERT(pasync_handle->is_header);
            // first element is the header, get its reference
            phdr = pasync_handle->pbuffer;
            hdr_len = pasync_handle->buffer_len;
        }
        else
        {
            // the data part can be contained in different buffers, but we're
            // interested in only 48 bytes. So copy upto 48 bytes into local
            // buffer
            if((offset + pasync_handle->buffer_len) >= 48)
            {
                buf_len = (48 - offset);
            }
            else
            {
                buf_len = pasync_handle->buffer_len;
            }

            OSM_MEMCPY(&pasync_ctx->buffer[0] + offset, 
                       pasync_handle->pbuffer,
                       buf_len);

            if((buf_len + offset) == 48)
            {
                // done collecting all the bytes we need, break out of the loop
                buf_len = buf_len + offset;
                break;
            }

            offset += buf_len;

            OSM_ASSERT(offset <= 48);
        }

        index++;
        plink = plink->next;
    }

    status = iscsi_process_async_pdu(pcontroller,
                                     CID_FROM_CRI(pcontroller, cri),
                                     phdr,
                                     hdr_len,
                                     &pasync_ctx->buffer,
                                     buf_len);

    if(status == ACIT_STATUS_SUCCESS)
    {
        // iscsi has finished processing the buffers, add it
        // to the free pool. This will remove the element
        // from wait list and add it to free list
        hwi_free_async_msg(pcontroller, cri);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_deliver_pending_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    ACIT_UEXACT32           status;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    if(pasync_ctx->async_entry[cri].wait_queue.hdr_received &&
       (pasync_ctx->async_entry[cri].wait_queue.bytes_received >= 
        pasync_ctx->async_entry[cri].wait_queue.bytes_received))
    {
        status = hwi_fwd_async_msg(pcontroller, 
                                   pasync_ctx, 
                                   (ACIT_UEXACT16)cri);

        if(status != ACIT_STATUS_INSUFFICIENT_RESOURCE)
        {
            // iscsi has finished processing the buffers, add it
            // to the free pool
            hwi_free_async_msg(pcontroller, cri);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_free_async_msg (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    PASYNC_PDU_HANDLE       pasync_handle;
    PACIT_DC_LIST           plink;
    ACIT_UEXACT32           i = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    plink = pasync_ctx->async_entry[cri].wait_queue.list.next;

    while(plink != &pasync_ctx->async_entry[cri].wait_queue.list)
    {
        pasync_handle = STRUCT_BASE(ASYNC_PDU_HANDLE, link, plink);
        OSM_ASSERT(pasync_handle);

        plink = plink->next;
        LIST_REMOVE_ELEMENT(&pasync_handle->link);

#if ACIT_CFG_DEBUG_BUILD
        OSM_MEMSET(pasync_handle->pbuffer,
                    0xbd,
                    pasync_ctx->buffer_size);
#endif
        
        if(i == 0)
        {
            LIST_ADDTO_TAIL(&pasync_handle->link, 
                            &pasync_ctx->async_header.free_list);
            pasync_ctx->async_header.free_entries++;
            i++;
        }
        else
        {
            LIST_ADDTO_TAIL(&pasync_handle->link, 
                            &pasync_ctx->async_data.free_list);
            pasync_ctx->async_data.free_entries++;
            i++;
        }
    }

    // reset the list header
    LIST_INIT_HEADER(&pasync_ctx->async_entry[cri].wait_queue.list);
    pasync_ctx->async_entry[cri].wait_queue.hdr_received = 0;
    pasync_ctx->async_entry[cri].wait_queue.bytes_received = 0;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_post_async_buffers (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_header)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    PASYNC_PDU_HANDLE       pasync_handle;
    PACIT_DC_LIST           pfree_link, pbusy_list;
    ACIT_UEXACT32           ring_id, num_entries;
    ACIT_UEXACT32           i = 0, host_write_ptr, writables;
    PPHYS_ADDR              pasync_sge;
    DEFAULT_PDU_DB          default_pdu_db;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    // The rule during posting buffers into the default PDU ring is that the
    // number of entries posted must be in multiples of 8.
    // There are 2 resources that matter, the free_entries and the writables.
    // free_entries is the count of number of buffers iSCSI has finished
    // processing and have posted them back. writables is the count of number
    // of buffers we know the EP firmware has consumed

    num_entries = pasync_ctx->num_entries;

    if(is_header)
    {
        writables = ACIT_MIN(pasync_ctx->async_header.writables,
                             pasync_ctx->async_header.free_entries);
        pfree_link = pasync_ctx->async_header.free_list.next;
        host_write_ptr = pasync_ctx->async_header.host_write_ptr;
        ring_id = phwi->default_pdu_hdr.id;
    }
    else
    {
        writables = ACIT_MIN(pasync_ctx->async_data.writables,
                             pasync_ctx->async_data.free_entries);
        pfree_link = pasync_ctx->async_data.free_list.next;
        host_write_ptr = pasync_ctx->async_data.host_write_ptr;
        ring_id = phwi->default_pdu_buf.id;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Before - is_hdr %d writables 0x%x free_entries 0x%x host_write_ptr 0x%x busy 0x%x\n",
                                                    is_header,
                                                    pasync_ctx->async_header.writables, 
                                                    pasync_ctx->async_header.free_entries,
                                                    pasync_ctx->async_header.host_write_ptr,
                                                    pasync_ctx->async_header.busy_entries));

    writables = (writables / 8) * 8;

    if(writables)
    {
        //OSM_ASSERT(writables == 8);

        for(i = 0; i < writables; i++)
        {
            pbusy_list = hwi_get_async_busy_list(pasync_ctx, 
                                                 is_header, 
                                                 host_write_ptr);

            pasync_handle = STRUCT_BASE(ASYNC_PDU_HANDLE, link, pfree_link);
            OSM_ASSERT(pasync_handle);
            pasync_handle->consumed = 0;
            pasync_handle->completed = 0;

            pfree_link = pfree_link->next;

            pasync_sge = hwi_get_ring_address(pasync_ctx, 
                                              is_header, 
                                              host_write_ptr);

            // Note! The hi and lo have to be reversed when submitting to the
            // hardware!
            ACIT_PUT_LE32(pasync_sge->hi, pasync_handle->pa.u.a32.address_lo);
            ACIT_PUT_LE32(pasync_sge->lo, pasync_handle->pa.u.a32.address_hi);

            if(IS_LIST_NOT_EMPTY(pbusy_list))
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("List growing horizontally at %d\n", host_write_ptr));
            }

            // remove from free list and add to busy list
            LIST_REMOVE_ELEMENT(&pasync_handle->link);
            LIST_ADDTO_HEAD(&pasync_handle->link,
                            pbusy_list);

            host_write_ptr ++;
            host_write_ptr = host_write_ptr % num_entries;
        }

        if(is_header)
        {
            pasync_ctx->async_header.host_write_ptr = host_write_ptr;
            pasync_ctx->async_header.free_entries -= writables;
            pasync_ctx->async_header.writables -= writables;
            pasync_ctx->async_header.busy_entries += writables;
            ACIT_UPDATE_ASYNC_HDR_POST_COUNT(pcontroller, writables);
        }
        else
        {
            pasync_ctx->async_data.host_write_ptr = host_write_ptr;
            pasync_ctx->async_data.free_entries -= writables;
            pasync_ctx->async_data.writables -= writables;
            pasync_ctx->async_data.busy_entries += writables;
            ACIT_UPDATE_ASYNC_DATA_POST_COUNT(pcontroller, writables);
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("After - is_hdr %d writables 0x%x free_entries 0x%x host_write_ptr 0x%x busy 0x%x\n",
                                                        is_header,
                                                        pasync_ctx->async_header.writables, 
                                                        pasync_ctx->async_header.free_entries,
                                                        pasync_ctx->async_header.host_write_ptr,
                                                        pasync_ctx->async_header.busy_entries));

		memset(&default_pdu_db,0,sizeof(default_pdu_db));
        default_pdu_db.qid = ring_id;
        default_pdu_db.rearm = 1;
        default_pdu_db.event = 0; // not an event queue
        default_pdu_db.cqproc = writables;

        ACIT_PD_WRITE (pcontroller, rxulp0_db, default_pdu_db.dw);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_update_async_writables (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx, 
    ACIT_UEXACT32           is_header, 
    ACIT_UEXACT32           cq_index)
{
    PACIT_DC_LIST           pbusy_list;
    PASYNC_PDU_HANDLE       pasync_handle;
    ACIT_UEXACT32           num_entries, writables = 0;
    PACIT_UEXACT32          pep_read_ptr, pwritables;

    // When we are here, its an indication that the hardware has consumed
    // a few of the buffers. cq_index is the last buffer we can assume to
    // have been consumed. So from the last ep_read_ptr upto cq_index, 
    // update all default_handles to have a consumed state. The value of
    // ep_read_ptr is also updated and so are the writables so as to
    // enable a post later

    if(is_header)
    {
        pep_read_ptr = &pasync_ctx->async_header.ep_read_ptr;
        pwritables = &pasync_ctx->async_header.writables;
    }
    else
    {
        pep_read_ptr = &pasync_ctx->async_data.ep_read_ptr;
        pwritables = &pasync_ctx->async_data.writables;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Before - is_hdr %d ep_read_ptr 0x%x writables 0x%x\n",
                                                    is_header, *pep_read_ptr, *pwritables));

    num_entries = pasync_ctx->num_entries;

    while((*pep_read_ptr) != cq_index)
    {
        // Pick only the first entry in the list for modification.
        // Though the default_list can grow horizontally, we only
        // need to update the element at the head. This is because
        // the list never grows horizontally without having the previous
        // entry consumed and the post routine always posts new ones to
        // the head of the list

        // the ep_read_ptr is pre-increment, note that its been initialized
        // to -1 in hwi_init_async_pdu_ctx
        (*pep_read_ptr)++;
        *pep_read_ptr = (*pep_read_ptr) % num_entries;

        pbusy_list = hwi_get_async_busy_list(pasync_ctx, is_header, *pep_read_ptr);

        if(writables == 0)
        {
            // if its the first time we're here inside the loop, the busy list
            // should never be empty
            OSM_ASSERT(IS_LIST_NOT_EMPTY(pbusy_list));
        }

        if(IS_LIST_NOT_EMPTY(pbusy_list))
        {
            // valid entry found, mark it consumed
            pasync_handle = STRUCT_BASE(ASYNC_PDU_HANDLE,
                                        link,
                                        pbusy_list->next);
            OSM_ASSERT(pasync_handle);
            pasync_handle->consumed = 1;
        }

        // increment the writables since its completed and we could re-post
        writables++;
    }

    if(writables > 1)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("writables > 1!! 0x%x\n", writables));
    }        

    // no updates were done?? This is a EP fw error!
    if(!writables)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Duplicate notification received - index 0x%x!!\n",
                                                cq_index));
        OSM_ASSERT(0);
    }

    // update the writables in the corresponding structures
    *pwritables = *pwritables + writables;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("After - is_hdr %d ep_read_ptr 0x%x writables 0x%x\n",
                                                    is_header, *pep_read_ptr, *pwritables));

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
hwi_process_default_pdu_ring (
    PACIT_CONTROLLER_WS     pcontroller,
    PI_T_DPDU_CQE           pdpdu_cqe)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    PASYNC_PDU_HANDLE       pasync_handle = ACIT_NULL;
    ACIT_UEXACT32           cq_index = ACIT_U32_MINUS_ONE;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    // The EP fw gives a CQ entry indicating completion on either the default 
    // PDU header ring the default PDU buffer ring. The CQ entry will contain
    // the index and the physical address of the entry. 
    // Get our reference from the CQ entry..
    pasync_handle = hwi_get_async_handle(pcontroller, pasync_ctx, pdpdu_cqe, &cq_index);
    OSM_ASSERT(pasync_handle);

    // If this entry is not yet consumed, then in addition to a completion,
    // it also means that buffers upto the cq_index have been consumed.
    // We can post fresh buffers back if there are atleast 8 free.
    // If this entry is already consumed, then this is only a completion. It
    // doesn't give us a chance to re-post any buffers
    if(pasync_handle->consumed == 0)
    {
        hwi_update_async_writables(pasync_ctx, 
                                   pasync_handle->is_header, 
                                   cq_index);
    }

    // gather the default entry for the connection and forward it to iSCSI for
    // processing. This function is where the element is removed from busy list
    // and added to the wait list
    hwi_gather_async_pdu(pcontroller, pasync_handle);

    // Attempt to post new entries back to the ring. Note that we call
    // the routine to post irrespective of the consumed or completed-only
    // status. This is because though a completed-only status does not
    // update the writables, it does free up some buffers which could
    // mean that we can post. 
    // (Look in hwi_post_async_buffers for more info on posting rules!)
    hwi_post_async_buffers(pcontroller, pasync_handle->is_header);

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
hwi_flush_default_pdu_buffer (
    PACIT_CONTROLLER_WS     pcontroller,
    PI_T_DPDU_CQE           pdpdu_cqe)
{
    PHWI_CONTROLLER_WS      phwi;
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx;
    PASYNC_PDU_HANDLE       pasync_handle = ACIT_NULL;
    ACIT_UEXACT32           cq_index = ACIT_U32_MINUS_ONE;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi);

    //
    // This function is called when there is a default PDU buffer digest 
    // (data digest error). The driver consumes the entry and drops
    // the contents
    //

    // The EP fw gives a CQ entry indicating completion on either the default 
    // PDU header ring the default PDU buffer ring. The CQ entry will contain
    // the index and the physical address of the entry. 
    // Get our reference from the CQ entry..
    pasync_handle = hwi_get_async_handle(pcontroller, pasync_ctx, pdpdu_cqe, &cq_index);
    OSM_ASSERT(pasync_handle);

    if(pasync_handle->is_header == 0)
    {

        // If this entry is not yet consumed, then in addition to a completion,
        // it also means that buffers upto the cq_index have been consumed.
        // We can post fresh buffers back if there are atleast 8 free.
        // If this entry is already consumed, then this is only a completion. It
        // doesn't give us a chance to re-post any buffers
        if(pasync_handle->consumed == 0)
        {
            hwi_update_async_writables(pasync_ctx, 
                                       pasync_handle->is_header, 
                                       cq_index);
        }

        // there will not be a consumer for this CQE. The connection's default PDU
        // header and buffer should be simply dropped. So release the buffers
        hwi_free_async_msg(pcontroller, pasync_handle->cri);

        // Attempt to post new entries back to the ring. Note that we call
        // the routine to post irrespective of the consumed or completed-only
        // status. This is because though a completed-only status does not
        // update the writables, it does free up some buffers which could
        // mean that we can post. 
        // (Look in hwi_post_async_buffers for more info on posting rules!)
        hwi_post_async_buffers(pcontroller, pasync_handle->is_header);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Looking for default PDU buffer, header found instead!!\n"));
        OSM_ASSERT(0);
    }

    return ACIT_STATUS_SUCCESS;
}
