/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
    iscsiwrb.cpp

@brief
    Shared code entry points for creating and managing iSCSI per-connection 
    resources such as CIDs ands WRB rings.

@author jflores

@note
    jflores  10/24/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/iscsiwrb_ll.c,v 1.27 2006/04/09 23:32:54 menz Exp $

*/

#include "pch.h"

BESTATUS
be_iscsi_wrb_queue_create (
    PBE_FUNCTION_OBJECT        function_object,    
    PSA_SGL                    sgl,
    u32                        length,
    PBE_ISCSI_WRB_QUEUE_OBJECT wrbq)
{
    // map the ring into the UT    
    BESTATUS status;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_ISCSI_WRBQ_CREATE *ioctl = NULL;

    FUNCTION_ASSERT (function_object);
    ASSERT (be_function_is_iscsi (function_object));
    ASSERT (wrbq);

    be_lock_wrb_post (function_object);

    sa_zero_mem (wrbq, sizeof (*wrbq));

    wrbq->magic = BE_ISCSI_WRB_QUEUE_MAGIC;
    wrbq->parent_function = function_object;
    wrbq->cid = 0xffffffff;
    wrbq->sgl = sgl;
    wrbq->length = length;

    _be_function_add_wrbq (function_object, wrbq);
    
    TRACE (DL_INFO, "Create iscsi wrb queue. length:%d pages:%d", length, sa_sgl_get_page_count (sgl));
   /* status = be_ut_map_ring (function_object, 
                             cid, 
                             CRA_SEND_TYPE, 
                             sgl,
                             length,
                             &wrbq->ut_mapping);*/

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_WRBQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create iscsi failed.");
        goto error;
    }

    wrbq->cid = ioctl->params.response.cid;

    if (status != BE_SUCCESS) {
        sa_zero_mem (wrbq, sizeof (*wrbq));
    }

error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_iscsi_wrb_queue_destroy (
    PBE_ISCSI_WRB_QUEUE_OBJECT wrbq)
{
    BESTATUS status = BE_SUCCESS;
    ISCSI_WRBQ_ASSERT (wrbq);

    // Send ioctl to destroy the queue.
    status = be_function_ring_destroy (wrbq->parent_function, wrbq->cid, IOCTL_RING_TYPE_ISCSI_WRBQ);
    ASSERT (status == 0);

    // Dereference
    _be_function_remove_wrbq (wrbq->parent_function, wrbq);

    // Zero tracking object
    sa_zero_mem (wrbq, sizeof (*wrbq));

    return status;
}


u32 
be_iscsi_wrbq_get_cid (
    IN PBE_ISCSI_WRB_QUEUE_OBJECT wrbq)
{
    ISCSI_WRBQ_ASSERT (wrbq);
    return wrbq->cid;
}

BESTATUS
be_iscsi_post_sgl_pages (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PSA_SGL sgl,
    IN u32 page_offset,
    IN u32 num_pages
    )
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_ISCSI_CFG_POST_SGL_PAGES *ioctl = NULL;
    u32 i = 0;
    u32 internal_page_offset = 0;
    u32 temp_num_pages = num_pages;

    if(num_pages == 0xff)
	    num_pages = 1;

    be_lock_wrb_post (function_object);

    // Tracking
    function_object->config.num_sgl += num_pages;

    // Post numerous ioctls to keep each one small enuf to be embedded in 
    // the WRB.
    do {

        wrb = be_function_peek_mcc_wrb (function_object);
        if (!wrb) {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_CFG_POST_SGL_PAGES);

        ioctl->params.request.num_pages = 
            MIN (num_pages, 
                 SA_NUMBER_OF_FIELD (IOCTL_COMMON_ISCSI_CFG_POST_SGL_PAGES, params.request.pages));

        TRACE (DL_INFO, "Create iSCSI map SGL ioctl. num_pages:%d page_offset:%d",
               ioctl->params.request.num_pages, page_offset);       
        ioctl->params.request.page_offset = page_offset;

        for (i=0; i<ioctl->params.request.num_pages; i++) {
            u64 pa = sa_sgl_get_page (sgl, internal_page_offset + i);

            ioctl->params.request.pages[i].lo = sa_lo (pa);
            ioctl->params.request.pages[i].hi = sa_hi (pa);
        }

        // Add this group to the offset.  Subtract from remaining pages to map.
        internal_page_offset += ioctl->params.request.num_pages;        
        page_offset += ioctl->params.request.num_pages;        
        num_pages -= ioctl->params.request.num_pages;

	if(temp_num_pages == 0xff)
		ioctl->params.request.num_pages = temp_num_pages;

        // Post the Ioctl
        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "ioctl to map iscsi frags failed.");
            goto Error;
        }

    } while (num_pages > 0);

Error:

    be_unlock_wrb_post (function_object);

    if (status != 0) {
        // Remove all pages that were successfully posted.
        be_iscsi_remove_sgl_pages (function_object);
    }

    return status;
}



BESTATUS
be_iscsi_remove_sgl_pages (
    PBE_FUNCTION_OBJECT function_object)
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_ISCSI_CFG_REMOVE_SGL_PAGES *ioctl = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_CFG_REMOVE_SGL_PAGES);

    // Tracking
    function_object->config.num_sgl = 0;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to unmap iscsi frags failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_iscsi_cxn_create (
    PBE_FUNCTION_OBJECT function_object,
    IN  PBE_ISCSI_WRB_QUEUE_OBJECT         wrbq,
    IN  PBE_DEFAULT_PDU_QUEUE_OBJECT       defq,
    IN  PBE_CQ_OBJECT                      cq,
    OUT PBE_ISCSI_CONNECTION_OBJECT        cxn)
{
    FUNCTION_ASSERT (function_object);
    ISCSI_WRBQ_ASSERT (wrbq);
    ISCSI_DEFQ_ASSERT (defq);
    CQ_ASSERT (cq);
    ASSERT (cxn);

    SA_ZERO_MEM (cxn);

    cxn->magic = BE_ISCSI_CONNECTION_MAGIC;
    cxn->cq = cq;
    cxn->defq = defq;
    cxn->parent_function = function_object;
    cxn->wrbq = wrbq;    

    be_cq_object_reference (cq);

    _be_function_add_iscsi_connection (function_object, cxn);

    return BE_SUCCESS;
}

void
be_iscsi_cxn_destroy (
    PBE_ISCSI_CONNECTION_OBJECT cxn)
{
    ISCSI_CXN_ASSERT (cxn);

    // dereference cq
    be_cq_object_dereference (cxn->cq);
    
    _be_function_remove_iscsi_connection (cxn->parent_function, cxn);

    sa_zero_mem (cxn, sizeof (*cxn));
}


BESTATUS 
be_iscsi_offload_from_tcp_stack (
    IN  PBE_ISCSI_WRB_QUEUE_OBJECT         wrbq,
    IN  PBE_DEFAULT_PDU_QUEUE_OBJECT       defq,
    IN  PBE_CQ_OBJECT                      cq,
    IN  IOCTL_LOWLEVEL_ISCSI_OFFLOAD*      va_for_ioctl, // ISCSI_CONTEXT, TEMPLATE_HEADER, TCP_CONTEXT & OFFLOAD_CXN_ENTRY must be set
    IN  u64                                pa_for_ioctl,
    OUT PBE_ISCSI_CONNECTION_OBJECT        cxn,
    IN  MCC_WRB_CQE_CALLBACK               callback,
    IN  PVOID                              callback_context)
{
    BESTATUS status = 0;
#if 0
    MCC_WRB* wrb = NULL;
    ISCSI_CONTEXT *context = NULL;
    BE_FUNCTION_OBJECT *function_object = NULL;
    
    ISCSI_WRBQ_ASSERT (wrbq);
    ISCSI_DEFQ_ASSERT (defq);
    CQ_ASSERT (cq);
    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);
    ASSERT (cxn);

    function_object = wrbq->parent_function;
    
    // Zero the cxn    
    status = be_iscsi_cxn_create (function_object, wrbq, defq, cq, cxn);
    if (status != BE_SUCCESS) {
        return status;
    }

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }
    
    BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, 
                                           wrb, 
                                           va_for_ioctl, 
                                           pa_for_ioctl, 
                                           LOWLEVEL_ISCSI_OFFLOAD,
                                           0);

    // Get the cid from the previously allocate SQ.
    va_for_ioctl->params.request.cid = be_iscsi_wrbq_get_cid (wrbq);

    // We fill in portions of iSCSI context
    context = &va_for_ioctl->params.request.context;

    SA_SET_BITFIELD (context->ring_context.o,      TRUE);
    SA_SET_BITFIELD (context->ring_context.pdid,   be_function_get_pd_number (function_object));
    SA_SET_BITFIELD (context->mode,                ISCSI_INITIATOR);
    SA_SET_BITFIELD (context->tx_wrbindex_drv_msg, 0xff);
    SA_SET_BITFIELD (context->rx_wrbindex_drv_msg, 0xff);
    SA_SET_BITFIELD (context->defaultQueueID,      be_iscsi_defq_get_id (defq));
    SA_SET_BITFIELD (context->complQueueID,        be_cq_get_id (cq));
    SA_SET_BITFIELD (context->domainID,            be_function_get_pd_number (function_object));

    SA_SET_BITFIELD (va_for_ioctl->params.request.filter.ct, ENUM_BE_RDMA_OFFLOAD);
    SA_SET_BITFIELD (va_for_ioctl->params.request.filter.valid, 1);

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, callback, callback_context, va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "iscsi offload ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
#endif 
    return status;
}


BESTATUS BECALL
be_iscsi_terminate (
    IN      PBE_ISCSI_CONNECTION_OBJECT cxn, 
    IN      MCC_WRB_CQE_CALLBACK        callback, 
    IN      PVOID                       callback_context)
{
    // BUG -- Poor man's implementation for iscsi cxn terminate

    return 0;
}





