/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
    rxf.cpp

@brief
    Manages access to BladeEngine RXF block.
    
@author jflores

@note
    jflores   8/03/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/rxf_ll.c,v 1.17 2006/04/09 23:32:54 menz Exp $

*/

#include "pch.h"

// LANCER TODO: Remove once LANCER specific defs find their right place
#include "lancer_temp.h"

/*!

@brief
    This routine adds, removes, or reads a WoL magic packet.

@param        
    FunctionObject  - Function object handle.

@param        
    Mac1            - Set to TRUE if this function will operate on the MAC 1 Magic Packet.
    
@param        
    Operation       - Operation to perform.
    
@param        
    Pattern         - Pattern to set, or the pattern returned.

@param        
    PatternMask     - Pattern mask to set, or the pattern mask returned.
            
@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL

*/                               
BESTATUS 
be_rxf_wake_on_lan_config (
    IN PBE_FUNCTION_OBJECT  function_object,
    IN boolean              enable_port0,
    IN boolean              enable_port1,
    IN boolean              magic_packet_enable,
    IN u32                  index,
    IN PVOID                pattern,
    IN PVOID                pattern_mask,
    IN MCC_WRB_CQE_CALLBACK callback,          // optional
    IN PVOID                callback_context,
    IN PBE_WAKE_ON_LAN_QUEUE_CONTEXT  queue_context      // optional
    )  // optional
{
    IOCTL_ETH_ACPI_CONFIG *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = BE_SUCCESS;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);

    // Valid range for index is 0-3.
    ASSERT (index < 4);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_ACPI_CONFIG);

    ioctl->params.request.index = index;
    ioctl->params.request.port0 = enable_port0;
    ioctl->params.request.port1 = enable_port1;
    ioctl->params.request.magic_packet = magic_packet_enable;

    if (enable_port0 || enable_port1) {

        if (pattern) {
            sa_memcpy (ioctl->params.request.byte_pattern, pattern, sizeof (ioctl->params.request.byte_pattern));
        }
        if (pattern_mask) {
            sa_memcpy (ioctl->params.request.bit_mask, pattern_mask, sizeof (ioctl->params.request.bit_mask));
        }

        // Track which indicies are enabled/disabled
        function_object->config.wol_bitmask |= (1 << index);
    } else {

        // Track which indicies are enabled/disabled
        function_object->config.wol_bitmask &= ~(1 << index);
    }
    
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "wake-on-lan ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return  status;
}



/*!
@brief
    This routine gets or sets a mac address for a domain  given the port and mac.
@param        
    FunctionObject  - Function object handle.
@param        
    Port1           - Set to TRUE if this function will set/get the Port 1 address.
                      Only the host may set this to TRUE.
@param        
    Mac1            - Set to TRUE if this function will set/get the MAC 1 address.
                      Only the host may set this to TRUE.
@param        
    Write           - Set to TRUE if this function should write the mac address.
@param        
    MacAddress      - Buffer of the mac address to read or write.
@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
@note
    IRQL: < DISPATCH_LEVEL
*/
BESTATUS 
be_rxf_mac_address_read_write (
    IN PBE_FUNCTION_OBJECT  function_object,
    IN     boolean              port1,            // VM must always set to false
    IN     boolean              mac1,             // VM must always set to false
    IN     boolean              mgmt,
    IN     boolean              write,
    IN     boolean              permanent,
    IN     PBE_RXIF_OBJECT      rxif_object,
    IN OUT PSA_MAC_ADDRESS      mac_address,
    IN     MCC_WRB_CQE_CALLBACK callback,         // optional
    IN     PVOID                callback_context) // optional
{
    BESTATUS status = BE_SUCCESS;
    union {
        IOCTL_COMMON_NTWK_MAC_QUERY *query;
        IOCTL_COMMON_NTWK_MAC_SET *set;
    } ioctl = {0};
    MCC_WRB *wrb = NULL;
    u32 type = 0;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    ASSERT (mac_address);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    if (mgmt) {
        type = MAC_ADDRESS_TYPE_MANAGEMENT;
    } else if (be_function_is_vm (function_object) &&
               be_function_is_networking (function_object)) {
        type = MAC_ADDRESS_TYPE_PD;
    } else {
        if (be_function_is_networking (function_object)) {
            type = MAC_ADDRESS_TYPE_NETWORK;
        } else {
            type = MAC_ADDRESS_TYPE_STORAGE;
        }
    }

    if (write) {
        ASSERT( rxif_object );

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl.set = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MAC_SET);
        ioctl.set->params.request.invalidate = 0;
        ioctl.set->params.request.type = type;
        ioctl.set->params.request.if_id = rxif_object->handle;

        // Copy the mac address to set.
        ioctl.set->params.request.mac.SizeOfStructure = sizeof (ioctl.set->params.request.mac);
        sa_memcpy (ioctl.set->params.request.mac.MACAddress, mac_address, sizeof(*mac_address));

        // Post the Ioctl
        status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 callback,
                 callback_context,
                 ioctl.set);
    
    } else {
        
        // Prepares an embedded ioctl, including request/response sizes.
        ioctl.query = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MAC_QUERY);
        ioctl.query->params.request.type = type;
        ioctl.query->params.request.permanent = permanent;
        if( !permanent ) {
            ASSERT(rxif_object);
            ioctl.query->params.request.if_id = rxif_object->handle;
        }

        // Post the Ioctl (with a copy for the response)
        status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL, // queue context
                 callback,
                 callback_context,
                 ioctl.query,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_MAC_QUERY, params.response.mac.MACAddress, mac_address));
    }

    if (status < 0) {
        TRACE (DL_ERR, "mac set/query failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

//TODO_BE2: Review
BESTATUS
be_rxf_read_write_pd_mac (
   IN PBE_FUNCTION_OBJECT                 function_object,
   IN BE_READ_WRITE_MAC_PARAMETERS*       parameters,
   IN OUT PSA_MAC_ADDRESS                 mac_address,
   IN MCC_WRB_CQE_CALLBACK                callback,
   IN PVOID                               callback_context,
   IN PBE_PD_READ_WRITE_MAC_QUEUE_CONTEXT queue_context
)
{
    BESTATUS status = BE_SUCCESS;
    union {
        IOCTL_COMMON_NTWK_MAC_QUERY *query;
        IOCTL_COMMON_NTWK_MAC_SET *set;
    } ioctl = {0};
    MCC_WRB *wrb = NULL;
    u32 type = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    ASSERT (mac_address);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    if (be_function_is_networking (function_object)) {
        type = MAC_ADDRESS_TYPE_NETWORK;
    } else {
        type = MAC_ADDRESS_TYPE_STORAGE;
    }

    if (parameters->write) {
        // Prepares an embedded ioctl, including request/response sizes.
        ioctl.set = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MAC_SET);

        ioctl.set->header.request.domain = parameters->pdid;
        ioctl.set->params.request.invalidate = 0;
        ioctl.set->params.request.type = type;
        
        // Copy the mac address to set.
        ioctl.set->params.request.mac.SizeOfStructure = sizeof (ioctl.set->params.request.mac);
        sa_memcpy (ioctl.set->params.request.mac.MACAddress, mac_address, sizeof(*mac_address));

        // Post the Ioctl
        status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl.set);
    
    } else {
        
        // Prepares an embedded ioctl, including request/response sizes.
        ioctl.query = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MAC_QUERY);
        
        ioctl.query->header.request.domain = parameters->pdid;
        ioctl.query->params.request.type = type;
        ioctl.query->params.request.permanent = parameters->permanent;

        // Post the Ioctl (with a copy for the response)
        status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context, // queue context
                 callback,
                 callback_context,
                 ioctl.query,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_MAC_QUERY, params.response.mac.MACAddress, mac_address));
    }

    if (status < 0) {
        TRACE (DL_ERR, "mac set/query failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

// retrieves the MAC Address corresponding to given mac_id.

BESTATUS
be_rxf_get_mac_with_id (
    IN PBE_FUNCTION_OBJECT      function_object,
    IN     u32                  mac_id,            // VM must always set to false
    IN     PBE_RXIF_OBJECT      rxif_object,
    IN OUT PSA_MAC_ADDRESS      mac_address,
    IN     MCC_WRB_CQE_CALLBACK callback,         // optional
    IN     PVOID                callback_context) // optional
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_NTWK_MAC_QUERY *ioctl;
    MCC_WRB *wrb = NULL; 
    FUNCTION_ASSERT (function_object);
    be_lock_wrb_post (function_object);
    ASSERT (mac_address);
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MAC_QUERY);
    ioctl->params.request.type = MAC_ADDRESS_TYPE_NETWORK;
    ioctl->params.request.permanent = 0;
    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->params.request.pmac_id = mac_id;
    // Post the Ioctl (with a copy for the response)
    status = be_function_post_mcc_wrb_with_copy (
                            function_object,
                            wrb,
                            NULL, // queue context
                            callback,
                            callback_context,
                            ioctl,
                            BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_MAC_QUERY, params.response.mac.MACAddress, mac_address));
    if (status < 0) {
        TRACE (DL_ERR, "COMMON_NTWK_MAC_QUERY failed.mac_id:%d",mac_id);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;

}



//---------------------------------------------------------------------------------------
// Function: be_rxf_multicast_config
//
//   Writes all multicast addresses for this protection domain. This function is called as
//   part of be_function_cleanup to remove all multicast filters. 
//
// function_object    - 
// rxif_object        - Receive object to use for the multicast addresses. 
// promiscuous        - TRUE, enables promiscuous mode where all multicast packets are
//                      passed up. 
// num                - Number of mac addresses in the mac_table list. There are a maximum
//                      of 32 valid addresses. Set num to 0 to delete multicast settings
//                      associated with the receive object. 
// mac_table          - List of multicast MAC addresses to associate with the receive
//                      object. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback function context. 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_multicast_config (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN PBE_RXIF_OBJECT             rxif_object,
   IN boolean                     promiscuous,
   IN u32                         num,
   IN SA_MAC_ADDRESS*             mac_table,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_MULTICAST_QUEUE_CONTEXT queue_context )
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_NTWK_MULTICAST_SET *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);
    RXIF_ASSERT (rxif_object);

    ASSERT (num <= SA_NUMBER_OF_FIELD (IOCTL_COMMON_NTWK_MULTICAST_SET, params.request.mac));
    
    if (num > SA_NUMBER_OF_FIELD (IOCTL_COMMON_NTWK_MULTICAST_SET, params.request.mac)) {
        TRACE (DL_ERR, 
               "Too many multicast addresses. BE supports %d.", 
               SA_NUMBER_OF_FIELD (IOCTL_COMMON_NTWK_MULTICAST_SET, params.request.mac));
        return BE_NOT_OK;
    }

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_MULTICAST_SET);

    TRACE (DL_INFO, "multicast config. function:%d promiscuous:%d num:%d",
           be_function_get_function_number (function_object),
           promiscuous,
           num);

    ioctl->params.request.promiscuous = promiscuous;
    ioctl->params.request.interface_id = rxif_object->handle;
    if (!promiscuous) {
        
        ioctl->params.request.num_mac = num;

        if (num > 0) {
            u32 i=0;

            ASSERT (mac_table);        
            sa_memcpy (ioctl->params.request.mac, mac_table, 6 * num);
            for (i=0; i<num; i++) {
                TRACE (DL_VERBOSE, " multicast address[%d]: " SA_MAC_FORMAT, SA_MAC_ARGS (mac_table[i]));
            }
        }

        // track number of addresses
        function_object->config.num_multicast = num;
    } else {
        // track number of addresses
        function_object->config.num_multicast = 0xFF; // infinite
    }
    
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object, 
                 wrb, 
                 generic_context, 
                 callback, 
                 callback_context, 
                 ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "multicast ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_rxf_vlan_config
//
//   Writes all vlan tags for this protection domain. This function is called as part of
//   be_function_cleanup to remove all vlan filters. 
//
// function_object    - 
// rxif_object        - Optional interface object that VLAN tags are applied to. If NULL,
//                      the default (i.e. first) RXF interface is used. For iSCSI, NULL is
//                      always suitable. 
// promiscuous        - TRUE, enables vlan promiscuous mode where all vlan packets are
//                      passed up the stack. Allows supporting >32 vlan tags. 
// untagged           - TRUE, enables untagged mode for the interface. FALSE, disables
//                      untagged mode for the interface. 
// num                - Number of vlan tags in the array.
// vlan_tag_array     - 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_vlan_config (
   IN PBE_FUNCTION_OBJECT    function_object,
   IN PBE_RXIF_OBJECT        rxif_object,
   IN boolean                promiscuous,
   IN boolean                untagged,
   IN u32                    num,
   IN u16*                   vlan_tag_array,
   IN MCC_WRB_CQE_CALLBACK   callback,
   IN PVOID                  callback_context,
   IN PBE_VLAN_QUEUE_CONTEXT queue_context )
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_NTWK_VLAN_CONFIG *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
   
    FUNCTION_ASSERT (function_object);
    
    ASSERT (num <= SA_NUMBER_OF_FIELD (IOCTL_COMMON_NTWK_VLAN_CONFIG, params.request.tags.normal_vlan));
    if (num > SA_NUMBER_OF_FIELD (IOCTL_COMMON_NTWK_VLAN_CONFIG, params.request.tags.normal_vlan)) {
        TRACE (DL_ERR, "Too many VLAN tags.");
        return BE_NOT_OK;
    }

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes. IOCTL_COMMON_NTWK_VLAN_CONFIG
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_VLAN_CONFIG);

    ioctl->header.request.domain = rxif_object->domain;

    TRACE (DL_INFO, "vlan config. function:%d promiscuous:%d num:%d",
           be_function_get_function_number (function_object),
           promiscuous,
           num);

    if (rxif_object) {
        ioctl->params.request.interface_id = rxif_object->handle;
    } else {
        ioctl->params.request.interface_id = _be_function_default_rxif_id (function_object);
    }
    ioctl->params.request.untagged = untagged;
    ioctl->params.request.promiscuous = promiscuous;
    if (!promiscuous) {

        ioctl->params.request.num_vlan = num;    

        if (num > 0) {
            ASSERT (vlan_tag_array);        
            sa_memcpy (ioctl->params.request.tags.normal_vlan, vlan_tag_array, num * sizeof (vlan_tag_array[0]));
        }

        // Track number of tags
        //function_object->config.num_vlan = num;

    } else {
        // promiscuous mode, tracking number is set inifinite
        //function_object->config.num_vlan = 0xFF;
    }

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "vlan ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
  


/*!
@brief
    This routine assigns a CPU number to an RSS table.
@param        
    FunctionObject      - Function object handle.
@param        
    RssEntry            - The RSS entry (bucket) to assign this CPU to.
@param        
    Cpu                 - The CPU number that incomming packets that hash to
                          the RssEntry bucket index will be directed to.
@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL
*/
BESTATUS 
be_rxf_rss_config (
    IN PBE_FUNCTION_OBJECT   function_object,
    IN PBE_RXIF_OBJECT       rxif_object,
    IN u32                   rss_type,         // use enumeration ENABLE_RSS_ENUM
    IN boolean               flush,            // generate flush completions                 
    IN PVOID                 hash,             // 16 bytes
    IN u32                   rss_cpu_table_length, // bytes (power of 2 from 2 to 128)
    IN u8*                   cpu_table,        // 2 to 128 bytes
    IN MCC_WRB_CQE_CALLBACK  callback,         // optional
    IN PVOID                 callback_context,
    IN PBE_RSS_QUEUE_CONTEXT queue_context)    // optional
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_ETH_RSS_CONFIG *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    u32 cpu_table_length = rss_cpu_table_length;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);
    
    ASSERT (!be_function_is_vm (function_object));
    ASSERT (rxif_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_RSS_CONFIG);    

    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->params.request.enable_rss = rss_type;
    ioctl->params.request.flush = flush;

    if (rss_type != RSS_ENABLE_NONE) {
        ASSERT (hash);
    }

    if (hash) {
        sa_memcpy (ioctl->params.request.hash, hash, sizeof (ioctl->params.request.hash));        
    }

    if (cpu_table_length > sizeof (ioctl->params.request.cpu_table) ) {
        cpu_table_length = sizeof (ioctl->params.request.cpu_table);
    }

    // Double check the table.  Each entry should be 0 to num_cq-1 corresponding to the
    // index in the cq_id_array to use.
    if (cpu_table) {

        ASSERT (cpu_table);
        ASSERT (cpu_table_length <= sizeof (ioctl->params.request.cpu_table));

        ioctl->params.request.cpu_table_size_log2 = cpu_table_length > 0 ? sa_log2 (cpu_table_length) : 1;
        ASSERT (ioctl->params.request.cpu_table_size_log2 > 0 &&
                ioctl->params.request.cpu_table_size_log2 <= 7);

        sa_memcpy (ioctl->params.request.cpu_table, cpu_table, sizeof (ioctl->params.request.cpu_table));
    }

    TRACE (DL_INFO, "RSS config. type:%s cpu_table_len:%d",
           ((ioctl->params.request.enable_rss == RSS_ENABLE_NONE           ? "none" :
            (ioctl->params.request.enable_rss == RSS_ENABLE_IPV4             ? "ip" :
            (ioctl->params.request.enable_rss == RSS_ENABLE_TCP_IPV4         ? "tcp" :
            (ioctl->params.request.enable_rss == (RSS_ENABLE_IPV4 | RSS_ENABLE_TCP_IPV4) ? "ip|tcp" : "??"))))
           ),
           cpu_table_length);

    TRACE (DL_INFO, "RSS hash. 0x%08x 0x%08x 0x%08x 0x%08x",
            ioctl->params.request.hash[0],
            ioctl->params.request.hash[1],
            ioctl->params.request.hash[2],
            ioctl->params.request.hash[3]);

    TRACE (DL_INFO, "RSS table. log2(len):%d 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
            ioctl->params.request.cpu_table_size_log2,
            ioctl->params.request.cpu_table[0],
            ioctl->params.request.cpu_table[1],
            ioctl->params.request.cpu_table[2],
            ioctl->params.request.cpu_table[3],
            ioctl->params.request.cpu_table[4],
            ioctl->params.request.cpu_table[5],
            ioctl->params.request.cpu_table[6],
            ioctl->params.request.cpu_table[7]);

    // software tracking
    function_object->config.rss_type = rss_type;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "rss ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_rxf_link_status (
    IN  PBE_FUNCTION_OBJECT           function_object,
    OUT BE_LINK_STATUS*               link_status,
    IN  MCC_WRB_CQE_CALLBACK          callback,      // optional
    IN  PVOID                         callback_context,
    IN  PBE_LINK_STATUS_QUEUE_CONTEXT queue_context) // optional
{
    IOCTL_COMMON_NTWK_LINK_STATUS_QUERY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (link_status);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_LINK_STATUS_QUERY);

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                generic_context, // Queue context
                callback,
                callback_context,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_LINK_STATUS_QUERY, params.response, link_status));

    if (status < 0) {
        TRACE (DL_ERR, "link status ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_rxf_link_status_v1 (
    IN  PBE_FUNCTION_OBJECT           function_object,
    OUT BE_LINK_STATUS*               link_status,
    IN  MCC_WRB_CQE_CALLBACK          callback,      // optional
    IN  PVOID                         callback_context,
    IN  PBE_LINK_STATUS_QUEUE_CONTEXT queue_context) // optional
{
    IOCTL_COMMON_NTWK_LINK_STATUS_QUERY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (link_status);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_LINK_STATUS_QUERY);

    ioctl->header.request.version = 1;

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                generic_context, // Queue context
                callback,
                callback_context,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_LINK_STATUS_QUERY, params.response, link_status));

    if (status < 0) {
        TRACE (DL_ERR, "link status ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_rxf_set_link_speed_v1 (
    IN  PBE_FUNCTION_OBJECT           function_object,
    IN  u32                           phy_port,
    IN  u16                           mac_speed
    ) 
{
    IOCTL_COMMON_NTWK_SET_LINK_SPEED_V1_V1* ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
    TRACE (DL_ERR, "MCC wrb peek failed.in be_rxf_set_link_speed_v1");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_SET_LINK_SPEED_V1_V1);

    ioctl->header.request.version = 1;
    ioctl->params.request.physical_port = (u8)phy_port;
    ioctl->params.request.mac_speed = mac_speed;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                     function_object,
                     wrb,
                     NULL, // callback
                     NULL, // context 
                     ioctl);

        if (status != 0) {
            TRACE (DL_ERR, "be_rxf_set_link_speed_v1 ioctl failed.");
            goto error;
        }

error:
    be_unlock_wrb_post (function_object);

    return status;
}


BESTATUS
be_rxf_query_eth_statistics (
    IN PBE_FUNCTION_OBJECT         function_object,
    IN IOCTL_ETH_GET_STATISTICS*   va_for_ioctl,
    IN u64                         pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK        callback,
    IN PVOID                       callback_context,
    IN PBE_ETH_STATS_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query eth stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.    
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, ETH_GET_STATISTICS, 0);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "eth stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

#ifdef BECLIB_BE3_SUPPORT

BESTATUS
be_rxf_query_eth_statistics_v1 (
    IN PBE_FUNCTION_OBJECT         function_object,
    IN IOCTL_ETH_GET_STATISTICS_V1*  va_for_ioctl,
    IN u64                         pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK        callback,
    IN PVOID                       callback_context,
    IN PBE_ETH_STATS_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query eth stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.    
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, ETH_GET_STATISTICS_V1, 1);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "eth stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

#endif



BESTATUS
be_rxf_promiscuous (
    IN PBE_FUNCTION_OBJECT           function_object,
    IN boolean                       enable_port0,
    IN boolean                       enable_port1,
    IN MCC_WRB_CQE_CALLBACK          callback,
    IN PVOID                         callback_context,
    IN PBE_PROMISCUOUS_QUEUE_CONTEXT queue_context)
{    
    IOCTL_ETH_PROMISCUOUS *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_PROMISCUOUS);

    ioctl->params.request.port0_promiscuous = enable_port0;
    ioctl->params.request.port1_promiscuous = enable_port1;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl
                 );

    if (status < 0) {
        TRACE (DL_ERR, "promiscuous ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

//<no_inbox>
BESTATUS
be_query_toe_statistics (
    IN PBE_FUNCTION_OBJECT         function_object,
    IN IOCTL_TOE_GET_STATISTICS*   va_for_ioctl,
    IN u64                         pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK        callback,
    IN PVOID                       callback_context,
    IN PBE_TOE_STATS_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query toe stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, TOE_GET_STATISTICS, 0);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "toe stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_query_tcp_statistics (
    IN PBE_FUNCTION_OBJECT              function_object,
    IN IOCTL_COMMON_TCP_GET_STATISTICS* va_for_ioctl,
    IN u64                              pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK             callback,
    IN PVOID                            callback_context,
    IN PBE_TCP_STATS_QUEUE_CONTEXT      queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query tcp stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, COMMON_TCP_GET_STATISTICS, 0);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "tcp stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
//</no_inbox>

//---------------------------------------------------------------------------------------
// Function: be_rxf_filter_config
//
//   Configures BladeEngine ethernet receive filter settings. 
//
// function_object    - 
// rxif_object        - Previously created RXIF object if receive interface options are
//                      changed. 
// settings           - Pointer to the requested filter settings. The response from
//                      BladeEngine will be placed back in this structure. 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_filter_config (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN OUT PBE_RXIF_OBJECT          rxif_object,
   IN OUT NTWK_RX_FILTER_SETTINGS* settings,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_RXF_FILTER_QUEUE_CONTEXT queue_context )
{
    IOCTL_COMMON_NTWK_RX_FILTER *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (settings);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_RX_FILTER);
    ioctl->header.request.domain = rxif_object->domain;

    sa_memcpy (&ioctl->params.request, settings, sizeof(*settings));
    if (ioctl->params.request.if_flags_mask) {
        // Receive interface options changing
        RXIF_ASSERT(rxif_object);        
        ioctl->params.request.if_id = rxif_object->handle;
    } else {
        ioctl->params.request.if_id = (UEXACT32)-1;
    }

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                generic_context, // Queue context
                callback,
                callback_context,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_RX_FILTER, params.response, settings));

    if (status < 0) {
        TRACE (DL_ERR, "RXF/ERX filter config ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

void be2_rxf_if_create_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_RXIF_OBJECT rxif_object; 
    BE_FUNCTION_OBJECT *function_object;
    IOCTL_COMMON_NTWK_INTERFACE_CREATE *ioctl = NULL;

    rxif_object = (PBE_RXIF_OBJECT)context;
    RXIF_ASSERT (rxif_object);

    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);

    if (status == BE_SUCCESS) {

        ASSERT (wrb);
        ASSERT (wrb->embedded);

        ioctl = (IOCTL_COMMON_NTWK_INTERFACE_CREATE *)&wrb->payload;

        rxif_object->handle = (u32)ioctl->params.response.interface_id;
        rxif_object->mac_id = (u32)ioctl->params.response.pmac_id;

        // Add the interface object to the function object
        _be_function_add_rxif (function_object, rxif_object);        
    } else {        
        TRACE (DL_WARN, "RXF interface create ioctl failed from callback.");
        SA_ZERO_MEM(rxif_object);
    }
}

//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_create
//
//   Creates a receive interface object. 
//
// function_object    - Previously created function object. 
// params             - Function-specific parameters. 
// rxif_object        - Pointer to an opaque object representing the interface 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_create (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN u32                          domain,
   IN PBE2_RXF_IF_CREATE_PARAMS    params,
   OUT PBE_RXIF_OBJECT             rxif_object,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT     *queue_context)
{
    IOCTL_COMMON_NTWK_INTERFACE_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};

    FUNCTION_ASSERT (function_object);
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_INTERFACE_CREATE);

    ioctl->header.request.domain = domain;

    sa_memcpy (&ioctl->params.request.rxf_if_create_params, params, sizeof(*params));

    // Init RXIF object.
    SA_ZERO_MEM (rxif_object);
    rxif_object->magic           = BE_RXIF_MAGIC;
    rxif_object->parent_function = function_object;
    rxif_object->capability_flags           = ioctl->params.request.rxf_if_create_params.capability_flags;
    rxif_object->enabled_flags              = ioctl->params.request.rxf_if_create_params.enable_flags;
    rxif_object->domain = domain;

    // Post the ioctl
    status = be_function_post_mcc_wrb_complete (
                function_object,
                wrb,
                generic_context,   // queue context
                callback,
                callback_context,
                be2_rxf_if_create_internal_callback,
                rxif_object, 
                ioctl,
                response_copy);

    if (status < 0) {
        TRACE (DL_ERR, "RXF interface create ioctl failed.");
        rxif_object->parent_function = NULL;
        rxif_object->handle = (u32)-1;
        rxif_object->mac_id = (u32)-1;
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

void be2_rxif_destroy_internal_cb (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_RXIF_OBJECT rxif_object; 
    BE_FUNCTION_OBJECT *function_object;

    rxif_object = (PBE_RXIF_OBJECT)context;
    RXIF_ASSERT (rxif_object);    

    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);
    
    // Remove the interface object from the function object
    _be_function_remove_rxif (function_object, rxif_object);
    SA_ZERO_MEM(rxif_object);        
}

//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_destroy
//
//   Deletes a receive interface object. 
//
// rxif_object        - Receive interface object to destroy. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_destroy (
   IN PBE_RXIF_OBJECT      rxif_object,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT *queue_context)
{
    PBE_FUNCTION_OBJECT  function_object = NULL;
    IOCTL_COMMON_NTWK_INTERFACE_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    RXIF_ASSERT (rxif_object);
    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_INTERFACE_DESTROY);
    ioctl->params.request.interface_id = rxif_object->handle;
    ioctl->header.request.domain = rxif_object->domain;

    // Post the ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                function_object,
                wrb,
                generic_context,
                callback,
                callback_context,
                be2_rxif_destroy_internal_cb,
                rxif_object,
                ioctl );    

Error:
    be_unlock_wrb_post (function_object);
    return status;
}




BESTATUS
be2_rxf_if_add_mac (
    IN PBE_RXIF_OBJECT      rxif_object,
    IN SA_MAC_ADDRESS       *mac,
    OUT u32                 *mac_id,    
    IN MCC_WRB_CQE_CALLBACK  callback,
    IN PVOID                 callback_context,
    IN PBE_IF_ADD_MAC_QUEUE_CONTEXT queue_context,
    IN u8                           domain)
{
    PBE_FUNCTION_OBJECT  function_object = NULL;
    IOCTL_COMMON_NTWK_PMAC_ADD *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    RXIF_ASSERT (rxif_object);
    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_PMAC_ADD);
    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->header.request.domain = domain;
    sa_mac_copy ((SA_MAC_ADDRESS*)ioctl->params.request.mac_address,
                 mac);

    // Post the Ioctl (with a copy for the response)
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context, // queue context
                 callback,
                 callback_context,
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_NTWK_PMAC_ADD, params.response.pmac_id, mac_id));

    if (status < 0) {
        TRACE (DL_ERR, "Add MAC ioctl failed status: 0x%x.",status);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);

    //TRACE (DL_ERR, "ADD MAC  New MAC:" SA_MAC_FORMAT, SA_MAC_ARGS (*mac));
    //TRACE (DL_ERR, "Add MAC ioctl status: 0x%x.,mac_id:0x%x",status,*mac_id);

    return status;


}

BESTATUS
be2_rxf_if_remove_mac (
    IN PBE_RXIF_OBJECT       rxif_object,
    IN u32                   mac_id,    
    IN MCC_WRB_CQE_CALLBACK  callback,
    IN PVOID                 callback_context,
    IN PBE_IF_ADD_MAC_QUEUE_CONTEXT queue_context,
    IN u8                           domain)
{
    PBE_FUNCTION_OBJECT  function_object = NULL;
    IOCTL_COMMON_NTWK_PMAC_DEL *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    RXIF_ASSERT (rxif_object);
    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

//    TRACE (DL_ERR, "DEL MAC ioctl mac_id:0x%x",mac_id);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NTWK_PMAC_DEL);
    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->params.request.pmac_id = mac_id;
    ioctl->header.request.domain = domain;

    // Post the Ioctl (with a copy for the response)
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context, // queue context
                 callback,
                 callback_context,
                 ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "Delete MAC ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;


}

// Sets given list of MAC Addresses for VF/domain.
// VFs can add/activate a MAC from the list of macs.

BESTATUS
be_set_maclist (
    IN PBE_FUNCTION_OBJECT                  function_object,
    IN u8                                   *mac,
    IN u8                                   num_macs,
    IN u8                                   domain,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_IF_SET_MACLIST_QUEUE_CONTEXT     queue_context,
    IN IOCTL_COMMON_SET_IFACE_MACLIST       *va_for_ioctl,
    IN u64                                  pa_for_ioctl)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    be_lock_wrb_post (function_object);
    
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context ) {
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    TRACE (DL_ERR, 
        "set_iface_mac_list. ioctl va:0x%x pa:0x%08x_%08x", 
        va_for_ioctl, 
        sa_hi (pa_for_ioctl), 
        sa_lo (pa_for_ioctl)); 
    // Prepares an embedded ioctl, including request/response sizes.    
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, COMMON_SET_IFACE_MACLIST, 0);
    //va_for_ioctl->header.request.version = 1;
    va_for_ioctl->header.request.domain = domain;
    va_for_ioctl->params.request.maclist_len = num_macs;
    if(num_macs) {
        //TODO currently only one mac in list?
        sa_mac_copy ((SA_MAC_ADDRESS*)&va_for_ioctl->params.request.mac_list[0][0], (SA_MAC_ADDRESS*)mac);
    }
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                function_object,
                wrb,
                generic_context,
                callback,
                callback_context,
                va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "set_iface_mac_list ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

// Retrieves list of mac_ids corresponding to list of macs added using set_maclist
// it also returns the status of each mac_id, wether its active or non-active.

BESTATUS
be_get_maclist (
    IN PBE_RXIF_OBJECT                      rxif_object,
    OUT u32                                 *mac_id,
    IN u8                                   domain,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_IF_GET_MACLIST_QUEUE_CONTEXT     queue_context,
    IN IOCTL_COMMON_GET_IFACE_MACLIST       *va_for_ioctl,
    IN u64                                  pa_for_ioctl)
{
    PBE_FUNCTION_OBJECT  function_object = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    RXIF_ASSERT (rxif_object);
    function_object = rxif_object->parent_function;
    FUNCTION_ASSERT (function_object);
    be_lock_wrb_post (function_object);
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    TRACE (DL_ERR, 
        "Get_iface_mac_list. ioctl va:0x%x pa:0x%08x_%08x", 
        va_for_ioctl, 
        sa_hi (pa_for_ioctl), 
        sa_lo (pa_for_ioctl)); 
    // Prepares an embedded ioctl, including request/response sizes.
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, COMMON_GET_IFACE_MACLIST, 0);
    va_for_ioctl->header.request.version = 1;
    va_for_ioctl->header.request.domain = domain;
    // Post the Ioctl (with a copy for the response)
    status = be_function_post_mcc_wrb_with_queue_context (
                            function_object,
                            wrb,
                            generic_context, // queue context
                            callback,
                            callback_context,
                            va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "GET_MACLIST ioctl failed status: 0x%x.",status);
        goto Error;
        }
Error:
    be_unlock_wrb_post (function_object);
    return status;

}


BESTATUS
be_query_eth_network_stats (
    IN PBE_FUNCTION_OBJECT                  function_object,
    IN u32                                  query_toe,
    OUT PNIC_TOE_STATS                      stats,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_ETH_NETWORK_STATS_QUEUE_CONTEXT  queue_context)
{
    IOCTL_ETH_NETWORK_STATS* ioctl;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_NETWORK_STATS);

    // Post the ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_ETH_NETWORK_STATS, params.response.stats, stats));


    if (status < 0) {
        TRACE (DL_ERR, "eth network stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}





BESTATUS
be_set_hsw_config (
   IN PBE_FUNCTION_OBJECT    function_object,
   IN PBE_RXIF_OBJECT        rxif_object, // proxied interface
   IN u8                     domain,
   IN u16                    vlan_tag,
   IN MCC_WRB_CQE_CALLBACK   callback,
   IN PVOID                  callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context )
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_SET_HSW_CONFIG_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
   
    FUNCTION_ASSERT (function_object);
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes. IOCTL_COMMON_NTWK_VLAN_CONFIG
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_HSW_CONFIG_V1);
    ioctl->header.request.domain = domain;
    ioctl->params.request.intf_pport_id = rxif_object->handle;
    ioctl->params.request.pvid_valid = 1;
    ioctl->params.request.pport = 0; // If this is 0 then intf_pport_id is interpreted as interface ID. If this is 1 then intf_pport_id is interpreted as a physical port number.
    ioctl->params.request.pvid = vlan_tag;
    
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "set hsw ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_get_hsw_config (
   IN PBE_FUNCTION_OBJECT    function_object,
   IN PBE_RXIF_OBJECT        rxif_object, // proxied interface
   IN u8                     domain,
   IN MCC_WRB_CQE_CALLBACK   callback,
   IN PVOID                  callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
   OUT GET_HSW_CONFIG_PARAMS_OUT_V1* response_cpoy
   )
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_GET_HSW_CONFIG_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
   
    FUNCTION_ASSERT (function_object);
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes. COMMON_GET_HSW_CONFIG_V1
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_HSW_CONFIG_V1);
    ioctl->header.request.domain = domain;
    ioctl->params.request.intf_pport_id = rxif_object->handle;
    ioctl->params.request.pvid_valid = 1;
    ioctl->params.request.pport = 0; // If this is 0 then intf_pport_id is interpreted as interface ID. If this is 1 then intf_pport_id is interpreted as a physical port number.

    // Post the Ioctl (with a copy for the response)
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_HSW_CONFIG_V1, params.response, response_cpoy));

    if (status < 0) {
        TRACE (DL_ERR, "get_hsw_config ioctl failed.");
        goto Error;
    }

    
Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
lancer_query_pport_statistics (
    IN PBE_FUNCTION_OBJECT               function_object,
    IN IOCTL_ETH_GET_PPORT_STATISTICS*   va_for_ioctl,
    IN u64                               pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK              callback,
    IN PVOID                             callback_context,
    IN PLANCER_PPORT_STATS_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query pport stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, ETH_GET_PPORT_STATISTICS, 0);

    // Command parameters
    va_for_ioctl->params.request.port_number = function_object->fw_cfg.phys_port;
    va_for_ioctl->params.request.reset_stats = 0;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "pport stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
lancer_query_vport_statistics (
    IN PBE_FUNCTION_OBJECT               function_object,
    IN IOCTL_ETH_GET_VPORT_STATISTICS*   va_for_ioctl,
    IN u64                               pa_for_ioctl,
    IN PBE_RXIF_OBJECT                   rxif_object,  // rxif object
    IN MCC_WRB_CQE_CALLBACK              callback,
    IN PVOID                             callback_context,
    IN PLANCER_VPORT_STATS_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, 
           "Query vport stats. ioctl va:0x%x pa:0x%08x_%08x", 
           va_for_ioctl, 
           sa_hi (pa_for_ioctl), 
           sa_lo (pa_for_ioctl)); 

    // Prepares an embedded ioctl, including request/response sizes.
    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, wrb, va_for_ioctl, pa_for_ioctl, ETH_GET_VPORT_STATISTICS, 0);

    // Command parameters
    va_for_ioctl->params.request.if_id = rxif_object->handle;
    va_for_ioctl->params.request.reset_stats = 0;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "vport stats ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


/*!
@brief
    This routine adds a filter on receive queue
@param        
    FunctionObject   - Function object handle.
@param        
    cq_id            - Value of 0xFFFF implies filters for Q-Group
                       Any other value implies filters for Q-Direct or adding
                       queues to Q-Group depending on filter_type
@param        
    filter_type      - For Q-Direct:
                         0: MAC 
                         1: VLAN (Not supported)
                         2: MAC_VLAN
                       For Q-Group:
                         3: PRIORITY
                         4: ETHER_TYPE
@param        
    filter_val       - Filter itself, value depends on filter_type mentioned
@param        
    mac_address      - MAC, if filter_type contains MAC Filter
@param        
    filter_id        - Filter ID assigned by HW upon successfull completion

@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL
*/
BESTATUS 
lancer_rxf_add_queue_filter (
    IN PBE_FUNCTION_OBJECT        function_object,
    IN PBE_RXIF_OBJECT            rxif_object,
    IN u16                        cq_id,
    IN u16                        filter_type,
    IN u16                        filter_val,
    IN PSA_MAC_ADDRESS            mac_address,
    OUT u32                       *filter_id,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context)
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);
    
    ASSERT (!be_function_is_vm (function_object));
    ASSERT (rxif_object);

    // Validate filters
    if (filter_type == FILTER_TYPE_VLAN) {
        TRACE (DL_ERR, "FILTER_TYPE_VLAN unsupported. Issue FILTER_TYPE_MAC_VLAN");
        goto Error;
    }

    if (filter_type > FILTER_TYPE_ETHERTYPE) {
        TRACE (DL_ERR, "Invalid filter_type 0x%x", filter_type);
        goto Error;
    }

    if (cq_id == 0xFFFF) {
        if (filter_type == FILTER_TYPE_PRIORITY || filter_type == FILTER_TYPE_ETHERTYPE) {
            TRACE (DL_ERR, "Wrong filter_type 0x%x for Q-Direct", filter_type);
            goto Error;
        }
    } else {
        if (filter_type == FILTER_TYPE_MAC || filter_type == FILTER_TYPE_MAC_VLAN) {
            TRACE (DL_INFO, "Setting Q-Direct filter");
            if (!mac_address) {
                TRACE (DL_ERR, "No MAC Address specified");
                goto Error;
            }
        } else {
            TRACE (DL_INFO, "Adding member to Q-Group");
        }
    }

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ADD_IFACE_QUEUE_FILTER);    

    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->params.request.cq_id = cq_id;
    ioctl->params.request.filter_type = filter_type;
    ioctl->params.request.filter_val = filter_val;
    if (mac_address) {
        ioctl->params.request.mac_lo[0] = mac_address->bytes[0];
        ioctl->params.request.mac_lo[1] = mac_address->bytes[1];
        ioctl->params.request.mac_hi[0] = mac_address->bytes[2];
        ioctl->params.request.mac_hi[1] = mac_address->bytes[3];
        ioctl->params.request.mac_hi[2] = mac_address->bytes[4];
        ioctl->params.request.mac_hi[3] = mac_address->bytes[5];
    }

    TRACE (DL_INFO, "Filter_type 0x%x, filter_val 0x%x, MAC %2x:%2x:%2x:%2x:%2x:%2x",
           ioctl->params.request.filter_type,
           ioctl->params.request.filter_val,
           ioctl->params.request.mac_lo[0],
           ioctl->params.request.mac_lo[1],
           ioctl->params.request.mac_hi[0],
           ioctl->params.request.mac_hi[1],
           ioctl->params.request.mac_hi[2],
           ioctl->params.request.mac_hi[3]);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context, // queue context
                 callback,
                 callback_context,
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER, params.response.filter_id, filter_id));

    if (status < 0) {
        TRACE (DL_ERR, "IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER failed");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


/*!
@brief
    This routine deletes a filters on receive queue
@param        
    FunctionObject   - Function object handle.
@param        
    filter_id        - Filter ID assigned by HW

@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
    
@note
    IRQL: < DISPATCH_LEVEL
*/
BESTATUS 
lancer_rxf_del_queue_filter (
    IN PBE_FUNCTION_OBJECT        function_object,
    IN PBE_RXIF_OBJECT            rxif_object,
    IN u32                        filter_id,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context)
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);
    
    ASSERT (!be_function_is_vm (function_object));
    ASSERT (rxif_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_DEL_IFACE_QUEUE_FILTER);    

    ioctl->params.request.if_id = rxif_object->handle;
    ioctl->params.request.filter_id = filter_id;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER failed");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
