/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_amap.h

@brief

    Array map macros and functions.
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_amap.h,v 1.5 2006/03/13 22:43:46 menz Exp $
*/
#ifndef __sa_amap_h__
#define __sa_amap_h__

#include "system/sa_types.h"

SA_CPP_HEADER

// Each bit is 1 byte in array maps.
typedef u8 AMAP_BIT;

// Using this function instead of casting forces a compiler error
// if the value cannot be casted to a u32 safely.  This is useful
// for the following macros that take an integer parameter.
STATIC INLINE u32 amap_to_dword (u32 value) { return value; }


// These macros get optimized due to constant folding.


// Returns the a bit mask for the register that is NOT shifted into location.
// That means return values always look like: 0x1, 0xFF, 0x7FF, etc... 
STATIC INLINE u32 amap_mask (u32 bit_size) {
    return (bit_size == 32 ? 0xFFFFFFFF : (1 << bit_size) - 1);
}

#define AMAP_BIT_MASK( _struct_, _register_ )       \
    amap_mask (AMAP_BIT_SIZE (_struct_, _register_))  \


// Returns the mask for clear the bitfield within a dword.  eg. 0x7F800
#define AMAP_SHIFTED_MASK(_struct_, _register_)                                   \
    (AMAP_BIT_MASK(_struct_, _register_) << AMAP_BIT_OFFSET(_struct_, _register_))  \


// Returns the given value, with the given bitfield within it cleared to zeros.
// To clear a bitfield and reset it, use it like this:
// contextMemory.dw[1] = AMAP_CLR_BITS (struct, field1, contextMemory.dw[1]) |
//                       AMAP_SET_BITS (struct, field1, 123);
#define AMAP_CLR_BITS(_struct_, _register_, _value_)                 \
    (amap_to_dword (_value_) & ~AMAP_SHIFTED_MASK(_struct_, _register_))  \


// Extracts bit field from the DWORD that contains it, shifting the value
// to the low-order bits.
// e.g. u32 value = AMAP_GET_BITS (struct, field1, contextMemory.dw[3]);
STATIC INLINE u32 amap_get_bits (u32 mask, u32 offset, u32 dword_in_struct){
    return mask & (dword_in_struct >> offset);
}

#define AMAP_GET_BITS( _struct_, _register_, _dw_in_struct_ )  \
    amap_get_bits (AMAP_BIT_MASK (_struct_, _register_),         \
                 AMAP_BIT_OFFSET (_struct_, _register_),       \
                 _dw_in_struct_)                             \


// Masks the bits (before shifting!) and then shifts to correct offset.
// contextMemory.dw[1] = AMAP_SET_BITS (struct, field1, 123) | 
//                       AMAP_SET_BITS (struct, field2, 456):
// Remember this does not clear bits!
STATIC INLINE u32 amap_set_bits (u32 mask, u32 offset, u32 new_value) {
    return (mask & new_value) << offset;
}
#define AMAP_SET_BITS( _struct_, _register_, _new_value_ )  \
    amap_set_bits (AMAP_BIT_MASK (_struct_, _register_),    \
                 AMAP_BIT_OFFSET (_struct_, _register_),    \
                 _new_value_)                               \

// Fancy, non-optimized set bits the first clears the bits and then assigns 'em.
// This does not require knowledge of the particular DWORD you are setting.
// e.g. AMAP_SET_BITS_PTR (struct, field1, &contextMemory, 123);
#ifdef SA_CPP
    SA_CPP_TRAILER // template requires C++ linkage
    
    // This forces ptr to be the correct type.
    template<class T> void amap_set (const T* ptr, u32 dwOffset, u32 mask, u32 offset, u32 value) {
#else
    STATIC INLINE void amap_set(void* ptr, u32 dwOffset, u32 mask, u32 offset, u32 value) {
#endif
    u32 *dw = (u32*)ptr;
    *(dw + dwOffset) &= ~(mask << offset);
    *(dw + dwOffset) |= (mask & value) << offset;
}

#define AMAP_SET_BITS_PTR( _struct_, _register_, _structPtr_, _value_)  \
    amap_set (_structPtr_,                                  \
            AMAP_WORD_OFFSET (_struct_, _register_),        \
            AMAP_BIT_MASK (_struct_, _register_),           \
            AMAP_BIT_OFFSET (_struct_, _register_),         \
            _value_)                                      \



// Non-optimized routine that gets the bits without knowing the correct DWORD.
// e.g. fieldValue = AMAP_GET_BITS_PTR (struct, field1, &contextMemory);
#ifdef SA_CPP
    // This forces ptr to be the correct type.
    template<class T> u32 amap_get (const T* ptr, u32 dwOffset, u32 mask, u32 offset) {
#else
    STATIC INLINE u32 amap_get (void* ptr, u32 dwOffset, u32 mask, u32 offset) {
#endif
    u32 *dw = (u32*)ptr;
    return mask & (*(dw + dwOffset) >> offset);
}
#define AMAP_GET_BITS_PTR( _struct_, _register_, _structPtr_)  \
    amap_get (_structPtr_,                                     \
            AMAP_WORD_OFFSET (_struct_, _register_),           \
            AMAP_BIT_MASK (_struct_, _register_),              \
            AMAP_BIT_OFFSET (_struct_, _register_))            \

#ifdef SA_CPP
    SA_CPP_HEADER // template requires C++ linkage
#endif

#define AMAP_SET AMAP_SET_BITS_PTR
#define AMAP_GET AMAP_GET_BITS_PTR

// Returns 0-31 representing bit offset within a DWORD of a bitfield.
#define AMAP_BIT_OFFSET( _struct_, _register_ )                  \
    (((SA_SIZE_T)&(((BE_ ## _struct_ ## _AMAP*)0)->_register_))%32)  \

// Returns 0-n representing byte offset of bitfield with the structure.
#define AMAP_BYTE_OFFSET( _struct_, _register_)                  \
    (((SA_SIZE_T)&(((BE_ ## _struct_ ## _AMAP *)0)->_register_))/8)  \

// Returns 0-n representing DWORD offset of bitfield within the structure.
#define AMAP_WORD_OFFSET( _struct_, _register_)  \
  (AMAP_BYTE_OFFSET( _struct_, _register_ )/4)   \

// Gets a pointer to a field within a structure
// The field must be byte aligned.
#define AMAP_GET_PTR( _struct_, _register_, _structPtr_)                     \
    (void *) ((u8*)(_structPtr_) + AMAP_BYTE_OFFSET(_struct_, _register_))  \

// Returns size of bitfield in bits.
#define AMAP_BIT_SIZE( _struct_, _register_ ) sizeof(((BE_ ## _struct_ ## _AMAP*)0)->_register_)

// Returns size of bitfield in bytes.
#define AMAP_BYTE_SIZE( _struct_ ) (sizeof( BE_ ## _struct_ ## _AMAP)/8)

// Returns size of bitfield in DWORDS.
#define AMAP_WORD_SIZE( _struct_ ) (AMAP_BYTE_SIZE( _struct_ )/4)

#define AMAP_FIELDS_EQUALS(_struct1_, _struct2_, _field_)                               \
    ((AMAP_BIT_SIZE (_struct1_, _field_) == AMAP_BIT_SIZE (_struct2_, _field_)) &&      \
     (AMAP_WORD_OFFSET (_struct1_, _field_) == AMAP_WORD_OFFSET (_struct2_, _field_)))  \

// Obsolete (legacy) names.  Stop using these names.
typedef AMAP_BIT BE_BIT;
#define BE_BIT_MASK     AMAP_BIT_MASK
#define BE_SHIFTED_MASK AMAP_SHIFTED_MASK
#define BE_CLR_BITS     AMAP_CLR_BITS
#define BE_GET_BITS     AMAP_GET_BITS
#define BE_SET_BITS     AMAP_SET_BITS
#define BE_SET_BITS_PTR AMAP_SET_BITS_PTR
#define BE_GET_BITS_PTR AMAP_GET_BITS_PTR
#define SET             AMAP_SET
#define GET             AMAP_GET
#define BE_BIT_OFFSET   AMAP_BIT_OFFSET
#define BE_BYTE_OFFSET  AMAP_BYTE_OFFSET
#define BE_WORD_OFFSET  AMAP_WORD_OFFSET
#define BE_GET_PTR      AMAP_GET_PTR
#define BE_BIT_SIZE     AMAP_BIT_SIZE
#define BE_BYTE_SIZE    AMAP_BYTE_SIZE
#define BE_WORD_SIZE    AMAP_WORD_SIZE

SA_CPP_TRAILER

#endif

