/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/****************************************************************************
* Description:
*   Does some base2 math, including sa_log2 and finding next higher
*   power of 2.
****************************************************************************/
#ifndef __sa_base2_h__
#define __sa_base2_h__

#include "system/sa_types.h"
SA_CPP_HEADER

//<no_inbox>
extern u8 sa_bits_in_byte_table[256];
//</no_inbox>

// Returns # of bits necessary to represent the input
// integer x.  If x is power of 2, this is sa_log2(x)+1.
// input->output
// 0->0, 1->1, 2->2, 3->2, 4->3, 5->3, 6->3, 7->3, 8->4, 9->4
STATIC
INLINE u32 sa_required_bits64 (u64 x) {
    u32 hi = 63, lo = 0;
    u32 guess = 32;

    if (x == 0) return 0;
    if (x & 0x8000000000000000ULL) return 64;
        
    // Binary search
    while (hi > lo) {
        if ((x >> guess) > 0) {
            // x is more than i bits in length
            lo = guess;
            guess = (hi + guess + 1) >> 1;        
        } else {
            hi = guess-1;
            guess = (lo + guess) >> 1;
        }
    }

    return hi + 1;
}

STATIC
INLINE u32 sa_required_bits (u32 x) {
    u32 hi = 31, lo = 0;
    u32 guess = 16;

    if (x == 0) return 0;
    if (x & 0x80000000) return 32;
        
    // Binary search
    while (hi > lo) {
        if ((x >> guess) > 0) {
            // x is more than i bits in length
            lo = guess;
            guess = (hi + guess + 1) >> 1;        
        } else {
            hi = guess-1;
            guess = (lo + guess) >> 1;
        }
    }

    return hi + 1;
}

// <no_inbox>
// Exact sa_log2
// input->output
// 0->-1.#INF, 1->0, 2->1, 3->1.58496, 4->2, 5->2.32193, 6->2.58496, 7->2.80735, 8->3, 9->3.16993
#if defined( SA_USER )

    #include <math.h>

    STATIC
    INLINE f64 sa_log2_exact (f64 x) {
        return log (x) / log (2.0);
    }

#endif
// </no_inbox>

// Returns the log base 2 for a number, always rounding down.
// sa_log2 example input->output
// 0->4294967295, 1->0, 2->1, 3->1, 4->2, 5->2, 6->2, 7->2, 8->3, 9->3
STATIC
INLINE u32 sa_log2 (u32 x) {
    return sa_required_bits (x) - 1;
}

// Returns the log base 2 for a number, always rounding down.
// sa_log2 example input->output
// 0->4294967295, 1->0, 2->1, 3->1, 4->2, 5->2, 6->2, 7->2, 8->3, 9->3
STATIC
INLINE u32 sa_log2_u64 (u64 x) {
    return sa_required_bits64 (x) - 1;
}


// <no_inbox>
// ------------------------------------------------------------------------------------
//      Number of Bits Set in u4 (nibble), u8, u32, or u64
// ------------------------------------------------------------------------------------
//
//  Each algorithm counts the number of set bits in the input.  For example:
//  0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
//
// The algorithms:
//
// Name                              Description
// sa_num_bits_set_uXX_loop          loops for each bit, optimized for small numbers
// sa_num_bits_set_uXX_add           adds all bits, optimized for pipelined processor, no looping
// sa_num_bits_set_uXX_nibble_table  uses 16 entry static lookup table (computes table first call).
// sa_num_bits_set_uXX_byte_table    Uses 256 entry static lookup table (computes table first call).
//
// Here are performance numbers in units of performance counter ticks for a 2.5GHz Intel Core2 Duo Quad.  
// This test is located in sa_runtime_text.c.  
// The numbers are for small inputs numbers, medium, and large numbers.
//sar_w32u info  sa_num_bits_set_u8_loop                 . ticks/iter small:  36 medium:  44 big:  65  
//sar_w32u info  sa_num_bits_set_u8_add                  . ticks/iter small:  17 medium:  17 big:  17  
//sar_w32u info  sa_num_bits_set_u8_byte_table           . ticks/iter small:  12 medium:  12 big:  12  
//sar_w32u info  sa_num_bits_set_u8_nibble_table         . ticks/iter small:  12 medium:  12 big:  12  
//
//sar_w32u info  sa_num_bits_set_u32_loop                . ticks/iter small:  67 medium: 138 big: 254
//sar_w32u info  sa_num_bits_set_u32_count               . ticks/iter small:  39 medium:  50 big:  69
//sar_w32u info  sa_num_bits_set_u32_fast_count          . ticks/iter small:  21 medium:  21 big:  21
//sar_w32u info  sa_num_bits_set_u32_nifty               . ticks/iter small:  21 medium:  21 big:  21
//sar_w32u info  sa_num_bits_set_u32_mit                 . ticks/iter small:  18 medium:  18 big:  19
//sar_w32u info  sa_num_bits_set_u32_add                 . ticks/iter small:  64 medium:  61 big:  61
//sar_w32u info  sa_num_bits_set_u32_byte_table          . ticks/iter small:  15 medium:  15 big:  15
//sar_w32u info  sa_num_bits_set_u32_nibble_table        . ticks/iter small:  28 medium:  28 big:  28
//
//sar_w32u info  sa_num_bits_set_u64_loop                . ticks/iter small: 130 medium: 245 big: 417 
//sar_w32u info  sa_num_bits_set_u64_count               . ticks/iter small:  79 medium: 117 big: 205 
//sar_w32u info  sa_num_bits_set_u64_fast_count          . ticks/iter small:  49 medium:  50 big:  49 
//sar_w32u info  sa_num_bits_set_u64_nifty               . ticks/iter small:  57 medium:  57 big:  59 
//sar_w32u info  sa_num_bits_set_u64_mit                 . ticks/iter small:  33 medium:  33 big:  34 
//sar_w32u info  sa_num_bits_set_u64_add                 . ticks/iter small: 112 medium: 112 big: 112 
//sar_w32u info  sa_num_bits_set_u64_byte_table          . ticks/iter small:  29 medium:  29 big:  29 
//sar_w32u info  sa_num_bits_set_u64_nibble_table        . ticks/iter small:  45 medium:  45 big:  45 

// These macros define a default algorithm to use for those not fascinated by a few clocks
// of performance difference.  Additional alogrithms are defined
// below.  Currently, the defaults avoid static variables since some environments may
// have problems using static variables.  (Need to investigate.)  
#ifndef sa_num_bits_set_u64
    #define sa_num_bits_set_u64        sa_num_bits_set_u64_nibble_table
#endif
#ifndef sa_num_bits_set_u32
    #define sa_num_bits_set_u32        sa_num_bits_set_u32_nibble_table
#endif
#ifndef sa_num_bits_set_u8
    #define sa_num_bits_set_u8         sa_num_bits_set_u8_nibble_table
#endif

//
// -------------------------- NIBBLE VERSIONS --------------------------
//
STATIC FORCEINLINE u32 sa_num_bits_set_u4_add (u8 x) {
    u32 y = (u32)x; // most processors will be faster using 32 ints
    return ((y >> 0) & 1) +
           ((y >> 1) & 1) +
           ((y >> 2) & 1) +
           ((y >> 3) & 1);
}

STATIC FORCEINLINE u32 sa_num_bits_set_u4_nibble_table (u8 x) {    
    return (u32)sa_bits_in_byte_table[x & 0xF];
}


//
// -------------------------- BYTE VERSIONS --------------------------
//
STATIC INLINE u32 sa_num_bits_set_u8_loop (u8 x) {
    u32 i = 0;
    while (x > 0) {
        i += (x & 1);
        x >>= 1;
    }
    return i;
}

STATIC FORCEINLINE u32 sa_num_bits_set_u8_add (u8 x) {

    u32 y = (u32)x; // most processors will be faster using 32 ints
    return ((y >> 0) & 1) +
           ((y >> 1) & 1) +
           ((y >> 2) & 1) +
           ((y >> 3) & 1) +
           ((y >> 4) & 1) +
           ((y >> 5) & 1) +
           ((y >> 6) & 1) +
           ((y >> 7) & 1);
}

STATIC INLINE u32 sa_num_bits_set_u8_byte_table (u8 x) {
    // reference the table
    return (u32)sa_bits_in_byte_table[x];
}

STATIC FORCEINLINE u32 sa_num_bits_set_u8_nibble_table (u8 x) {
   return sa_num_bits_set_u4_nibble_table (x) +
          sa_num_bits_set_u4_nibble_table (x >> 4);
}

#ifdef SA_USER
    STATIC INLINE void
    sa_print_bit_tables (void) {
        u32 i;        
        sa_printf ("STATIC u8 sa_bits_in_nibble_table[] = {");
        for (i=0; i<16; i++) {
            sa_printf ("%d, ", (u32)sa_num_bits_set_u4_add((u8)i));        
        }
        sa_printf ("};\n");

        sa_printf ("STATIC u8 sa_bits_in_byte_table[] = {");
        for (i=0; i<256; i++) {
            if (i % 16 == 0) {
                sa_printf ("\n");
            }
            sa_printf ("%d, ", (u32)sa_num_bits_set_u8_add((u8)i));
        }
        sa_printf ("};\n");
    }
#else
    #define sa_print_bit_tables() 
#endif



//
// -------------------------- u32 VERSIONS --------------------------
//

// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
// Optimized for small code size. Most efficient for smaller numbers.
STATIC INLINE u32 sa_num_bits_set_u32_loop (u32 x) {
    u32 i = 0;
    while (x > 0) {
        i += (u32)(x & 1);
        x >>= 1;
    }
    return i;
}


// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
// Optimized for sparse number set bits.
STATIC INLINE u32 sa_num_bits_set_u32_count (u32 x) {
    u32 i = 0;
    while (x > 0) {
        x &= (x-1);
        ++i;        
    }
    return i;
}

#define SA_FASTBIT_TWO(c)     (0x1u << (c))
#define SA_FASTBIT_MASK(c)    (((u32)(-1)) / (SA_FASTBIT_TWO(SA_FASTBIT_TWO(c)) + 1UL))
#define SA_FASTBIT_COUNT(x,c) ((x) & SA_FASTBIT_MASK(c)) + (((x) >> (SA_FASTBIT_TWO(c))) & SA_FASTBIT_MASK(c))
 
STATIC INLINE u32 sa_num_bits_set_u32_fast_count (u32 n)
{
    n = SA_FASTBIT_COUNT (n, 0) ;
    n = SA_FASTBIT_COUNT (n, 1) ;
    n = SA_FASTBIT_COUNT (n, 2) ;
    n = SA_FASTBIT_COUNT (n, 3) ;
    n = SA_FASTBIT_COUNT (n, 4) ;
    return n ;
}

#define SA_NIFTY_MASK_01010101 (((u32)(-1))/3)
#define SA_NIFTY_MASK_00110011 (((u32)(-1))/5)
#define SA_NIFTY_MASK_00001111 (((u32)(-1))/17)

STATIC INLINE u32 sa_num_bits_set_u32_nifty (u32 n)
{
    n = (n & SA_NIFTY_MASK_01010101) + ((n >> 1) & SA_NIFTY_MASK_01010101) ; 
    n = (n & SA_NIFTY_MASK_00110011) + ((n >> 2) & SA_NIFTY_MASK_00110011) ; 
    n = (n & SA_NIFTY_MASK_00001111) + ((n >> 4) & SA_NIFTY_MASK_00001111) ; 
    return (u32)(n % 255);
}

STATIC INLINE u32 sa_num_bits_set_u32_mit (u32 n)
{
    u32 tmp;

    tmp = n - ((n >> 1) & 033333333333)
            - ((n >> 2) & 011111111111);

    return ((tmp + (tmp >> 3)) & 030707070707) % 63;
}


// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
// Add the bits without looping to optimize for pipelined architectures (no possibility
// for incorrect branch prediction).
STATIC
INLINE u32 sa_num_bits_set_u32_add (u32 x) {
    return sa_num_bits_set_u8_add ((u8)(x>>0)) +
           sa_num_bits_set_u8_add ((u8)(x>>8)) +
           sa_num_bits_set_u8_add ((u8)(x>>16)) +
           sa_num_bits_set_u8_add ((u8)(x>>24));
}

// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
// Add the bits without looping to optimize for pipelined architectures (no possibility
// for incorrect branch prediction).  Uses lookup table to optimal repeated performance.
STATIC
INLINE u32 sa_num_bits_set_u32_byte_table (u32 x) {
    return sa_num_bits_set_u8_byte_table ((u8)(x>>0)) +
           sa_num_bits_set_u8_byte_table ((u8)(x>>8)) +
           sa_num_bits_set_u8_byte_table ((u8)(x>>16)) +
           sa_num_bits_set_u8_byte_table ((u8)(x>>24));
}

// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2
// Add the bits without looping to optimize for pipelined architectures (no possibility
// for incorrect branch prediction).
STATIC
INLINE u32 sa_num_bits_set_u32_nibble_table (u32 x) {
    return sa_num_bits_set_u4_nibble_table ((u8)(x>>0)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>4)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>8)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>12)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>16)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>20)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>24)) +
           sa_num_bits_set_u4_nibble_table ((u8)(x>>28));
}

//
// -------------------------- u64 VERSIONS --------------------------
//
STATIC INLINE u32 sa_num_bits_set_u64_loop (u64 x) {
    u32 i = 0;
    while (x > 0) {
        i += (u32)(x & 1);
        x >>= 1;
    }
    return i;
}

STATIC INLINE u32 sa_num_bits_set_u64_count (u64 x) {
    u32 i = 0;
    while (x > 0) {
        x &= (x-1);
        ++i;        
    }
    return i;
}

#define SA_FASTBIT64_TWO(c)     ((u64)1 << (c))
#define SA_FASTBIT64_MASK(c)    (((u64)(-1)) / (SA_FASTBIT64_TWO(SA_FASTBIT64_TWO(c)) + (u64)1))
#define SA_FASTBIT64_COUNT(x,c) ((x) & SA_FASTBIT64_MASK(c)) + ((((x) >> SA_FASTBIT64_TWO(c)) & SA_FASTBIT64_MASK(c)))

STATIC INLINE u32 sa_num_bits_set_u64_fast_count (u64 n)
{
    n = SA_FASTBIT64_COUNT (n, 0) ;
    n = SA_FASTBIT64_COUNT (n, 1) ;
    n = SA_FASTBIT64_COUNT (n, 2) ;
    n = SA_FASTBIT64_COUNT (n, 3) ;
    n = SA_FASTBIT64_COUNT (n, 4) ;
    n = SA_FASTBIT64_COUNT (n, 5) ;
    return (u32)n ;
}

#define SA_NIFTY64_MASK_01010101 (((u64)(-1))/3)
#define SA_NIFTY64_MASK_00110011 (((u64)(-1))/5)
#define SA_NIFTY64_MASK_00001111 (((u64)(-1))/17)

STATIC INLINE u32 sa_num_bits_set_u64_nifty (u64 n)
{
    n = (n & SA_NIFTY64_MASK_01010101) + ((n >> 1) & SA_NIFTY64_MASK_01010101) ; 
    n = (n & SA_NIFTY64_MASK_00110011) + ((n >> 2) & SA_NIFTY64_MASK_00110011) ; 
    n = (n & SA_NIFTY64_MASK_00001111) + ((n >> 4) & SA_NIFTY64_MASK_00001111) ; 
    return (u32)(n % 255);
}


STATIC INLINE u32 sa_num_bits_set_u64_mit (u64 n)
{
    return sa_num_bits_set_u32_mit ((u32)n) +
           sa_num_bits_set_u32_mit ((u32)(n>>32));
}


STATIC INLINE u32 sa_num_bits_set_u64_add (u64 x) {
    return sa_num_bits_set_u32_add ((u32)x) +
           sa_num_bits_set_u32_add ((u32)(x>>32));
}

STATIC INLINE u32 sa_num_bits_set_u64_byte_table (u64 x) {
    return sa_num_bits_set_u32_byte_table ((u32)x) +
           sa_num_bits_set_u32_byte_table ((u32)(x>>32));
}

STATIC INLINE u32 sa_num_bits_set_u64_nibble_table (u64 x) {
    return sa_num_bits_set_u32_nibble_table ((u32)x) +
           sa_num_bits_set_u32_nibble_table ((u32)(x>>32));
}

#if 0
// Use the MIT version for inbox. 
// </no_inbox>
// Returns the number of bits set in x.
// 0->0, 1->1, 2->1, 3->2, 4->1, 5->2, 6->2, 7->3, 8->1, 9->2

STATIC INLINE u32 sa_num_bits_set_u32 (u32 n)
{
    u32 tmp;
    tmp = n - ((n >> 1) & 033333333333)
            - ((n >> 2) & 011111111111);
    return ((tmp + (tmp >> 3)) & 030707070707) % 63;
}

STATIC INLINE u32 sa_num_bits_set_u64 (u64 n)
{
    return sa_num_bits_set_u32 ((u32)n) +
            sa_num_bits_set_u32 ((u32)(n>>32));
}

//<no_inbox>
#endif
//</no_inbox>


// Returns true if the number is a power of 2.
// input->output
// 0->0, 1->1, 2->1, 3->0, 4->1, 5->0, 6->0, 7->0, 8->1, 9->0
STATIC
INLINE boolean sa_is_power2 (u32 x) {    
    return (boolean)(sa_num_bits_set_u32 (x) == 1);        
}

STATIC
INLINE boolean sa_is_power2_u64 (u64 x) {    
    return (boolean)(sa_num_bits_set_u64 (x) == 1);        
}

// Returns true if the number is a power of 2. Uses binary search.
// input->output
// 0->0, 1->1, 2->1, 3->0, 4->1, 5->0, 6->0, 7->0, 8->1, 9->0
STATIC
INLINE boolean sa_is_power2_binary (u32 x) {    
    u32 bits = sa_required_bits (x);
    return (((1<<(bits-1))^x)==0);
}

// "greater than power of 2"
// Returns the next higher power of 2 -- even if input is a power of 2.
// input->output
// 0->1, 1->2, 2->4, 3->4, 4->8, 5->8, 6->8, 7->8, 8->16, 9->16
STATIC
INLINE u32 sa_gt_power2 (u32 x) {
    x = sa_required_bits (x);
    
    if (x >= 32) // error -- cannot represent this number in 32 bits
        return (u32)0xFFFFFFFF;  

    return (1 << x);
}


// "greater than or equal power of 2"
// Returns the next higher power of 2, but if the 
// input is a power of 2 it returns the same value!
// input->output
// 0->1, 1->1, 2->2, 3->4, 4->4, 5->8, 6->8, 7->8, 8->8, 9->16
STATIC
INLINE u32 sa_ge_power2 (u32 x) {
    if (sa_is_power2 (x)) 
        return x;

    return sa_gt_power2 (x);
}


SA_CPP_TRAILER
#endif

