/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/****************************************************************************
* Description:
*   Some integer math routines.
****************************************************************************/
#ifndef __sa_int_math_h__
#define __sa_int_math_h__

#include "system/sa_types.h"
#include "system/sa_assert.h"
#include "runtime/sa_base2.h"
SA_CPP_HEADER

STATIC
INLINE
u32 sa_abs (i32 x) {
    if (x < 0) 
        return -1 * x;
    else
        return x;
}

// Returns x/y plus 1 if there is a remainder.
STATIC
INLINE u32 sa_ceiling (u32 x, u32 y) {
    ASSERT (y > 0);    
    return (x+y-1)/y;
}

// Included for completeness and the assert.
// Simply returns x/y.
STATIC
INLINE u32 sa_floor (u32 x, u32 y) {
    ASSERT (y>0);
    return (x/y);
}

// Round Up Increment
// Returns number to add to x to round up to next multiple
// of y.  If x is multiple of y it returns 0.
STATIC
INLINE u32 sa_round_up_inc (u32 x, u32 y) {
    u32 z;
    ASSERT (y > 0);
    z = x%y;
    return (z > 0 ? y - z : 0);
}

STATIC
INLINE u32 sa_round_up_inc_power2 (u32 x, u32 power2) {
    u32 z;
    ASSERT (sa_is_power2 (power2)); // this assert is a bit expensive!
    z = ((power2-1) & x);
    return (z > 0 ? power2 - z : 0);
}


// Returns x/y plus 1 if there is a remainder.
STATIC
INLINE u64 sa_ceiling64 (u64 x, u64 y) {
    ASSERT (y > 0);
    return (x+y-1)/y;
}

// Round Up Increment
// Returns number to add to x to round up to next multiple
// of y.  If x is multiple of y it returns 0.
STATIC
INLINE u32 sa_round_up_inc64 (u64 x, u32 y) {
    u32 z;
    ASSERT (y > 0);
    z = (u32)x%y;
    return (z > 0 ? y - z : 0);
}


// Returns x rounded to next multiple of y.  If x is a multiple
// of y, it returns x.
STATIC
INLINE u32 sa_round_up (u32 x, u32 y)
{
    return x + sa_round_up_inc (x,y);
}

// Round down decrement
// Return number to subtract from x to round down to next multiple of y.
// If x is multiple of y it returns 0.
STATIC
INLINE u32 sa_round_down_dec (u32 x, u32 y)
{
    ASSERT (y > 0);
    return (x%y);
}

// Return x rounded to next lower multiple of y. If x is a multiple of y, 
// it returns x. 
STATIC
INLINE u32 sa_round_down (u32 x, u32 y)
{    
    return x - sa_round_down_dec (x, y);
}

// These macros handle comparison of a u32 "a" with another u32 b.
// They treat the circular 2G range before "a" as less than a, 
// and the circular 2GB range after a as greater than a.  
// Example:
//   C_LT(0x100, 0x0)        == FALSE
//    ...                    == FALSE
//   C_LT(0x100, 0x100)      == FALSE
//   C_LT(0x100, 0x101)      == TRUE
//    ...                    == TRUE
//   C_LT(0x100, 0x80000100) == TRUE
//   C_LT(0x100, 0x80000101) == FALSE
//    ...                    == FALSE
//   C_LT(0x100, 0xFFFFFFFF) == FALSE
//
// The "C" stands for circular.
// This is used for tcp sequence number.
#define C_LT(a,b)   ((i32)((a)-(b)) < 0)
#define C_LE(a,b)   ((i32)((a)-(b)) <= 0)
#define C_GT(a,b)   ((i32)((a)-(b)) > 0)
#define C_GE(a,b)   ((i32)((a)-(b)) >= 0)


// circular subtract.
// Returns a - b assuming a circular number system, where a and b are in range (0, 
// maxValue-1). If a==b, zero is returned so the highest value possible with this
// subtraction is maxValue-1.
STATIC
INLINE u32 sa_subc (u32 a, u32 b, u32 max) {
    ASSERT (a <= max && b <= max);
    ASSERT (max > 0);    
    return (a >= b ? (a-b) : (max - b + a));
}

// Optimized version if max is power of 2.
STATIC
INLINE u32 sa_subc_power2 (u32 a, u32 b, u32 max) {
    ASSERT (a < max && b < max);
    ASSERT (sa_is_power2 (max)); // this assert is a bit expensive!
    return ((a-b) & (max-1));
}

// circular add.
// Returns a + b assuming a circular number system, where a and b are in range (0, 
// maxValue-1). 
STATIC
INLINE u32 sa_addc_mod (u32 a, u32 b, u32 max) {    
    ASSERT (a < max);
    ASSERT (max > 0);
    return (a + b) % max;    
}

STATIC
INLINE u32 sa_addc_if (u32 a, u32 b, u32 max) {    
    ASSERT (a < max);
    ASSERT (max > 0);   
    return ((max - a > b) ? (a+b) : (b + a - max));
}

// Optimized version if max is power of 2.
STATIC
INLINE u32 sa_addc_power2 (u32 a, u32 b, u32 max) {    
    ASSERT (a < max);
    ASSERT (sa_is_power2 (max)); // this assert is a bit expensive!
    return ((a+b) & (max-1));
}

// Define the sa_addc function as the if version based on the sa_test.exe performance numbers.
// sar_w32u info  addc_mod avg ticks/loop: 45
// sar_w32u info  addc_if avg ticks/loop: 7
// sar_w32u info  addc_power2 avg ticks/loop: 6
#ifndef sa_addc
    #define sa_addc  sa_addc_if
#endif


SA_CPP_TRAILER
#endif

