/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   spstaticmem.h

@brief
   
    Defines a class for dividing up a memory chunk.  The chunk is divided into
    regions.  The user does these steps:
        1. allocate a static memory allocator (SMA) and initialize it
        2. register the size and count of each region in a memory chunk
        3. query the total size of memory chunk required
        4. allocate and assign the memory chunk to the allocator
        5. use the allocator to retrieve pointers to individual regions

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_static_mem.h,v 1.4 2005/11/16 18:30:05 menz Exp $
*/
#ifndef __sa_static_memory_allocator_h__
#define __sa_static_memory_allocator_h__

#include "system/sa_system.h"
#include "runtime/sa_ring.h"

SA_CPP_HEADER

// 
// Constant value for magic field.
//
#define SA_SMA_MAGIC 0x01010203
#define SA_SMA_ASSERT(_sa_sma_)                 \
    ASSERT (_sa_sma_);                          \
    ASSERT ((_sa_sma_)->magic == SA_SMA_MAGIC)  \

//
// One large memory chunk is divided in regions.  Each region looks is represented
// with this structure.
//
typedef struct _SA_STATIC_REGION {
    
    u32 offset;
    u32 count;
    u32 size;
    u32 alignment;

    #if defined (SA_DEBUG)
        const c8* name;
    #endif

} SA_STATIC_REGION, *PSA_STATIC_REGION;

//
// One large memory chunk is divided in regions.  This structure 
// represents one memory chunk that requires dividing up.
//
typedef struct _SA_STATIC_MEMORY_ALLOCATOR {

    u32 magic;
    u32 maxRegions;    
    PVOID base_va;
    u64 base_pa;
    u32 total_size;

    SA_STATIC_REGION *regions;

} SA_STATIC_MEMORY_ALLOCATOR, *PSA_STATIC_MEMORY_ALLOCATOR;


/*!
@brief
    Initialize the SMA with the max number of regions.
@param
    sma - object is init
@param
    maxRegions - number of regions (not all may be used)
@param
    regions - array of maxRegions SA_STATIC_REGION structs
@return 
    none
*/
STATIC INLINE
void
sa_sma_init (PSA_STATIC_MEMORY_ALLOCATOR sma, 
             u32 maxRegions, 
             SA_STATIC_REGION *regions) 
{
    u32 i;

    ASSERT (sma);
    ASSERT (regions);

    sa_memset (sma, 0, sizeof (SA_STATIC_MEMORY_ALLOCATOR));
    sa_memset (regions, 0, sizeof (SA_STATIC_REGION) * maxRegions);

    sma->maxRegions = maxRegions;
    sma->regions = regions;
    sma->magic = SA_SMA_MAGIC;
    sma->total_size = 0;

    for (i=0; i<maxRegions; i++) {
        sma->regions[i].alignment = 1;
    }
}

STATIC INLINE
void
sa_sma_destroy (
    PSA_STATIC_MEMORY_ALLOCATOR sma)
{
    SA_SMA_ASSERT (sma);
    sa_memset (sma->regions, 0, sizeof (SA_STATIC_REGION) * sma->maxRegions);
    sa_memset (sma, 0, sizeof (SA_STATIC_MEMORY_ALLOCATOR));    
}

STATIC INLINE
PVOID
sa_sma_get_base_va (PSA_STATIC_MEMORY_ALLOCATOR sma) {
    return sma->base_va;
}

STATIC INLINE
u64
sa_sma_get_base_pa (PSA_STATIC_MEMORY_ALLOCATOR sma) {
    SA_SMA_ASSERT (sma);
    return sma->base_pa;
}


/*!
@brief
    Registers the size of a given region.  
@param
    sma - static memory allocator object
@param
    region - index to a region
@param
    size - byte size of one item in region
@param
    count - number of items in region
@param
    alignment - byte alignment for starting address.
@param
    name - optional name for region for debug prints
@return 
    none
*/
STATIC INLINE
void
sa_sma_register (PSA_STATIC_MEMORY_ALLOCATOR sma, 
                 u32 region, 
                 u32 size, 
                 u32 count,
                 u32 alignment,
                 const c8* name) {

    SA_SMA_ASSERT (sma);
    ASSERT (region < sma->maxRegions);
    ASSERT (alignment > 0);

    sma->total_size = 0; // Clear the cached total size

    sma->regions[region].count = count;
    sma->regions[region].size = size;
    sma->regions[region].alignment = alignment;
    #if defined(SA_DEBUG)
        sma->regions[region].name = name;
    #endif

    TRACE (DL_INFO, "Register region: %s size:%d count:%d totalKB:%d", name, size, count, (count*size/1024));
}


STATIC INLINE u32
sa_sma_get_region_size (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region) {

    SA_SMA_ASSERT (sma);
    ASSERT (region < sma->maxRegions);          
    return sma->regions[region].size;
}

STATIC INLINE u32
sa_sma_get_region_count (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region) {

    SA_SMA_ASSERT (sma);
    ASSERT (region < sma->maxRegions);          
    return sma->regions[region].count;
}

STATIC INLINE u32
sa_sma_get_region_alignment (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region) {

    SA_SMA_ASSERT (sma);
    ASSERT (region < sma->maxRegions);          
    return sma->regions[region].alignment;
}


STATIC INLINE u32
sa_sma_get_region_offset (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region, u32 index) {
    
    u32 offset = 0;
    
    SA_SMA_ASSERT (sma);
    ASSERT (sma->total_size > 0);
    ASSERT (region < sma->maxRegions);  
    ASSERT (index < sma->regions[region].count);
    ASSERT (sma->regions[region].alignment > 0);
    ASSERT (sma->regions[region].size > 0);
    ASSERT (sma->regions[region].count > 0);    
    
    ASSERT (sma->base_va || sma->base_pa);

    offset = sma->regions[region].offset + sma->regions[region].size * index;    

    return offset;
}


STATIC 
INLINE
PVOID
sa_sma_get_region_va (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region, u32 index) {
    u64 va = 0;
    
    SA_SMA_ASSERT (sma);
    ASSERT (region < sma->maxRegions);  

    if (sma->regions[region].size * sma->regions[region].count == 0) {
        return NULL;
    }

    if (index >= sma->regions[region].count) {
        return NULL;
    }

    va = SA_PTR_TO_U64 (sma->base_va) + sa_sma_get_region_offset (sma, region, index);
    ASSERT (va % sma->regions[region].alignment == 0);

    return SA_U64_TO_PTR (va);
}

STATIC 
INLINE
u64
sa_sma_get_region_pa (PSA_STATIC_MEMORY_ALLOCATOR sma, u32 region, u32 index) {
    u64 pa = 0;

    SA_SMA_ASSERT (sma);    
    ASSERT (region < sma->maxRegions);  

    if (sma->regions[region].size * sma->regions[region].count == 0) {
        return 0;
    }

    if (index >= sma->regions[region].count) {
        return 0;
    }
    
    pa = sma->base_pa + sa_sma_get_region_offset (sma, region, index);
    ASSERT (pa % sma->regions[region].alignment == 0);

    return pa;
}



/*!
@brief
    Gives the required size of the memory chunk to fit all the registered regions.
@param
    sma - static memory allocator object
@return 
    size of entire memory chunk
*/
STATIC 
INLINE
u32
sa_sma_get_total_size (PSA_STATIC_MEMORY_ALLOCATOR sma) {
    
    u32 totalSize = 0;    
    u32 i;

    SA_SMA_ASSERT (sma);

    if (sma->total_size != 0) 
        return sma->total_size;

    //
    // The alignment of the base is assumed to be random, so initialize the
    // size based on the required alignment of the first region.
    //
    if (sma->maxRegions > 0) {
        totalSize = sma->regions[0].alignment + sma->regions[0].size * sma->regions[0].count;
    }

    //
    // Compute the number of bytes required to support the remaining blocks.
    //
    for (i = 1; i < sma->maxRegions; i++) {                
        if (sma->regions[i].size) {
            ASSERT(sma->regions[i].alignment);
            if (totalSize % sma->regions[i].alignment) {
                totalSize += sma->regions[i].alignment - (totalSize % sma->regions[i].alignment);
            }
            totalSize += sma->regions[i].size * sma->regions[i].count;
        }
    }

    sma->total_size = totalSize;

    TRACE (DL_INFO, "TotalSize: 0x%x (%d)", totalSize, totalSize);

    return totalSize;
}

/*!
@brief
    Assigns a memory chunk to the allocator.  Call this before calling
    the allocate methods.  The base pointer must be >= sa_sma_GetTotalSize.
@param
    sma - static memory allocator object
@return 
    none
*/
STATIC 
INLINE
void
sa_sma_assign_base (PSA_STATIC_MEMORY_ALLOCATOR sma, PVOID base_va, u64 base_pa) {
    
    u64 va = SA_PTR_TO_U64 (base_va);
    u32 pad;
    u32 offset = 0;
    u32 i;

    SA_SMA_ASSERT (sma);
    ASSERT (base_va || base_pa);

    sma->base_va = base_va;
    sma->base_pa = base_pa;
    
    for (i = 0; i < sma->maxRegions; i++) {
        
        pad = 0;
        if (sma->regions[i].alignment) {
            if (va % sma->regions[i].alignment) {
                pad = sma->regions[i].alignment - ((u32)va % sma->regions[i].alignment);
            } 
        }
        
        sma->regions[i].offset = offset + pad;

        offset += pad + sma->regions[i].size * sma->regions[i].count;
        va     += (unsigned)(pad + (unsigned)(sma->regions[i].size * sma->regions[i].count));

        ASSERT (offset <= sma->total_size);
    }

}


SA_CPP_TRAILER

#endif

