/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_.h

@brief
   
    Includes device abstraction for reading registers, mapping bus
    addresses, etc.

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_dev.h,v 1.11 2006/03/09 17:51:19 menz Exp $
*/
#ifndef __sa_dev_h__
#define __sa_dev_h__

#include "sa_system.h"

SA_CPP_HEADER

typedef enum 
{
    SA_BAR_TYPE_INVALID     = 0,

    SA_BAR_TYPE_BOOT,
    SA_BAR_TYPE_CSR,
    SA_BAR_TYPE_PD,
    SA_BAR_TYPE_PCI,

    SA_BAR_TYPE_MAX

} SA_DEV_BAR_TYPE;


//
// MemOrIOMapped has the following two values
// for reference purpose
//
typedef enum 
{
    SA_IO_MAPPED = 0x01,
    SA_MEM_MAPPED = 0x02

} SA_DEV_MAPPED_TYPE;


typedef struct _SA_DEV_BAR_LOCATIONS
{
    void*               base_va;  // Virtual Address
    u64                 base_pa;  // Physical Address
    u32                 length;   // Length of register space
    
    SA_DEV_MAPPED_TYPE  mem_or_io_mapped; // io or memory mapped
    SA_DEV_BAR_TYPE     type;             // What is this BAR used for.

} SA_DEV_BAR_LOCATIONS, *PSA_DEV_BAR_LOCATIONS;


typedef u64 SA_DEV_BUS_ADDRESS, *PSA_DEV_BUS_ADDRESS;

#if defined(SA_STORAGE_MINIPORT)
    #include "sa_dev_sport.h"
#elif defined(SA_NDIS)
    #include "sa_dev_ndis.h"
#elif defined(SA_WIN_USER)
    #include "sa_dev_win32u.h"
#elif defined(SA_SIM)
    #include "sim/sa_dev_sim.h"
#elif defined(SA_LINUXU)
    #include "linuxu/sa_dev_linuxu.h"
#elif defined(SA_LINUXK)
    #include "linuxk/sa_dev_linuxk.h"
#elif defined(SA_WIN_KERNEL)
    #include "win32k/sa_dev_win32k.h"
#else
    #error "No OS dev include file defined."
#endif


SA_DEV*
sa_dev_alloc (
    SA_DEV_BAR_LOCATIONS* bars,
    u32                   num_bars,
    SA_DEV_OS_HANDLE      os_handle);

void
sa_dev_free (
    SA_DEV* dev);

SA_STATUS
sa_dev_create (
    SA_DEV_BAR_LOCATIONS* bars,
    u32                   num_bars,
    SA_DEV_OS_HANDLE      os_handle, // device extension pointer for storport/ndis
    SA_DEV*               device);

void
sa_dev_destroy (
    SA_DEV      *dev);


// This assumes that each type defines the members "bars" and "num_bars"
// Returns the bar number, or ~0 if the given type doesn't exist.
STATIC INLINE u32
sa_get_bar_num (SA_DEV *dev,
                SA_DEV_BAR_TYPE type)
{
    u32 i;
    for (i=0; i<dev->num_bars; i++) {
        if (type == dev->bars[i].type) {
            return i;
        }
    }

    return ~(u32)0; // 
}

// Internal function.
STATIC INLINE void
_sa_cache_bar_nums (SA_DEV *dev) {
    dev->csr_bar_num = sa_get_bar_num (dev, SA_BAR_TYPE_CSR);
    dev->pd_bar_num  = sa_get_bar_num (dev, SA_BAR_TYPE_PD);
    dev->pci_bar_num = sa_get_bar_num (dev, SA_BAR_TYPE_PCI);
}

// Assume these values are cached in the sa_dev.
STATIC FORCEINLINE u32 sa_get_csr_bar (SA_DEV *dev) { return dev->csr_bar_num; }
STATIC FORCEINLINE u32 sa_get_pd_bar (SA_DEV *dev) { return dev->pd_bar_num; }
STATIC FORCEINLINE u32 sa_get_pci_bar (SA_DEV *dev) { return dev->pci_bar_num; }

// Assume the bar array exists in the sa_dev object.
STATIC INLINE PVOID sa_dev_get_bar_offset (SA_DEV *dev, u32 barnum) {
    ASSERT (barnum < dev->num_bars);
    return dev->bars[barnum].base_va;
}

u8  sa_dev_read_u8  (SA_DEV *dev, u32 bar, u32 offset);
u16 sa_dev_read_u16 (SA_DEV *dev, u32 bar, u32 offset);
u32 sa_dev_read_u32 (SA_DEV *dev, u32 bar, u32 offset);

void sa_dev_write_u8  (SA_DEV *dev, u32 bar, u32 offset, u8  value);
void sa_dev_write_u16 (SA_DEV *dev, u32 bar, u32 offset, u16 value);
void sa_dev_write_u32 (SA_DEV *dev, u32 bar, u32 offset, u32 value);

SA_STATUS sa_dev_read_buffer_u8 (SA_DEV* dev, u32 bar, u32 offset, u8* buffer, u32 count);
SA_STATUS sa_dev_read_buffer_u16 (SA_DEV* dev, u32 bar, u32 offset, u16* buffer, u32 count);
SA_STATUS sa_dev_read_buffer_u32 (SA_DEV* dev, u32 bar, u32 offset, u32* buffer, u32 count);

SA_STATUS sa_dev_write_buffer_u8  (SA_DEV* dev, u32 bar, u32 offset, u8* buffer, u32 count);
SA_STATUS sa_dev_write_buffer_u16 (SA_DEV *dev, u32 bar, u32 offset, u16* buffer, u32 count);
SA_STATUS sa_dev_write_buffer_u32 (SA_DEV* dev, u32 bar, u32 offset, u32* buffer, u32 count);

SA_PHYSICAL_ADDRESS
sa_dev_get_bus_address (
    SA_DEV*                     dev,
    PSA_DEV_BUS_ADDRESS_CONTEXT context, // e.g. SRB if in the SRB extension
    void*                       va,
    u32                         minimum_length);

void
sa_dev_stall (
    SA_DEV* dev,
    u32     us_to_stall);
    
void
sa_dev_log_error (
    SA_DEV*                   dev,
    PSA_DEV_LOG_ERROR_CONTEXT context,
    u32                       error_code,
    u32                       unique_code);


SA_CPP_TRAILER

#endif



