/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_endian.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_endian.h,v 1.8 2006/03/13 19:42:46 menz Exp $
*/
#ifndef __sa_endian_h__
#define __sa_endian_h__

#include "sa_types.h"

SA_CPP_HEADER

// Assume all platforms are little endian right now!!
#define SA_HOST_LITTLE_ENDIAN
#undef  SA_HOST_BIG_ENDIAN

//
// Macros for swapping bytes. 
// Use the INLINE functions below for potential performance boost!
//
#define SA_SWAP16(_x_)   (u16) (                        \
                        (((u16)(_x_) & 0x00FF) << 8) |  \
                        (((u16)(_x_) & 0xFF00) >> 8) )  \


#define SA_SWAP24(_x_)   (u32) (                                 \
                            (((u32)(_x_) & 0x000000FFUL) << 16) |  \
                            (((u32)(_x_) & 0x0000FF00UL)      ) |  \
                            (((u32)(_x_) & 0x00FF0000UL) >> 16) )  \

#define SA_SWAP32(_x_)   (u32) (                                 \
                            (((u32)(_x_) & 0x000000FFUL) << 24) |  \
                            (((u32)(_x_) & 0x0000FF00UL) << 8)  |  \
                            (((u32)(_x_) & 0x00FF0000UL) >> 8)  |  \
                            (((u32)(_x_) & 0xFF000000UL) >> 24) )  \

#define SA_SWAP64(_x_)   (u64) (                                         \
                            (((u64)(_x_) & 0x00000000000000FFULL) << 56) |  \
                            (((u64)(_x_) & 0x000000000000FF00ULL) << 40) |  \
                            (((u64)(_x_) & 0x0000000000FF0000ULL) << 24) |  \
                            (((u64)(_x_) & 0x00000000FF000000ULL) << 8)  |  \
                            (((u64)(_x_) & 0x000000FF00000000ULL) >> 8)  |  \
                            (((u64)(_x_) & 0x0000FF0000000000ULL) >> 24) |  \
                            (((u64)(_x_) & 0x00FF000000000000ULL) >> 40) |  \
                            (((u64)(_x_) & 0xFF00000000000000ULL) >> 56) )  \


// ----------------------------------------------------------------------------
// These versions use assembly when possible.
// ----------------------------------------------------------------------------

// This function does a byte swap on u16 quantity.
STATIC FORCEINLINE u16 sa_asm_swap16 (u16 data)
{
    u16 return_value;

    #if defined(SA_X86) && defined(SA_MICROSOFT)
        _asm mov ax, data
        _asm xchg al, ah
        _asm mov return_value, ax
    #else 
        return_value = SA_SWAP16(data);
    #endif

    return return_value;
}

// This function does a byte swap on an u32 quantity.
STATIC FORCEINLINE u32 sa_asm_swap32 (u32 data)
{
    #if defined(SA_X86) && defined(SA_MICROSOFT)
        _asm mov eax, data
        _asm bswap eax
        _asm mov data, eax
    #elif defined(SA_X86) && defined(SA_LINUX)
        __asm__ __volatile__ ("bswap %0" : "=r" (data) : "0" (data));    
    #else
        data = SA_SWAP32 (data);
    #endif
    
    return data;
}

// This function does a byte swap on an u32 quantity.
STATIC FORCEINLINE u64 sa_asm_swap64 (u64 data)
{
    u64 return_value;
    return_value = ((u64) sa_asm_swap32 ((u32)data) << 32) |
                          sa_asm_swap32 ((u32)(data >> 32));
    return return_value;
}

// ----------------------------------------------------------------------------
// Define these macros based on current performance measurements.
// On a Pentium D 2.8GHz, Dell SC430.
//
//   sar_w32u info  u16 swap C-code avg ticks/loop: 14
//   sar_w32u info  u16 swap ASM-code avg ticks/loop: 24
//   sar_w32u info  u32 swap C-code avg ticks/loop: 18
//   sar_w32u info  u32 swap ASM-code avg ticks/loop: 17
//   sar_w32u info  u64 swap C-code avg ticks/loop: 62
//   sar_w32u info  u64 swap ASM-code avg ticks/loop: 28
//
// Note: There is a substantial dependence on processor for gaining 
// performance with inline assembly so by default I'm using the C code versions.
//
// ----------------------------------------------------------------------------
#ifndef sa_swap16
    #define sa_swap16 SA_SWAP16
#endif
#ifndef sa_swap32
    #define sa_swap32 SA_SWAP32
#endif
#ifndef sa_swap64
    #define sa_swap64 SA_SWAP64
#endif


// For a little-endian system, swaps network (n) to/from host (h) for
// either 32 bit longs (l) or 16 bit shorts (s).
#if defined(SA_HOST_LITTLE_ENDIAN)

    // Host and Network byte order are different.

    // Linux defines these as inline functions -- we will use its native definition.
    #if !defined(SA_LINUX_KERNEL) && !defined(SA_LINUX_USER)
        #if !defined(ntohl)
            #define ntohl sa_swap32
        #endif

        #if !defined(ntohs)
            #define ntohs sa_swap16
        #endif

        #if !defined(htonl)
            #define htonl sa_swap32
        #endif

        #if !defined(htons)
            #define htons sa_swap16
        #endif
    #endif

    // Define host-to-little-endian and host-to-big-endian macros.
    #define SA_HOST_TO_LE16(_x_) (_x_)
    #define SA_HOST_TO_LE32(_x_) (_x_)
    #define SA_HOST_TO_LE64(_x_) (_x_)   

    #define SA_HOST_TO_BE16  sa_swap16
    #define SA_HOST_TO_BE32  sa_swap32
    #define SA_HOST_TO_BE64  sa_swap64   

    #define SA_LE16_TO_HOST(_x_) (_x_)
    #define SA_LE32_TO_HOST(_x_) (_x_)
    #define SA_LE64_TO_HOST(_x_) (_x_)     

    #define SA_BE16_TO_HOST sa_swap16
    #define SA_BE32_TO_HOST sa_swap32
    #define SA_BE64_TO_HOST sa_swap64

    // Define a query function for endianess.
    STATIC FORCEINLINE boolean sa_is_host_little_endian (void) { 
        return TRUE; 
    }
    STATIC FORCEINLINE boolean sa_is_host_big_endian (void)    { 
        return !sa_is_host_little_endian (); 
    }

#else 

    // Big-Endian machine.
    // Host and Network order are the same.
    #if !defined(SA_LINUX)
        #if !defined(ntohl)
            #define ntohl(_x_) (_x_)
        #endif

        #if !defined(ntohs)
            #define ntohs(_x_) (_x_)
        #endif

        #if !defined(htonl)
            #define htonl(_x_) (_x_)
        #endif

        #if !defined(htons)
            #define htons(_x_) (_x_)
        #endif
    #endif

    // Define host-to-little-endian and host-to-big-endian macros.
    #define SA_HOST_TO_LE16 sa_swap16
    #define SA_HOST_TO_LE32 sa_swap32
    #define SA_HOST_TO_LE64 sa_swap64      

    #define SA_HOST_TO_BE16(_x_) (_x_)
    #define SA_HOST_TO_BE32(_x_) (_x_)
    #define SA_HOST_TO_BE64(_x_) (_x_)

    #define SA_LE16_TO_HOST sa_swap16
    #define SA_LE32_TO_HOST sa_swap32
    #define SA_LE64_TO_HOST sa_swap64      

    #define SA_BE16_TO_HOST(_x_) (_x_)
    #define SA_BE32_TO_HOST(_x_) (_x_)
    #define SA_BE64_TO_HOST(_x_) (_x_)

    // Define a query function for endianess.
    STATIC FORCEINLINE boolean sa_is_host_little_endian () { 
        return FALSE; 
    }
    STATIC FORCEINLINE boolean sa_is_host_big_endian ()    { 
        return !sa_is_host_little_endian (); 
    }

#endif


// Functions for converting an array of bytes to an integer, or 
// placing an integer into a byte array.  Currently, these functions assume
// that integers can start on any byte alignment, so the array pointer can
// be safely casted to the appropriate integer type.

// Put as native endianess
#define SA_PUT8(_ptr_, _value_)  ((*(u8*)(_ptr_))   = (u8)(_value_))
#define SA_PUT16(_ptr_, _value_)  ((*(u16*)(_ptr_)) = ((u16)(_value_)))
#define SA_PUT32(_ptr_, _value_)  ((*(u32*)(_ptr_)) = ((u32)(_value_)))
#define SA_PUT64(_ptr_, _value_)  ((*(u64*)(_ptr_)) = ((u64)(_value_)))

// Put as little endian
#define SA_PUT_LE8(_ptr_, _value_)  ((*(u8*)(_ptr_)) = (u8)(_value_))
#define SA_PUT_LE16(_ptr_, _value_)  ((*(u16*)(_ptr_)) = SA_HOST_TO_LE16 ((u16)(_value_)))
#define SA_PUT_LE32(_ptr_, _value_)  ((*(u32*)(_ptr_)) = SA_HOST_TO_LE32 ((u32)(_value_)))
#define SA_PUT_LE64(_ptr_, _value_)  ((*(u64*)(_ptr_)) = SA_HOST_TO_LE64 ((u64)(_value_)))

// Put as big endian
#define SA_PUT_BE8(_ptr_, _value_)  ((*(u8*)(_ptr_)) = (u8)(_value_))
#define SA_PUT_BE16(_ptr_, _value_)  ((*(u16*)(_ptr_)) = SA_HOST_TO_BE16 (((u16)(_value_)))
#define SA_PUT_BE32(_ptr_, _value_)  ((*(u32*)(_ptr_)) = SA_HOST_TO_BE32 ((u32)(_value_)))
#define SA_PUT_BE64(_ptr_, _value_)  ((*(u64*)(_ptr_)) = SA_HOST_TO_BE64 ((u64)(_value_)))

// Get as native endianess
#define SA_GET8(_ptr_)  (*((u8*)_ptr_))
#define SA_GET16(_ptr_) ((*((u16*)_ptr_)))
#define SA_GET32(_ptr_) ((*((u32*)_ptr_)))
#define SA_GET64(_ptr_) ((*((u64*)_ptr_)))

// Get as little endian
#define SA_GET_LE8(_ptr_)  (*((u8*)_ptr_))
#define SA_GET_LE16(_ptr_) (SA_HOST_TO_LE16(*((u16*)_ptr_)))
#define SA_GET_LE32(_ptr_) (SA_HOST_TO_LE32(*((u32*)_ptr_)))
#define SA_GET_LE64(_ptr_) (SA_HOST_TO_LE64(*((u64*)_ptr_)))

// Get as big endian
#define SA_GET_BE8(_ptr_)  (*((u8*)_ptr_))
#define SA_GET_BE16(_ptr_) (SA_HOST_TO_BE16(*((u16*)_ptr_)))
#define SA_GET_BE32(_ptr_) (SA_HOST_TO_BE32(*((u32*)_ptr_)))
#define SA_GET_BE64(_ptr_) (SA_HOST_TO_BE64(*((u64*)_ptr_)))

//
// Functions for reversing the byte, words, or dwords in an array.
//
STATIC INLINE
void sa_reverse8 (u8* buffer, u32 num) {
    u32 i;
    u8 tmp;
    for (i=0; i<num/2; i++) {
        tmp = buffer[i];
        buffer[i] = buffer[num - 1 - i];
        buffer[num - 1 - i] = tmp;
    }
}

STATIC INLINE
void sa_reverse16 (u16* buffer, u32 num) {
    u32 i;
    u16 tmp;
    for (i=0; i<num/2; i++) {
        tmp = buffer[i];
        buffer[i] = buffer[num - 1 - i];
        buffer[num - 1 - i] = tmp;
    }
}

STATIC INLINE
void sa_reverse32 (u32* buffer, u32 num) {
    u32 i;
    u32 tmp;
    for (i=0; i<num/2; i++) {
        tmp = buffer[i];
        buffer[i] = buffer[num - 1 - i];
        buffer[num - 1 - i] = tmp;
    }
}

//
// Reverses order of bits in x.
//
STATIC
INLINE u32 sa_bit_reverse (u32 x)
{
    u32 out = 0, shift = 31;

    while (x) {
        out |= (x&1) << shift;
        shift--;
        x >>= 1;
    }

    return out;
}


SA_CPP_TRAILER

#endif // __sa_endian_h__

