/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_mem.h

@brief
    
    Includes operating system memory management abstractions.

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_mem.h,v 1.19 2006/03/10 19:13:08 menz Exp $
*/
#ifndef __sa_mem_h__
#define __sa_mem_h__

#include "sa_types.h"
#include "sa_assert.h"
#include "sa_std.h"

SA_CPP_HEADER

#if defined(SA_STORAGE_MINIPORT)

    #include "sa_mem_sport.h"
    typedef u32 SA_MEM_ALLOC_CONTEXT;

#elif defined(SA_NDIS)

    #include "sa_mem_ndis.h"
    typedef u32 SA_MEM_ALLOC_CONTEXT;

#elif defined(SA_KERNEL) && defined(SA_MICROSOFT)

    #include "win32k/sa_mem_win32k.h"      
    typedef u32 SA_MEM_ALLOC_CONTEXT;

#elif defined(SA_LINUXK)

    #include "linuxk/sa_mem_linuxk.h"
    
    typedef struct _SA_MEM_ALLOC_CONTEXT { 
        u32   flags;
        PVOID ptr;
	u32   len;
    } SA_MEM_ALLOC_CONTEXT, *PSA_MEM_ALLOC_CONTEXT;

#else
    
    // user mode memory management
    typedef u32 SA_MEM_ALLOC_CONTEXT;
    #include "sim/sa_mem_user.h"
    
#endif


//============================================================================
//                              P R O T O T Y P E S
//============================================================================

// Used for multi-characters tags.  Such as SA_TAG("abcd").  Tag MUST be 4 chars.
#define SA_TAG sa_tag

#if defined(SA_MICROSOFT)
    // Includes an assert in debug to verify number of characters.
    // Microsoft compiler optimizes this to a constant.
    /*lint -e571 */
    STATIC FORCEINLINE u32
    sa_tag (const c8* tag) {
        ASSERT (tag && tag[0] && tag[1] && tag[2] && tag[3] && (tag[4]==0));
        return (((u32)tag[3] << 24) | ((u32)tag[2] << 16) | ((u32)tag[1] << 8) | ((u32)tag[0]));
    }
    /*lint -restore */
#else

    // GCC won't optimize the inline function.  Plus, this is even better.  It uses
    // a block expression in a macro, allowing a simple compile time assert to verify
    // the length of the input string.
    // The gcc compiler optimizes the entire thing to a constant with -02.
    #define sa_tag(__four_char_string__) ({                                           \
        SA_GLOBAL_C_ASSERT (tag_must_be_4_char, sizeof (__four_char_string__) == 5);  \
        (((u32)(__four_char_string__[3])<<24) |                                       \
        ((u32)(__four_char_string__[2])<<16) |                                        \
        ((u32)(__four_char_string__[1])<<8)  |                                        \
        ((u32)(__four_char_string__[0])));                                            \
    })                                                                                \

#endif


// Zero a struct, automatically determine the size of the struct with sizeof.
// This avoids a common bug, like sa_zero_mem (ptr, sizeof(ptr)) where you 
// accidentally only zero the first 4 bytes -- the size of the ptr instead of
// the size of the structure.
#define SA_ZERO_MEM(_ptr_) sa_zero_mem ((_ptr_), sizeof (*(_ptr_)))

// Efficently zero objects of a given size by casting to integers and assigining zero.
#define SA_ZERO_16(_ptr_)  (*((u16*)(_ptr_)) = 0)
#define SA_ZERO_32(_ptr_)  (*((u32*)(_ptr_)) = 0UL)
#define SA_ZERO_64(_ptr_)  (*((u64*)(_ptr_)) = 0ULL)
#define SA_ZERO_96(_ptr_)  { SA_ZERO_64(_ptr_); SA_ZERO_32((u64*)_ptr_ + 1); }
#define SA_ZERO_128(_ptr_) { SA_ZERO_64(_ptr_); SA_ZERO_64((u64*)_ptr_ + 1); }

// Efficiently zero 32 struct by casting to an integer and assigning. 
// Verifies that the object is bigger than the request size.
#define SA_ZERO_STRUCT32(_pstruct_) {      \
    SA_C_ASSERT(sizeof(*_pstruct_) >= 4);  \
    SA_ZERO_32 (_pstruct_); }              \

#define SA_ZERO_STRUCT64(_pstruct_) {      \
    SA_C_ASSERT(sizeof(*_pstruct_) >= 8);  \
    SA_ZERO_64 (_pstruct_); }              \

#define SA_ZERO_STRUCT96(_pstruct_) {       \
    SA_C_ASSERT(sizeof(*_pstruct_) >= 12);  \
    SA_ZERO_96 (_pstruct_); }               \

#define SA_ZERO_STRUCT128(_pstruct_) {      \
    SA_C_ASSERT(sizeof(*_pstruct_) >= 16);  \
    SA_ZERO_128 (_pstruct_); }              \


// Optimized as the kernel version for the given platform.
STATIC INLINE void*
sa_zero_mem (PVOID ptr, u32 bytes)
{
    #if defined(SA_WIN_KERNEL)
        return RtlZeroMemory (ptr, (SA_SIZE_T)bytes);
    #else
        return memset (ptr, 0, (SA_SIZE_T)bytes);
    #endif
}

// Defined as the intrinsic for windows, or the standard C runtime function.
STATIC INLINE void*
sa_memset_zero (PVOID ptr, u32 bytes)
{
    return memset (ptr, 0, (SA_SIZE_T)bytes);
}


STATIC INLINE void*
sa_memset (PVOID ptr, u32 value, u32 bytes)
{
    return memset (ptr, value, (SA_SIZE_T)bytes);
}

STATIC INLINE void*
sa_memcpy (PVOID dest, const void* src, u32 bytes)
{
    return memcpy (dest, src, (SA_SIZE_T)bytes);
}


// Returns value of dest.
STATIC INLINE void*
sa_memmove (PVOID dest, const void* src, u32 bytes)
{
    return (void*) memmove (dest, src, (SA_SIZE_T)bytes);
}

STATIC INLINE i32
sa_memcmp (const void* a, const void* b, u32 bytes)
{
    return memcmp (a, b, (SA_SIZE_T)bytes);
}


// Flags for the malloc functions.
#define  SA_MALLOC_NORMAL       0x00000001    // Allocates pinned kernel memory
#define  SA_MALLOC_PAGEABLE     0x00000002    // Allocates pageable kernel memory

//
//  Allocates memory.  This function is generally used to allocate
//  software owned (control) data structures and not for memory shared with
//  the controller.
//
//  Size        - Size in bytes to allocate
//  Flag        - SA_MALLOC_NORMAL or SA_MALLOC_PAGEABLE to allocate pageable memory
//  Tag         - Tracking tag for this memory allocation
//  Context     - Optional (opaque) struct pointer that SA uses to remember the allocation
//                parameters.  To be truly portable, a pointer to a context object should be
//                used.  This may be on the stack. It is safe to copy the resulting object
//                using sa_memcpy -- for example it may be copied to the beginning of the newly
//                allocated memory buffer.  The resulting copied version may be used during
//                freeing the memory.
//                Many environments do not require the SA_MEM_ALLOC_CONTEXT object and 
//                it is acceptable to pass NULL for this parameter if the code does not need
//                to be completely portable.
//
//  Returns a pointer to newly allocated virtual memory or NULL on failure.
//
//  DO NOT call this function directly.  Use the sa_malloc #define with the same 
//  prototype to select either the debug of release version.
//
PVOID _sa_malloc_release (u32 size, u32 flag, u32 tag, SA_MEM_ALLOC_CONTEXT *optional_context);    


//  Frees memory allocated by _sa_malloc_release.
//
//  DO NOT call this function directly.  Use the sa_free #define with the same 
//  prototype to select either the debug or release version.
//
void _sa_free_release (IN PVOID ptr, SA_MEM_ALLOC_CONTEXT *optional_context);

#if defined(SA_MEM_DEBUG)

    // Redefine the malloc function in terms of the debug. 
    // These assume a NULL context, which is not always portable for all "flag" settings and environments.
    #define sa_malloc(_size, _flag, _tag) _sa_malloc_debug (_size, _flag, _tag, NULL, __FILE__, __LINE__)
    #define sa_free(_ptr)                 _sa_free_debug (_ptr, NULL)

    // These all include a context, which makes them portable to all environments.
    #define sa_malloc_with_context(_size, _flag, _tag, _pcontext) _sa_malloc_debug (_size, _flag, _tag, _pcontext, __FILE__, __LINE__)
    #define sa_free_with_context                                  _sa_free_debug    

    // Debug version of sa_free.  Do not call these directly.
    PVOID _sa_malloc_debug (u32 size, u32 flag, u32 tag, SA_MEM_ALLOC_CONTEXT *optional_context, const c8* filename, u32 line);    
    void _sa_free_debug (IN PVOID ptr, SA_MEM_ALLOC_CONTEXT *optional_context);
        
#else

    #define sa_malloc(_size, _flag, _tag) _sa_malloc_release(_size, _flag, _tag, NULL)
    #define sa_free(_ptr)                 _sa_free_release(_ptr, NULL)    

    #define sa_malloc_with_context        _sa_malloc_release
    #define sa_free_with_context          _sa_free_release    

#endif


#if defined(SA_MEM_DEBUG)

    // Initializes debug tracking.  Tracking is initialized by default during the
    // first sa_malloc.  However, in a multi-threaded environment the initialization
    // must be done only once.  That means either this call to init or the first
    // call to sa_malloc must be serialized with respect to all threads.  (In a single
    // threaded environment, all calls to sa_malloc are serialized, so there is no problem.)
    void _sa_mem_debug_init (void);

    // Prints allocated memory regions to the debugger.
    void _sa_mem_debug_dump (void);

    // Scans and check the magic fields in each memory allocation, attempting
    // to spot any memory corruption.  Prints total allocations and byte count.
    void _sa_mem_debug_scan_magic (void);

    // Define these macros equal to the function calls during memory debugging.
    #define sa_mem_debug_init       _sa_mem_debug_init
    #define sa_mem_debug_dump       _sa_mem_debug_dump
    #define sa_mem_debug_scan_magic _sa_mem_debug_scan_magic

#else

    // Define the memory debug functions are nothing for release builds.
    #define sa_mem_debug_init()
    #define sa_mem_debug_dump()
    #define sa_mem_debug_scan_magic()

#endif


// Allocates an SGL and the requested size in physical pages. 
// The pages allocated are generally shared with the controller for DMA.
// The resulting pages may NOT have a corresponding virtual address.  The use should 
// call sa_sgl_map before requesting the base PA. 
// 
//    size                       - Size in bytes to allocate
//    highest_acceptable_address - The highest physical address that returned in the SGL
//                                                                                                      
//    Returns a pointer to newly allocated virtual memory
PSA_SGL      
sa_physalloc (
    IN u32                 size,
    IN SA_PHYSICAL_ADDRESS highest_acceptable_address);

// Frees physical memory allocated by sa_physalloc, and it frees the sgl itself.    
void            
sa_physfree (
    IN PSA_SGL Sgl);

// Allocates a SA_SGL, correctly sizing it for the given VA and length.
// For some environments this will include creating the page frame array. (windows)
// After this function completes, it is safe to use the sa_sgl_get_page function to
// query physical page addresses for the SGL.
//
//  virtual_address - previously allocated virtual address
//  length - size in bytes of buffer
//
//  Returns the allocate SG_SGL structure, ready to use.
PSA_SGL
sa_sgl_allocate_for_va (
    IN PVOID virtual_address,
    IN u32   length);

// Frees a SA_SGL data structure allocated by sa_sgl_allocate()
// This guy does not free or touch page frames described by the SGL
void
sa_sgl_free (
    IN PSA_SGL sgl);


// This creates a virtual address mapping of the physical pages described
// by sgl.  The target address can be either UserSpace or KernelSpace. Only
// Kernel space need be implemented on non-windows platforms for now.
//
// If a mapping already exists, it should be returned w/o creating a new mapping.
// It is up to the caller to track mappings, should nested mapping attempts occur.
//
// This function is generally used with sa_physalloc.  However, not all allocated
// SGLs can be successfully mapped due to size limitations.
PVOID
sa_sgl_map (
    IN PSA_SGL Sgl,
    IN boolean UserSpace);

// Undoes the VA mapping done by sa_sgl_map.  
// This function is automatically called by sa_sgl_free if a mapping exists at that point.
void
sa_sgl_unmap (PSA_SGL sgl);

// This function creates an SGL describing a physically and virtually contiguous region.
// This function is not supported in all environments -- check the return status.
// The SA_SGL structure must be previously allocated.
SA_STATUS
sa_sgl_create_contiguous (
    PVOID                virtual_address,
    SA_PHYSICAL_ADDRESS  physical_address,
    u32                  length,
    PSA_SGL              sgl);

// Destroys an sgl created with sa_sgl_create_contiguous.
void
sa_sgl_destroy_contiguous (PSA_SGL sgl);


// Translates a virtual address to physical - virt_to_phys
SA_PHYSICAL_ADDRESS
sa_get_physical_address (
    PVOID virtual_address );


// Used to generate a mapping from a physical address to a virtual
// address.  This assumes a physically contiguous buffer.
PVOID
sa_map_page(
    IN      SA_PHYSICAL_ADDRESS  PhysicalAddress,
    IN      SA_SIZE_T            NumberOfBytes );

// Undoes the mapping done by sa_map_page
void
sa_unmap_page(
    IN      PVOID                VirtualAddress,
    IN      SA_SIZE_T            NumberOfBytes );


SA_CPP_TRAILER

#endif // __sa_mem_h__

