/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_sync.h

@brief

    Synchronization objects.
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_sync.h,v 1.10 2006/03/10 20:22:40 menz Exp $
*/
#ifndef __sa_sync_h__
#define __sa_sync_h__

#include "sa_types.h"
#include "sa_status.h"
#include "sa_time.h"

SA_CPP_HEADER

// ----------------------------------------------------------------------
// Define the macro below to enable spinlock debug code.
// This can be defined in the client makefile without recompiling SA.
// ----------------------------------------------------------------------
// #define SA_SPINLOCK_DEBUG

#if defined(SA_STORAGE_MINIPORT)

    typedef u32   SA_REAL_SPINLOCK, *PSA_REAL_SPINLOCK;
    typedef PVOID SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef PVOID SA_EVENT,      *PSA_EVENT;
    typedef u32   SA_IRQ,        *PSA_IRQ;

#elif defined(SA_NDIS)
    
    typedef NDIS_SPIN_LOCK SA_REAL_SPINLOCK,   *PSA_REAL_SPINLOCK;
    typedef PVOID          SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef NDIS_EVENT     SA_EVENT,      *PSA_EVENT;
    typedef u32            SA_IRQ,        *PSA_IRQ;    

    #include "sa_sync_ndis.h" // inline definitions

#elif defined(SA_LINUX_KERNEL)

    #include <linux/spinlock.h>
    #include <linux/errno.h>
    #include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
    #include <asm/semaphore.h>
#else
    #include <linux/semaphore.h>
#endif    
    typedef spinlock_t        SA_REAL_SPINLOCK,   *PSA_REAL_SPINLOCK;
    typedef struct semaphore  SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef struct semaphore  SA_EVENT,      *PSA_EVENT;
    typedef unsigned long     SA_IRQ,        *PSA_IRQ;
    
#elif defined(SA_WIN_USER)

    typedef HANDLE SA_REAL_SPINLOCK,   *PSA_REAL_SPINLOCK;
    typedef HANDLE SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef HANDLE SA_EVENT,      *PSA_EVENT;
    typedef u32    SA_IRQ,        *PSA_IRQ;

#elif defined(SA_WIN_KERNEL)

    typedef KSPIN_LOCK SA_REAL_SPINLOCK,   *PSA_REAL_SPINLOCK;
    typedef FAST_MUTEX SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef KEVENT     SA_EVENT,      *PSA_EVENT;
    typedef KIRQL      SA_IRQ,        *PSA_IRQ;

#elif defined(SA_LINUXU)

    #include <pthread.h>

    typedef u32 SA_REAL_SPINLOCK, *PSA_REAL_SPINLOCK;
    typedef pthread_mutex_t SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef pthread_mutex_t SA_EVENT,      *PSA_EVENT;
    typedef u32 SA_IRQ,        *PSA_IRQ;

#else
    
    typedef u32 SA_REAL_SPINLOCK, *PSA_REAL_SPINLOCK;
    typedef u32 SA_FAST_MUTEX, *PSA_FAST_MUTEX;
    typedef u32 SA_EVENT,      *PSA_EVENT;
    typedef u32 SA_IRQ,        *PSA_IRQ;

#endif

// Debug spinlock struct that maintains stats.
typedef struct _SA_DEBUG_SPINLOCK {
    SA_REAL_SPINLOCK lock;
    u64 spin_ticks;
    u64 acquires;
    u32 is_acquired;
} SA_DEBUG_SPINLOCK, *PSA_DEBUG_SPINLOCK;

#define SA_INFINITE  ((u32)-1)



#ifndef SA_ATOMIC_FUNCTIONS_INLINED
    /*!
    @brief
        Subtracts 1 from Value in a single atomic operation
    @param    
        Value           - Pointer to the value to subtract 1 from
    @return
        The current value after the operation   
    @note
    */
    u32 sa_atomic_decrement (u32* value);


    /*!
    @brief
        Adds 1 from Value in a single atomic operation
    @param    
        Value           - Pointer to the value to add 1 from
    @return
        The current value after the operation   
    @note
    */
    u32 
    sa_atomic_increment( 
        IN      u32*                 Value );


    // Atomically does this: *v += i;
    void sa_atomic_add (u32 i, u32 *v);

    // Atomically does this: *v -= i;
    void sa_atomic_sub (u32 i, u32 *v);

    /*!
    @brief
        Does this atomically:

        PVOID current = *destination;
        if (current == compare) 
            (*destination) = exchange;
        return current; 

    @return
        Original value of (*destination)
    */
    PVOID
    sa_atomic_compare_exchange_ptr (
        OUT PVOID *destination, 
        IN  PVOID exchange,
        IN  PVOID compare);

#endif // SA_ATOMIC_DECLARED_INLINE

/*!
@brief
    Tells if the current thread is in a state where it may block
@return
    TRUE if the current thread may block. FALSE if not
@note
*/
boolean
sa_can_block (void);

boolean
sa_is_passive (void);

boolean
sa_is_dispatch (void);

boolean 
sa_is_not_dirql (void);

// Initializes a previously allocated spinlock object.
void 
sa_init_spinlock_real ( 
    IN      PSA_REAL_SPINLOCK         SpinLock );

#ifndef SA_ACQUIRE_SPINLOCK_REAL_INLINED
    // Acquires a previously initialized spinlock object.
    void 
    sa_acquire_spinlock_real (
        IN  PSA_REAL_SPINLOCK SpinLock,
        OUT PSA_IRQ           Irq);

    void 
    sa_acquire_spinlock_noirq_real (
        IN PSA_REAL_SPINLOCK SpinLock);

#endif

#ifndef SA_RELEASE_SPINLOCK_REAL_INLINED
    // Releases a previously acquired spinlock object.
    void 
    sa_release_spinlock_real (
        IN PSA_REAL_SPINLOCK SpinLock,
        IN PSA_IRQ           Irq);

    void 
    sa_release_spinlock_noirq_real (
        IN PSA_REAL_SPINLOCK SpinLock);

#endif

// Returns TRUE if the spin lock is currently held.
// Not implemented on all platforms
boolean
sa_is_spin_lock_held_real ( PSA_REAL_SPINLOCK lock );

// Returns FALSE if the spin lock is currently held.
// Not implemented on all platforms
boolean
sa_is_spinlock_not_held_real ( PSA_REAL_SPINLOCK lock);


//
// Debug spinlock code to detect contention.
//
STATIC INLINE void 
sa_init_spinlock_debug ( 
    IN      PSA_DEBUG_SPINLOCK         lock)
{
    // Can't use sa_zero_mem since spinlock.h is included in sa_mem.h!
    lock->acquires = 0;
    lock->is_acquired = 0;
    lock->spin_ticks = 0;
    sa_init_spinlock_real (&lock->lock);
}

STATIC INLINE void 
sa_acquire_spinlock_debug ( 
    IN      PSA_DEBUG_SPINLOCK   lock, 
    OUT     PSA_IRQ              irq )
{
    u64 ticks;

    lock->acquires++;

    ticks = sa_timestamp ();
    sa_acquire_spinlock_real (&lock->lock, irq);
    lock->is_acquired = TRUE;
    lock->spin_ticks += (sa_timestamp () - ticks);
}

STATIC INLINE void 
sa_release_spinlock_debug ( 
    IN      PSA_DEBUG_SPINLOCK   lock, 
    IN      PSA_IRQ              Irq )
{
    lock->is_acquired = FALSE;
    sa_release_spinlock_real (&lock->lock, Irq);
}


STATIC INLINE void 
sa_acquire_spinlock_noirq_debug ( 
    IN      PSA_DEBUG_SPINLOCK   lock)
{
    u64 ticks;

    lock->acquires++;

    ticks = sa_timestamp ();
    sa_acquire_spinlock_noirq_real (&lock->lock);
    lock->is_acquired = TRUE;
    lock->spin_ticks += (sa_timestamp () - ticks);
}

STATIC INLINE void 
sa_release_spinlock_noirq_debug ( 
    IN      PSA_DEBUG_SPINLOCK   lock)
{
    lock->is_acquired = FALSE;
    sa_release_spinlock_noirq_real (&lock->lock);
}

STATIC INLINE boolean
sa_is_spin_lock_held_debug (PSA_DEBUG_SPINLOCK  lock) {
    return (boolean)lock->is_acquired;
}

STATIC INLINE boolean
sa_is_spinlock_not_held_debug (PSA_DEBUG_SPINLOCK lock) {
    return (boolean)!lock->is_acquired;
}

#ifdef SA_SPINLOCK_DEBUG
    typedef SA_DEBUG_SPINLOCK          SA_SPINLOCK;
    typedef PSA_DEBUG_SPINLOCK         PSA_SPINLOCK;
    #define sa_init_spinlock           sa_init_spinlock_debug
    #define sa_acquire_spinlock        sa_acquire_spinlock_debug
    #define sa_release_spinlock        sa_release_spinlock_debug
    #define sa_acquire_spinlock_noirq  sa_acquire_spinlock_noirq_debug
    #define sa_release_spinlock_noirq  sa_release_spinlock_noirq_debug
    #define sa_is_spin_lock_held       sa_is_spin_lock_held_debug
    #define sa_is_spinlock_not_held    sa_is_spinlock_not_held_debug
#else
    typedef SA_REAL_SPINLOCK           SA_SPINLOCK;
    typedef PSA_REAL_SPINLOCK          PSA_SPINLOCK;
    #define sa_init_spinlock           sa_init_spinlock_real
    #define sa_acquire_spinlock        sa_acquire_spinlock_real
    #define sa_release_spinlock        sa_release_spinlock_real
    #define sa_acquire_spinlock_noirq  sa_acquire_spinlock_noirq_real
    #define sa_release_spinlock_noirq  sa_release_spinlock_noirq_real
    #define sa_is_spin_lock_held       sa_is_spin_lock_held_real
    #define sa_is_spinlock_not_held    sa_is_spinlock_not_held_real
#endif

// Create windows-centric names for the dispatch level routines.
#define sa_acquire_spinlock_at_dispatch  sa_acquire_spinlock_noirq
#define sa_release_spinlock_at_dispatch  sa_release_spinlock_noirq


/*!

@brief
    Initializes a FastMutex 
    
@param    
    FastMutex           - Pointer to a FastMutex
    
@return
   

@note

*/
void 
sa_initialize_fast_mutex(
    IN      PSA_FAST_MUTEX       FastMutex );

/*!

@brief
    Acquires a FastMutex
    
@param    
    FastMutex        - Pointer to a FastMutex
    
@return

@note
    Current thread must be in a blockable state

*/
void 
sa_acquire_fast_mutex( 
    IN      PSA_FAST_MUTEX       FastMutex );

/*!

@brief
    Releases a FastMutex
    
@param    
    FastMutex        - Pointer to a FastMutex
    
@return

@note

*/
void 
sa_release_fast_mutex( 
    IN      PSA_FAST_MUTEX       FastMutex );



//
// You can skip reading this section
//
// These routines are really only used in the MPU code. If the MPU is not
// being used (as will be the case for device bringup) these functions need
// not be implemented correctly but must have a body in order to link.
// Implementing as SA_BREAKPOINT_MSG("Not implemented"); is fine.
//
// The semi-high probability that all the MPU code may need to be rewritten to
// accomidate the ARM characteristics may render all these functions unnecessary.
//


//
// Initializes a watiable event. The event can either 1) wake all threads blocked
// on it when signaled, or 2) wake all threads
// 
// Replaces KeInitializeEvent
//
void
sa_event_init (
    IN      PSA_EVENT            Event,
    IN      boolean                 WakeAll );

// 
// This function signals the event unblocking waiting threads.  The number of
// threads unblocked depends on the WakeAll argument that was passed during
// sa_EventInit()
//
// Replaces KeSetEvent
//
void
sa_event_set (
    IN PSA_EVENT Event);

// Closes a previously "init" event.
void
sa_event_close (
    PSA_EVENT  event);

// 
// This function signals the event unblocking waiting threads.  The number of
// threads unblocked depends on the WakeAll argument that was passed during
// sa_EventInit()
//
// A timeout of SA_INFINITE means wait forever.  A timeout of 0 means check to see
// if the object is signaled, and then return but don't block.
//
// Replaces KeWaitForSingleObject
//
// Return values include
//      STATUS_SUCCESS - Object was signaled
//      STATUS_ALERTED - The wait was completed because of an alert/signal to the thread. 
//      STATUS_TIMEOUT - The object was not signaled
//
SA_STATUS
sa_wait_for_single_object (
    IN      PVOID  object,
    IN      u32    timeout_in_milliseconds );



SA_CPP_TRAILER

#endif

