/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __sa_time_h__
#define __sa_time_h__

#include "sa_types.h"
#include "sa_assert.h"

SA_CPP_HEADER

#ifndef __GNUC__
#pragma warning (disable:4090)
#pragma warning (disable:4328)
#endif

#if defined(SA_NDIS)
    
    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {
        UNALIGNED LARGE_INTEGER l;                
        NdisGetCurrentSystemTime ( (UNALIGNED LARGE_INTEGER *) &l); 
        return l.QuadPart; // Returns  units of 100ns ticks.
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {
        u64 diff = (end - start) / 10; // convert to us.
        return diff;
    }

    #ifndef NDIS_MINIPORT_DRIVER
        EXPORT VOID NdisMSleep (ULONG MicrosecondsToSleep);
    #endif

    // IRQL < DISPATCH only!
    STATIC FORCEINLINE void sa_sleep(u32 microseconds) {
        NdisMSleep (microseconds);
    }

    // IRQL <= DISPATCH
    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        NdisStallExecution (microseconds);
    }

#elif defined(SA_STORPORT) || defined(SA_SCSIPORT)

    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {
        LARGE_INTEGER l = {0};       
        #ifdef ScsiPortQuerySystemTime 
            ScsiPortQuerySystemTime (&l); // Returns  units of 100ns ticks.
        #else
            // Not supported for w2k scsiport
            KeQuerySystemTime (&l); // Returns  units of 100ns ticks.
        #endif
        return l.QuadPart;
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {
        u64 diff = (end - start) / 10; // convert to us.
        return diff;
    }

    // IRQL <= DISPATCH
    STATIC FORCEINLINE void sa_sleep (u32 microseconds) {
        ScsiPortStallExecution  (microseconds); // spins -- sleeping not allowed in sports
    }

    // IRQL <= DISPATCH
    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        ScsiPortStallExecution  (microseconds);
    }

#elif defined(SA_WIN32K)

    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {
        LARGE_INTEGER l;                
        KeQuerySystemTime (&l); 
        return l.QuadPart; // Returns  units of 100ns ticks.
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {
        u64 diff = (end - start) / 10; // convert to us.
        return diff;
    }

    // IRQL < DISPATCH only!
    STATIC FORCEINLINE void sa_sleep (u32 microseconds) {
        LARGE_INTEGER t;

        // units of 100ns ticks. negative for relative time.
        t.QuadPart = -1 * ((u64)microseconds * 10); 
        KeDelayExecutionThread (KernelMode, 
                                FALSE, // alertable
                                &t);   // time to wait
    }

    // IRQL <= DISPATCH
    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        KeStallExecutionProcessor (microseconds);
    }

#elif defined(SA_WIN32U) || defined(SA_WIN_USER)

    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {        
        LARGE_INTEGER t;
        QueryPerformanceCounter(&t);
        return t.QuadPart;
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {
        STATIC LARGE_INTEGER f = {0};
        u64 diff;

        if (f.QuadPart == 0) {
            // Query the frequency.
            BOOL SupportsPerfCounter = QueryPerformanceFrequency (&f);
            ASSERT (SupportsPerfCounter);           
            ASSERT (f.QuadPart > 0);
            f.QuadPart = MAX (1, f.QuadPart); // Avoid divide by zero.
        }

        diff = (end - start) * 1000000 / f.QuadPart;
        return diff;
    }

    // IRQL < DISPATCH only!
    STATIC FORCEINLINE void sa_sleep (u32 microseconds) {
        Sleep (microseconds / 1000); // convert to milli
    }

    // IRQL <= DISPATCH
    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        u64 start = sa_get_system_time ();

        if (microseconds > 10000) {
            // Sleep for the majority of the time.
            sa_sleep ((microseconds - 10000) / 1000); 
        }
        
        // Spin for the rest
        while (sa_diff_system_time (start, sa_get_system_time ()) < (u64)microseconds)
            ;
    }

#elif defined(SA_LINUXK)

    #include <linux/delay.h>
    #include <linux/sched.h>
    
    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {        
        return (u64)jiffies;
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {        
        return (end - start) * (1000000 / HZ); // convert to us
    }

    STATIC FORCEINLINE void sa_sleep (u32 microseconds) {
        if (microseconds <= 1000)
            udelay (microseconds);
        else
            mdelay (microseconds / 1000);

    }

    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        udelay (microseconds);
    }

#elif defined(SA_DOS) ||  defined(SA_LINUXU)

    #include <sys/time.h> // gettimeofday
    #include <unistd.h>   // for usleep

    // Returns arbitrary units.  Use sa_diff_system_time to convert 
    // to us.  This allows us to handle rollovers in the system time
    // clock while preserving time differences.
    STATIC FORCEINLINE u64 sa_get_system_time (void) {        
        struct timeval t;
        struct timezone z;
        gettimeofday (&t, &z);
        return ((u64)t.tv_sec * 1000000 + (u64)t.tv_usec); // us
    }

    // Returns the difference between two system time values in us.
    STATIC FORCEINLINE u64 sa_diff_system_time (u64 start, u64 end) {        
        return (end - start);
    }

    STATIC FORCEINLINE void sa_sleep (u32 microseconds) {
        usleep (microseconds);
    }

    STATIC FORCEINLINE void sa_spin (u32 microseconds) {
        u64 start = sa_get_system_time ();

        if (microseconds > 10000) {
            // Sleep for the majority of the time.
            sa_sleep ((microseconds - 10000) / 1000); 
        }
        
        // Spin for the rest
        while (sa_diff_system_time (start, sa_get_system_time ()) < (u64)microseconds)
            ;
    }

#else

    #error Undefined platform in sa_time.h.

#endif



// Define microsecond versions
#define sa_usleep(_microseconds) sa_sleep(_microseconds)
#define sa_uspin(_microseconds)  sa_spin(_microseconds)

// Define millisecond versions
#define sa_msleep(_milliseconds) sa_sleep((_milliseconds)*1000)
#define sa_mspin(_milliseconds)  sa_spin((_milliseconds)*1000)



#if defined(SA_X86)

    #if defined(SA_MICROSOFT) // includes NDIS, storport, win kernel environs

        #pragma warning( disable : 4035 )   // Turn off no return value warning
        
        // ---------------------------------------------------------------------------
        // Procedure   sa_memory_barrier
        // Synopsis    Creates a memory barrier such that all previous reads/write complete to 
        //             main memory before future reads/writes.  The serializes memory access
        //             without serializing entire instruction pipeline such as cpuid.
        // Arguments   None
        // Returns     None 
        // ---------------------------------------------------------------------------

        #if defined(SA_NDIS)
            STATIC FORCEINLINE void sa_memory_barrier (void) {
                KeMemoryBarrier ();
            }
        #else
            STATIC FORCEINLINE void sa_memory_barrier (void) {
                #if defined(SA_KERNEL)
                    // MemoryBarrier ();
                    BREAKPOINT (); // Not defined...
                #else
                    // _ReadWriteBarrier (); // Use this intrinsic? 
                    __asm { mfence };
                    // _mm_mfence ();
                #endif
            }
        #endif
                    
        // ---------------------------------------------------------------------------
        // Procedure   RDMSR
        // Synopsis    Reads a given MSR
        // Arguments   Model Specific Register to read from.
        // Returns     Value read. See x86 manuals
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdmsr(int MSR) {
            _asm mov ecx, MSR
            _asm _emit 0xf
            _asm _emit 0x32
        }

        // ---------------------------------------------------------------------------
        // Procedure   RDPMC
        // Synopsis    Read the PMC
        // Arguments   perfCtr  - Performance Counter to read
        // Returns     Value read.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdpmc(int perfCtr) {
            _asm mov ecx, perfCtr
            _asm _emit 0xf
            _asm _emit 0x33
        }

        // ---------------------------------------------------------------------------
        // Procedure   RDTSC
        // Synopsis    Read the processor timestamp
        // Arguments   None
        // Returns     Processor timestamp. This is not serialized w/ respect 
        //             currently pipelined instructions.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdtsc (void) {
            // _asm rdtsc
            _asm _emit 0xf
            _asm _emit 0x31
        }

        // ---------------------------------------------------------------------------
        // Procedure   CPUID
        // Synopsis    Read CPU info.
        // Arguments   None
        // Returns     Normally eax is programmed as input to determine which info to
        //             return.  This function ignores output, using cpuid as 
        //             serializing function.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE void sa_cpuid (void) {            
            _asm cpuid
        }

        #pragma warning( default : 4035 ) // Turn on no return value warning

    #else

        // Linux

        // TODO: Uncomment once it is tested...
        // ---------------------------------------------------------------------------
        // Procedure   sa_memory_barrier
        // Synopsis    Creates a memory barrier such that all previous reads/write complete to 
        //             main memory before future reads/writes.  The serializes memory access
        //             without serializing entire instruction pipeline such as cpuid.
        // Arguments   None
        // Returns     None 
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE void sa_memory_barrier (void) {
            //smp_mb ();
            BREAKPOINT (); // not implemented
        }

        // ---------------------------------------------------------------------------
        // Procedure   RDMSR
        // Synopsis    Reads a given MSR
        // Arguments   Model Specific Register to read from.
        // Returns     Value read. See x86 manuals
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdmsr(int MSR) {
            SA_BREAKPOINT_MSG ("Not Implemented");
            return 0;
        }

        // ---------------------------------------------------------------------------
        // Procedure   RDPMC
        // Synopsis    Read the PMC
        // Arguments   perfCtr  - Performance Counter to read
        // Returns     Value read.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdpmc(int perfCtr) {
            SA_BREAKPOINT_MSG ("Not Implemented");
            return 0;
        }

        // ---------------------------------------------------------------------------
        // Procedure   RDTSC
        // Synopsis    Read the processor timestamp
        // Arguments   None
        // Returns     Processor timestamp. This is not serialized w/ respect 
        //             currently pipelined instructions.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdtsc (void) {
            u64 val;
            __asm__ __volatile__("rdtsc" : "=A" (val));
            return val;
        }

        // ---------------------------------------------------------------------------
        // Procedure   CPUID
        // Synopsis    Read CPU info.
        // Arguments   None
        // Returns     Normally eax is programmed as input to determine which info to
        //             return.  This function ignores output, using cpuid as 
        //             serializing function.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE void sa_cpuid (void) {
            __asm__ __volatile__ (
               "xorl %%eax,%%eax\n\t" // eax inputs which cpuid info to read.  We don't care, set to zero.
               "cpuid"                // serializes the execution pipeline
               : : : "eax", "ebx", "ecx", "edx" ); // clobbers these 4 registers
        }

    #endif

    // Returns processor timestamp
    STATIC FORCEINLINE u64 sa_timestamp(void) { 
        sa_cpuid ();        // serialize instructions
        return sa_rdtsc (); // return timestamp processor
    }

#elif defined(SA_AMD64)    

    // ---------------------------------------------------------------------------
    // Procedure   RDTSC
    // Synopsis    Read the processor timestamp
    // Arguments   None
    // Returns     Processor timestamp. This is not serialized w/ respect 
    //             currently pipelined instructions.
    // ---------------------------------------------------------------------------
    #pragma intrinsic(__rdtsc)

    // ---------------------------------------------------------------------------
    // Procedure   sa_memory_barrier
    // Synopsis    Creates a memory barrier such that all previous reads/write complete to 
    //             main memory before future reads/writes.  The serializes memory access
    //             without serializing entire instruction pipeline such as cpuid.
    // Arguments   None
    // Returns     None 
    // ---------------------------------------------------------------------------
        #if defined(SA_NDIS)
            STATIC FORCEINLINE void sa_memory_barrier (void) {
                KeMemoryBarrier ();
            }
        #else
            STATIC FORCEINLINE void sa_memory_barrier (void) {
                #if defined(SA_KERNEL)
                    //FIXME KeMemoryBarrier ();  
                    BREAKPOINT ();
                #else
                    // _ReadWriteBarrier (); // Use this intrinsic? 
                    _mm_mfence ();                    
                #endif
            }
        #endif

    STATIC FORCEINLINE u64 sa_rdtsc (void) {        
        // unsigned __int64 __rdtsc( void );
        return __rdtsc (); // calls a compiler intrinsic -- Does it require serializing?
    }

    // ---------------------------------------------------------------------------
    // Procedure   CPUID
    // Synopsis    Read CPU info.
    // Arguments   None
    // Returns     Normally eax is programmed as input to determine which info to
    //             return.  This function ignores output, using cpuid as 
    //             serializing function.
    // ---------------------------------------------------------------------------
    STATIC FORCEINLINE void sa_cpuid (void) {            
        //int CpuInfo[4];
        //__cpuid (&CpuInfo, 0);
    }
    
    // Returns processor timestamp
    STATIC FORCEINLINE u64 sa_timestamp(void) { 
        // sa_cpuid ();        // serialize instructions
        return sa_rdtsc (); // return timestamp processor
    }

#elif defined(SA_IA64)    

    #if defined(SA_MICROSOFT) // includes NDIS, storport, win kernel environs

        #define _IA64_REG_AR_ITC 3116

        unsigned __int64 __getReg(int number);

        // ---------------------------------------------------------------------------
        // Procedure   RDTSC
        // Synopsis    Read the processor timestamp
        // Arguments   None
        // Returns     Processor timestamp. This is not serialized w/ respect 
        //             currently pipelined instructions.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE u64 sa_rdtsc (void) {        
            return __getReg(_IA64_REG_AR_ITC);
        }

        // ---------------------------------------------------------------------------
        // Procedure   sa_memory_barrier
        // Synopsis    Creates a memory barrier such that all previous reads/write complete to 
        //             main memory before future reads/writes.  The serializes memory access
        //             without serializing entire instruction pipeline such as cpuid.
        // Arguments   None
        // Returns     None 
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE void sa_memory_barrier (void) {
            __mf();    
        }


        // ---------------------------------------------------------------------------
        // Procedure   CPUID
        // Synopsis    Read CPU info.
        // Arguments   None
        // Returns     Normally eax is programmed as input to determine which info to
        //             return.  This function ignores output, using cpuid as 
        //             serializing function.
        // ---------------------------------------------------------------------------
        STATIC FORCEINLINE void sa_cpuid (void) {            
            //int CpuInfo[4];
            //__cpuid (&CpuInfo, 0);
        }
        
        // Returns processor timestamp
        STATIC FORCEINLINE u64 sa_timestamp(void) { 
            // sa_cpuid ();        // serialize instructions
            return sa_rdtsc(); // return timestamp processor
        }

    #else

        //
        // IA64, non-Microsoft
        //

        STATIC FORCEINLINE u64 sa_rdtsc (void) {   
            BREAKPOINT (); // not implemented
            return 0;
        }

        STATIC FORCEINLINE void sa_memory_barrier (void) {
            mb();    
        }

        STATIC FORCEINLINE void sa_cpuid (void) {            
            // nop??
        }

        STATIC FORCEINLINE u64 sa_timestamp(void) {             
            return sa_rdtsc(); // return timestamp processor
        }

    #endif

#elif defined(SA_XSCALE) || defined(SA_PPC)

    #ifndef SA_LINUXK
    #error "As of now only SA_LINUXK compiles for SA_XSCALE/SA_PPC"
    #endif

    // ---------------------------------------------------------------------------
    // Procedure   sa_memory_barrier
    // Synopsis    Creates a memory barrier such that all previous reads/write complete to 
    //             main memory before future reads/writes.  The serializes memory access
    //             without serializing entire instruction pipeline such as cpuid.
    // Arguments   None
    // Returns     None 
    // ---------------------------------------------------------------------------
    STATIC FORCEINLINE void sa_memory_barrier (void) {
        SA_BREAKPOINT_MSG("NOT IMPLEMENTED");
    }

    STATIC FORCEINLINE u64 sa_rdtsc (void) {        
        SA_BREAKPOINT_MSG("NOT IMPLEMENTED");
        return 0;
    }

    // ---------------------------------------------------------------------------
    // Procedure   CPUID
    // Synopsis    Read CPU info.
    // Arguments   None
    // Returns     Normally eax is programmed as input to determine which info to
    //             return.  This function ignores output, using cpuid as 
    //             serializing function.
    // ---------------------------------------------------------------------------
    STATIC FORCEINLINE void sa_cpuid (void) {            
    SA_BREAKPOINT_MSG("NOT IMPLEMENTED");
    }
   
    #define sa_timestamp sa_get_system_time
 
#else 

    // Not X86 or AMD64 or IA64 or XSCALE   
    #define sa_timestamp sa_get_system_time

#endif

SA_CPP_TRAILER

#endif  // __sa_time_h__
