/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file 
    $RCSfile: sa_index_allocator.c,v $
    
@brief 
    Implements generic index and object allocators that use a 
    free list to efficiently allocate/deallocate items.  Supports
    both multi-thread safe, and non-safe methods.
    
@notes

CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_index_allocator.c,v 1.8 2006/03/13 19:42:46 menz Exp $
*/

#include "sa_index_allocator.h"
#include "sa_mem.h"
#include "sa_sync.h"


PSA_INDEX_ALLOCATOR
sa_indexalloc_allocate (u32 num, 
                        u32 alloc_flags, // typically SA_MALLOC_NORMAL
                        const c8* optional_name)
{
    boolean *dbg_vector = NULL;
    SA_STATUS status = SA_STATUS_SUCCESS;

    // Allocate the object itself.
    SA_INDEX_ALLOCATOR *ia = sa_malloc (sizeof (SA_INDEX_ALLOCATOR), 
                                         alloc_flags, 
                                         SA_TAG ("ia.."));
    
    // Allocate the entries.
    SA_INDEX_ALLOCATOR_ENTRY* entries = sa_malloc (sizeof (SA_INDEX_ALLOCATOR_ENTRY) * num, 
                                                    alloc_flags, 
                                                    SA_TAG("iae."));

    #if defined(SA_DEBUG)
        // Allocate the debug vector to track double frees, etc.
        dbg_vector = sa_malloc (sizeof (boolean) * num, 
                                 alloc_flags, 
                                 SA_TAG("iabv"));
    #endif

    if (ia == NULL || entries == NULL || dbg_vector == NULL) {
        status = SA_STATUS_NOT_OK;
    } else {
        status = sa_indexalloc_create (ia, entries, num, dbg_vector, optional_name);
    }

    if (status != SA_STATUS_SUCCESS) {
        // sa_free handles null pointers.
        sa_free (ia);
        sa_free (entries);
        sa_free (dbg_vector);
        return NULL;
    }

    return ia;
}


void
sa_indexalloc_free (PSA_INDEX_ALLOCATOR ia)
{
    if (ia == NULL) {
        return;
    }

    sa_indexalloc_destroy (ia);

    // sa_free handles null pointers.
    sa_free (ia->entries);
    sa_free (ia->dbg_vector);
    sa_free (ia);
}


/*!
@brief 
    Initializes a free list with all the list entries.

@param 
    ia - pointer to index allocator to intialize
@param 
    entries - pointer to array of SLIST_ENTRY objects 
@param
    num - number of SLIST_ENTRY objects in array. 
@param
    optionalDbgVector - optional array of BOOLEANs used only during
        debug builds for tracking double frees.  This vector must
        have "num" elements.
                      
@return
    None
*/
SA_STATUS
sa_indexalloc_create (PSA_INDEX_ALLOCATOR ia, 
                      SA_INDEX_ALLOCATOR_ENTRY *entries, 
                      u32 num, 
                      boolean *optionalDbgVector, 
                      const c8* optionalName)
{
    u32 i;

    ASSERT (ia);
    ASSERT (entries);
    ASSERT (num > 0);

    SA_NOT_USED (optionalDbgVector);

    //
    // Initialize values of the ia structure.
    //
    sa_zero_mem (ia, sizeof (SA_INDEX_ALLOCATOR));
    ia->magic = SA_IA_MAGIC;
    ia->num = num;
    ia->num_free = num;
    ia->entries = entries;
    
    //
    // Init singly linked list.
    //     
    ia->free_list.next = entries;    
    for (i = 0; i < num - 1; i++) {
        entries[i].next = &entries[i+1];
    }
    entries[num - 1].next = NULL;

#if defined(SA_DEBUG)
    
    //
    // Optionally, provide tracking for double frees.
    //
    if (optionalDbgVector) {
        sa_zero_mem (optionalDbgVector, sizeof (boolean) * num);
        ia->dbg_vector = optionalDbgVector;
    }

#endif

    ia->dbg_name = optionalName;
    
    TRACE (DL_VERBOSE, "init index allocator %s with %d entries", optionalName, num);

    //
    // Current implementation always succeeds since it doesn't 
    // allocate any memory.
    //
    return SA_STATUS_SUCCESS; 
}


/*!
@brief
    Destroys a index allocator object.  It is safe to
    call this function even if the iaInit fails.
@param
    ia - Pointer to index allocator object.  
@return
    none
*/
void
sa_indexalloc_destroy (PSA_INDEX_ALLOCATOR ia) {
    
    SA_IA_ASSERT (ia);
    
    ia->magic = 0;
}

/*!
@brief 
    Allocates an index value, removing it from the free list.
    If no values are available, IA_BAD_INDEX is returned. 
    This version is not multithread safe.

@param 
    ia - pointer to index allocator
                      
@return
    allocated index on success
    IA_BAD_INDEX on failure
*/
u32
sa_indexalloc_allocate_index (PSA_INDEX_ALLOCATOR ia) {

    u32 index;
    SA_INDEX_ALLOCATOR_ENTRY *head;

    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (ia->entries);

    //
    // Remove the head SLIST_ENTRY item from the free_list 
    //    
    head = ia->free_list.next;

    //
    // Check if list is empty.
    //
    if (head == NULL) {
        return SA_IA_BAD_INDEX;
    } 

    // Free list now points to second entry.
    ia->free_list.next = head->next;

    //
    // Compute an index for this element.
    //
    index = (u32)(((u8*)head - (u8*)ia->entries) / (sizeof (*head)));
    ASSERT (index < ia->num);

    // 
    // Maintain free count.
    //
    ia->num_free--;

#if defined(SA_DEBUG)
    //
    // Optional debug checks
    //
    if (ia->dbg_vector) {
        ASSERT (ia->dbg_vector[index] == FALSE);
        ia->dbg_vector[index] = TRUE;
    }
    TRACE (DL_VERBOSE, "Allocate %s index %d.", ia->dbg_name, index);
#endif
    

     return index;
}

/*!
@brief 
    Allocates the given index value, removing it from the free list.
    If the given index is not available, IA_BAD_INDEX is returned. 
    This version is not multithread safe.  It is used during initialization only.

@param 
    ia - pointer to index allocator

@return
    allocated index on success
    IA_BAD_INDEX on failure
*/
u32
sa_indexalloc_index_allocate_exact (PSA_INDEX_ALLOCATOR ia, u32 exactIndex) {

    SA_INDEX_ALLOCATOR_ENTRY *previous = NULL;
    SA_INDEX_ALLOCATOR_ENTRY *head = NULL;
    u32 index;

    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (ia->entries);

    //
    // Remove the head SLIST_ENTRY item from the free_list 
    //
    head = ia->free_list.next;
    
    while (head) {
        //
        // Compute an index for this element.
        //
        index = (u32)(((u8*)head - (u8*)ia->entries) / (sizeof (*head)));
        ASSERT (index < ia->num);
        
        if (index == exactIndex) {

            if (previous == NULL) {
                //
                // Remove head.
                //
                ia->free_list.next = head->next;
            } else {
                previous->next = head->next;
            }

            #ifdef SA_DEBUG
                //
                // Optional debug checks
                //
                if (ia->dbg_vector) {
                    ASSERT (ia->dbg_vector[index] == FALSE);
                    ia->dbg_vector[index] = TRUE;
                }
                TRACE (DL_VERBOSE, "Allocate %s exact index %d.", ia->dbg_name, index);
            #endif

            //
            // Maintain free count
            //
            ia->num_free--;

            return exactIndex;

        } else {
            // Walk the free list
            previous = head;
            head = head->next;
        }
    }

    return SA_IA_BAD_INDEX;
}

/*
@brief
    Thread safe allocator.
*/
u32
sa_indexalloc_allocate_index_interlocked (PSA_INDEX_ALLOCATOR ia) {
    
    SA_INDEX_ALLOCATOR_ENTRY *head;

    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (ia->entries);

    head = ia->free_list.next;

    while (head) {

        if (sa_atomic_compare_exchange_ptr ((PVOID*)&ia->free_list.next, head->next, head) == head) {
            //
            // Exchange succeeded.  Head element was removed from list. 
            // Compute an index for this element.
            //
            u32 index = (u32)(((u8*)head - (u8*)ia->entries) / (sizeof (*head)));
            ASSERT (index < ia->num);

            #if defined(SA_DEBUG)
                //
                // Optional debug checks
                //
                if (ia->dbg_vector) {
                    ASSERT (ia->dbg_vector[index] == FALSE);
                    ia->dbg_vector[index] = TRUE;
                }

                TRACE (DL_VERBOSE, "InterlockedAllocate %s index %d.", ia->dbg_name, index);
            #endif

            //
            // Maintain free count
            //
            sa_atomic_decrement (&ia->num_free);

            return index;
            
        } else {
            //
            // Another thread removed the head.  Try again.
            //
            ASSERT (head != ia->free_list.next);
            head = ia->free_list.next;                        
        }
    }
    
    return SA_IA_BAD_INDEX;
}


/*!
@brief 
    Frees an index, adding it back to the free list

@param 
    ia - pointer to index allocator
@param
    index - value to free
                      
@return
    none
*/
void
sa_indexalloc_free_index (PSA_INDEX_ALLOCATOR ia, u32 index) {
    
    SA_INDEX_ALLOCATOR_ENTRY *entry;

    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (index < ia->num);

    //
    // Optional debug check for double frees.
    //
#if defined(SA_DEBUG)
    if (ia->dbg_vector) {
        ASSERT (ia->dbg_vector[index] == TRUE);
        ia->dbg_vector[index] = FALSE;
    }
    TRACE (DL_VERBOSE, "Free index %d for %s", index, ia->dbg_name);
#endif

    //
    // Get a pointer to the SLIST_ENTRY at the given index
    //
    entry = &ia->entries[index];

    //
    // Maintain free count
    //
    ia->num_free++;

    entry->next = ia->free_list.next;
    ia->free_list.next = entry;        
    return;
}

/*
@brief
    Thread safe allocator.
*/
void
sa_indexalloc_free_index_interlocked (PSA_INDEX_ALLOCATOR ia, u32 index) {
    
    SA_INDEX_ALLOCATOR_ENTRY *entry = NULL;
    SA_INDEX_ALLOCATOR_ENTRY *head = NULL;

    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (index < ia->num);

    #if defined (SA_DEBUG)
        if (ia->dbg_vector) {
            ASSERT (ia->dbg_vector[index] == TRUE);
            ia->dbg_vector[index] = FALSE;
        }
        TRACE (DL_VERBOSE, "InterlockedFree index %d for %s", index, ia->dbg_name);
    #endif

    //
    // Get a pointer to the SLIST_ENTRY at the given index
    //
    entry = &ia->entries[index];
    head = ia->free_list.next;
    entry->next = head; // link this entry to current head entry

    while (sa_atomic_compare_exchange_ptr ((PVOID*)&ia->free_list.next, entry, head) != head) {
        //
        // The head pointer was modified before we pushed this back onto the free list.
        // Try again.
        //
        head = ia->free_list.next;
        entry->next = head; // link this entry to current head entry
    }

    //
    // Maintain free count
    //
    sa_atomic_increment (&ia->num_free);

    return;
}


#if !defined(SA_OMIT_TEST_CODE)

    void
    _sa_indexalloc_test_internal (const c8* name,
                                u32 (*ialloc)(SA_INDEX_ALLOCATOR*),
                                void (*ifree)(SA_INDEX_ALLOCATOR*, u32))
    {
        SA_INDEX_ALLOCATOR ia;
        SA_INDEX_ALLOCATOR_ENTRY entries[4];
        boolean dbg[4];
        u32 a,b,c,d,e;

        sa_indexalloc_create (&ia, entries, 4, dbg, name);

        a = ialloc (&ia);    
        ASSERT (a == 0);

        b = ialloc (&ia);    
        ASSERT (b == 1);

        c = ialloc (&ia);
        ASSERT (c == 2);

        d = ialloc (&ia);
        ASSERT (d == 3);

        e = ialloc (&ia); // should fail
        ASSERT (e == SA_IA_BAD_INDEX);
        
        ifree (&ia, c);  // 2
        ifree (&ia, b);  // 1
        ifree (&ia, a);  // 0

        a = ialloc (&ia);
        ASSERT (a == 0);

        b = ialloc (&ia);
        ASSERT (b == 1);

        c = ialloc (&ia);
        ASSERT (c == 2);

        e = ialloc (&ia); // should fail
        ASSERT (e == SA_IA_BAD_INDEX);

        ifree (&ia, a);
        ifree (&ia, b);
        ifree (&ia, c);
        ifree (&ia, d);

        sa_indexalloc_destroy (&ia);    
    }

    void
    sa_indexalloc_test (void)
    {
        _sa_indexalloc_test_internal ("ia-test", 
                                    sa_indexalloc_allocate_index, 
                                    sa_indexalloc_free_index);

        _sa_indexalloc_test_internal ("ia-interlocked-test", 
                                    sa_indexalloc_allocate_index_interlocked, 
                                    sa_indexalloc_free_index_interlocked);

    }

#endif

