/*
* This code implements the md5 message-digest algorithm.
* The algorithm is due to Ron Rivest.  This code was
* written by Colin Plumb in 1993, no copyright is claimed.
* This code is in the public domain; do with it what you wish.
*
* Equivalent code is available from RSA Data Security, Inc.
* This code has been tested against that, and is equivalent,
* except that you don't need to include two pages of legalese
* with every copy.
*
* To compute the message digest of a chunk of bytes, declare an
* md5Context structure, pass it to md5Init, call md5Update as
* needed on buffers full of bytes, and then call md5Final, which
* will fill a supplied 16-byte array with the digest.
*/

/* Brutally hacked by John Walker back from ANSI C to K&R (no
prototypes) to maintain the tradition that Netfone will compile
with Sun's original "cc". */


#include "sa_md5.h"
#include "sa_endian.h"
#include "sa_mem.h"
#include "sa_text.h"

#ifndef SA_BIG_ENDIAN

#define sa_byte_reverse(buf, len)   // Nothing

#else

void sa_byte_reverse (u8 *buf, u32 longs)
{
    u32 t;
    do {
        t = (u32) ((u32) buf[3] << 8 | buf[2]) << 16 |
            ((u32) buf[1] << 8 | buf[0]);
        *(u32 *) buf = t;
        buf += 4;
    } while (--longs);
}

#endif

/*
* Start md5 accumulation.  Set bit count to 0 and buffer to mysterious
* initialization constants.
*/
void sa_md5_init(PSA_MD5_CONTEXT ctx)
{
    ctx->buf[0] = 0x67452301;
    ctx->buf[1] = 0xefcdab89;
    ctx->buf[2] = 0x98badcfe;
    ctx->buf[3] = 0x10325476;

    ctx->bits[0] = 0;
    ctx->bits[1] = 0;
}

/*
* Update context to reflect the concatenation of another buffer full
* of bytes.
*/
void sa_md5_update(PSA_MD5_CONTEXT ctx, const u8* buf, u32 len)

{
    u32 t;

    /* Update bitcount */

    t = ctx->bits[0];
    if ((ctx->bits[0] = t + ((u32) len << 3)) < t)
        ctx->bits[1]++;     /* Carry from low to high */
    ctx->bits[1] += len >> 29;

    t = (t >> 3) & 0x3f;    /* Bytes already in shsInfo->data */

    /* Handle any leading odd-sized chunks */

    if (t) {
        u8 *p = (u8 *) ctx->in + t;

        t = 64 - t;
        if (len < t) {
            sa_memcpy(p, buf, len);
            return;
        }
        sa_memcpy(p, buf, t);
        sa_byte_reverse(ctx->in, 16);
        sa_md5_transform(ctx->buf, (u32 *) ctx->in);
        buf += t;
        len -= t;
    }
    /* Process data in 64-byte chunks */

    while (len >= 64) {
        sa_memcpy(ctx->in, buf, 64);
        sa_byte_reverse(ctx->in, 16);
        sa_md5_transform(ctx->buf, (u32 *) ctx->in);
        buf += 64;
        len -= 64;
    }

    /* Handle any remaining bytes of data. */

    sa_memcpy(ctx->in, buf, len);
}

/*
* Final wrapup - pad to 64-byte boundary with the bit pattern 
* 1 0* (64-bit count of bits processed, MSB-first)
*/
void sa_md5_final(u8 digest[16], PSA_MD5_CONTEXT ctx)    
{
    u32 count;
    u8 *p;

    /* Compute number of bytes mod 64 */
    count = (ctx->bits[0] >> 3) & 0x3F;

    /* Set the first c8 of padding to 0x80.  This is safe since there is
    always at least one byte free */
    p = ctx->in + count;
    *p++ = 0x80;

    /* Bytes of padding needed to make 64 bytes */
    count = 64 - 1 - count;

    /* Pad out to 56 mod 64 */
    if (count < 8) {
        /* Two lots of padding:  Pad the first block to 64 bytes */
        sa_memset(p, 0, count);
        sa_byte_reverse(ctx->in, 16);
        sa_md5_transform(ctx->buf, (u32 *) ctx->in);

        /* Now fill the next block with 56 bytes */
        sa_memset(ctx->in, 0, 56);
    } else {
        /* Pad block to 56 bytes */
        sa_memset(p, 0, count - 8);
    }
    sa_byte_reverse(ctx->in, 14);

    /* Append length in bits and transform */
    ((u32 *) ctx->in)[14] = ctx->bits[0];
    ((u32 *) ctx->in)[15] = ctx->bits[1];

    sa_md5_transform(ctx->buf, (u32 *) ctx->in);
    sa_byte_reverse((u8 *) ctx->buf, 4);
    sa_memcpy(digest, ctx->buf, 16);
    sa_memset(ctx, 0, sizeof(*ctx));        /* In case it's sensitive */
}


/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define SA_MD5_F1(x, y, z) (z ^ (x & (y ^ z)))
#define SA_MD5_F2(x, y, z) SA_MD5_F1(z, x, y)
#define SA_MD5_F3(x, y, z) (x ^ y ^ z)
#define SA_MD5_F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the md5 algorithm. */
#define SA_MD5_STEP(f, w, x, y, z, data, s) \
    ( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )

/*
* The core of the md5 algorithm, this alters an existing md5 hash to
* reflect the addition of 16 longwords of new data.  md5Update blocks
* the data and converts bytes into longwords for this routine.
*/
void sa_md5_transform(u32 buf[4], u32 in[16])    
{
    register u32 a, b, c, d;

    a = buf[0];
    b = buf[1];
    c = buf[2];
    d = buf[3];

    SA_MD5_STEP(SA_MD5_F1, a, b, c, d, in[0] + 0xd76aa478, 7);
    SA_MD5_STEP(SA_MD5_F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
    SA_MD5_STEP(SA_MD5_F1, c, d, a, b, in[2] + 0x242070db, 17);
    SA_MD5_STEP(SA_MD5_F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
    SA_MD5_STEP(SA_MD5_F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
    SA_MD5_STEP(SA_MD5_F1, d, a, b, c, in[5] + 0x4787c62a, 12);
    SA_MD5_STEP(SA_MD5_F1, c, d, a, b, in[6] + 0xa8304613, 17);
    SA_MD5_STEP(SA_MD5_F1, b, c, d, a, in[7] + 0xfd469501, 22);
    SA_MD5_STEP(SA_MD5_F1, a, b, c, d, in[8] + 0x698098d8, 7);
    SA_MD5_STEP(SA_MD5_F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
    SA_MD5_STEP(SA_MD5_F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
    SA_MD5_STEP(SA_MD5_F1, b, c, d, a, in[11] + 0x895cd7be, 22);
    SA_MD5_STEP(SA_MD5_F1, a, b, c, d, in[12] + 0x6b901122, 7);
    SA_MD5_STEP(SA_MD5_F1, d, a, b, c, in[13] + 0xfd987193, 12);
    SA_MD5_STEP(SA_MD5_F1, c, d, a, b, in[14] + 0xa679438e, 17);
    SA_MD5_STEP(SA_MD5_F1, b, c, d, a, in[15] + 0x49b40821, 22);

    SA_MD5_STEP(SA_MD5_F2, a, b, c, d, in[1] + 0xf61e2562, 5);
    SA_MD5_STEP(SA_MD5_F2, d, a, b, c, in[6] + 0xc040b340, 9);
    SA_MD5_STEP(SA_MD5_F2, c, d, a, b, in[11] + 0x265e5a51, 14);
    SA_MD5_STEP(SA_MD5_F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
    SA_MD5_STEP(SA_MD5_F2, a, b, c, d, in[5] + 0xd62f105d, 5);
    SA_MD5_STEP(SA_MD5_F2, d, a, b, c, in[10] + 0x02441453, 9);
    SA_MD5_STEP(SA_MD5_F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
    SA_MD5_STEP(SA_MD5_F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
    SA_MD5_STEP(SA_MD5_F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
    SA_MD5_STEP(SA_MD5_F2, d, a, b, c, in[14] + 0xc33707d6, 9);
    SA_MD5_STEP(SA_MD5_F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
    SA_MD5_STEP(SA_MD5_F2, b, c, d, a, in[8] + 0x455a14ed, 20);
    SA_MD5_STEP(SA_MD5_F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
    SA_MD5_STEP(SA_MD5_F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
    SA_MD5_STEP(SA_MD5_F2, c, d, a, b, in[7] + 0x676f02d9, 14);
    SA_MD5_STEP(SA_MD5_F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

    SA_MD5_STEP(SA_MD5_F3, a, b, c, d, in[5] + 0xfffa3942, 4);
    SA_MD5_STEP(SA_MD5_F3, d, a, b, c, in[8] + 0x8771f681, 11);
    SA_MD5_STEP(SA_MD5_F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
    SA_MD5_STEP(SA_MD5_F3, b, c, d, a, in[14] + 0xfde5380c, 23);
    SA_MD5_STEP(SA_MD5_F3, a, b, c, d, in[1] + 0xa4beea44, 4);
    SA_MD5_STEP(SA_MD5_F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
    SA_MD5_STEP(SA_MD5_F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
    SA_MD5_STEP(SA_MD5_F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
    SA_MD5_STEP(SA_MD5_F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
    SA_MD5_STEP(SA_MD5_F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
    SA_MD5_STEP(SA_MD5_F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
    SA_MD5_STEP(SA_MD5_F3, b, c, d, a, in[6] + 0x04881d05, 23);
    SA_MD5_STEP(SA_MD5_F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
    SA_MD5_STEP(SA_MD5_F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
    SA_MD5_STEP(SA_MD5_F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
    SA_MD5_STEP(SA_MD5_F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

    SA_MD5_STEP(SA_MD5_F4, a, b, c, d, in[0] + 0xf4292244, 6);
    SA_MD5_STEP(SA_MD5_F4, d, a, b, c, in[7] + 0x432aff97, 10);
    SA_MD5_STEP(SA_MD5_F4, c, d, a, b, in[14] + 0xab9423a7, 15);
    SA_MD5_STEP(SA_MD5_F4, b, c, d, a, in[5] + 0xfc93a039, 21);
    SA_MD5_STEP(SA_MD5_F4, a, b, c, d, in[12] + 0x655b59c3, 6);
    SA_MD5_STEP(SA_MD5_F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
    SA_MD5_STEP(SA_MD5_F4, c, d, a, b, in[10] + 0xffeff47d, 15);
    SA_MD5_STEP(SA_MD5_F4, b, c, d, a, in[1] + 0x85845dd1, 21);
    SA_MD5_STEP(SA_MD5_F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
    SA_MD5_STEP(SA_MD5_F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
    SA_MD5_STEP(SA_MD5_F4, c, d, a, b, in[6] + 0xa3014314, 15);
    SA_MD5_STEP(SA_MD5_F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
    SA_MD5_STEP(SA_MD5_F4, a, b, c, d, in[4] + 0xf7537e82, 6);
    SA_MD5_STEP(SA_MD5_F4, d, a, b, c, in[11] + 0xbd3af235, 10);
    SA_MD5_STEP(SA_MD5_F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
    SA_MD5_STEP(SA_MD5_F4, b, c, d, a, in[9] + 0xeb86d391, 21);

    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
}

void sa_md5_string (const c8* str, u8 *response) {

    u32 len = sa_strlen (str);

    SA_MD5_CONTEXT ctx;
    sa_md5_init (&ctx);
    sa_md5_update (&ctx, (u8*)str, len);      // shared secret
    sa_md5_final (response, &ctx);
}


#if !defined(SA_OMIT_TEST_CODE)


    /* Digests a reference suite of strings and prints the results.
    */
    boolean sa_md5_test_suite ()
    {
        u8 response[16];
        u8 answer[16];
        u32 i;
        boolean pass = TRUE;

        // (challenge, response) pairs
        const c8* pairs [] =
        {
            "",                                                                                 "0xd41d8cd98f00b204e9800998ecf8427e",
            "a",                                                                                "0x0cc175b9c0f1b6a831c399e269772661",
            "abc",                                                                              "0x900150983cd24fb0d6963f7d28e17f72",
            "message digest",                                                                   "0xf96b697d7cb7938d525a2f31aaf161d0",
            "abcdefghijklmnopqrstuvwxyz",                                                       "0xc3fcd3d76192e4007dfb496cca67e13b",
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",                   "0xd174ab98d277d9f5a5611c2c9f419d9f",
            "12345678901234567890123456789012345678901234567890123456789012345678901234567890", "0x57edf4a22be3c955ac49da2e2107b67a"
        };

        for (i=0; i < sizeof(pairs) / sizeof(pairs[0]); i += 2) 
        {
            sa_md5_string (pairs[i], response); 
            sa_text_hex_str_to_byte_array (pairs[i+1], sa_strlen (pairs[i+1]), answer, sizeof(answer));

            if (sa_memcmp (response, answer, 16) != 0) {
                pass = FALSE;
                SA_BREAKPOINT_MSG ("MD5 test failure.");
            }        
        }

        return pass;
    }

#endif

