/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_mem_storport.c

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_mem_debug.c,v 1.7 2006/03/13 19:42:46 menz Exp $
*/

#include "sa_types.h"
#include "sa_trace.h"
#include "sa_assert.h"
#include "sa_mem.h"
#include "runtime/sa_text.h"
#include "sa_list.h"
#include "sa_sync.h"

// In release builds we need to avoid declaring any static variables so it would be
// nice to avoid this code. However, if the user wants memory debugging in release mode we are
// sunk. They would have to modify both theirs and SA's build to enable it.
#if defined(SA_MEM_DEBUG) || defined(SA_DEBUG) || defined(_SA_MEM_DEBUG_COMPILE)

    #define SA_MAGIC_HEAD (0xA0B01020)
    #define SA_MAGIC_TAIL (0xF040D030)

    // Each allocation includes this header
    typedef struct _SA_MEM_HEADER {
        u32 magic_head;
        SA_LIST_ENTRY next;
        u32 tag;
        const c8 *filename;
        u32 line;
        u32 length;
        u32 magic_tail;
    } SA_MEM_HEADER, *PSA_MEM_HEADER;

    // Each allocation includes this tail
    typedef struct _SA_MEM_TAIL {
        u32 magic_head;
        u32 magic_tail;
    } SA_MEM_TAIL, *PSA_MEM_TAIL;

    // Global context for debug information.
    typedef struct _SA_MEM_DEBUG_CONTEXT {

        boolean       enabled;           // indicates if the context was initialized
        SA_SPINLOCK   lock;              // protects the linked list
        SA_LIST_ENTRY allocated_memory;  // linked list of all allocations

    } SA_MEM_DEBUG_CONTEXT, *PSA_MEM_DEBUG_CONTEXT;


    // Global context for the allocated memory list.
    STATIC SA_MEM_DEBUG_CONTEXT sa_mem_debug_context = {0};


    // Initializes tracking for memory.  Should be explicitly 
    // called once in a multi-threaded environment, before any
    // allocations.
    void
    _sa_mem_debug_init (void)
    {
        if (!sa_mem_debug_context.enabled) {
            sa_mem_debug_context.enabled = TRUE;
            sa_init_spinlock (&sa_mem_debug_context.lock);
            sa_initialize_list_head (&sa_mem_debug_context.allocated_memory);
        }
    }

    /*!
    @brief
        Allocates virtual memory.  This function is generally used to allocate
        software owned (control) data structures and not for memory shared with
        the controller.
    @param  Size        - Size in bytes to allocate
    @param  Flag        - SA_MALLOC_NORMAL or SA_MALLOC_PAGEABLE to allocate pageable memory
    @param  Tag         - Tracking tag for this memory allocation
    @return Returns a pointer to newly allocated virtual memory
    @note
    */
    PVOID 
    _sa_malloc_debug (u32 size, u32 flag, u32 tag, SA_MEM_ALLOC_CONTEXT *optional_context, const c8* filename, u32 line)
    {    
        PVOID va;

        PSA_MEM_HEADER hdr;
        PSA_MEM_TAIL tail;
        SA_IRQ irq;

        size += sizeof (SA_MEM_HEADER) + sizeof (SA_MEM_TAIL);
        
        va = _sa_malloc_release (size, flag, tag, optional_context);

        if (va) {
            hdr = (PSA_MEM_HEADER)va;
            tail = (PSA_MEM_TAIL)((u8*)va + size - sizeof (SA_MEM_TAIL));
            va =  (u8*)va + sizeof (SA_MEM_HEADER);
            
            hdr->magic_head = tail->magic_head = SA_MAGIC_HEAD;
            hdr->magic_tail = tail->magic_tail = SA_MAGIC_TAIL;
            hdr->filename = filename;
            hdr->line = line;
            hdr->tag = tag;
            hdr->length = size - sizeof (SA_MEM_HEADER) - sizeof (SA_MEM_TAIL);
            
            if (!sa_mem_debug_context.enabled) {
                // This is dangerous, since 2 threads may try to init things simulatenously.
                // In a multi-threaded environment, force users to call sa_mem_debug_init
                _sa_mem_debug_init ();
            }

            sa_acquire_spinlock (&sa_mem_debug_context.lock, &irq);
            sa_insert_tail_list (&sa_mem_debug_context.allocated_memory, &hdr->next);
            sa_release_spinlock (&sa_mem_debug_context.lock, &irq);
        }

        return va;
    }

    /*!
    @brief Frees virtual memory allocated by sa_KMAlloc.  
    @param Ptr         - Memory to be freed
    */        
    void            
    _sa_free_debug (PVOID ptr, SA_MEM_ALLOC_CONTEXT *optional_context)
    {    
        if (ptr) {
            PSA_MEM_HEADER hdr = (PSA_MEM_HEADER)((u8*)ptr - sizeof (SA_MEM_HEADER));
            PSA_MEM_TAIL tail;
            SA_IRQ irq;        

            SA_RELEASE_ASSERT (hdr->magic_head == SA_MAGIC_HEAD);
            SA_RELEASE_ASSERT (hdr->magic_tail == SA_MAGIC_TAIL);
            SA_RELEASE_ASSERT (hdr->next.Flink && hdr->next.Blink);

            hdr->magic_head = ~SA_MAGIC_HEAD; // helps to detect double frees
            
            tail = (PSA_MEM_TAIL)((u8*)ptr + hdr->length);
            SA_RELEASE_ASSERT (tail->magic_head == SA_MAGIC_HEAD);
            SA_RELEASE_ASSERT (tail->magic_tail == SA_MAGIC_TAIL);

            ptr = hdr;

            sa_acquire_spinlock (&sa_mem_debug_context.lock, &irq);
            sa_remove_entry_list (&hdr->next);
            sa_release_spinlock (&sa_mem_debug_context.lock, &irq);

            _sa_free_release (ptr, optional_context);
        }
    }

    void
    _sa_mem_debug_dump (void)
    {
        PSA_MEM_HEADER hdr;
        PSA_MEM_TAIL tail;
        SA_IRQ irq;
        u32 num = 0, size = 0;

        if (!sa_mem_debug_context.enabled) {
            _sa_mem_debug_init ();
        }

        sa_acquire_spinlock (&sa_mem_debug_context.lock, &irq);
        SA_FOR_EACH_LIST_ENTRY (hdr, sa_mem_debug_context.allocated_memory, SA_MEM_HEADER, next) {
            c8 *tag = (c8*)&hdr->tag;        
            
            ASSERT (hdr);
            SA_RELEASE_ASSERT (hdr->magic_head == SA_MAGIC_HEAD); // Check for memory corruption
            SA_RELEASE_ASSERT (hdr->magic_tail == SA_MAGIC_TAIL); // Check for memory corruption

            tail = (PSA_MEM_TAIL)((u8*)hdr + sizeof(SA_MEM_HEADER) + hdr->length);
            SA_RELEASE_ASSERT (tail->magic_head == SA_MAGIC_HEAD); // Check for memory corruption (array bounds overflow)
            SA_RELEASE_ASSERT (tail->magic_tail == SA_MAGIC_TAIL); // Check for memory corruption (array bounds overflow)

            num++;
            size += hdr->length;

            sa_trace (DL_INFO, 
                "tag 0x%x (%c%c%c%c) @ %p is %d bytes. File:%s (%d)", 
                hdr->tag, 
                sa_text_toprintable (tag[0]), 
                sa_text_toprintable (tag[1]), 
                sa_text_toprintable (tag[2]), 
                sa_text_toprintable (tag[3]), 
                (PVOID) (hdr + 1), 
                hdr->length, 
                hdr->filename, 
                hdr->line);
        }    
        sa_release_spinlock (&sa_mem_debug_context.lock, &irq);

        sa_trace (DL_INFO, "Current memory allocations:%d bytes:%d", num, size);
    }

    void
    _sa_mem_debug_scan_magic (void)
    {
        PSA_MEM_HEADER hdr;
        PSA_MEM_TAIL tail;
        SA_IRQ irq;
        u32 num = 0, size = 0;

        if (!sa_mem_debug_context.enabled) {
            _sa_mem_debug_init ();
        }

        sa_acquire_spinlock (&sa_mem_debug_context.lock, &irq);
        SA_FOR_EACH_LIST_ENTRY (hdr, sa_mem_debug_context.allocated_memory, SA_MEM_HEADER, next) {
            
            ASSERT (hdr);
            SA_RELEASE_ASSERT (hdr->magic_head == SA_MAGIC_HEAD); // Check for memory corruption
            SA_RELEASE_ASSERT (hdr->magic_tail == SA_MAGIC_TAIL); // Check for memory corruption

            tail = (PSA_MEM_TAIL)((u8*)hdr + sizeof(SA_MEM_HEADER) + hdr->length);
            SA_RELEASE_ASSERT (tail->magic_head == SA_MAGIC_HEAD); // Check for memory corruption (array bounds overflow)
            SA_RELEASE_ASSERT (tail->magic_tail == SA_MAGIC_TAIL); // Check for memory corruption (array bounds overflow)

            num++;
            size += hdr->length;
        }    
        sa_release_spinlock (&sa_mem_debug_context.lock, &irq);

        sa_trace (DL_INFO, "Magic scan done. Current allocations:%d bytes:%d", num, size);
    }

#endif

