/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

//
//    Implements generic index and object allocators that use a 
//    free list to efficiently allocate/deallocate items.  Supports
//    both multi-thread safe, and non-safe methods.
//
#include "sa_ring_index_allocator.h"
#include "sa_mem.h"
#include "sa_int_math.h"
#include "sa_assert.h"

PSA_RING_ALLOCATOR 
sa_ringalloc_allocate (u32 num,
                       u32 alloc_flags,
                       const c8* optional_name)
{
    SA_STATUS status = SA_STATUS_SUCCESS;
    boolean *dbg_vector = NULL;

    // Allocate the object itself.
    PSA_RING_ALLOCATOR ra = sa_malloc (sizeof (SA_RING_ALLOCATOR),
                                        alloc_flags,
                                        SA_TAG ("ra__"));

    // Allocate ring indices
    u32 *ring = sa_malloc (sizeof (u32) * (num+1),
                            alloc_flags,
                            SA_TAG ("rarg"));

    #if defined(SA_DEBUG)
        // Allocate the debug vector to track double frees, etc.
        dbg_vector = sa_malloc (sizeof (boolean) * num, 
                                 alloc_flags, 
                                 SA_TAG ("radv"));
    #endif

    if (!ra || !ring || !dbg_vector) {
        status = SA_STATUS_NOT_OK;
    } else {
        status = sa_ringalloc_create (ra, ring, num, dbg_vector, optional_name);
    }

    if (status != SA_STATUS_SUCCESS) {
        sa_free (ra);
        sa_free (ring);
        sa_free (dbg_vector);

        return NULL;
    }

    return ra;
}


void
sa_ringalloc_free (PSA_RING_ALLOCATOR ra)
{
    if (ra == NULL) {
        return;
    }

    sa_ringalloc_destroy (ra);

    sa_free (ra->ring);
    sa_free (ra->dbg_vector);
    sa_free (ra);
}



/*!
@brief 
    Initializes a ring used for allocating the index values.

@param 
    ra - pointer to ring allocator to intialize
@param 
    ring - buffer with num+1 elements
@param
    num - number of indices to allocate from 0 to n-1.
@param
    optionalDbgVector - optional array of BOOLEANs used only during
        debug builds for tracking double frees.  This vector must
        have "num" elements.
@param
    optionalName - optional name used for debug prints
                      
@return
    None
*/
SA_STATUS
sa_ringalloc_create (PSA_RING_ALLOCATOR ra,         // object to init
                     u32 *ring,                  // buffer with num+1 elements
                     u32 num,                    // num of elements that can be allocated
                     boolean *optional_dbg_vector, // optional dbg vector with num elements
                     const c8* optional_name)   
{
    u32 i;

    ASSERT (ra);
    ASSERT (ring);
    ASSERT (num > 0);

    SA_NOT_USED (optional_dbg_vector);

    //
    // Initialize values of the ia structure.
    //
    memset (ra, 0, sizeof (SA_RING_ALLOCATOR));

    ra->magic = SA_RA_MAGIC;
    ra->num = num;
    ra->ring = ring;
    ra->free_index = num;
    ra->allocate_index = 0;
        
    for (i = 0; i < num; i++) {
        ring[i] = i;
    }
    ring[num] = SA_RA_BAD_INDEX;

#if defined(SA_DEBUG)
    //
    // Optionally, provide tracking for double frees.
    //
    if (optional_dbg_vector) {
        memset (optional_dbg_vector, 0, sizeof (boolean) * num);
        ra->dbg_vector = optional_dbg_vector;
    }
#endif

    ra->dbg_name = optional_name;
    TRACE (DL_INFO, "init index allocator %s with %d entries", SA_DBG_CSTR(optional_name), num);

    return SA_STATUS_SUCCESS;
}


void
sa_ringalloc_destroy (PSA_RING_ALLOCATOR ra)
{
    ra->magic = 0;
    // nop
}


/*!
@brief 
    Allocates an index value.
    If no values are available, SA_RA_BAD_INDEX is returned. 
    It is safe to allocate from one thread and free from
    another simultaneously.  It is NOT safe for two threads
    to simultaneously allocate, or two threads to simultaneously
    free.

@param 
    ra - pointer to ring allocator
                      
@return
    allocated index on success
    SA_RA_BAD_INDEX on failure
*/
u32
sa_ringalloc_allocate_index (PSA_RING_ALLOCATOR ra) {

    u32 index;

    ASSERT (ra);
    ASSERT (ra->magic == SA_RA_MAGIC);
    ASSERT (ra->ring);

    if (sa_ringalloc_num_free (ra) > 0) 
    {
        index = ra->ring[ra->allocate_index];

        SA_DBG_ONLY (ra->ring[ra->allocate_index] = SA_RA_BAD_INDEX);

        ra->allocate_index = sa_addc (ra->allocate_index, 1, ra->num+1);

    } else {
        TRACE (DL_INFO, "Allocate %s index failed.", SA_DBG_CSTR(ra->dbg_name));
        index = SA_RA_BAD_INDEX;  
        return index;
    }

    ASSERT (index < ra->num);

#if defined(SA_DEBUG)
    //
    // Optional debug checks
    //
    if (ra->dbg_vector) {
        ASSERT (ra->dbg_vector[index] == FALSE);
        ra->dbg_vector[index] = TRUE;
    }
    TRACE (DL_VERBOSE, "Allocate %s index %d.", SA_DBG_CSTR (ra->dbg_name), index);
#endif
    
    return index;
}


/*!
@brief 
    Frees an index.
    It is safe to allocate from one thread and free from
    another simultaneously.  It is NOT safe for two threads
    to simultaneously allocate, or two threads to simultaneously
    free.

@param 
    ra - pointer to ring allocator
@param
    index - value to free
                      
@return
    none
*/
void
sa_ringalloc_free_index (PSA_RING_ALLOCATOR ra, u32 index) {

    ASSERT (ra);
    ASSERT (ra->magic == SA_RA_MAGIC);
    ASSERT (ra->ring);
    ASSERT (index < ra->num);
    ASSERT (sa_ringalloc_num_allocated (ra) > 0);    

    //
    // Optional debug check for double frees.
    //
#if defined(SA_DEBUG)
    if (ra->dbg_vector) {
        ASSERT (ra->dbg_vector[index] == TRUE);
        ra->dbg_vector[index] = FALSE;
    }
    TRACE (DL_VERBOSE, "Free index %d for %s", index, SA_DBG_CSTR (ra->dbg_name));
#endif

    ASSERT (ra->ring[ra->free_index] == SA_RA_BAD_INDEX);
    ra->ring[ra->free_index] = index;
    ra->free_index = sa_addc (ra->free_index, 1, ra->num+1);
    return;
}

#if !defined(SA_OMIT_TEST_CODE)

    void
    sa_ringalloc_test ()
    {
        SA_RING_ALLOCATOR ra;
        u32 ring[5];
        boolean dbg[4];
        u32 a,b,c,d,e;


        sa_ringalloc_create (&ra, ring, 4, dbg, "test");

        a = sa_ringalloc_allocate_index (&ra);
        ASSERT (a == 0);

        b = sa_ringalloc_allocate_index (&ra);    
        ASSERT (b == 1);

        c = sa_ringalloc_allocate_index (&ra);
        ASSERT (c == 2);

        d = sa_ringalloc_allocate_index (&ra);
        ASSERT (d == 3);

        e = sa_ringalloc_allocate_index (&ra); // should fail
        ASSERT (e == SA_RA_BAD_INDEX);
        
        sa_ringalloc_free_index (&ra, c);
        sa_ringalloc_free_index (&ra, b);
        sa_ringalloc_free_index (&ra, a);

        a = sa_ringalloc_allocate_index (&ra);
        ASSERT (a == 2);

        b = sa_ringalloc_allocate_index (&ra);
        ASSERT (b == 1);

        c = sa_ringalloc_allocate_index (&ra);
        ASSERT (c == 0);

        e = sa_ringalloc_allocate_index (&ra); // should fail
        ASSERT (e == SA_RA_BAD_INDEX);

        sa_ringalloc_free_index (&ra, a);
        sa_ringalloc_free_index (&ra, b);
        sa_ringalloc_free_index (&ra, c);
        sa_ringalloc_free_index (&ra, d);

        sa_ringalloc_destroy (&ra);

        return;
    }

#endif

