/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!   

@file
    $RCSfile: sa_text.c,v $

@brief
    String manipulation routines needed the driver.

@notes

CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_text.c,v 1.3 2006/03/01 22:04:12 menz Exp $
*/  

#include "sa_text.h"
#include "sa_assert.h"
#include "sa_mem.h"

/*!
@brief
    Same as strlen, but it scans for NULL only up to maxLength.  If
    the string is not terminated within maxLength, it returns 0.
@param
    str - string to get length
@param
    maxLength - maximum length to scan for NULL
@return
    length of string, 0 on error
*/
u32
sa_text_len (const c8* str, u32 maxLength)
{    
    u32 i = 0;

    while (i < maxLength && str[i] != 0) {
        i++;        
    }

    if (str[i] != 0) {
        // String not terminated before maxLength
        return 0; 
    }

    return i;
}

u32
sa_wtext_len (const c16* str, u32 maxLength)
{    
    u32 i = 0;

    while (i < maxLength && str[i] != 0) {
        i++;        
    }

    if (str[i] != 0) {
        // String not terminated before maxLength
        return 0; 
    }

    return i;
}

/*!
@brief
    Same as textLen, but it scans for any character.
@param
    str - string to get length
@param
    maxLength - maximum length to scan for NULL
@return
    length of string, 0 on error
*/
u32 
sa_text_find_char (const c8* str, c8 c, u32 maxLength)
{
    u32 i = 0;

    while (i < maxLength && str[i] != c) {
        i++;        
    }

    if (str[i] != c) {
        // Char not found before maxLength
        return 0; 
    }

    return i;
}

u32 
sa_wtext_find_char (const c16* str, c16 c, u32 maxLength)
{
    u32 i = 0;

    while (i < maxLength && str[i] != c) {
        i++;        
    }

    if (str[i] != c) {
        // Char not found before maxLength
        return 0; 
    }

    return i;
}


/*!
@brief
    Simple hash function for a string.
@param
    str - string to hash
@param
    maxLength - number of characters to hash
@return
    hash value
*/
u32
textHash (c8* str, u32 len)
{
   u32 hash = 1315423911;
   u32 i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash ^= ((hash << 5) + sa_text_toupper (*str) + (hash >> 2));
   }

   return hash;
}

/*!
@brief
    Compares strings independent of case, making sure strings are the same length.

@param
    a - string to compare
@param
    b - string to compare

@return
    TRUE if 'a' equals b, independent of case, for all characters
*/
boolean
sa_text_compare_no_case (const char* a, const char *b, u32 maxLength) {

    char A,B;
    u32 i = 0;

    ASSERT (a && b);
    
    while (*a && *b && (i++ < maxLength)) {

        A = sa_text_toupper (*a);
        B = sa_text_toupper (*b);
        if (A != B) {
            return FALSE;
        }
        a++;
        b++;
    }

    if ((*b | *a) != 0) {
        // Strings are not the same length. Or maxLength < the real length.
        return FALSE;
    }

    //
    // Strings match each character
    //
    return TRUE;
}


boolean
sa_wtext_compare_no_case (const c16* a, const c16 *b, u32 maxLength) {

    c16 A,B;
    u32 i = 0;

    ASSERT (a && b);
    
    while (*a && *b && (i++ < maxLength)) {

        A = sa_wtext_toupper (*a);
        B = sa_wtext_toupper (*b);
        if (A != B) {
            return FALSE;
        }
        a++;
        b++;
    }

    if ((*b | *a) != 0) {
        // Strings are not the same length. Or maxLength < the real length.
        return FALSE;
    }

    //
    // Strings match each character
    //
    return TRUE;
}


/*!
@brief
    Compares strings independent of case up to strlen(b) characters.

@param
    a - string to compare (it may be longer than b)
@param
    b - string to compare, all its characters must match

@return
    TRUE if 'a' starts with string b, independent of case
*/
boolean
sa_text_starts_with (const char* a, const char *b) {

    char A,B;
    const char *bEnd;

    ASSERT (a && b);

    bEnd = b + strlen (b);

    while (*a && *b) {

        A = sa_text_toupper (*a);
        B = sa_text_toupper (*b);
        if (A != B) {
            return FALSE;
        }
        a++;
        b++;
    }

    if (b != bEnd) {
        return FALSE; // String a is shorter than b so the match fails.
    }

    //
    // Strings match each character up to the shorter of the two strings.
    //
    return TRUE;
}


/*!
@brief 
    Converts with a hex string (starting with "0x") or a decimal
    string into a u32.
            
@param
    str - Pointer to a Numeric ASCII String
@param
    size - Size of the string buffer.
@param
    value - pointer to numeric value returned
    
@return 
    True on success
*/
boolean
sa_text_str_to_int (const char* str, 
                    u32 size, 
                    u32 *value)
{
    ASSERT (str);
    ASSERT (value);

    if (size == 0) {        
        return FALSE;
    }

    if ((size > 2) && (str[0] == '0') && (sa_text_toupper (str[1]) == 'X')) {
        return sa_text_hex_str_to_int (&str[2], size-2, value);
    } else {
        return sa_text_dec_str_to_int (str, size, value);
    }
}    


boolean
sa_wtext_str_to_int (const c16* str, 
                     u32 size, 
                     u32 *value)
{
    ASSERT (str);
    ASSERT (value);

    if (size == 0) {        
        return FALSE;
    }

    if ((size > 2) && (str[0] == L'0') && (sa_wtext_toupper (str[1]) == L'X')) {
        return sa_wtext_hex_str_to_int (&str[2], size-2, value);
    } else {
        return sa_wtext_dec_str_to_int (str, size, value);
    }
} 


/*!
@brief 
    Converts a hexadecimal string value into a u32 numeric format.
            
@param
    str - Pointer to a Numeric ASCII String
@param
    size - size of the string buffer.
@param
    value - resulting value

@return 
    TRUE on success
*/
boolean
sa_text_hex_str_to_int (const char* str, 
                        u32 size, 
                        u32 *value)
{
    return sa_text_base_str_to_int (str, size, 16, value);
}

boolean
sa_wtext_hex_str_to_int (const c16* str, 
                         u32 size, 
                         u32 *value)
{
    return sa_wtext_base_str_to_int (str, size, 16, value);
}


/*!
@brief 
    Converts a decimal (base 10) string value into a u32 numeric format.
            
@param
    str - Pointer to a Numeric ASCII String
@param
    size - size of the string buffer.
@param
    value - resulting value

@return 
    TRUE on success
*/
boolean
sa_text_dec_str_to_int (const char* str, 
                        u32 size, 
                        u32 *value)
{
    return sa_text_base_str_to_int (str, size, 10, value);
}

boolean
sa_wtext_dec_str_to_int (const c16* str, 
                         u32 size, 
                         u32 *value)
{
    return sa_wtext_base_str_to_int (str, size, 10, value);
}


/*!
@brief 
    Converts any byte array into a hex string

@param
    str - Pointer to a Numeric ASCII String
@param
    strLength - Size of the string buffer.
@param
    value - byte array for result
@param
    valueLength - length of byte array
        
@return 
    number of char stored str including NULL.  zero on failure.
*/
u32
sa_text_byte_array_to_hex_str (c8* str, 
                               u32 strLength, 
                               u8 *value, 
                               u32 valueLength) 
{
    u32 i;
    u8 digit;
    const c8* hexDigits = "0123456789ABCDEF";

    ASSERT (str);
    ASSERT (value);

    if (valueLength == 0) {
        return 0;
    }

    // Resulting string starts with 0x and is terminated.
    if (strLength < valueLength*2 + 3) {
        return 0;
    }
    
    *str++ = '0';
    *str++ = 'x';
    
    //
    // Iterate for byte array
    //
    for (i=0; i<valueLength; i++) {
        
        digit = value[i];
        *str++ = hexDigits[(digit >> 4) & 0xF];
        *str++ = hexDigits[digit & 0xF];
    }

    *str = 0; // terminate

    return valueLength*2 + 3;
}


u32
sa_wtext_byte_array_to_hex_str (c16* str, 
                                u32 strLength, 
                                u8 *value, 
                                u32 valueLength) 
{
    u32 i;
    u8 digit;
    c16* hexDigits = (c16*)L"0123456789ABCDEF";

    ASSERT (str);
    ASSERT (value);

    if (valueLength == 0) {
        return 0;
    }

    // Resulting string starts with 0x and is terminated.
    if (strLength < valueLength*2 + 3) {
        return 0;
    }
    
    *str++ = L'0';
    *str++ = L'x';
    
    //
    // Iterate for byte array
    //
    for (i=0; i<valueLength; i++) {
        
        digit = value[i];
        *str++ = hexDigits[(digit >> 4) & 0xF];
        *str++ = hexDigits[digit & 0xF];
    }

    *str = 0; // terminate

    return valueLength*2 + 3;
}


/*!
@brief 
    Converts any byte array into a base64 string

@param
    str - Pointer to a Numeric ASCII String
@param
    strLength - Size of the string buffer.
@param
    value - byte array for result
@param
    valueLength - length of byte array
        
@return 
    number of char stored str including NULL.  zero on failure.
*/
u32
sa_text_byte_array_to_base64_str (c8* str, 
                                  u32 strLength, 
                                  u8 *value, 
                                  u32 valueLength) 
{
    u32 i;
    u32 digit;
    u32 outlen = 4 * (valueLength+2)/3 + 3;

    // 6 bits per char
    const c8* base64Digits = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

    ASSERT (str);
    ASSERT (value);

    if (valueLength == 0) {
        return 0;
    }

    // Resulting string starts with 0b and is terminated.
    
    if (strLength < outlen) {
        return 0;
    }
    
    *str++ = '0';
    *str++ = 'b';
    
    //
    // 4 chars per 3 bytes.
    //
    for (i=0; i < 3*(valueLength/3); i+=3) {
        digit = (value[i]<<16) | (value[i+1]<<8) | value[i+2];

        *str++ = base64Digits[(digit & 0xFC0000) >> 18];
        *str++ = base64Digits[(digit & 0x03F000) >> 12];
        *str++ = base64Digits[(digit & 0x000FC0) >> 6];
        *str++ = base64Digits[(digit & 0x00003F)];
    }


    if (valueLength%3 == 1) {

        *str++ = base64Digits[(value[i] & 0xFC) >> 2];
		*str++ = base64Digits[((value[i] & 0x03) << 4)];
		*str++ = base64Digits[64];
		*str++ = base64Digits[64];
		
	} else if (valueLength%3 == 2) {

		*str++ = base64Digits[(value[i]   & 0xFC) >> 2];
		*str++ = base64Digits[((value[i]  & 0x03) << 4) | ((value[i+1] & 0xF0) >> 4)];
		*str++ = base64Digits[(value[i+1] & 0x0F) << 2];
		*str++ = base64Digits[64];		
	}

    *str = 0; // terminate

    return outlen;
}

/*!
@brief 
    Maps a character from a base64 string to the value it represents.
            
@param
    c - character to map
@param
    value - output value.  Valid values are 0-63. Pad characters will return 65.
            Invalid characters will return something >65.
        
@return 
    TRUE if conversion is ok.  FALSE if the character
    is greater than 64.
*/
boolean
sa_text_to_base64 (char c, u8 *value) {

    const char offset = 43;
    const char *reverseMap = "i~~~j_`abcdefgh~~~k~~~+,-./0123456789:;<=>?@ABCD~~~~~~EFGHIJKLMNOPQRSTUVWXYZ[\\]^";                                  
    const char len = 80;

    if (c < offset || c-offset > len)
        *value = 66; // illegal value
    else
        *value = reverseMap[c - offset] - offset;

    if (*value > 65) {
        TRACE (DL_ERR, "Illegal char base64 string: %c", c);        
    }

    if (*value > 64) {
        return FALSE;
    }

    return TRUE;
}

/*!
@brief 
    Converts a base64 string into a byte array following RFC 3548.
            
@param
    str - Input ASCII String
@param
    strLength - Length of string
@param
    value - output byte array
@param
    valueLength - length of byte array
        
@return 
    number of bytes stored value array.  zero on failure.
*/
u32
sa_text_base64_str_to_byte_array (const char* str, 
                                  u32 strLength, 
                                  u8 *value, 
                                  u32 valueLength)
{  
    u8    byte[4]; 
    u32   word;
    u32   i;    
    boolean  success;
    u32 bytesOut = 0;

    ASSERT (str);
    ASSERT (value);


    if ((strLength > 2) && (str[0] == '0') && (sa_text_toupper (str[1]) == 'B')) {
        str += 2;
        strLength -= 2;
    }

    if (strLength == 0) {
        return 0;
    }

    if (strLength % 4 != 0) {
        TRACE (DL_ERR, "Bad base64 string length:%d", strLength);
        return 0;
    }

    bytesOut = strLength / 4 * 3;
    if (str[strLength - 1] == '=') {
        bytesOut--;
        if (str[strLength - 2] == '=') {	    
            bytesOut--;
        }
    }

    if (valueLength < bytesOut) {
        TRACE (DL_ERR, "Byte array too short (%d < %d).", valueLength, bytesOut);
        return 0;
    }

    // 4 chars into 3 bytes
    for (i = 0; i< 3 * (bytesOut/3); i+=3) {
        
        success  = sa_text_to_base64 (*str++, &byte[0]);
        success |= sa_text_to_base64 (*str++, &byte[1]);
        success |= sa_text_to_base64 (*str++, &byte[2]);
        success |= sa_text_to_base64 (*str++, &byte[3]);

        if (!success) {
            TRACE (DL_ERR, "Unexpected character base64 string.");
            return 0;
        }

        word  = (byte[0] << 18) | 
                (byte[1] << 12) |
                (byte[2] << 6)  |
                (byte[3] << 0);

        value[i]     = (u8)((word & 0xFF0000) >> 16);
		value[i + 1] = (u8)((word & 0x00FF00) >> 8);
		value[i + 2] = (u8)(word & 0x0000FF);
    }    
    
    if (bytesOut % 3 == 2) {

        //
        // two more bytes, requires 3 chars
        //
        
        success  = sa_text_to_base64 (*str++, &byte[0]);
        success |= sa_text_to_base64 (*str++, &byte[1]);
        success |= sa_text_to_base64 (*str++, &byte[2]);

        if (!success) {
            TRACE (DL_ERR, "Unexpected character base64 string.");
            return 0;
        }

        word  = (byte[0] << 18) | 
                (byte[1] << 12) |
                (byte[2] << 6);

        value[i]     = (u8)((word & 0xFF0000) >> 16);
		value[i + 1] = (u8)((word & 0x00FF00) >> 8);


    } else if (bytesOut % 3 == 1) {
    
        //
        // one more byte, requires 2 chars
        //
        
        success  = sa_text_to_base64 (*str++, &byte[0]);
        success |= sa_text_to_base64 (*str++, &byte[1]);

        if (!success) {
            TRACE (DL_ERR, "Unexpected character base64 string.");
            return 0;
        }

        word  = (byte[0] << 18) | 
                (byte[1] << 12);

        value[i]     = (u8)((word & 0xFF0000) >> 16);
    }
      

    return bytesOut;
}

/*!
@brief 
    Converts a base64 or base16 (hex) string into a byte array following RFC 3548.
    The base of the string is determined by the first 2 characters, 0x for hex
    or 0b for base64.
            
@param
    str - Input ASCII String
@param
    strLength - Length of string
@param
    value - output byte array
@param
    valueLength - length of byte array
        
@return 
    number of bytes stored value array.  zero on failure.
*/
u32
sa_text_str_to_byte_array  (const char* str, 
                            u32 strLength, 
                            u8 *value, 
                            u32 valueLength)
{

    if (str[0] == '0' && sa_text_toupper (str[1]) == 'B') {
        return sa_text_base64_str_to_byte_array (str, strLength, value, valueLength);
    } else if (str[0] == '0' && sa_text_toupper (str[1]) == 'X') {
        return sa_text_hex_str_to_byte_array (str, strLength, value, valueLength);
    } else {
        TRACE (DL_ERR, "Unrecognized byte array string started with: %c%c", str[0], str[1]);
        return 0;
    }
}

/*!
@brief 
    Converts any base number into a u32 numeric format.
            
@param
    str - Pointer to a Numeric ASCII String
@param
    strLength - Size of the string buffer.
@param
    value - byte array for result
@param
    valueLength - length of byte array
        
@return 
    number of bytes stored value array.  zero on failure.
*/
u32
sa_text_hex_str_to_byte_array (const char* str, 
                               u32 strLength, 
                               u8 *value, 
                               u32 valueLength)
{   
    u8    Key;   
    u32   i;    

    ASSERT (str);
    ASSERT (value);   

    if ((strLength > 2) && (str[0] == '0') && (sa_text_toupper (str[1]) == 'X')) {
        str += 2;
        strLength -= 2;
    }

    if (valueLength*2 < strLength) {
        TRACE (DL_ERR, "Byte array too short (%d < %d).", valueLength*2, strLength);
        return 0;
    }

    if (strLength == 0) {
        return 0;
    }

    //
    // Clear output
    //
    sa_memset (value, 0, valueLength); 
    
    for (i = 0; i < strLength; i++) {
    
        Key = *str;
        str++;
        ASSERT (Key != 0); // indicates str is terminated before strLength
        
        //
        // Convert to upper case for hex letters.
        //
        Key = sa_text_toupper (Key);

        //
        // Get numerical value of the character.
        //
        if ((Key >= '0') && (Key <= '9')) {
            Key = Key - '0';
        } else if ((Key >= 'A') && (Key <= 'F')) {
            Key = Key - 'A' + 10;
        } else {
            TRACE (DL_ERR, "Invalid numeric char:%c string", Key);
            return 0;
        }

        //
        // If odd index, this character is high 4 bits the byte.
        //
        if ((i & 1) == 0) {
            Key <<= 4;
        }
        value[i>>1] |= Key;
    } 
    
    return (i+1)>>1;
}   

/*!
@brief 
    Converts any base number into a u32 numeric format.
            
@param
    str - Pointer to a Numeric ASCII String
@param
    size - Size of the string buffer.
@param
    base - Number system base (e.g. 16 for hex)
@param
    value - resulting value
        
@return 
    TRUE on success
*/
boolean
sa_text_base_str_to_int (const char* str, 
                         u32 size, 
                         u32 base, 
                         u32 *value)
{
    u32   ConvFac = 1;    
    u32   RetVal  = 0x00;
    u8    Key;   
    i32     i;

    ASSERT (str);
    ASSERT (value);
    ASSERT (base <= 16);
    
    for (i=(size-1);i>-1;i--) {      

        Key = *(str+i);
        ASSERT (Key != 0);
        
        //
        // Convert to upper case for hex letters.
        //
        Key = sa_text_toupper (Key);

        //
        // Get numerical value of the character.
        //
        if ((Key >= '0') && (Key <= '9')) {
            Key = Key - '0';
        } else if ((Key >= 'A') && (Key <= 'F')) {
            Key = Key - 'A' + 10;
        } else {
            TRACE (DL_ERR, "Invalid numeric char:%c string", Key);
            return FALSE;
        }

        if (Key >= base) {
            TRACE (DL_ERR, "Invalid char string during conversion to int, base %d", base);
            return FALSE;
        }
        
        RetVal += Key * ConvFac;        
        ConvFac = ConvFac * base;
    } 

    *value = RetVal;
    
    return TRUE;
}    


boolean
sa_wtext_base_str_to_int (const c16* str, 
                          u32 size, 
                          u32 base, 
                          u32 *value)
{
    u32   ConvFac = 1;    
    u32   RetVal  = 0x00;
    c16   Key;   
    i32   i;

    ASSERT (str);
    ASSERT (value);
    ASSERT (base <= 16);
    
    for (i=(size-1);i>-1;i--) {      

        Key = str[i];
        ASSERT (Key != 0);
        
        //
        // Convert to upper case for hex letters.
        //
        Key = sa_wtext_toupper (Key);

        //
        // Get numerical value of the character.
        //
        if ((Key >= L'0') && (Key <= L'9')) {
            Key = Key - L'0';
        } else if ((Key >= L'A') && (Key <= L'F')) {
            Key = Key - L'A' + 10;
        } else {
            TRACE (DL_ERR, "Invalid numeric char:%c (%d)", Key, (u32)Key);
            return FALSE;
        }

        if (Key >= base) {
            TRACE (DL_ERR, "Invalid char string during conversion to int, base %d", base);
            return FALSE;
        }
        
        RetVal += Key * ConvFac;        
        ConvFac = ConvFac * base;
    } 

    *value = RetVal;
    
    return TRUE;
}

/*!
@brief 
    Converts an unsigned integer into a string, writing up
    to strLength-1 characters to the str buffer.  The NULL
    is always included.  The maximum length of the string is
    10 bytes.

@param
    str - Pointer to a Numeric ASCII String, Output
@param
    number - number to convert.
@param
    strLength - Length of str buffer.
@return 
    The size of the numeric string excluding the NULL terminater.
    On error returns 0.
*/
u32
sa_text_int_to_base_str (c8* str, 
                         u32 strLength, 
                         u32 base, 
                         u32 number, 
                         u32 fixedDigits) 
{
    u32     x = 0;
    u32     y;
    u32     digit;
    #define kInternalStrLength 33 // Base 2 would yield 33 characters.
    u8      string[kInternalStrLength]; 

    ASSERT (str);

    str[0] = 0;

    if (strLength == 0) {
        return 0;
    }

    if (base < 2 || base > 16) {
        TRACE (DL_ERR, "Base base %d for int to str conversion.", base);
        return 0;
    }

    if (fixedDigits > kInternalStrLength-1) {
        TRACE (DL_ERR, "Bad fixed digits.");
        return 0;
    }

    if (number == 0) {
        string[x] = '0';
        x++;
    }   

    while (number > 0 && x < strLength) {
        
        ASSERT (x < kInternalStrLength);

        digit = number % base;

        if (digit < 10) {
            string[x] = (u8)digit + '0';
        } else {
            string[x] = (u8)(digit-10) + 'A';
        }

        number /= base;
        x++;
    }

    ASSERT (x < kInternalStrLength);
    while (fixedDigits > x) {
        string[x++] = '0'; // add preceeding zeros (for MAC addresses & fixed length decimal strings)
    }
    ASSERT (x < kInternalStrLength);
    string[x] = 0; // terminate

    if (x > strLength) {
        BREAKPOINT ();
        TRACE (DL_ERR, 
               "Too short a string (%d bytes) for int (%d) to string base (%d)",
               strLength, 
               number, 
               base);
        return 0;
    }

    // invert and copy into output, this includes termination    
    for (y = 0; y < x; y++) {
        str[y] = string[x - y - 1];
    }
    str[y] = 0; // terminate

    return y;
}

u32
sa_wtext_int_to_base_str (c16* str, 
                          u32 strLength, 
                          u32 base, 
                          u32 number, 
                          u32 fixedDigits) 
{
    u32     x = 0;
    u32     y;
    u32     digit;
    #define kInternalStrLength 33 // Base 2 would yield 33 characters.
    c16     string[kInternalStrLength]; 

    ASSERT (str);

    str[0] = 0;

    if (strLength == 0) {
        return 0;
    }

    if (base < 2 || base > 16) {
        TRACE (DL_ERR, "Base base %d for int to str conversion.", base);
        return 0;
    }

    if (fixedDigits > kInternalStrLength-1) {
        TRACE (DL_ERR, "Bad fixed digits.");
        return 0;
    }

    if (number == 0) {
        string[x] = L'0';
        x++;
    }   

    while (number > 0 && x < strLength) {
        
        ASSERT (x < kInternalStrLength);

        digit = number % base;

        if (digit < 10) {
            string[x] = (u8)digit + L'0';
        } else {
            string[x] = (u8)(digit-10) + L'A';
        }

        number /= base;
        x++;
    }

    ASSERT (x < kInternalStrLength);
    while (fixedDigits > x) {
        string[x++] = L'0'; // add preceeding zeros (for MAC addresses & fixed length decimal strings)
    }
    ASSERT (x < kInternalStrLength);
    string[x] = 0; // terminate

    if (x > strLength) {
        BREAKPOINT ();
        TRACE (DL_ERR, 
               "Too short a string (%d bytes) for int (%d) to string base (%d)",
               strLength, 
               number, 
               base);
        return 0;
    }

    // invert and copy into output, this includes termination    
    for (y = 0; y < x; y++) {
        str[y] = string[x - y - 1];
    }
    str[y] = 0; // terminate

    return y;
}

/*!
@brief 
    Converts an unsigned integer into a hex string, writing up
    to strLength-1 characters to the str buffer.  The NULL
    is always included. 

@param
    str - Pointer to a Numeric ASCII String, Output
@param
    number - number to convert.
@param
    strLength - Length of str buffer.
@return 
    The size of the numeric string excluding the NULL terminater.
*/
u32
sa_text_int_to_hex_str (c8* str, 
                        u32 strLength, 
                        u32 number, 
                        u32 fixedDigits) 
{
    return sa_text_int_to_base_str (str, strLength, 16, number, fixedDigits);
}

u32
sa_wtext_int_to_hex_str (c16* str, 
                         u32 strLength, 
                         u32 number, 
                         u32 fixedDigits) 
{
    return sa_wtext_int_to_base_str (str, strLength, 16, number, fixedDigits);
}

/*!
@brief 
    Converts an unsigned integer into a decimal string, writing up
    to strLength-1 characters to the str buffer.  The NULL
    is always included.  

@param
    str - Pointer to a Numeric ASCII String, Output
@param
    number - number to convert.
@param
    strLength - Length of str buffer.
@return 
    The size of the numeric string excluding the NULL terminater.
*/
u32
sa_text_int_to_dec_str (c8* str, 
                        u32 strLength, 
                        u32 number, 
                        u32 fixedDigits) 
{
    return sa_text_int_to_base_str (str, strLength, 10, number, fixedDigits);
}

u32
sa_wtext_int_to_dec_str (c16* str, 
                         u32 strLength, 
                         u32 number, 
                         u32 fixedDigits) 
{
    return sa_wtext_int_to_base_str (str, strLength, 10, number, fixedDigits);
}


/*!
@brief 
    Converts MAC address string to numeric values.
            
@param
    str - pointer to MAC address string
@param
    size - size of string
@param
    macAddr - array of sizeof (SA_MAC_ADDRESS) bytes mac address
    
@return 
    TRUE on success
*/
boolean 
sa_text_str_to_mac (const char* str, 
                 u32 strLength, 
                 SA_MAC_ADDRESS* mac)
{
    u32 offset = 0;
    u32 j;
    u32 digit;

    ASSERT (str);
    ASSERT (mac);
    
    if (strLength < sizeof (SA_MAC_ADDRESS) * 2) {
        return FALSE;
    }

    sa_zero_mem (mac, sizeof (SA_MAC_ADDRESS));

    // Convert each byte.
    for (j=0; offset <= strLength-2 && j < sizeof (SA_MAC_ADDRESS); j++, offset += 2) {
        
        if (sa_text_hex_str_to_int ((const char*)&str[offset], 2, &digit) == FALSE) {
            sa_zero_mem (mac, sizeof (SA_MAC_ADDRESS));
            return FALSE;
        }

        mac->bytes[j] = (u8)digit;

        if ((str[offset+2] == ':')||(str[offset+2] == '-')) {
            offset++;
        }
    }    
   
    return TRUE;
}

boolean 
sa_wtext_str_to_mac (c16* wstr, 
                     u32 strLength, 
                     SA_MAC_ADDRESS* mac)
{
    c8 convert[20];
    i32 len = sa_text_unicode_to_ansi_str (wstr, convert, sizeof (convert), strLength);

    if (len <= 0) {
        return FALSE;
    }

    return sa_text_str_to_mac (convert, (u32)len, mac);
}

/*!
@brief 
    Converts an integer to an MAC address hex string with ':' delimiters.
            
@param
    str - Pointer to MAC String
@param
    size - Size of string.
@param
    mac - MAC address
    
@return 
    Number of characters string, excluding NULL.
*/
u32 
sa_text_mac_to_str (c8* str, 
                    u32 size, 
                    SA_MAC_ADDRESS *mac)
{
    u32 offset = 0;
    u32 i;
    
    for (i=0; i<sizeof (SA_MAC_ADDRESS) && offset < size; i++) {
        offset += sa_text_int_to_hex_str (&str[offset], size - offset, mac->bytes[sizeof (SA_MAC_ADDRESS)-1-i], 2);
                
        if (i<(sizeof (SA_MAC_ADDRESS)-1) && offset < size) {
            str[offset++] = ':';
        }
    }
    
    if (offset >= size) {
        // the input string was too short 
        BREAKPOINT ();
        offset = 0;
    }

    // Always terminate string
    str[offset] = 0;
    return offset;
}

u32 
sa_wtext_mac_to_str (c16* str, 
                     u32 size, 
                     SA_MAC_ADDRESS *mac)
{
    c8 convert[20];
    u32 len = sa_text_mac_to_str (convert, size, mac);

    if (len <= 0) {
        return 0;
    }

    return sa_text_ansi_to_unicode (convert, str, size);
}

/*!
@brief 
    Converts an IPv4 address String into u32 numeric format.
    Assumpiton is that the IP String is Decimal format.    
            
@param
    str - Pointer to IP String
@param
    size - Size of string.
@param
    IpAddr - returned IP address
    
@return 
    TRUE and the Converted IpAddress parameter 3 on Success or
    FALSE on error
*/
boolean 
sa_text_str_to_ip (const char* str, 
                   u32 strLength, 
                   u32* IpAddr)
{                      
    boolean retVal  = TRUE;     // good success
    char   wrkStr[5];
    u32  i;
    i32 j;
    u32   Shift   = 24;       // shift 24 bits
        
    *(IpAddr)   = 0;            // init to zero.    
    sa_memset(&wrkStr[0], 0, 5);
        
    // size+1 to process the last charater
    for (i=0,j=0;i<(strLength+1);i++,j++ ) {        
        if (j > 3) {    
            // account for '.' delimiter
            //TRACE(DL_ERR, "\n  IP Parsing error, sub section size=%d\n",j);
            retVal = FALSE;
            break;    
        }
        
        if ( (*(str+i) == '.') || (*(str+i) == 0x00) ) {
            
            // Convert next byte to number.
            u32 digit = 0;
            if (!sa_text_str_to_int (&wrkStr[0], j, &digit)) {
                TRACE (DL_ERR, "IP Address parsing error: %s", str);
                retVal = FALSE;
                break;
            }

            *(IpAddr) |= ((digit & 0xff) << Shift);
            j = -1; // reset sub string size
            sa_memset(&wrkStr[0], 0, 5);
            
            Shift -= 8;
            continue;
        }
        
        wrkStr[j] = *(str+i);
    }
           
    return(retVal);
} 


/*!
@brief 
    Converts an integer to an IPv4 address decimal string with '.' delimiters.
            
@param
    str - Pointer to IP String
@param
    size - Size of string.
@param
    IpAddr - returned IP address
    
@return 
    Number of characters string, excluding NULL.
*/
u32 
sa_text_ip_to_str (c8* str, 
                   u32 size, 
                   u32 ip)
{
    u32 offset = 0;
    i32 shift = 24;
    u32 i;

    for (i=0; i<4 && offset < size; i++) {
        offset += sa_text_int_to_dec_str (&str[offset], size - offset, ((ip >> shift) & 0xff), 0);
        shift -= 8;
                
        if (i<3 && offset < size) {
            str[offset++] = '.';
        }
    }
    
    if (offset >= size) {
        // the input string was too short 
        BREAKPOINT ();
        offset = 0;
    }

    // Always terminate string
    str[offset] = 0;
    return offset;
}



/*!
@brief
    Copies the src string into destination pointer.  At most
    it copies destLength-1 characters, always maintaining
    a NULL terminated string.
@param
    destination - pointer to a destination string
@param
    destLength - maximum number of characters to copy
@param
    src - Source string to copy. Must be NULL terminated.
@return
   Actual number of characters copied, excluding NULL.
*/
u32
sa_text_copy (c8* dest, 
              const c8* src, 
              u32 destLength)
{
    u32 copy_length;

    ASSERT (dest);
    ASSERT (src);
    ASSERT (destLength > 0);

    if (destLength == 0)
        return 0;

    copy_length = MIN (destLength-1, sa_strlen (src));
    sa_memcpy (dest, src, copy_length);
    dest[copy_length] = 0; // always terminate string
    return copy_length;
}


u32
sa_wtext_copy (c16* dest, 
               const c16* src, 
               u32 destLength)
{
    u32 copy_length;

    ASSERT (dest);
    ASSERT (src);
    ASSERT (destLength > 0);

    if (destLength == 0)
        return 0;

    copy_length = MIN (destLength-1, sa_wtext_len (src, 0xFFFFFFFF));
    sa_memcpy (dest, src, copy_length * sizeof (c16));
    dest[copy_length] = 0; // always terminate string
    return copy_length;
}

/*!
@brief
    Concatenates src onto end of dest string. Always returns NULL
    terminated string.  It is truncated if destLength is too short.
@param
    dest - NULL terminated destination string
@param
    src - NULL terminated  

@return 
    Actual number of characters destination, excluding NULL.
*/
u32 
sa_text_cat (c8* dest, 
             const c8* src, 
             u32 destLength)
{
    u32 offset, len;

    ASSERT (dest);
    ASSERT (src);
    ASSERT (destLength > 0);

    if (destLength == 0)
        return 0;

    offset = sa_strlen (dest);
    len = sa_strlen (src);

    // Return immediately if the destination string if full.
    ASSERT (destLength > offset);    
    if (destLength <= offset) {
        dest[destLength-1] = 0;
        return destLength-1;
    }

    offset += sa_text_copy (&dest[offset], src, MIN (len, destLength -  offset));
    return offset;
}

/*!
@brief
    Copies the key and value strings into the destination string.  At most
    it copies destLength-1 characters, always maintaining
    a NULL terminated string.
@param
    destination - pointer to a destination string
@param
    destLength - maximum number of characters to copy
@param
    src - Source string to copy. Must be NULL terminated.
@return
   Actual number of characters copied, including NULL.
*/
u32 
sa_text_key_str_copy (c8* dest, 
                      const c8* key, 
                      const c8* value, 
                      u32 destLength)
{
    u32 i;

    ASSERT (dest);
    ASSERT (key);
    ASSERT (value);
    ASSERT (destLength > 0);

    i  = sa_text_copy (dest, key, destLength);
    i += sa_text_copy (&dest[i], value, destLength - i);
    return i + 1; // return length of key, value, and the final NULL
}


/*!
@brief
    Copies the key and decimal value strings into the destination string.  At most
    it copies destLength-1 characters, always maintaining
    a NULL terminated string.
@param
    destination - pointer to a destination string
@param
    destLength - maximum number of characters to copy
@param
    src - Source string to copy. Must be NULL terminated.
@return
   Actual number of characters copied, including the NULL.
*/
u32 
sa_text_key_int_copy (c8* dest, 
                      const c8* key, 
                      u32 value, 
                      u32 destLength)
{
    u32 i;

    ASSERT (dest);
    ASSERT (key);
    ASSERT (destLength > 0);

    i  = sa_text_copy (dest, key, destLength);
    ASSERT (destLength - i > 0);
    i += sa_text_int_to_dec_str (&dest[i], destLength - i, value, 0);
    return i + 1; // return length of key, value, and the final NULL
}

/*!
@brief 
    Search a text buffer for a string match and return a pointer to the
    character after the string match.  For example, searching the 
    input string "input=1" for the find string "input=" returns a 
    pointer to the '1' character.
                     
@param 
    input - String to search, not NULL terminated
@param 
    find - Null terminated string to find.
@param 
    inputLength - Length of string to search
                      
@return
    Pointer to the character _after_ the end of the found string the buffer.
    NULL if the search fails.
*/
const char*
sa_text_find_end (const char* input, 
                  const char* find, 
                  u32 inputLength)
{
    u32 compareLength;

    ASSERT (input);
    ASSERT (find);    
    
    if (inputLength == 0)
        return NULL;

    compareLength = sa_strlen (find);
    
    if ((compareLength == 0) || (compareLength > inputLength))
        return NULL;

    //
    // Match sure we never do a string comparison past the end 
    // of the entire input buffer.
    //
    inputLength = inputLength - compareLength + 1;

    while (inputLength--) { 
        
        //
        // Compare only first character of the find string until we get a match.
        //
        if (sa_text_toupper (*input) == sa_text_toupper (*find)) {
            
            //
            // Compare entire string
            //
            if (sa_text_starts_with (input, find)) {

                //
                // Found a match
                //
                return input + sa_strlen (find);
            }
        }

        input++;
    }

    return NULL;
}


/*!
@brief
    Converts a Unicode UTF16 string with size to an ANSI string.

@param  
    wStr - the Source Unicode string
@param  
    aStr - the ANSI destination string
@param  
    aLen - the size of the Output ANSI string chars
@param  
    wSiz - Input size, the maximum length WCHARs of the source UNICODE string,

@return - the size of the ANSI string including the a terminating NULL character

*/
i32 
sa_text_unicode_to_ansi_str (c16* wStr, 
                             c8*  aStr, 
                             i32   aLen, 
                             i32   wSiz)
{
    i32 i, x, wlen;
    i32 wchar;
    i32 wchar2;     // used on Unicode surrogates
    c8* pchar = 0;
 
    ASSERT(wStr);
    pchar = (c8*)aStr;
    
    for (i = 0, x = 0, wlen = wSiz; wlen != 0 && wStr[i]; i++)
    {
        wchar = wStr[i];

        if ((wchar >= 0xd800) && (wchar < 0xdc00) && (wlen > 1)) {
            //
            // Possible UTF16 surrogate pair found, process it
            // see unicode spec for details at www.unicode.org
            // if unpaired surrogate, pass it through unmodified.
            //
            wchar2 = wStr[i+1];
            if (wchar2 >= 0xdc00 && wchar2 < 0xe000) {
                //
                // surrogate pair found
                //
                i++;
                wlen--;     // account for the double characters wmi string.
                // 
                // Convert the two 32 bit unicode char into a single c16 
                // and process normally.
                //
                wchar = ((wchar - 0xd800) * 0x400 + (wchar2 - 0xdc00) + 0x10000);
            }
        }
        if (wchar < 0x80) {
            x += 1;
            if (x <= aLen) {
                *pchar++ = (u8) wchar;
            }
        } else if (wchar < 0x800) {
            x += 2;
            if (x <= aLen) {
                *pchar++ = (u8) (0xC0 | (wchar >> 6));
                *pchar++ = (u8) (0x80 | (wchar & 0x3F));
            }
        } else if (wchar < 0x10000) {
            x += 3;
            if (x <= aLen) {
                *pchar++ = (u8) (0xE0 | (wchar >> 12));
                *pchar++ = (u8) (0x80 | (wchar >> 6 & 0x3F));
                *pchar++ = (u8) (0x80 | (wchar & 0x3F));
            }
        } else if (wchar < 0x200000) {
            x += 4;
            if (x <= aLen) {
                *pchar++ = (u8) (0xF0 | (wchar >> 18));
                *pchar++ = (u8) (0x80 | (wchar >> 12 & 0x3F));
                *pchar++ = (u8) (0x80 | (wchar >>  6 & 0x3F));
                *pchar++ = (u8) (0x80 | (wchar & 0x3F));
            }
        }
        wlen--;
    }

    x++;    // Account for terminating NULL char
    if ( x <= aLen ) {
        *pchar = '\0';        // insure Null terminated string.
    }

    return x;
}

/*!
@brief
    Converts a WMI UNICODE string into an ANSI string.

@param  
    wstr - the Source Unicode string
@param  
    astr - the ANSI destination string
@param  
    alen - the size of the Output ANSI string chars

@return 
    BESTAUS

*/
SA_STATUS
sa_text_wmi_string_to_ansi (c16* wmiStr, 
                            PVOID  ansiStr, 
                            u32* ansiLen)
{                
    SA_STATUS    status = SA_SUCCESS; 
    i32        alen;           // ansi string len
    i32        wlen;           // wmi string len
    
    ASSERT(wmiStr);
    ASSERT(ansiStr);
    ASSERT(ansiLen);
    
    sa_memset (ansiStr, 0, *ansiLen);               // clear the string
    wlen = (i32)(wmiStr[0] / sizeof(c16));   // get number of WCHARs
    
    //
    // convert the Unicode string skipping the length field
    //
    alen = sa_text_unicode_to_ansi_str (&wmiStr[1], (c8*)ansiStr, (i32)*ansiLen, wlen);
        
    if (alen > (i32)(*ansiLen)) {
        status = SA_STATUS_NOT_OK;
    }             
    
    *ansiLen = alen;        // return length required.
    
    return(status);
}                        
/*
SA_STATUS
sa_text_ansi_string_to_wmi(c16* wmiStr,
                    u32  wmiStrLen,
                    PVOID  ansiStr,
                    u32  aStrLen)
{                  
    UNICODE_STRING uStr;
    ANSI_STRING    aStr; 
    SA_STATUS    beStat = SA_SUCCESS; 
    
    TRACE_IN(DL_WMI);
    
    TRACE(DL_WMI,"   wmi@=0x%p, wStrlen=0x%x, str@=0x%p, strlen=0x%x",
                wmiStr, wmiStrLen, ansiStr, aStrLen);
                
    if (KeGetCurrentIrql () <= PASSIVE_LEVEL)  { 
        TRACE(DL_ERR,"  Running at passive Level or less");
        uStr.Length = (u16)wmiStrLen;
        uStr.MaximumLength = (u16)wmiStrLen;
        uStr.Buffer = (PWSTR)&wmiStr[1];
        aStr.Length = (u16)aStrLen;
        aStr.MaximumLength = (u16)aStrLen;
        aStr.Buffer = (c8*)ansiStr;
        
        beStat = RtlAnsiStringToUnicodeString(&uStr, &aStr, FALSE);
        if (SA_SUCCESS == beStat) {
            u16 size = uStr.Length;
            TRACE(DL_WMI,"  Size of WMI buffer = %d", size);
            wmiStr[0] = size;       // save the size WMI string.
        }
    } else {
        TRACE(DL_ERR,"  Not at passive level");
        beStat = BE_NOT_OK;
    }
    
    TRACE_OUT(DL_WMI);
    return(beStat);
} 
*/

SA_STATUS
sa_text_ansi_to_unicode (c8*  ansiStr, 
                         c16* uStr, 
                         u32    uStrLen
                         )
{                  
    u32 i = 0;
    c8 *in = ansiStr;
    c16 *out = uStr;

    ASSERT (uStr);
    ASSERT (uStrLen >= 2);
    ASSERT (ansiStr);

    while (*in && i<uStrLen/2) {
        *out++ = (c16)*in++;
        i++;
    }

    if (i < uStrLen/2) {
        *out = SA_UNICODE_NULL;
        return SA_SUCCESS;
    } else if (uStrLen >= 2) {
        // Terminate if possible.
        uStr[uStrLen/2 - 1] = SA_UNICODE_NULL;
    }

    TRACE (DL_ERR, "Truncated unicode string: %s", ansiStr);
    return SA_STATUS_NOT_OK;
}    


/*!
@brief
    Returns the real length of a WMI string, not including any terminating NULLs.
@param 
    wmiStr - Counted WMI string where the first 2 byte are the length and the 
        remaining WCHARs form the string.  It may or may not be NULL terminated.
@return
    Actual length of the WMI string bytes, not including any NULLs.
*/
u32 
sa_text_wmi_string_length (c16* wmiStr) {

    u16 len = *((u16*)wmiStr);

    ASSERT ((len & 1) == 0); // length bytes cannot be odd number
    
    if (len < 2)
        return 0;    

    while (len>=2 && wmiStr[len/2]==0) {
        len -= 2;
    }

    return len;
}

/*!
@brief
    Tests if two WMI strings are equal, regardless of terminating NULLs.
@param 
    one - first WMI string to test
@param
    two - second WMI string to test
@return
    TRUE if they are equal
*/
boolean
sa_text_wmi_str_equal (c16* one, c16* two) {

    u32 len1, len2;

    ASSERT (one && two);

    len1 = sa_text_wmi_string_length (one);
    len2 = sa_text_wmi_string_length (two);

    if (len1 != len2)
        return FALSE;

    return (sa_memcmp (&one[1], &two[1], len1) == 0);
}


