/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
// display, modification, or disclosure of this program is strictly
// prohibited.
//
#include "runtime/sa_tiny_log.h"
#include "runtime/sa_text.h"

static u32 sa_tlog_sequence = 0;

STATIC FORCEINLINE void
sa_tlog_lock (PSA_TLOG log) {
    SA_TLOG_ASSERT (log);
    if (!(log->options & SA_TLOG_OPTION_NO_LOCKING)) {
        sa_acquire_spinlock (&log->lock, &log->irq);
    }
}

STATIC FORCEINLINE void
sa_tlog_unlock (PSA_TLOG log) {
    SA_TLOG_ASSERT (log);
    if (!(log->options & SA_TLOG_OPTION_NO_LOCKING)) {
        sa_release_spinlock (&log->lock, &log->irq);
    }
}

void sa_tlog_print_message (PSA_TLOG log, SA_OUTPUT *out, SA_TLOG_ENTRY *e);

PSA_TLOG sa_tlog_allocate (u32 num_msg, u32 options)
{
    SA_STATUS status;
    PVOID buffer = NULL;
    PSA_TLOG log = NULL;
    u32 bytes = (num_msg+1) * (sizeof (SA_TLOG_ENTRY));

    ASSERT (num_msg > 0);

    log = sa_malloc (
        sizeof (SA_TLOG), 
        SA_MALLOC_NORMAL, 
        SA_TAG ("logt"));
    if (!log) {
        status = SA_STATUS_NO_SYSTEM_RESOURCES;
        goto Error;
    }
   
    buffer = sa_malloc (bytes, SA_MALLOC_NORMAL, SA_TAG ("logb"));
    if (!buffer) {
        status = SA_STATUS_NO_SYSTEM_RESOURCES;
        goto Error;
    }    

    status = sa_tlog_create (log, buffer, bytes, options);
    
Error:

    if (status != SA_SUCCESS) {
        if (buffer)
            sa_free (buffer);

        if (log)
            sa_free (log);

        log = NULL;
    }

    return log;
}

void sa_tlog_free (PSA_TLOG log)
{   
    if (log) {
        PVOID buffer = log->buffer;

        SA_TLOG_ASSERT (log);
        sa_tlog_destroy (log);

        sa_free (buffer);
        sa_free (log);
    }
}

SA_STATUS sa_tlog_create (PSA_TLOG log, PVOID buffer, u32 buffer_bytes, u32 options)
{
    ASSERT (buffer);
    ASSERT (log);
    ASSERT (buffer_bytes > 0);

    sa_zero_mem (log, sizeof (SA_TLOG));
    sa_zero_mem (buffer, buffer_bytes);

    log->magic = SA_TLOG_MAGIC;
    log->buffer = buffer;
    log->buffer_bytes = buffer_bytes;
    log->options = options & (~SA_TLOG_OPTION_SALOG_TRACE);    
    log->num_entries = buffer_bytes / sizeof (SA_TLOG_ENTRY);
    log->prefix[0] = 0;

    ASSERT (log->num_entries > 0);
    sa_ring_create (&log->ring, log->num_entries, sizeof (SA_TLOG_ENTRY), buffer);

    if (!(options & SA_TLOG_OPTION_NO_LOCKING)) {
        sa_init_spinlock (&log->lock);
    }

    return SA_SUCCESS;
}

void sa_tlog_destroy (PSA_TLOG log) {
    if (log) {
        sa_zero_mem (log, sizeof (SA_TLOG));
    }
}

void sa_tlog_clear (PSA_TLOG log) {
    if (log) {
        SA_TLOG_ASSERT (log);

        sa_tlog_lock (log);
        sa_ring_clear (&log->ring);
        sa_tlog_unlock (log);
    }
}


void sa_tlog_assign_names (PSA_TLOG log, const char* names[], u32 max_names)
{
    if (log) {
        SA_TLOG_ASSERT (log);
        log->names = names;
        log->max_names = max_names;
    }
}

void sa_tlog_print_message (PSA_TLOG log, SA_OUTPUT *out, SA_TLOG_ENTRY *e)
{
//<no_inbox>
    c8 buffer[100];    
    u32 len;       
    u32 num_args = 0;    
    u32 i;

    if (log->names && 
        e->code < log->max_names && 
        log->names[e->code]) 
    {
        const char *format = log->names[e->code];

        len = (u32)strlen (format);
        for (i=0; i<len; i++) {
            if (format[i] == '%') {
                num_args++;
            }
        }

        switch (num_args)
        {        
        case 0: 
            len = SA_SNPRINTF_S (buffer, sizeof (buffer), sizeof (buffer)-1, format);
            break;
        case 1:
            len = SA_SNPRINTF_S (buffer, sizeof (buffer), sizeof (buffer)-1, format, e->args[0]);
            break;
        case 2:
             len = SA_SNPRINTF_S (buffer, 
                                  sizeof (buffer),
                                  sizeof (buffer)-1, 
                                  format, 
                                  (u32) (e->args[0] & 0xFFFFFFFF), // 32-bit
                                  (u32) (e->args[0]>>32));         // 32-bit
            break;
        case 3:
            len = SA_SNPRINTF_S (buffer, 
                                 sizeof (buffer),
                                 sizeof (buffer)-1, 
                                 format, 
                                 (u32) (e->args[0] & 0xFFFF),         // 16-bit
                                 (u32) ((e->args[0]>>16) & 0xFFFF),   // 16-bit
                                 (u32) (e->args[0]>>32)               // 32-bit
                                 );
            break;

        case 4:
            len = SA_SNPRINTF_S (buffer, 
                                 sizeof (buffer),
                                 sizeof (buffer)-1, 
                                 format, 
                                 (u32) ((e->args[0]>>0)  & 0xFFFF),   // 16-bit
                                 (u32) ((e->args[0]>>16) & 0xFFFF),   // 16-bit
                                 (u32) ((e->args[0]>>32) & 0xFFFF),   // 16-bit
                                 (u32) ((e->args[0]>>48) & 0xFFFF)    // 16-bit
                                 );
            break;

        case 7: // u16 + MAC
            {
                SA_MAC64 m64;
                m64.quad = e->args[0];
                len = SA_SNPRINTF_S (buffer, 
                                     sizeof (buffer),
                                     sizeof (buffer)-1, 
                                     format, 
                                     (u32) ((e->args[0] >> 48) & 0xFFFF), // 16-bit
                                     SA_MAC_ARGS (m64.mac)                // mac
                                     );
            }
            break;

        default:
            len = SA_SNPRINTF_S (buffer, sizeof (buffer), sizeof (buffer)-1, "Unsupported # args. code:0x%08llx format:%s arg:%llx", (u64)e->code, format, (u64)e->args[0]);
            break;
        }


    } else {
        len = SA_SNPRINTF_S (buffer, sizeof (buffer), sizeof (buffer)-1, "Unknown code: 0x%08x", e->code);
    }

    buffer[sizeof(buffer) - 1] = 0;
    
    sa_output_printf (
        out,
        "%s%s\n", 
        log->prefix,
        buffer
        );
//</no_inbox>
}

void sa_tlog_print (PSA_TLOG log, SA_OUTPUT *out)
{
    u32 index;    

    SA_TLOG_ASSERT (log);
    sa_tlog_lock (log);

    if (sa_ring_num_pending (&log->ring) > 0) {

        // Index to print first.
        index = sa_ring_consumer_index (&log->ring);

        for (; 
             index != sa_ring_producer_index (&log->ring);
             index = sa_addc (index, 1, sa_ring_num (&log->ring)))
        {
            PSA_TLOG_ENTRY e = sa_ring_item (&log->ring, index);

            sa_tlog_print_message (log, out, e);
        }
    }

    sa_output_printf (out, 
                      "TinyLog contains %d messages. \n", 
                      sa_ring_num_pending (&log->ring));

    sa_tlog_unlock (log);
}

void
sa_tlog_test (void)
{
    PSA_TLOG log;

    const char* names[] = {
        "This is message zero.",
        "This is message two with code %p."
    };
    
    log = sa_tlog_allocate (3, SA_TLOG_OPTION_NONE);
    if (log) {

        sa_tlog_assign_names (log, names, 2);

        sa_tlog_trace64 (log, 0, (0xdead0));
        sa_tlog_trace64 (log, 1, (0xdead1));
        sa_tlog_trace64 (log, 2, (0xdead2)); // unknown code
        
        sa_tlog_print (log, sa_stdout); // 3 entries will print

        sa_tlog_trace64 (log, 3, (0xdead3)); // unknown code
        sa_tlog_print (log, sa_stdout);  // 3 entries will print (code 0 will disappear)

        sa_tlog_clear (log);  

        sa_tlog_print (log, sa_stdout);  // 0 entries will print

        sa_tlog_free (log);
    }
}

void 
sa_tlog_set_prefix (
    PSA_TLOG tlog, 
    char *prefix)
{
    if (tlog) {
        sa_text_copy (tlog->prefix, prefix, sizeof (tlog->prefix));
    }
}

void 
_sa_tlog_trace (PSA_TLOG log, u32 code, u64 arg0) {
    SA_TLOG_ENTRY *entry;

    if (!log) {
        return;
    }

    SA_TLOG_ASSERT (log);
    sa_tlog_lock (log);

    if (sa_ring_num_empty (&log->ring) == 0) {
        sa_ring_consume (&log->ring);
    }

    ASSERT (sa_ring_num_empty (&log->ring) > 0);

    entry = (SA_TLOG_ENTRY*)sa_ring_producer_ptr (&log->ring);
    ASSERT (entry);

    entry->code = code;
    entry->sequence = sa_atomic_increment (&sa_tlog_sequence);
    entry->args[0] = arg0;
    
    if (log->options & SA_TLOG_OPTION_DEBUG_TRACE) {
        sa_tlog_print_message (log, sa_stdtrace, entry);
    } 
    if (log->options & SA_TLOG_OPTION_SALOG_TRACE) {
        sa_tlog_print_message (log, &log->salog_output, entry);
    }

    sa_ring_produce (&log->ring);

    sa_tlog_unlock (log);
}


void 
sa_tlog_assign_log (PSA_TLOG tlog, SA_LOG *log, u32 category, u32 level) {

    if (!tlog) {
        return;
    }

    // nop for inbox driver <no_inbox>    
    SA_TLOG_ASSERT (tlog);
    sa_tlog_lock (tlog);
    tlog->options |= SA_TLOG_OPTION_SALOG_TRACE;
    sa_output_salog_create (&tlog->salog_output, 
                            log,
                            category,
                            level);
    sa_tlog_unlock (tlog);
    // </no_inbox>
}

