/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file
   toedq_ll.c

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/toedq_ll.c,v 1.20 2006/04/17 00:49:17 menz Exp $
*/

#include "pch.h"

/*!

@brief
    This routine creates a TOE default receive ring.          

@param
    function_object      - handle to a function object                           
@param
    recv_queue_va         - base VA for a the TOE default receive ring
@param
    ring_size_power_of_two  - number of entries in the ring (power of two). 64K = 0 
@param
    pp_toe_default_queue   - pointer to an opqaue handle where a TOE default receive 
                          object is returned.                          

@return
    BE_SUCCESS if successfull, , otherwise a useful BESTATUS error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    this function allocates a BE_ISCSI_CONNECTION_OBJECT object.

*/
BESTATUS 
be_toe_dq_create (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PSA_SGL                         sgl, 
    IN      u32                             length, 
    IN      u32                             default_buffer_length,
    IN      u32                             ulp_id,
    OUT     PBE_TOE_DQ_OBJECT               toe_dq)
{

    IOCTL_TOE_DEFQ_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    FUNCTION_ASSERT (function_object);    
    ASSERT (be_function_is_networking (function_object));
    ASSERT (sgl);
    ASSERT (toe_dq);

    be_lock_wrb_post (function_object);
    
    if (!sa_is_power2 (default_buffer_length)) {
        TRACE (DL_ERR, 
               "bad toe_dq default_buffer_length:%d  It must be power of 2.", 
               default_buffer_length);
        status = BE_NOT_OK;
        goto Error;
    }
    
    SA_ZERO_MEM (toe_dq);
    toe_dq->magic                 = BE_TOE_DQ_MAGIC;
    toe_dq->parent_function       = function_object;
    toe_dq->sgl                   = sgl;
    toe_dq->default_buffer_length = default_buffer_length;
    toe_dq->id                    = 0xFFFFFFFF;    

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_DEFQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16)SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    ioctl->params.request.context.ring_size         = be_ring_length_to_encoding (length, sizeof(TOE_DEFBUFFER));
    ioctl->params.request.context.pci_func          = be_function_get_function_number (function_object);
    ioctl->params.request.context.pdid              = be_function_get_pd_number (function_object);
    ioctl->params.request.context.pdid_valid        = TRUE;    

    if (function_object->fw_cfg.ulp[0].ulp_mode & TOE_MODE) {
        ioctl->params.request.ulp_num = 0;
    } else {
        ASSERT (function_object->fw_cfg.ulp[1].ulp_mode & TOE_MODE);
        ioctl->params.request.ulp_num = 1;
    }

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, ioctl->params.request.pages, SA_NUMBER_OF (ioctl->params.request.pages));

    // SA_TRACE_BUFFER (DL_WARN, ioctl, sizeof (*ioctl));

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create toe defq failed.");
        goto Error;
    }

    // Save the returned id.
    toe_dq->id = ioctl->params.response.id;

    // Add object to the function object.
    _be_function_add_toe_dq (function_object, toe_dq);

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

/*!

@brief
    this routine destroys a TOE default receive queue     

@param    
    toe_dq - TOE default receive queue handle returned from toe_default_receive_object_create

@return   
    returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note
    this function frees resourcs allocated by toe_default_receive_object_create.
    IRQL: < DISPATCH_LEVEL

*/

BESTATUS 
be_toe_dq_destroy (
    PBE_TOE_DQ_OBJECT toe_dq)
{
    PBE_FUNCTION_OBJECT function_object;
    IOCTL_TOE_DEFQ_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    u32 id = toe_dq->id;
    
    TOE_DQ_ASSERT (toe_dq);
    
    function_object = toe_dq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy TOE DQ ring id:%d", id);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in destroy TOE DQ ring.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_DEFQ_DESTROY);

    ioctl->params.request.id = id;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 ioctl);
    
    ASSERT (status != BE_PENDING);
    if (status != BE_PENDING) {

        // Remove from function object
        _be_function_remove_toe_dq (function_object, toe_dq);

        // Zero the tracking object
        SA_ZERO_MEM (toe_dq);

    }

Error:
    ASSERT (0 == status || function_object->ue);
    be_unlock_wrb_post (function_object);
    return status;
}

/*!
@brief
    this routine returns the TOE default queue ring identifier. 
@param    
    toe_dq - TOE default receive queue handle returned from toe_default_receive_object_create
@return   
    returns the ring ID. 
*/
u32
be_toe_dq_get_id (
    IN PBE_TOE_DQ_OBJECT toe_dq)
{
    TOE_DQ_ASSERT (toe_dq);
    return toe_dq->id;
}

u32
be_toe_dq_get_buffer_length (
    IN PBE_TOE_DQ_OBJECT toe_dq)
{
    TOE_DQ_ASSERT (toe_dq);
    return toe_dq->default_buffer_length;
}


