/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        acitmain.c

    Purpose:

        This file implements all functions that are exported by the ACIT 
        library for internal use. iSCSI and Management modules that are 
        internal to ACIT library may use these exported functions.
        

--*/


#include "acitpch.h"

ACIT_UEXACT32
ap_flag_busy_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           reason)
{
    /*
     * This function can be called for many reasons.
     * If Unavailable IOHandle, Unavailable hw resources, Link-Down etc.
     */

    pcontroller->stats.controller_busy++;

    ACIT_BIT_SET(pcontroller->busy, reason);

    /*
     * Notify the OSM that the controller is currently busy.
     */
    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_CONTROLLER,
                    ACIT_EVENT_CONTROLLER_BUSY,
                    0);
    return 0;
}



ACIT_UEXACT32
ap_flag_ready_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           reason)
{
    /*
     * This function can be called for many reasons.
     * If Available IOHandle, Available hw resources, Link-Up etc.
     */

    ACIT_BIT_CLR(pcontroller->busy, reason);

    pcontroller->stats.controller_ready++;

    /*
     * Check if the controller is no longer busy, notify the OSM.
     */
    if(!pcontroller->busy)
    {
        if(pcontroller->counters.queued_ios)
        {
            /*
             * IOs are currently queued in our list for delivery, 
             * set flag to flush them
             */
            pcontroller->state.flush_queued_ios = ACIT_TRUE;
        }

        if(pcontroller->counters.queued_tmfs &&
           ap_ready_for_tmf(pcontroller, ACIT_NULL) == ACIT_STATUS_SUCCESS)
        {
            pcontroller->state.flush_queued_tmfs = ACIT_TRUE;
        }

        OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                        ACIT_EVENT_CLASS_CONTROLLER,
                        ACIT_EVENT_CONTROLLER_READY,
                        0);
    }
    return 0;
}


ACIT_UEXACT32
ap_flag_busy_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           reason)
{
    ACIT_UEXACT32           index;

    /*
     * This function can be called for many reasons.
     * If iSCSI detects Command Window closed, Target missing, 
     * session recovery in progress etc.
     */

    pcontroller->stats.target_busy++;

    ACIT_BIT_SET(ptarget->busy, reason);

    index = ptarget->target_table_index;

    /* 
     * Notify OSM about target busy
     */
    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_TARGET,
                    ACIT_EVENT_TARGET_BUSY,
                    index);

    //
    // This cannot be allowed since all warnings
    // will be logged. Have to be commented out
    //
//    ap_log_warning(pcontroller,
//                  ACIT_SUBMODULE_ACITGENERIC,
//                  ACIT_PHASE_IO,
//                  ACIT_FLAG_TARGET_BUSY_STATUS,
//                  reason);
//
    return 0;
}


ACIT_UEXACT32
ap_flag_ready_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           reason)
{
    ACIT_UEXACT32           index;

    /*
     * This function can be called for many reasons.
     * If iSCSI detects target has become available after the Command Window
     * opens again or if session recover completes successfully.
     */

    pcontroller->stats.target_ready++;

    ACIT_BIT_CLR(ptarget->busy, reason);

    index = ptarget->target_table_index;

    if(!ptarget->busy)
    {
        if(pcontroller->counters.queued_ios)
        {
            /*
             * IOs are currently queued in our list for delivery, 
             * set flag to flush them
             */
            pcontroller->state.flush_queued_ios = ACIT_TRUE;
        }

        if(pcontroller->counters.queued_tmfs &&
           ap_ready_for_tmf(pcontroller, ACIT_NULL) == ACIT_STATUS_SUCCESS)
        {
            pcontroller->state.flush_queued_tmfs = ACIT_TRUE;
        }

        OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                        ACIT_EVENT_CLASS_TARGET,
                        ACIT_EVENT_TARGET_READY,
                        index);
    }

    return 0;
}


ACIT_UEXACT32
ap_flag_busy_lun(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           reason)
{
    /*
     * This function can be called for many reasons.
     * If a SCSI status of queue full or busy is returned by the LU.
     */

    plun->busy = reason;
    return 0;
}


ACIT_UEXACT32
ap_flag_ready_lun(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           reason)
{
    /*
     * This function can be called for many reasons.
     * If a SCSI status of non-queuefull or non-busy is returned by the LU.
     */

    plun->busy = 0;
    return 0;
}


ACIT_UEXACT32
ap_set_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           state)
{

    if (ACIT_IS_BIT_CLR (ptarget->event_flag, state)) {

        ACIT_BIT_SET(ptarget->event_flag, state);

        /*
        * if target is not already on the event link, add it now
        */
        if(IS_LINK_INVALID(&ptarget->event_link))
        {
            LIST_ADDTO_TAIL(&ptarget->event_link, &pcontroller->target_event_list);
        }
    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_clear_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           state)
{
    if (ACIT_IS_BIT_SET (ptarget->event_flag, state)) {

        ACIT_BIT_CLR(ptarget->event_flag, state);
    }

    /*
    * check if there are no more pending states
    */
    if(!ptarget->event_flag)
    {
        LIST_REMOVE_ELEMENT(&ptarget->event_link);
        LIST_INIT_ELEMENT(&ptarget->event_link);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_update_lun_tmf_stats(
    PACIT_LUN_WS        plun,
    ACIT_UEXACT32       tmf_status)
{
    ACIT_UEXACT32       idx;

    if(tmf_status != ACIT_TMF_STATUS_SUCCESS)
    {
        idx = tmf_status - ACIT_TMF_STATUS_ERROR + 1;
    }
    else
    {
        idx = tmf_status;
    }

    plun->eh_stats.tmf_err_stats[idx]++;
}

ACIT_VOID
ap_update_tgt_tmf_stats(
    PACIT_TARGET_WS     ptarget,
    ACIT_UEXACT32       tmf_status)
{
    ACIT_UEXACT32       idx;

    if(tmf_status != ACIT_TMF_STATUS_SUCCESS)
    {
        idx = tmf_status - ACIT_TMF_STATUS_ERROR + 1;
    }
    else
    {
        idx = tmf_status;
    }

    ptarget->eh_stats.tmf_err_stats[idx]++;
}

ACIT_VOID
ap_update_ctl_tmf_stats(
    PACIT_CONTROLLER_WS pcontroller,
    ACIT_UEXACT32       tmf_status)
{
    ACIT_UEXACT32       idx;

    if(tmf_status != ACIT_TMF_STATUS_SUCCESS)
    {
        idx = tmf_status - ACIT_TMF_STATUS_ERROR + 1;
    }
    else
    {
        idx = tmf_status;
    }

    pcontroller->eh_stats.tmf_err_stats[idx]++;
}

ACIT_UEXACT32
ap_reinit_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle)
{
    ACIT_UEXACT16           cid, connection_handle, cq_id, def_hdr_id, def_data_id;
    PHYS_ADDR               template_address = {{{0,0}}};
    ACIT_UEXACT32           status;

    //
    // Reset the login pending state in target table
    //
    pcontroller->ptarget_table->target_entry[ptarget->target_table_index].is_login_pending = 0;
    
    ptarget->flags = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Calling reinit for target 0x%p session 0x%x\n", ptarget, session_handle));

    iscsi_reinit_target_ws(pcontroller, 
                           ptarget,
                           &connection_handle,
                           &cid,
                           &template_address);

    cq_id = (ACIT_UEXACT16)HWI_GET_CQ_ID(pcontroller, 0);
    def_hdr_id = (ACIT_UEXACT16)HWI_GET_DEF_HDRQ_ID(pcontroller);
    def_data_id = (ACIT_UEXACT16)HWI_GET_DEF_BUFQ_ID(pcontroller);

    status = mgmt_init_target_ws(pcontroller, 
                                 ptarget, 
                                 session_handle,
                                 connection_handle,
                                 cid,
                                 cq_id,
                                 def_hdr_id,
                                 def_data_id,
                                 &template_address,
                                 ACIT_FALSE);   // asynchronous

    if(status != ACIT_STATUS_SUCCESS)
    {
        if(status == ACIT_STATUS_INSUFFICIENT_RESOURCE)
        {
            // this will be cleared when ETO is stopped..
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Insufficient resource!! retry init target 0x%p session 0x%x\n", ptarget, session_handle));
            ap_set_target_event(pcontroller, ptarget, EVENT_TGT_WS_INIT_PENDING);
        }

        return status;
    }

    return ACIT_STATUS_SUCCESS;
}

INLINE
ACIT_UEXACT32
ap_decide_io_disposition(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_WS             pio)
{
//    ACIT_UEXACT32           port_num;
//    port_num = ISCSI_GET_TARGET_PORT(ptarget);

    if(IS_CONTROLLER_LDTO_EXPIRED(pcontroller, 0))
    {
        return ACIT_STATUS_LINK_DOWN_TIMEOUT; 
    }

    if(IS_TARGET_ETO_EXPIRED(ptarget))
    {
        return ACIT_STATUS_TARGET_TIMEOUT;
    }

    if(pio->io_retries >= pcontroller->params.io_retries_b4_err)
    {
        return ACIT_STATUS_ERROR;
    }

//    if(pio->available_ticks >= GET_TARGET_EXEC_TICKS(ptarget, 0) +
//                              pcontroller->state.invalidation_ticks)

    if(pio->available_ticks > (ACIT_SEXACT32)(GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget) +
                              pcontroller->state.invalidation_ticks))
    {
        OSM_ASSERT(pio->available_ticks > pcontroller->state.invalidation_ticks);
        return ACIT_STATUS_SUCCESS;
    }

    return ACIT_STATUS_INSUFFICIENT_TIMEOUT; 
}

ACIT_UEXACT32
ap_process_io_tick(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_IO_WS             pio;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_DC_LIST            *pio_link, *tmp, *pioh_link;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           status = 0, invs_allowed;
    ACIT_DC_LIST            timeout_list;
#if ACIT_CFG_DEBUG_BUILD
    ACIT_UEXACT32           count = 0, ios_skipped = 0;
#endif

    LIST_INIT_HEADER(&timeout_list);

    invs_allowed = (pcontroller->params.ioctls_per_ctrl - (pcontroller->counters.pending_ioctls + 1)) * 
                        BE_MAX_ITTS_PER_INVALIDATE;

    if(!invs_allowed)
    {
        pcontroller->state.defer_invalidations = ACIT_TRUE;
    }

    /*
     * Lets start with counting I/O ticks and see if any one of 
     * them has expired.
     */
    /* First traverse the list of ios which are submitted to the hardware */

    for(pioh_link = pcontroller->io_handle_busy_list.next, tmp = pioh_link->next;
        pioh_link != &pcontroller->io_handle_busy_list;
        pioh_link = tmp, tmp = pioh_link->next) {

        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                             controller_link,
                             pioh_link);

        OSM_ASSERT(pio_handle);
        OSM_ASSERT(pio_handle->psgl_handle);

        ptarget = pio_handle->ptarget;

#if ACIT_CFG_DEBUG_BUILD
        OSM_ASSERT(count < pcontroller->counters.pending_ios);
        count ++;
#endif

        pio = pio_handle->pio;
        OSM_ASSERT(pio);
        OSM_ASSERT(pio->available_ticks);
        //OSM_ASSERT(pio->internal_available_ticks);

         /* Decrement the available ticks for this I/O */
        if(pio->available_ticks > 0)
        {
            pio->available_ticks --;
        }

        /* Decrement the ticks for internal IO timeout */
        if(pio->internal_available_ticks > 0)
        {
            pio->internal_available_ticks --;
        }

        if(ACIT_IS_BIT_SET(ptarget->flags, TGT_FLAG_SKIP_TIMEOUT_TICKS_COUNTDOWN))
        {
#if ACIT_CFG_DEBUG_BUILD
            ios_skipped++;
#endif
            continue;
        }

        /* Internal IO Timeout condition */
        if(pio->internal_available_ticks <= 0)
        {
            if(!pcontroller->state.defer_invalidations)
            {
                /* Internal timeout, invalidate the IO */
                LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
                LIST_ADDTO_TAIL(&pio_handle->controller_link, &timeout_list);
                pcontroller->eh_stats.io_timeouts++;
                invs_allowed--;
                if(!invs_allowed)
                {
                    pcontroller->state.defer_invalidations = ACIT_TRUE;
                    pcontroller->eh_stats.invs_deferred++;
                    pcontroller->eh_stats.invs_deferred_no_ioctls++;
                }
            }
            else
            {
                // set internal_available_ticks to 1 so that the next
                // timer tick will time out this IO again
                pio->internal_available_ticks = 1;
                pcontroller->eh_stats.invs_deferred++;
                pcontroller->eh_stats.invs_deferred_no_ioctls++;
            }
        }
    }

    if(IS_LIST_NOT_EMPTY(&timeout_list))
    {
#if ACIT_CFG_DEBUG_BUILD
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Skipped IOS %d\n", ios_skipped));
#endif

        /* Directly call the invalidate function that takes in a list */
        ap_invalidate_cmd_list(pcontroller, 
                               &timeout_list, 
                               ACIT_INVH_TYPE_IO,
                               CONTROLLER_QUEUE);

        //
        // if IOs still persist in timeout_list then invalidations weren't complete. Put
        // them back to the IO handle busy list for later processing
        //
        if(IS_LIST_NOT_EMPTY(&timeout_list))
        {
            OSM_ASSERT(0);
            LIST_MOVETO_TAIL(&pcontroller->io_handle_busy_list, &timeout_list);
        }
    }
	    
    /* Traverse the ios which are not submitted to the hardware */
    for(pio_link = pcontroller->io_queued_list.next, tmp = pio_link->next; 
        pio_link != &pcontroller->io_queued_list; 
        pio_link = tmp, tmp= pio_link->next) {

        pio = STRUCT_BASE(ACIT_IO_WS,
                          controller_link,
                          pio_link);

        ptarget = GET_TARGET_WS(pio->pio_request->ptarget_key);

        /*
         * Decrement the available ticks for this I/O
         */
        if(pio->available_ticks > 0)
        {
            pio->available_ticks --;
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("IOWS 0x%p Ticks - Original %d Available %d Internal %d\n"),
                pio, pio->original_ticks, pio->available_ticks, pio->internal_available_ticks));

            OSM_ASSERT(pio->available_ticks > 0);
        }

        /* Just to be safe, return the IO slightly ahead of more time */
        //if(pio->available_ticks > (ACIT_SEXACT32)(pcontroller->state.invalidation_ticks))
        //    continue;

        //
        // Bug 25401 : Do not hold on to the IO until command timer reaches
        // invalidation ticks. Return when the minimum command time
        // threshold is reached. We get LUN reset around the time of command timeouts
        // and this approach should work on all OS
        //
        if(pio->available_ticks > (ACIT_SEXACT32)(GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget) +
                                  pcontroller->state.invalidation_ticks))
        {
            continue;
        }

        switch (ap_decide_io_disposition(pcontroller, 
                                         ptarget,
                                         pio))
        {
            case ACIT_STATUS_INSUFFICIENT_TIMEOUT:

                status = ACIT_IO_STATUS_IO_TIMEOUT;
                break;

            case ACIT_STATUS_LINK_DOWN_TIMEOUT:
        
                status = ACIT_IO_STATUS_LINK_DOWN_TIMEOUT;
                break;

            case ACIT_STATUS_TARGET_TIMEOUT:

                status = ACIT_IO_STATUS_TARGET_TIMEOUT;
                break;

            case ACIT_STATUS_ERROR:

                status = ACIT_IO_STATUS_ERROR;
                break;

            default:

                // we should not be here. We called io_disposition routine for
                // a timeout and the disposition routine returned success
                status = ACIT_STATUS_INSUFFICIENT_TIMEOUT;
                OSM_ASSERT(0);
                break;
        }

        ap_return_io(pcontroller, pio, status, ACIT_TRUE);
    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_process_tmf_tick(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_TARGET_WS         ptarget = ACIT_NULL;
    PACIT_LUN_WS            plun = ACIT_NULL;
    PACIT_DC_LIST           ptmf_link;
    PACIT_TARGET_ENTRY      ptarget_entry;
    PACIT_TMF_HANDLE        ptmf_handle;
    ACIT_UEXACT32           index;
    PACIT_TMF_WS            ptmf_ws;

    if(pcontroller->counters.pending_tmfs)
    {
        ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

        for(index = 0; 
            index < pcontroller->params.targets_per_controller;
            index ++)
        {
            if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                                ACIT_TFLAG_IN_USE))
            {
                ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

                if(!ptarget)
                    continue;

                if(ACIT_IS_BIT_SET(ptarget->flags, TGT_FLAG_SKIP_TIMEOUT_TICKS_COUNTDOWN))
                    continue;

                // check for pending TMFs with hardware
                ptmf_link = ptarget->tmf_handle_busy_list.next;

                while(ptmf_link != &ptarget->tmf_handle_busy_list)
                {
                    ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                                target_link,
                                                ptmf_link);

                    OSM_ASSERT(ptmf_handle);
                    ptmf_handle->available_ticks --;

                    ptmf_link = ptmf_link->next;

                    if(!ptmf_handle->available_ticks)
                    {
                        // IOSince and TMFSince
                        ACIT_UEXACT32 ios = 0xffffffff, tmfs = 0xffffffff;

                        pcontroller->eh_stats.tmf_timeouts++;
                        ptarget->eh_stats.tmf_timeouts++;

                        if(ptmf_handle->tmf_function == ACIT_TMF_ABORT_TASK)
                        {
                            pcontroller->eh_stats.tmf_abort_timeouts++;
                            ptarget->eh_stats.tmf_abort_timeouts++;

                            ios = ptmf_handle->pio_handle->starting_tick_count;
                        }
                        else if(ptmf_handle->tmf_function == ACIT_TMF_LUN_RESET)
                        {
                            pcontroller->eh_stats.tmf_lunrst_timeouts++;
                            ptarget->eh_stats.tmf_lunrst_timeouts++;
                        }
                        else 
                        {
                            pcontroller->eh_stats.tmf_tgtrst_timeouts++;
                            ptarget->eh_stats.tmf_tgtrst_timeouts++;
                        }

                        tmfs = ptmf_handle->starting_tick_count;

                        ptmf_handle->flags = FLAG_TMFH_TMF_TIMEDOUT;
                        //
                        // Set the TMF_TIMEOUT flag. This will be used in invalidation
                        // completion to trigger session recovery
                        //
                        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_TIMEOUT);

                        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF in target 0x%p timeout, will trigger session recovery...\n"),
                                                                ptmf_handle->ptarget));

						//
						// When TMF timesout then ACIT initiates recovery and in this case we will have to update the
						// statistics counters. If session recovery was already in progress then connection already
						// timed out and we would have updated statistics counter thus no need to update them here.
						//
						if (!(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY))) {
							PISCSI_TARGET_WS        piscsi_target;
						    ACIT_UEXACT32           i;

							piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
							pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_CXNTIMEOUT_ERRORS;
							pcontroller->eh_stats.ssn_cxn_timeout_errors++;
							pcontroller->eh_stats.ssn_failure_count++;
							ptarget->eh_stats.ssn_cxn_timeout_errors++;
							for(i=0; i<MAX_CXNS_PER_SESSION; i++)
								if(piscsi_target->pcxn_list[i])
									piscsi_target->pcxn_list[i]->timeout_errors++;

                            // we can assume the recovery reason is a TMF timeout

                            // This will be used when setting the eto timer
                            ptarget->recovery_reason.tmf_timeout = 1;

                            // The io_resp_missing_since will give the amount
                            // of time the target is found to be not responding
                            
                            // Between the IO and TMF, use the one that has a starting_tick
                            // which is earlier
                            ptarget->recovery_reason.io_resp_missing_since = ACIT_MIN(ios, tmfs);

                            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("0x%p TMFTO IOResponseMissingSince %d (Current %d)\n"),
                                                                ptarget,
                                                                ptarget->recovery_reason.io_resp_missing_since,
                                                                CURRENT_TICK(pcontroller)));
						}

						iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
                        break;
                    }
                }
            }
        }
    }

    // check external TMFs sitting in internal queue
    if(pcontroller->counters.queued_tmfs)
    {
        ptmf_link = pcontroller->tmf_queued_list.next;

        while(ptmf_link != &pcontroller->tmf_queued_list)
        {
            ACIT_UEXACT32 tmf_function = 0, tmfs = 0;

            ptmf_ws = STRUCT_BASE(ACIT_TMF_WS, controller_link, ptmf_link);
            OSM_ASSERT(ptmf_ws);

            ptmf_link = ptmf_link->next;
            ptmf_ws->available_ticks--;

            //if(!(ptmf_ws->available_ticks))
            // Bug 25401 : Return the queued TMFs slightly ahead of time
            if(ptmf_ws->available_ticks <= (ACIT_UEXACT32)(pcontroller->state.invalidation_ticks * 2))
            {
                tmf_function = ptmf_ws->ptmf_request->tmf_function;
                tmfs = ptmf_ws->original_ticks;

                ptarget = GET_TARGET_WS(ptmf_ws->ptmf_request->ptarget_key);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF in target 0x%p timeout waiting in queue...\n"),
                                                        ptarget));

                if(tmf_function == ACIT_TMF_LUN_RESET)
                {
                    OSM_ASSERT(ptmf_ws->ptmf_request->plun_key);
                    plun = GET_LUN_WS(ptmf_ws->ptmf_request->plun_key);
		    
                    ap_return_queued_lun_tmfs(pcontroller, 
                                              ptarget, 
                                              plun, 
                                              ACIT_TMF_STATUS_SUCCESS,
                                              ACIT_FALSE);
                }
                else //if((tmf_function == ACIT_TMF_TARGET_WARM_RESET) ||
                     //   (tmf_function == ACIT_TMF_TARGET_COLD_RESET))
                {
                    ap_return_queued_target_tmfs(pcontroller,
                                                 ptarget,
                                                 ACIT_TMF_STATUS_SUCCESS,
                                                 ACIT_FALSE);
                }
                /*
                else
                {
                    PACIT_IO_WS pio;
                    PACIT_IO_HANDLE pio_handle;
                    plun = GET_LUN_WS(ptmf_ws->ptmf_request->plun_key);
                    pio = GET_IO_WS(ptmf_ws->ptmf_request->pio_key);
                    OSM_ASSERT(pio);
                    OSM_ASSERT(pio->pio_handle);

                    ap_dequeue_tmf(pcontroller, ptarget, plun, pio_handle);
                    OSM_ASSERT(pio_handle->ptmf_ws);
                    ap_return_tmf(pcontroller, 
                                  ptmf_ws->ptmf_request,
                                  ACIT_TMF_STATUS_SUCCESS);

                    ap_free_io_resources(pcontroller, pio_handle);
                }
                */

                pcontroller->eh_stats.tmf_timeouts++;
                pcontroller->eh_stats.tmf_internal_timeouts++;

                if(tmf_function == ACIT_TMF_ABORT_TASK)
                {
                    pcontroller->eh_stats.tmf_abort_internal_timeouts++;
                }
                else if(tmf_function == ACIT_TMF_LUN_RESET)
                {
                    pcontroller->eh_stats.tmf_lunrst_internal_timeouts++;
                }
                else 
                {
                    pcontroller->eh_stats.tmf_tgtrst_internal_timeouts++;
                }

                if(ptarget->initialized)
                {
                    //
                    // since TMF timedout and there may be associated IO handles that are
                    // not freed, trigger session recovery to free those handles
                    //
            
                    //
                    // Set the TMF_TIMEOUT flag. This will be used in invalidation
                    // completion to trigger session recovery
                    //
                    ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_TIMEOUT);

                    ptarget->eh_stats.tmf_timeouts++;
                    ptarget->eh_stats.tmf_internal_timeouts++;

                   if(tmf_function == ACIT_TMF_ABORT_TASK)
                    {
                        ptarget->eh_stats.tmf_abort_internal_timeouts++;
                    }
                    else if(tmf_function == ACIT_TMF_LUN_RESET)
                    {
                        ptarget->eh_stats.tmf_lunrst_internal_timeouts++;
                    }
                    else 
                    {
                        ptarget->eh_stats.tmf_tgtrst_internal_timeouts++;
                    }

                    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF in target 0x%p timeout internally, will trigger session recovery...\n"),
                                                            ptarget));

					//
					// When TMF timesout then ACIT initiates recovery and in this case we will have to update the
					// statistics counters. If session recovery was already in progress then connection already
					// timed out and we would have updated statistics counter thus no need to update them here.
					//
					if (!(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY))) {
						PISCSI_TARGET_WS        piscsi_target;
					    ACIT_UEXACT32           i;

						piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
						pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_CXNTIMEOUT_ERRORS;
						pcontroller->eh_stats.ssn_cxn_timeout_errors++;
						pcontroller->eh_stats.ssn_failure_count++;
						ptarget->eh_stats.ssn_cxn_timeout_errors++;
						for(i=0; i<MAX_CXNS_PER_SESSION; i++)
							if(piscsi_target->pcxn_list[i])
								piscsi_target->pcxn_list[i]->timeout_errors++;

                        // we can assume the recovery reason is a TMF timeout

                        // This will be used when setting the eto timer
                        ptarget->recovery_reason.tmf_timeout = 1;

                        // The io_resp_missing_since will give the amount
                        // of time the target is found to be not responding
                        ptarget->recovery_reason.io_resp_missing_since = 
                            CURRENT_TICK(pcontroller) - tmfs;

                        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("0x%p TMFTO IOResponseMissingSince %d (Current %d)\n"),
                                                            ptarget,
                                                            ptarget->recovery_reason.io_resp_missing_since,
                                                            CURRENT_TICK(pcontroller)));
					}

                    iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
                }
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_process_logout_tick(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_LOGOUT_HANDLE     plogout_handle;
    ACIT_DC_LIST            *plogout_link;

    // Walk the logout_handle_busy_list
    if(IS_LIST_NOT_EMPTY(&pcontroller->logout_handle_busy_list))
    {
        plogout_link = pcontroller->logout_handle_busy_list.next;

        // Loop and travel this linked list.
        while(plogout_link != &pcontroller->logout_handle_busy_list)
        {
            plogout_handle = STRUCT_BASE(ACIT_LOGOUT_HANDLE,
                                         controller_link,
                                         plogout_link);

            plogout_link = plogout_link->next;

             // Decrement the available ticks
            plogout_handle->available_ticks --;

            if(!plogout_handle->available_ticks)
            {
                LIST_REMOVE_ELEMENT(&plogout_handle->controller_link);

                //
                // if logout timeouts, then directly jump to completion with a
                // timeout. This will trigger a connection invalidation with
                // RST bit set. We don't have to invalidate the logout handle 
                //as such
                ap_complete_logout(pcontroller, 
                                   plogout_handle, 
                                   ISCSI_PDU_LOGOUT_RESPONSE_FAIL, // not used if timeout
                                   ACIT_TRUE); // timeout = Yes
            }
        }
    }

    return ACIT_STATUS_SUCCESS;

}

ACIT_UEXACT32
ap_process_target_tick(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_DC_LIST           ptarget_link;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           async = 1;

    if(IS_LIST_NOT_EMPTY(&pcontroller->target_event_list))
    {
        ptarget_link = pcontroller->target_event_list.next;

        /*
         * Loop and travel this linked list.
         */
        while(ptarget_link != &pcontroller->target_event_list)
        {
            ptarget = STRUCT_BASE(ACIT_TARGET_WS,
                                  event_link,
                                  ptarget_link);

            ptarget_link = ptarget_link->next;

            // if interrupts are disabled, target events need to be processed
            // synchronously
            if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS) ||
                pcontroller->state.interrupt_enabled == 0)
            {
                async = 0;
            }

            ap_process_target_event(pcontroller, ptarget, async);
        }
    }

    return ACIT_STATUS_SUCCESS;

}

ACIT_UEXACT32
ap_complete_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PI_SOL_CQE              psol)
{
    PACIT_IO_REQUEST        pio_request;
    ACIT_UEXACT32           invalidated = ACIT_TRUE;

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_ap_complete_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_handle,
                        (pio_handle->pio)?(ACIT_UEXACT64)pio_handle->pio->pio_request:(ACIT_UEXACT64)0,
                        pcontroller->counters.pending_ios);

    //
    // Bug 27997/28601. We were returning the IO back which completed via CQE, but had
    // a pending abort. The abort could not be issued because session recovery was in
    // progress.

    // if commands are being invalidated or has completed invalidation or
    // has seen a CQ error the completions will not be provided to OSM.
    if(ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING) &&
        ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_INVALIDATE_COMPLETE) &&
        ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_CQ_ERROR) &&
        (pio_handle->ptmf_ws == ACIT_NULL))
    {
        invalidated = ACIT_FALSE;
    }

    iscsi_complete_io(pcontroller, 
                      pio_handle,
                      psol,
                      invalidated);

    if(!invalidated)
    {
        PACIT_TARGET_WS         ptarget;

        // IO WS is valid only if command is not invalidated yet
        pio_request = pio_handle->pio->pio_request;

        OSM_ASSERT(pio_request);

        if(pio_request->io_status == ACIT_STATUS_SUCCESS)
        {
            //
            // Linux x86 systems on some arch have trouble mapping VA in 
            // IO request structure. So this could be unreliable on those
            // systems. We've added a new API for those OSMs to set the
            // device_type field later via acit_update_lun_info
            //

            // Snoop into data buffer to determine the SCSI device type
            // only if this is an Inquiry command with EVPD of 0.
            if( (pio_request->cdb[0] == ACIT_SCSI_CMD_INQUIRY) &&
                (!(pio_request->cdb[1] & ACIT_SCSI_CMD_INQUIRY_EVPD_BIT)) )
            {
                if(pio_request->data_transfer_length && 
                   (pio_request->scsi_status == ACIT_SCSI_STATUS_GOOD) &&
                   (pio_request->pdata_buffer != ACIT_NULL))
                {
                    PACIT_LUN_WS plun;
                    plun = pio_handle->plun;
                    plun->dev_type = *((PACIT_UEXACT8)pio_request->pdata_buffer);
                    ap_update_lun_retry_param(pcontroller, plun);
                }
            }

            // Update ticks taken for this I/O in target workspace

            ptarget = pio_handle->ptarget;

#if ACIT_CFG_INTERRUPT_REDIRECTION
            ptarget->stats.io_completion[pio_handle->message_id]++;
#else
            ptarget->stats.io_completion[0]++;
#endif

#if 0
            // As per the new ITO scheme, the following is not required

            //
            // The exponential backoff scheme is disabled for eto
            // value of 0. That means for MPIO configuration, the
            // target will always have the same internal timeouts
            // irrespective of completion times
            //
            if(ACIT_IS_ETO_ENABLED(ptarget))
            {
                ticks_taken = (pio_handle->pio->original_ticks - 
                                            pio_handle->pio->available_ticks);

                if(ticks_taken > ptarget->default_cmd_ticks)
                {
                    // set max ticks upto a factor of ETO ticks
                    if(ticks_taken > ptarget->eto.original_ticks / 3)
                        ticks_taken = ptarget->eto.original_ticks / 3;

                    ptarget->last_cmd_ticks = ticks_taken;
                }
                else
                {
                    ptarget->last_cmd_ticks = ptarget->default_cmd_ticks;
                }
            }
#endif
        }

        pio_handle->pio->pio_handle = ACIT_NULL;

        // update stats before pio_request is freed
        ACIT_UPDATE_IOSTATUS_COUNT(pcontroller, pio_request->io_status);

        // Notify OSM that the I/O operation completed
        if(pio_request->ppost_routine)
            (pio_request->ppost_routine)(pcontroller->pcontroller_key,
                                        pio_request);
        else
            OSM_COMPLETE_IO(pcontroller->pcontroller_key,
                            pio_request);

        ACIT_INC_IOS_COMPLETED(pcontroller);

        ap_free_io_resources(pcontroller, pio_handle);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IO handle 0x%p waiting for invalidation completed successfully\n"),
                                            pio_handle));
    }

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_ap_complete_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64)pio_handle,
                        (pio_handle->pio)?(ACIT_UEXACT64)pio_handle->pio->pio_request:(ACIT_UEXACT64)0,
                        pcontroller->counters.pending_ios);
    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_complete_io_error(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    ACIT_UEXACT32           ioh_flag)
{
    // This API is called from HWI when EP firmware has reported this command
    // to have an error. Invalidate this command and issue an abort so as to
    // plug any holes in cmd_sn.

    // Set internal available ticks value to invalidation ticks. When the timer
    // runs the next time it will invalidate this IO and send out an
    // abort

    ACIT_BIT_SET(pio_handle->flags, ioh_flag);

    if(pio_handle->pio->internal_available_ticks >
        (pcontroller->state.invalidation_ticks - 1))
    {
        pio_handle->pio->internal_available_ticks = 
                (pcontroller->state.invalidation_ticks - 1);
    }
}

ACIT_UEXACT32
ap_return_io(   
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           io_status,
    ACIT_UEXACT32           unlink)
{
    /*
     * This API must be called to return an IO request that
     * is pending in ACIT internal queue. This function removes
     * the IO WS from ACIT internal queues and returns the IO
     * request back to OSMs.
     */

    /*
     * if the request pointer is valid we still have the OS request
     */
    OSM_ASSERT(pio);

    /*
     * If this IO was queued in the IOWS, then unlink it and
     * decrement the queued_ios count
     */
    if(unlink)
    {
        OSM_ASSERT(pio->queued);
        LIST_REMOVE_ELEMENT(&pio->controller_link);
        LIST_REMOVE_ELEMENT(&pio->target_link);
        LIST_REMOVE_ELEMENT(&pio->lun_link);

        OSM_ASSERT(pcontroller->counters.queued_ios);
        pcontroller->counters.queued_ios --;

        // IO Handle MUST be NULL
        OSM_ASSERT(!pio->pio_handle);
    }

    // Bug 32811 : Check for pio->pio_handle. In the acit_start_io path,
    // ap_return_io is called AND unlink is NOT set. Without the check
    // there is a crash here.
    if(pio->pio_handle)
    {
        pio->pio_handle->pio = ACIT_NULL;
    }

    pio->pio_handle = ACIT_NULL;

    pio->queued = 0;

    pio->pio_request->io_status = io_status;

    ACIT_UPDATE_IOSTATUS_COUNT(pcontroller, pio->pio_request->io_status);
    ACIT_UPDATE_TARGET_IOSTATUS_COUNT(GET_TARGET_WS(pio->pio_request->ptarget_key),
                                      pio->pio_request->io_status);
    ACIT_UPDATE_LUN_IOSTATUS_COUNT(GET_LUN_WS(pio->pio_request->plun_key),
                                      pio->pio_request->io_status);

    if(pio->pio_request->ppost_routine)
        (pio->pio_request->ppost_routine)(pcontroller->pcontroller_key,
                                     pio->pio_request);
    else
        OSM_COMPLETE_IO(pcontroller->pcontroller_key,
                        pio->pio_request);

    ACIT_INC_IOS_COMPLETED(pcontroller);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_return_hiber_io(   
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request)
{
    PACIT_TARGET_WS         ptarget;

    ptarget = GET_TARGET_WS(pio_request->ptarget_key);

    /*
     * This function returns the IO request back to OSMs.
     */

    /*
     * if the request pointer is valid we still have the OS request
     */
    if(pio_request)
    {
        if(pcontroller->params.crash_dump.sync_completion)
        {
            if(pio_request->ppost_routine)
                (pio_request->ppost_routine)(pcontroller->pcontroller_key,
                                             pio_request);
            else
                OSM_COMPLETE_IO(pcontroller->pcontroller_key,
                                pio_request);

            ptarget->plast_hiber_io = ACIT_NULL;
        }
        else
        {
            // wait for the interrupt to be called and then return the request 
            // back to os.
            ptarget->plast_hiber_io = pio_request;
        }
    }
    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_send_logout (
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    ACIT_UEXACT32               cri,
    ACIT_UEXACT32               reason_code)
{
    PACIT_LOGOUT_HANDLE         plogout_handle;

    if(ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY))
    {
        if(ptarget->osm_notification.u.bits.remove_pending == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("** Setting Target 0x%p state to REMOVE_PENDING \n"), ptarget));
            OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                            ACIT_EVENT_CLASS_TARGET,
                            ACIT_EVENT_TARGET_REMOVE_PENDING,
                            ptarget->target_table_index);

            ptarget->osm_notification.u.bits.remove_pending = 1;
        }
    }

    plogout_handle = ap_alloc_logout_resources(pcontroller, ptarget, cri, reason_code);

    if (plogout_handle)
    {
        // Clear any pending logout event.
        ap_clear_target_event (pcontroller, ptarget, EVENT_TGT_LOGOUT_PENDING);
        ap_clear_target_event (pcontroller, ptarget, EVENT_TGT_RECOVERY_LOGOUT_PENDING);

        // Flag that we are waiting for the logout response.
        //ACIT_BIT_SET(ptarget->flags, FLAG_TGT_LOGOUT_RESPONSE_PENDING);

        iscsi_start_logout(pcontroller, plogout_handle);
        hwi_fill_logout_request (pcontroller->phwi_ws, plogout_handle);
        hwi_submit_request (pcontroller, 
                            plogout_handle->pwrb_handle, 
                            plogout_handle->cri);
        
        return ACIT_STATUS_SUCCESS;
    }

    // No resources available.  Pend the logout.

    // Bug 25953 : If the logout is for session recovery, then
    // maintain different event flags. Else the logout event
    // will override the recovery event and we will not attempt
    // a recovery.
    if (ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY)) 
    {
        ap_set_target_event (pcontroller, 
                             ptarget, 
                             EVENT_TGT_RECOVERY_LOGOUT_PENDING);
    }
    else
    {
        ap_set_target_event (pcontroller, 
                             ptarget, 
                             EVENT_TGT_LOGOUT_PENDING);
    }

    return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}


ACIT_UEXACT32
ap_complete_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT8            timeout)
{
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           cri;
    ACIT_UEXACT32           reason_code;
    
    //ACIT_BIT_CLR (plogout_handle->ptarget->flags, FLAG_TGT_LOGOUT_RESPONSE_PENDING);
    //
    //Save target and cri before calling ap_free_logout_resource.
    //This routine otherwise will clear it up
    //
    ptarget = plogout_handle->ptarget;
    cri = plogout_handle->cri;
    reason_code = plogout_handle->reason_code;

     
    // Free up resources allocated for logout handle only if there is no timeout
    if(!timeout)
    {
        ap_free_logout_resources(pcontroller, plogout_handle);
    }

    iscsi_complete_logout(pcontroller,
                          ptarget,
                          cri,
                          iresponse,
                          reason_code,
                          timeout);
    return ACIT_STATUS_SUCCESS;
}

PACIT_NOP_HANDLE
ap_alloc_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    PACIT_VOID              ppdu,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PACIT_NOP_HANDLE        pnop_handle;
    ACIT_STATUS             status;

    //if( ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_EH_SGL_HANDLE) |
    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_ALL_SESSIONS_RECOVERY) |
        ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_LINK_DOWN) |
        ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS) |
        IS_TARGET_MISSING(ptarget) |
        ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_SESSION_LOGOUT) |
        ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY) |
        ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_WRB_HANDLE))
    {
        return ACIT_NULL;
    }

    pnop_handle = ap_alloc_nop_handle(pcontroller);

    if(pnop_handle)
    {
        pnop_handle->ptarget = ptarget;
        pnop_handle->cri = (ACIT_UEXACT16)cri;

        status = hwi_alloc_nop_resources(pcontroller, pnop_handle, cri, poptional_handle);

        if(status != ACIT_STATUS_SUCCESS)
        {
            ap_free_nop_handle(pcontroller, pnop_handle);
            pnop_handle = ACIT_NULL;
            
            return pnop_handle;
        }
        // Use the TMF Timeouts for NOP Outs as well
        //pnop_handle->available_ticks = GET_TARGET_EXEC_TICKS(ptarget, 0);
        pnop_handle->available_ticks = GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget);

        pnop_handle->starting_tick_count = CURRENT_TICK(pcontroller);

        iscsi_start_nopout(pcontroller, pnop_handle, ppdu, poptional_handle);
        LIST_ADDTO_TAIL(&pnop_handle->target_link, &pnop_handle->ptarget->nop_handle_list);
        pcontroller->counters.pending_nops++;
    }

    return pnop_handle;
}

ACIT_VOID
ap_free_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle)
{
    hwi_free_nop_resources(pcontroller, pnop_handle);
    ap_free_nop_handle(pcontroller, pnop_handle);
    pcontroller->counters.pending_nops--;
}

ACIT_UEXACT32
ap_send_nopout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    PACIT_VOID              pdu,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PACIT_NOP_HANDLE        pnop_handle = ACIT_NULL;

    pnop_handle = ap_alloc_nop_resources(pcontroller, 
                                         ptarget,
                                         cri, 
                                         pdu,
                                         poptional_handle);

    if (!pnop_handle)
    {
        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    // Format and send the reply
    hwi_fill_nop_request (pcontroller->phwi_ws, 
                          pnop_handle,
                          poptional_handle);
    
    hwi_submit_drvr_request (pcontroller, 
                             pnop_handle->pwrb_handle, 
                             pnop_handle->cri);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_setup_ioh_for_tmf(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_IO_HANDLE         pio_handle)
{
    //
    // This function is to be called after invalidation is complete and
    // the IO handle needs to be setup for TMF
    // Some of the fields in IO handle need not be setup for IO submission.
    // In order to save some cycles in the fast path, those fields that are
    // not needed unless there is a TMF are setup here
    //

    pio_handle->lun.u.lun_complete = pio_handle->pio->pio_request->lun.u.lun_complete;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_ready_for_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    ACIT_UEXACT32           status = ACIT_STATUS_INSUFFICIENT_RESOURCE;
    /*
     * This API will determine if it is OK to send out a TMF
     * on pcontroller and ptarget.
     */

    if(IS_CONTROLLER_UE(pcontroller))
    {
        return ACIT_STATUS_ERROR;
    }

    if( ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_EH_SGL_HANDLE) &&
        pcontroller->counters.tmf_handles_available)
    {
        status = ACIT_STATUS_SUCCESS;
    }

    if(!status && ptarget)
    {
        if(IS_TARGET_MISSING(ptarget))
        {
            return ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }

        if(ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_SESSION_LOGOUT) &&
            ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY) &&
            ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_WRB_HANDLE) &&
            ACIT_IS_BIT_CLR(ptarget->flags, TGT_FLAG_TMF_QUEUE_FULL) )
        {
            status = ACIT_STATUS_SUCCESS;
        }
        else
        {
            status = ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
    }

    return status;
}


ACIT_VOID
ap_free_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PACIT_IO_HANDLE         pio_handle;

    if(ptmf_handle->pio_handle)
    {
        pio_handle = ptmf_handle->pio_handle;

        // Double free?
        OSM_ASSERT(pio_handle->pwrb_handle);
        OSM_ASSERT(pio_handle->psgl_handle);

        //
        // At this point, the I/O handle has been invalidated. Free the
        // resources now
        //
        ap_free_io_resources(pcontroller, pio_handle);
    }

    iscsi_free_tmf_resources(pcontroller, ptmf_handle);
    hwi_free_tmf_resources(pcontroller, ptmf_handle);
    ap_free_tmf_handle(pcontroller, ptmf_handle);
    pcontroller->counters.pending_tmfs --;
}

ACIT_UEXACT32
ap_verify_reset_hierarchy(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           tmf_function)
{
    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_LOGOUT) ||
        ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_RECOVERY))
    {
        return ACIT_STATUS_RESET_IN_PROGRESS;
    }

    switch(tmf_function)
    {
    case ACIT_TMF_ABORT_TASK:

        OSM_ASSERT(plun);

        if(plun->plun_reset_tmf_request)
        {
            return ACIT_STATUS_RESET_IN_PROGRESS;
        }

        if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET))
            return ACIT_STATUS_RESET_IN_PROGRESS;

        break;

    case ACIT_TMF_LUN_RESET:

        if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET))
            return ACIT_STATUS_RESET_IN_PROGRESS;

        break;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_check_and_issue_abort(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle)
{
    ACIT_UEXACT32           status = 0;

    if(ap_verify_reset_hierarchy(pcontroller, 
                                  ptarget, 
                                  plun, 
                                  ACIT_TMF_ABORT_TASK) == ACIT_STATUS_RESET_IN_PROGRESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("abort IOH 0x%p for target 0x%p lun 0x%p already in reset\n", 
                                    pio_handle, ptarget, plun));

        return ACIT_STATUS_RESET_IN_PROGRESS;
    }

    if(ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING) &&
       ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_INVALIDATE_COMPLETE))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("invalidation needed for IOH 0x%p\n", pio_handle));

        status = ap_invalidate_pending_ios(pcontroller, 
                                            ptarget, 
                                            plun, 
                                            pio_handle, 
                                            ACIT_U32_MINUS_ONE);
    }
    else if((ACIT_IS_BIT_SET(pio_handle->flags, FLAG_IOH_INVALIDATE_COMPLETE)) &&
             ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_TMF_ABORT))
    {
        if(pio_handle->ptmf_ws && IS_LINK_VALID(&pio_handle->ptmf_ws->controller_link))
        {
            // this is an external TMF request, dequeue it before we issue an abort
            ap_dequeue_tmf(pcontroller, ptarget, plun, pio_handle);
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Attempting to issue abort after invalidation/error for IOH 0x%p\n", pio_handle));
        ap_abort_io(pcontroller, ptarget, plun, pio_handle);
        status = ACIT_STATUS_SUCCESS;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Abort request for a IO in invalidation or already in queue!! 0x%p 0x%x, ignoring\n",
            pio_handle, pio_handle->flags));
        status = ACIT_STATUS_PENDING;
    }

    return status;
}

//
// This routine checks if a LU or a Target reset can be issued. If a reset was
// issued, it will return ACIT_STATUS_SUCCESS. Otherwise ACIT_STATUS_PENDING
//
ACIT_UEXACT32
ap_check_and_issue_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    ACIT_UEXACT32           status;

    if(plun == ACIT_NULL)
    {
        OSM_ASSERT(ptarget->ptarget_reset_tmf_request);

        if(ap_verify_reset_hierarchy(pcontroller, 
                                     ptarget, 
                                     ACIT_NULL,
                                     ptarget->ptarget_reset_tmf_request->tmf_function) == ACIT_STATUS_RESET_IN_PROGRESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reset request for target 0x%p already in progress\n", 
                                        ptarget));

            return ACIT_STATUS_RESET_IN_PROGRESS;
        }

        status = ap_setup_target_for_cleanup(pcontroller, 
                                            ptarget, 
                                            ACIT_U32_MINUS_ONE,
                                            TGT_CLEANUP_FLAG_FOR_TMF_RESET,
                                            ACIT_FALSE);

        //OSM_ASSERT(!ptarget->counters.queued_tmfs);
        if(status == ACIT_STATUS_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target 0x%p in TMF TargetReset, queued TMFs %d...\n"),
                                                ptarget, ptarget->counters.queued_tmfs));

            ap_reset_target(pcontroller, 
                            ptarget, 
                            ptarget->ptarget_reset_tmf_request->tmf_function);
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target 0x%p waiting for %d invalidations or pending TMFs before issuing ResetTMF...\n"),
                ptarget, ptarget->counters.num_io_invs));

            ap_queue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
        }
    }
    else
    {
        if(ap_verify_reset_hierarchy(pcontroller, 
                                     ptarget, 
                                     plun,
                                     ACIT_TMF_LUN_RESET) == ACIT_STATUS_RESET_IN_PROGRESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reset request for target 0x%p lun 0x%p already in progress\n", 
                                        ptarget, plun));

            return ACIT_STATUS_RESET_IN_PROGRESS;
        }

        status = ap_setup_lun_for_reset(pcontroller, ptarget, plun);

        if(status == ACIT_STATUS_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("LUN 0x%p in target 0x%p in TMF LunReset...\n"),
                                                plun, ptarget));

            ap_reset_lun(pcontroller, ptarget, plun);
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("LUN 0x%p waiting for %d invalidations or pending TMFs before issuing ResetTMF...\n"),
                                                plun, plun->counters.num_io_invs));

            ap_queue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
        }
    }

    return status;
}

ACIT_UEXACT32
ap_start_abort_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_LUN_WS            plun;
    PACIT_TMF_WS            ptmf;
    PACIT_IO_WS             pio;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_UEXACT32           status;

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block, Aborts);

    OSM_ASSERT(ptmf_request);
    OSM_ASSERT(ptmf_request->ptarget_key);
    OSM_ASSERT(ptmf_request->plun_key);
    OSM_ASSERT(ptmf_request->ptmf_key);
    OSM_ASSERT(ptmf_request->pio_request);
    OSM_ASSERT(ptmf_request->pio_request->pio_key);

    ptarget = GET_TARGET_WS(ptmf_request->ptarget_key);
    plun    = GET_LUN_WS(ptmf_request->plun_key);
    ptmf    = GET_TMF_WS(ptmf_request->ptmf_key);
    pio     = GET_IO_WS(ptmf_request->pio_request->pio_key);

    OSM_ASSERT(pio);

    pio_handle = pio->pio_handle;

    if(!pio_handle)
    {
        ACIT_UEXACT32   tmf_status;
        /* 
         * This request is either still in ACIT internal queue or not
         * even with ACIT. Check and return the IO back now
         */
        if(pio->queued)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning IOWS 0x%p queued. target 0x%p lun 0x%p\n",
                                               pio, ptarget, plun));
            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_ABORTED, 
                         ACIT_TRUE);

            tmf_status = ACIT_TMF_STATUS_SUCCESS;
        }
        else
        {

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("IOWS 0x%p contains no IOH, abort failure target 0x%p lun 0x%p\n", pio, ptarget, plun));
            tmf_status = ACIT_TMF_STATUS_INVALID_TASK;

            ap_log_osm_event(pcontroller,
                            UNI_CFG_LOG_LEVEL_CRITICAL,
                            UNI_SUBMODULE_ACIT_GENERIC,
                            UNI_PHASE_TMF,
                            0,
                            ACIT_EC_TASK_INVALID_FOR_OSM_ABORT,
                            0, ACIT_NULL,
                            ACIT_NULL);
        }

        ap_return_tmf(pcontroller, ptmf_request, tmf_status);

        return ACIT_STATUS_SUCCESS;
    }

    OSM_ASSERT(!pio->queued);

    /*
     * Setup context for completion later
     */
    pio_handle->ptmf_ws = ptmf;
    
    status = ap_check_and_issue_abort(pcontroller, 
                                 ptarget, 
                                 plun, 
                                 pio_handle);

    switch(status)
    {

    case ACIT_STATUS_SUCCESS:
        
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issued abort for IOH 0x%p TMFRequest 0x%p TMFWS 0x%p, nothing more to be done!\n", pio_handle, ptmf_request, ptmf));
        break;
  
    case ACIT_STATUS_RESET_IN_PROGRESS:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Could not abortIOH 0x%p TMFRequest 0x%p TMFWS 0x%p, reset in progress!\n", pio_handle, ptmf_request, ptmf));

        //
        // This is to only to timeout the TMF if its in the internal queue for
        // too long
        //
        if(IS_LINK_INVALID(&pio_handle->ptmf_ws->controller_link))
        {
            LIST_ADDTO_TAIL(&pio_handle->ptmf_ws->controller_link, 
                            &pcontroller->tmf_queued_list);
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Queued TMF Abort for IOH 0x%p TMFRequest 0x%p TMFWS 0x%p in internal queue\n", pio_handle, ptmf_request, ptmf));
        }
        else
        {
            OSM_ASSERT(0);
        }

        break;

    case ACIT_STATUS_INSUFFICIENT_RESOURCE:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Could not abortIOH 0x%p TMFRequest 0x%p TMFWS 0x%p, no resource!\n", pio_handle, ptmf_request, ptmf));

        // its a bit tricky now. The invalidation could not be issued since
        // there was not enough resource. But the TMF has to be queued. The
        // rule for calling ap_queue_tmf is that the IOH should've been removed
        // from the busy/invalidate list prior to the call. So we explicitly
        // remove the IO handle from the lists and call ap_queue_tmf. Normally
        // this would be done as part of invalidation itself..
        LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
        LIST_INIT_ELEMENT(&pio_handle->controller_link);
        LIST_REMOVE_ELEMENT(&pio_handle->target_link);
        LIST_INIT_ELEMENT(&pio_handle->target_link);
        LIST_REMOVE_ELEMENT(&pio_handle->lun_link);
        LIST_INIT_ELEMENT(&pio_handle->lun_link);

        ap_queue_tmf(pcontroller, ptarget, plun, pio_handle);
        break;

    case ACIT_STATUS_PENDING:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Could not abortIOH 0x%p TMFRequest 0x%p TMFWS 0x%p, pending!\n", pio_handle, ptmf_request, ptmf));

        break;

    default:

          OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Unknown status 0x%x\n", status));
          OSM_ASSERT(0);
          break;

    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_start_lun_reset(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TMF_REQUEST           ptmf_request)
{
    PACIT_TARGET_WS             ptarget;
    PACIT_LUN_WS                plun;
    PACIT_TMF_WS                ptmf;

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block, LuResets);

    ptarget = GET_TARGET_WS(ptmf_request->ptarget_key);
    plun    = GET_LUN_WS(ptmf_request->plun_key);
    ptmf    = GET_TMF_WS(ptmf_request->ptmf_key);

    if((ACIT_IS_BIT_SET(plun->busy, LUN_BUSY_RESET_WAITING)) ||
        (ACIT_IS_BIT_SET(plun->busy, LUN_BUSY_RESET_WAIT_FOR_RESOURCE)) ||
        (ACIT_IS_BIT_SET(plun->busy, LUN_BUSY_RESET_SUBMITTED)) ||
        (ACIT_IS_BIT_SET(plun->busy, LUN_BUSY_RESET_IN_PROGRESS)))
    {
        // Reset already in progress, return this request
        ap_return_tmf(pcontroller, ptmf_request, ACIT_TMF_STATUS_IN_PROGRESS);
    }
    else
    {
        plun->eh_stats.tmf_lun_resets++;
        ap_flag_busy_lun(pcontroller, plun, LUN_BUSY_RESET_IN_PROGRESS);

        // OS has issued a LUN reset. So return all IOs queued on this LUN
        ap_return_queued_ios(pcontroller, 
                                ptarget, 
                                plun, 
                                ACIT_IO_STATUS_ABORTED,
                                ACIT_U32_MINUS_ONE);

        plun->plun_reset_tmf_request = ptmf_request;
        if(ap_check_and_issue_reset(pcontroller, 
                                    ptarget, 
                                    plun) != ACIT_STATUS_SUCCESS)
        {
            ap_queue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_start_target_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_TMF_WS            ptmf;

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block, TgtResets);

    ptarget = GET_TARGET_WS(ptmf_request->ptarget_key);
    ptmf    = GET_TMF_WS(ptmf_request->ptmf_key);

    if(ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_TMF_RESET_IN_PROGRESS))
    {
        // Reset already in progress, return this request
        ap_return_tmf(pcontroller, ptmf_request, ACIT_TMF_STATUS_IN_PROGRESS);
    }
    else
    {
        ptarget->eh_stats.tmf_target_resets++;
        ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_TMF_RESET_IN_PROGRESS);

        // OS has issued a target reset. So return all IOs queued on this LUN
        ap_return_queued_ios(pcontroller, 
                                ptarget, 
                                ACIT_NULL, 
                                ACIT_IO_STATUS_ABORTED,
                                ACIT_U32_MINUS_ONE);

        ptarget->ptarget_reset_tmf_request = ptmf_request;
        if(ap_check_and_issue_reset(pcontroller, 
                                    ptarget, 
                                    ACIT_NULL) != ACIT_STATUS_SUCCESS)
        {
            ap_queue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_start_abort_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request)
{

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_return_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request,
    ACIT_UEXACT32           tmf_status)
{
    PACIT_TMF_WS            ptmf_ws;
    PACIT_TARGET_WS         ptarget;
    PACIT_LUN_WS            plun;

    ptmf_ws = GET_TMF_WS(ptmf_request->ptmf_key);
    LIST_REMOVE_ELEMENT(&ptmf_ws->controller_link);

    ptarget = GET_TARGET_WS_SAFE(ptmf_request->ptarget_key);

    switch(ptmf_request->tmf_function)
    {
    case ACIT_TMF_ABORT_TASK:
        plun = GET_LUN_WS_SAFE(ptmf_request->plun_key);
        ap_update_lun_tmf_stats(plun, tmf_status);
        ap_update_tgt_tmf_stats(ptarget, tmf_status);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning Abort TMF 0x%p target 0x%p lun 0x%p with status 0x%x\n",
                                        ptmf_ws, ptarget, plun, tmf_status));

        break;
    case ACIT_TMF_LUN_RESET:

        plun = GET_LUN_WS_SAFE(ptmf_request->plun_key);
        OSM_ASSERT(plun->plun_reset_tmf_request);
        if(tmf_status != ACIT_TMF_STATUS_IN_PROGRESS)
        {
            plun->flags = 0;
            ap_flag_ready_lun(pcontroller, plun, LUN_BUSY_RESET_IN_PROGRESS);
            plun->plun_reset_tmf_request = ACIT_NULL;
        }

        ap_update_lun_tmf_stats(plun, tmf_status);
        ap_update_tgt_tmf_stats(ptarget, tmf_status);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning LunRst target 0x%p lun 0x%p with status 0x%x\n",
                                        ptarget, plun, tmf_status));

        break;

    case ACIT_TMF_TARGET_WARM_RESET:
    case ACIT_TMF_TARGET_COLD_RESET:

        OSM_ASSERT(ptarget->ptarget_reset_tmf_request);

        if(tmf_status != ACIT_TMF_STATUS_IN_PROGRESS)
        {
            ACIT_BIT_CLR(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET);
            ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_TMF_RESET_IN_PROGRESS);
            ptarget->ptarget_reset_tmf_request = ACIT_NULL;
        }

        ap_update_tgt_tmf_stats(ptarget, tmf_status);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning TgtRst target 0x%p with status 0x%x\n",
                ptarget, tmf_status));

        break;

    default:
        //OSM_ASSERT(0);
        break;
    }

    ptmf_request->tmf_status = tmf_status;

    ap_update_ctl_tmf_stats(pcontroller, tmf_status);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning TMF 0x%p status 0x%x TMFWS 0x%p\n", ptmf_request, tmf_status, ptmf_ws));

    if(ptmf_request->ppost_routine)
        (ptmf_request->ppost_routine(pcontroller->pcontroller_key,
                                     ptmf_request));
    else
        OSM_COMPLETE_TMF(pcontroller->pcontroller_key,
                        ptmf_request);
}

ACIT_UEXACT32
ap_complete_abort_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT32           tmf_status)
{
    /*
     * This API is called when TMF response is received for abort
     * task. If the TMF WS is available in tmf handle, then the I/O
     * and TMF is returned back to the OSM.
     */

    OSM_ASSERT(ptmf_handle->ptarget);
    OSM_ASSERT(ptmf_handle->plun);
    OSM_ASSERT(ptmf_handle->pio_handle);

    // The IOWs may be freed here, but the IO Handle is always valid on return
    ap_decide_io_disposition_in_cleanup(pcontroller, ptmf_handle->pio_handle);

    // Bug 106047 :
    // If an external Abort request came in after a TMF handle was instantiated
    // an abort was already issued to the target, the pio_handle->ptmf_ws is
    // the only reference to the external abort. In that case, ptmf_handle->ptmf
    // can be NULL. So use the pio_handle->ptmf_ws reference to check for
    // external TMF
    //if(ptmf_handle->ptmf)
    if(ptmf_handle->pio_handle->ptmf_ws)
    {
        OSM_ASSERT(ptmf_handle->pio_handle->ptmf_ws->ptmf_request);
        ap_return_tmf(pcontroller, ptmf_handle->pio_handle->ptmf_ws->ptmf_request, tmf_status);
        ptmf_handle->pio_handle->ptmf_ws = ACIT_NULL;
    }

    /*
     * Lastly free the TMF resources
     */
    ap_free_tmf_resources(pcontroller, ptmf_handle);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_complete_abort_multiple_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT32           tmf_status)
{

    return ACIT_STATUS_SUCCESS;


}

ACIT_UEXACT32
ap_complete_nop_in(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle)
{
    OSM_ASSERT(0);
    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_complete_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT8            max_cmd_sn_offset,
    ACIT_UEXACT32           exp_cmd_sn)
{
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           status;

    ptarget = ptmf_handle->ptarget;

    status = iscsi_complete_tmf(pcontroller, 
                                ptmf_handle,
                                iresponse,
                                max_cmd_sn_offset,
                                exp_cmd_sn);

    switch(ptmf_handle->tmf_function)
    {
    case ACIT_TMF_ABORT_TASK:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF completion for AbortTMF 0x%p - IOH 0x%p...\n"),
                                            ptmf_handle, ptmf_handle->pio_handle));

        ap_complete_abort_io(pcontroller, ptmf_handle, status);
        break;

    case ACIT_TMF_LUN_RESET:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF completion for LUNResetTMF - LUN 0x%p...\n"),
                                            ptmf_handle->plun));

        ap_free_pending_handles(pcontroller, 
                                ptmf_handle->ptarget,
                                ptmf_handle->plun,
                                ACIT_U32_MINUS_ONE);

        break;

    case ACIT_TMF_TARGET_WARM_RESET:
    case ACIT_TMF_TARGET_COLD_RESET:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF completion for TargetResetTMF - target 0x%p...\n"),
                                            ptmf_handle->ptarget));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_RESET_DONE);
        ap_free_pending_handles(pcontroller,
                                ptarget,
                                ACIT_NULL,
                                ACIT_U32_MINUS_ONE);

        break;

    case ACIT_TMF_ABORT_MULTIPLE_IOS:

        ap_complete_abort_multiple_ios(pcontroller, ptmf_handle, status);
        break;
    }

    ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_FREE);

    if((status == ACIT_TMF_STATUS_FUNCTION_REJECTED) ||
        (status == ACIT_TMF_STATUS_NOT_SUPPORTED))
    {
        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_ERROR);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("TMF in target 0x%p error, will trigger session recovery..\n"),
                                                ptarget));

        pcontroller->eh_stats.tmf_errors++;
        ptarget->eh_stats.tmf_errors++;
        iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_abort_io(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun,
    PACIT_IO_HANDLE             pio_handle)
{
    PACIT_TMF_HANDLE            ptmf_handle;

    ptmf_handle = ap_alloc_tmf_resources(pcontroller,
                                         ptarget, 
                                         plun,
                                         pio_handle,
                                         pio_handle->ptmf_ws,
                                         ACIT_TMF_ABORT_TASK);
    if(ptmf_handle)
    {
        ptmf_handle->flags = FLAG_TMFH_TMF_STARTED;

        iscsi_start_tmf(pcontroller, ptmf_handle);
        hwi_fill_tmf_request(pcontroller->phwi_ws, ptmf_handle);
        hwi_submit_request(pcontroller, ptmf_handle->pwrb_handle, ptmf_handle->cri);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Issued TMF abort for IOH 0x%p and TMFH 0x%p\n", pio_handle, ptmf_handle));
        ACIT_BIT_SET(pio_handle->flags, FLAG_IOH_TMF_ABORT);
    }
    else
    {
        ap_queue_tmf(pcontroller, ptarget, plun, pio_handle);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("LUN 0x%p has %d TMFs queued\n", plun, plun->counters.queued_tmfs));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s exit CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
                        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));

    OSM_ASSERT(ptarget->counters.queued_tmfs <= pcontroller->counters.queued_tmfs);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_reset_lun(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun)
{
    PACIT_TMF_HANDLE            ptmf_handle;
    PACIT_TMF_WS                ptmf = ACIT_NULL;

    //OSM_ASSERT(!plun->counters.queued_tmfs);

    if(plun->plun_reset_tmf_request)
    {
        ptmf = GET_TMF_WS(plun->plun_reset_tmf_request->ptmf_key);
    }

    ptmf_handle = ap_alloc_tmf_resources(pcontroller,
                                         ptarget, 
                                         plun,
                                         ACIT_NULL,
                                         ptmf,
                                         ACIT_TMF_LUN_RESET);

    if(ptmf_handle)
    {
        ptmf_handle->flags = FLAG_TMFH_TMF_STARTED;

        iscsi_start_tmf(pcontroller, ptmf_handle);
        hwi_fill_tmf_request(pcontroller->phwi_ws, ptmf_handle);
        hwi_submit_request(pcontroller, ptmf_handle->pwrb_handle, ptmf_handle->cri);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Issued TMF LunReset for LUN 0x%p TMFH 0x%p\n", plun, ptmf_handle));
        ap_flag_busy_lun(pcontroller, plun, LUN_BUSY_RESET_SUBMITTED);
    }
    else
    {
        ap_flag_busy_lun(pcontroller, plun, LUN_BUSY_RESET_WAIT_FOR_RESOURCE);
        // Resources were unavailable, queue this in controller's list and return
        ap_queue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("LUN 0x%p has %d TMFs queued\n", plun, plun->counters.queued_tmfs));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s exit CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
                        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));

    OSM_ASSERT(ptarget->counters.queued_tmfs <= pcontroller->counters.queued_tmfs);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_reset_target(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    ACIT_UEXACT32               tmf_function)
{
    PACIT_TMF_HANDLE            ptmf_handle;
    PACIT_TMF_WS                ptmf = ACIT_NULL;

    if(ptarget->ptarget_reset_tmf_request)
    {
        ptmf = GET_TMF_WS(ptarget->ptarget_reset_tmf_request->ptmf_key);
    }

    //OSM_ASSERT(!ptarget->counters.queued_tmfs);

    ptmf_handle = ap_alloc_tmf_resources(pcontroller,
                                         ptarget, 
                                         ACIT_NULL,
                                         ACIT_NULL,
                                         ptmf,
                                         tmf_function);
    if(ptmf_handle)
    {
        ptmf_handle->flags = FLAG_TMFH_TMF_STARTED;

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_RESET_START);
        iscsi_start_tmf(pcontroller, ptmf_handle);
        hwi_fill_tmf_request(pcontroller->phwi_ws, ptmf_handle);
        hwi_submit_request(pcontroller, ptmf_handle->pwrb_handle, ptmf_handle->cri);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Issued TMF TargetReset for TMFH 0x%p\n", ptmf_handle));
    }
    else
    {
        // Resources were unavailable, queue this in controller's list and return
        ap_queue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s exit CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
                        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));
    OSM_ASSERT(ptarget->counters.queued_tmfs <= pcontroller->counters.queued_tmfs);

    return ACIT_STATUS_SUCCESS;
}

PACIT_INVALIDATE_HANDLE
ap_alloc_inv_resources(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_INVALIDATE_HANDLE pinv_handle;
    ACIT_UEXACT8            status;

    pinv_handle = ap_alloc_inv_handle(pcontroller);

    if(pinv_handle)
    {
        status = mgmt_peek_icd_inv_resources(pcontroller);

        if(status != MGMT_STATUS_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("No free MGMT resources for invalidate\n"));
            ap_free_inv_handle(pcontroller, pinv_handle);
            pinv_handle = ACIT_NULL;
            pcontroller->eh_stats.invs_deferred_no_ioctls++;
        }
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("No free invalidate handles\n"));
        pcontroller->eh_stats.invs_deferred_no_invh++;
    }

    return pinv_handle;
}

ACIT_VOID
ap_free_inv_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle)
{
    ap_free_inv_handle(pcontroller, pinv_handle);
}

#if ACIT_CFG_VASE_ENABLED

ACIT_UEXACT32
ap_prepare_vase(
    PACIT_CONTROLLER_WS         pcontroller)
{
    VASE_PROFILE                vase_profile;

    OSM_MEMSET(&vase_profile, 0, sizeof(VASE_PROFILE));

    vase_profile.Context = (ACIT_UEXACT64) pcontroller;
    /*
     * vase_profile.ConfigFlags = pcontroller->params.vase_blocks;
     * vase_profile.TimeStampFrequency = 0;
     * vase_profile.FuncBlockSize = pcontroller->params.vase_funcblk_size;
     */
    
    vase_profile.FuncBlockSize = 1000;
    vase_profile.ConfigFlags = 0xFFFFFFFF;

    DbgLogPreInit(  &pcontroller->vase_master_block,
                    &pcontroller->vase_mem_needed,
                    &vase_profile);

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_initialize_vase(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_UEXACT32              pmemory)
{
    ACIT_UEXACT64               timedelta;

    OSM_ASSERT(pmemory);

    timedelta = OSM_GETCLOCK();
    OSM_DELAY_TIMER(pcontroller->pcontroller_key, 10000);
    timedelta = (ACIT_UEXACT32)(OSM_GETCLOCK() - timedelta);

    DbgLogInitialize(&pcontroller->vase_master_block, pmemory, (ACIT_UEXACT32)timedelta);

    return ACIT_STATUS_SUCCESS;
}


#endif  /* For #if ACIT_CFG_VASE_ENABLED*/

ACIT_VOID
ap_set_mem_const (
    PACIT_CONTROLLER_WS     pcontroller,
    const PACIT_UEXACT8     name,
    ACIT_UEXACT32           index,
    ACIT_UEXACT32           size_in_bytes,
    ACIT_UEXACT32           alignment_mask,
    ACIT_UEXACT32           category)
{

    PACIT_MEM_DESCRIPTOR    pmem;

    //OSM_ASSERT(size_in_bytes);
    OSM_ASSERT(index < pcontroller->acit_var_desc_index.acit_mem_max);
    
    pmem = &pcontroller->mem_descriptors[index];

    pmem->virtual_address = 0;
    pmem->bus_address.u.a64.address = 0;

    pmem->alignment_mask = alignment_mask;
    pmem->size_in_bytes = size_in_bytes;
    pmem->index = index;
    ACIT_BIT_SET (pmem->category, category);

    OSM_DEBUG_PRINT (ACIT_DBGMSG_TRACE, 
                      ("%d)Mem Descriptor:%12s Bytes:0x%08x (%8d) Cached?:%d\n", 
                      index,
                      SA_DBG_CSTR (name), 
                       size_in_bytes, 
                       size_in_bytes, 
                       ACIT_IS_BIT_SET (pmem->category, ACIT_MEM_CATEGORY_CACHED)));
}


ACIT_VOID
ap_set_mem_variable (
    PACIT_CONTROLLER_WS         pcontroller,
    const PACIT_UEXACT8         name,
    ACIT_UEXACT32               index,
    ACIT_UEXACT32               num_elements,
    ACIT_UEXACT32               element_size,
    ACIT_UEXACT32               alignment_mask,
    ACIT_UEXACT32               category)
{
    ACIT_UEXACT32               num_desc, i, elements_per_desc, desc_sz;
    ACIT_UEXACT32               bytes_left, elements_left;
    PACIT_MEM_DESCRIPTOR        pmem;

    if (element_size == 0)
    {
        pmem = &pcontroller->mem_descriptors[index];

        pmem->virtual_address = 0;
        pmem->bus_address.u.a64.address = 0;

        pmem->alignment_mask = alignment_mask;
        pmem->index = index;
        ACIT_BIT_SET (pmem->category, category);

        pmem->size_in_bytes = 0;
        pmem->num_elements = 0;
        return;
    }
    
    OSM_ASSERT(element_size);
    OSM_ASSERT(num_elements);

    elements_per_desc = ACIT_CFG_MAX_CONTIGUOUS_MEM/element_size;

    desc_sz = num_elements * element_size;
    num_desc = sa_ceiling(desc_sz,ACIT_CFG_MAX_CONTIGUOUS_MEM);
    
    elements_left = num_elements;
    bytes_left = num_elements * element_size;
    
    for(i=0; i<num_desc; i++)
    {
        OSM_ASSERT(index < pcontroller->acit_var_desc_index.acit_mem_max);

        pmem = &pcontroller->mem_descriptors[index];

        pmem->virtual_address = 0;
        pmem->bus_address.u.a64.address = 0;

        pmem->alignment_mask = alignment_mask;
        pmem->index = index;
        ACIT_BIT_SET (pmem->category, category);

        /* Calculate in size and num_elements */
        if(elements_left > elements_per_desc)
        {
            pmem->size_in_bytes = element_size * elements_per_desc;
            pmem->num_elements = elements_per_desc;
        }
        else
        {
            pmem->size_in_bytes = bytes_left;
            pmem->num_elements = elements_left;
        }

        OSM_ASSERT(pmem->size_in_bytes);

        OSM_DEBUG_PRINT (ACIT_DBGMSG_TRACE, 
                          ("Mem Descriptor:%12s Bytes:0x%08x (%8d) Cached?:%d\n", 
                           SA_DBG_CSTR (name), 
                           pmem->size_in_bytes, 
                           pmem->size_in_bytes, 
                           ACIT_IS_BIT_SET (pmem->category, ACIT_MEM_CATEGORY_CACHED)));

        bytes_left -= pmem->size_in_bytes;
        elements_left -= pmem->num_elements;
        index++;
    }
}

ACIT_VOID
ap_init_var_desc_index(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT32           num_ioh_sz, num_cxn_sz, num_frag_sz;
    ACIT_UEXACT32           num_wrb_sz, num_wrb_sz_per_cxn;
    ACIT_UEXACT32           max_phys = ACIT_CFG_MAX_CONTIGUOUS_MEM;
    PACIT_VAR_DESC_INDEX    pvar = &pc->acit_var_desc_index;

    /* Initialize num_index values based on config params*/

    num_ioh_sz = sizeof(struct _ACIT_IO_HANDLE)     * p->ios_per_controller;
    num_cxn_sz = sizeof(struct _ISCSI_CONNECTION)   * p->num_cxn;
    num_frag_sz = sizeof(SGE)* p->num_sge * p->icds_per_controller;

    num_wrb_sz_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    num_wrb_sz = num_wrb_sz_per_cxn * p->num_cxn;

    pvar->num.ioh  = (ACIT_UEXACT8)sa_ceiling(num_ioh_sz, max_phys);
    pvar->num.cxn  = (ACIT_UEXACT8)sa_ceiling(num_cxn_sz, max_phys);
    pvar->num.frag = (ACIT_UEXACT8)sa_ceiling(num_frag_sz, max_phys);
    pvar->num.wrb  = (ACIT_UEXACT8)sa_ceiling(num_wrb_sz, max_phys);
    pvar->num.wrbh  = (ACIT_UEXACT8)sa_ceiling(p->num_cxn, 8);

    /* Initialize starting values for different descriptors */
    pvar->range.ioh = AP_MEM_MAX;
    pvar->range.cxn = pvar->range.ioh + pvar->num.ioh;
    pvar->range.wrb = pvar->range.cxn + pvar->num.cxn;
    pvar->range.frag= pvar->range.wrb + pvar->num.wrb;
    pvar->range.wrbh = pvar->range.frag + pvar->num.frag;
    
    /* Initialize the max value */
    pvar->acit_mem_max = pvar->range.wrbh + pvar->num.wrbh;

    /* Initialize the runtime default. This can be overridden later */
    pcontroller->acit_var_desc_index.run_time.frag = pvar->num.frag;
}


ACIT_UEXACT32
ap_init_all_mem_descriptors(
    PACIT_CONTROLLER_WS         pcontroller)
{
    ap_init_var_desc_index(pcontroller);

    ap_init_mem_descriptors(pcontroller);
    iscsi_init_mem_descriptors(pcontroller);
    mgmt_init_mem_descriptors(pcontroller);
    if (pcontroller->state.operating_mem_level == ACIT_MEM_LEVEL_MINIMUM) 
        hwi_init_mailbox_descriptor(pcontroller);
    else
        hwi_init_mem_descriptors(pcontroller);

#ifdef BE_EMULATION
    emul_init_mem_descriptors(pcontroller);
#endif

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_init_mem_descriptors (
    PACIT_CONTROLLER_WS     pcontroller)
{
    // Abbreviations.

    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    ACIT_UEXACT32           num_invs;

    const ACIT_UEXACT32 a8 = ACIT_ALIGN_8BYTE;
    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;

    /*
     * ACIT main related descriptors
     */
    ap_set_mem_const (pc, (PUEXACT8)"vase_log",AP_MEM_VASE, pc->vase_mem_needed, a8, cached);
    ap_set_mem_const (pc, (PUEXACT8)"logout_handle",AP_MEM_LOGOUT_HANDLE,p->logouts_per_controller  *sizeof(ACIT_LOGOUT_HANDLE), a8, cached);
    ap_set_mem_const (pc, (PUEXACT8)"tmf_handle",   AP_MEM_TMF_HANDLE,   p->tmfs_per_ctrl           *sizeof(ACIT_TMF_HANDLE),    a8, cached);
    ap_set_mem_const (pc, (PUEXACT8)"nop_handle",   AP_MEM_NOP_HANDLE,   p->num_async_pdus          *sizeof(ACIT_NOP_HANDLE),    a8, cached);

    num_invs = p->ios_per_controller + p->tmfs_per_ctrl;
    ap_set_mem_const (pc, (PUEXACT8)"inv_handle",   AP_MEM_INV_HANDLE,   num_invs                   *sizeof(ACIT_INVALIDATE_HANDLE),    a8, cached);
    ap_set_mem_const (pc, (PUEXACT8)"target_key",   AP_MEM_TARGET_KEY,   p->targets_per_controller  *sizeof(ACIT_TARGET_KEY),    a8, cached);
    ap_set_mem_const (pc, (PUEXACT8)"target_table", AP_MEM_TARGET_TABLE, (p->targets_per_controller - 1) *
                                                                         sizeof(ACIT_TARGET_ENTRY) + 
                                                                         sizeof(ACIT_TARGET_TABLE),                    a8, cached);


    pcontroller->ioh_mem_index = pc->acit_var_desc_index.range.ioh;
    ap_set_mem_variable (pc, (PUEXACT8)"io_handle", pcontroller->ioh_mem_index, p->ios_per_controller, sizeof(ACIT_IO_HANDLE), a8, cached);

    return ACIT_STATUS_SUCCESS;
}


ACIT_VOID
ap_init_io_handle (
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_IO_HANDLE             pio_handle;
    PACIT_MEM_DESCRIPTOR        pmem_desc;
    ACIT_UEXACT32               index, desc_index, ios, i;

    ios = 0;
    index = 0;
    desc_index = pcontroller->ioh_mem_index;

    while(ios < pcontroller->params.ios_per_controller)
    {
        pmem_desc = &pcontroller->mem_descriptors[desc_index];
        pcontroller->pio_handle_base[index] = pmem_desc->virtual_address;
        pio_handle = (PACIT_IO_HANDLE) pcontroller->pio_handle_base[index];

        OSM_ASSERT(pmem_desc->num_elements);

        for(i = 0; i < pmem_desc->num_elements; i++)
        {
            LIST_ADDTO_TAIL(&pio_handle->controller_link,
                            &pcontroller->io_handle_free_list);

            LIST_INIT_ELEMENT(&pio_handle->target_link);
            LIST_INIT_ELEMENT(&pio_handle->lun_link);

            pcontroller->counters.io_handles_available ++;

            pio_handle ++;

            ios++;

            if(ios >= pcontroller->params.ios_per_controller)
            {
                break;
            }
        }

        index++;
        desc_index++;
    }

    OSM_ASSERT(pcontroller->counters.io_handles_available == pcontroller->params.ios_per_controller);
}

ACIT_VOID
ap_init_logout_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_LOGOUT_HANDLE     plogout_handle;
    ACIT_UEXACT32           i;

    plogout_handle = (PACIT_LOGOUT_HANDLE) pcontroller->plogout_handle_base;

    for(i = 0; i < pcontroller->params.logouts_per_controller; i++)
    {
        plogout_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;
        plogout_handle->bhs_pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                                     ACIT_NULL,
                                                     &plogout_handle->noncmd_bhs,
                                                     &plogout_handle->bhs_len);
    
        /* Reset the size back to whatever is allocated on the virtual space */
        plogout_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;

        OSM_ASSERT(plogout_handle->bhs_pa.u.a64.address);

        LIST_ADDTO_TAIL(&plogout_handle->controller_link,
                        &pcontroller->logout_handle_free_list);

        pcontroller->counters.logout_handles_available ++;
        plogout_handle ++;
    }
}

ACIT_VOID
ap_init_tmf_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_TMF_HANDLE        ptmf_handle;
    ACIT_UEXACT32           i;

    ptmf_handle = (PACIT_TMF_HANDLE) pcontroller->ptmf_handle_base;

    for(i = 0; i < pcontroller->params.tmfs_per_ctrl; i++)
    {
        ptmf_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;
        ptmf_handle->bhs_pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                                  ACIT_NULL,
                                                  &ptmf_handle->noncmd_bhs,
                                                  &ptmf_handle->bhs_len);

        /* Reset the size back to whatever is allocated on the virtual space */
        ptmf_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;

        OSM_ASSERT(ptmf_handle->bhs_pa.u.a64.address);

        LIST_ADDTO_TAIL(&ptmf_handle->controller_link,
                        &pcontroller->tmf_handle_free_list);

        pcontroller->counters.tmf_handles_available ++;

        ptmf_handle ++;
    }
}

ACIT_VOID
ap_init_nop_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_NOP_HANDLE        pnop_handle;
    ACIT_UEXACT32           i;

    /*
     * Initialize all NOP handles
     */
    pnop_handle = (PACIT_NOP_HANDLE) pcontroller->pnop_handle_base;

    for(i = 0; i < pcontroller->params.num_async_pdus; i++)
    {

        pnop_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;
        pnop_handle->bhs_pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                                  ACIT_NULL,
                                                  &pnop_handle->noncmd_bhs,
                                                  &pnop_handle->bhs_len);

        /* Reset the size back to whatever is allocated on the virtual space */
        pnop_handle->bhs_len = ACIT_NON_CMD_BHS_LEN;

        OSM_ASSERT(pnop_handle->bhs_pa.u.a64.address);

        LIST_ADDTO_TAIL(&pnop_handle->controller_link,
                        &pcontroller->nop_handle_free_list);

        pcontroller->counters.nop_handles_available ++;

        pnop_handle ++;
    }
}

ACIT_VOID
ap_init_invalidate_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_INVALIDATE_HANDLE pinv_handle;
    ACIT_UEXACT32           i, num_invs;

    // number of invalidations allowed depends on the number of MCC WRBs
    // we can post
    num_invs = pcontroller->params.ioctls_per_ctrl;

    pinv_handle = (PACIT_INVALIDATE_HANDLE) pcontroller->pinv_handle_base;

    for(i = 0; i < num_invs; i++)
    {
        LIST_ADDTO_TAIL(&pinv_handle->controller_link,
                        &pcontroller->inv_handle_free_list);

        LIST_INIT_HEADER(&pinv_handle->list_header);
        pinv_handle->index = (ACIT_UEXACT16)i;

        pcontroller->counters.inv_handles_available ++;
        pinv_handle ++;
    }
}


PACIT_IO_HANDLE
ap_alloc_io_handle (
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_DC_LIST               plink;
    PACIT_IO_HANDLE             pio_handle;

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&pcontroller->io_handle_free_list));

    LIST_REMOVEFROM_HEAD (&plink, 
                        &pcontroller->io_handle_free_list);
    pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                controller_link, 
                                plink);
    
    LIST_ADDTO_TAIL(&pio_handle->controller_link,
                    &pcontroller->io_handle_busy_list);

    pcontroller->counters.io_handles_available --;

    if(!pcontroller->counters.io_handles_available)
    {
        /*
         * No more IO Handles available, notify OSM that controller is busy
         */
        ap_flag_busy_controller(pcontroller, 
                                CONTROLLER_BUSY_IO_HANDLE);
    }

    pio_handle->flags = 0;

    return pio_handle;
}


ACIT_VOID
ap_assign_sglh2_ioh(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_DC_LIST   plink;
    PACIT_IO_HANDLE pio_handle = ACIT_NULL;

    // this routine allocates SGL handle and associates it with
    // an IO handle. This way SGL handles for IO need not be
    // allocated at run-time

    plink = pcontroller->io_handle_free_list.next;

    while(plink != &pcontroller->io_handle_free_list)
    {
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                 controller_link, 
                                 plink);

        plink = plink->next;

        pio_handle->psgl_handle = hwi_alloc_io_sgl_handle(pcontroller->phwi_ws);
        OSM_ASSERT(pio_handle->psgl_handle);
    }

    // controller should go busy since all IO SGL handles are pre-allocated
    OSM_ASSERT(pcontroller->busy & CONTROLLER_BUSY_IO_SGL_HANDLE);

    // flag it ready
    ap_flag_ready_controller(pcontroller, CONTROLLER_BUSY_IO_SGL_HANDLE);

}


ACIT_VOID
ap_free_io_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    // double free??
    OSM_ASSERT(ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_IO_FREE));

    LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
    LIST_REMOVE_ELEMENT(&pio_handle->target_link);
    LIST_REMOVE_ELEMENT(&pio_handle->lun_link);

    LIST_INIT_ELEMENT(&pio_handle->target_link);
    LIST_INIT_ELEMENT(&pio_handle->lun_link);

    pcontroller->counters.io_handles_available ++;
    
    pio_handle->flags = FLAG_IOH_IO_FREE;

    if(pio_handle->ptmf_ws)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("TMFWS 0x%p in IOH 0x%p LUN 0x%p Target 0x%p not freed!\n",
            pio_handle->ptmf_ws, pio_handle, pio_handle->lun, pio_handle->ptarget));
    }

    pio_handle->ptmf_ws = ACIT_NULL;
    pio_handle->inv_start_tick = 0;
    
    //TODO Add Free state to IO handle
    LIST_ADDTO_TAIL(&pio_handle->controller_link,
                    &pcontroller->io_handle_free_list);

    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_IO_HANDLE))
    {
        /*
         * New IO Handles are just being made available and there
         * may be I/Os waiting in the queue for it.
         */
        ap_flag_ready_controller(   pcontroller,
                                    CONTROLLER_BUSY_IO_HANDLE);
    }

}

PACIT_LOGOUT_HANDLE
ap_alloc_logout_handle(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_DC_LIST               plink;
    PACIT_LOGOUT_HANDLE         plogout_handle = ACIT_NULL;

    if(IS_LIST_NOT_EMPTY(&pcontroller->logout_handle_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                            &pcontroller->logout_handle_free_list);

        plogout_handle = STRUCT_BASE(ACIT_LOGOUT_HANDLE,
                                     controller_link,
                                     plink);
        
        LIST_ADDTO_TAIL(&plogout_handle->controller_link,
                        &pcontroller->logout_handle_busy_list);

        pcontroller->counters.logout_handles_available --;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("LOGOUT Handle not available"));
    }

    /*
     * Do not busy controller if logout handle is not available
     */

    return plogout_handle;
}

ACIT_VOID
ap_free_logout_handle(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_LOGOUT_HANDLE         plogout_handle)
{
    /*
     * ADPmp86793 - double freeing of controller_link.
     * Logout invalidation will already take out the
     * logout handle, so it does not have to be done again
     */
    if(IS_LINK_VALID(&plogout_handle->controller_link))
        LIST_REMOVE_ELEMENT(&plogout_handle->controller_link);

    if(IS_LINK_VALID(&plogout_handle->target_link))
        LIST_REMOVE_ELEMENT(&plogout_handle->target_link);

    LIST_ADDTO_TAIL(&plogout_handle->controller_link,
                    &pcontroller->logout_handle_free_list);

    pcontroller->counters.logout_handles_available ++;


}

PACIT_TMF_HANDLE
ap_alloc_tmf_handle(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_DC_LIST               plink;
    PACIT_TMF_HANDLE            ptmf_handle = ACIT_NULL;

    if(IS_LIST_NOT_EMPTY(&pcontroller->tmf_handle_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                            &pcontroller->tmf_handle_free_list);

        ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                    controller_link,
                                    plink);
        
        LIST_ADDTO_TAIL(&ptmf_handle->controller_link,
                        &pcontroller->tmf_handle_busy_list);

        LIST_INIT_ELEMENT(&ptmf_handle->target_link);
        LIST_INIT_ELEMENT(&ptmf_handle->lun_link);

        pcontroller->counters.tmf_handles_available --;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("TMF Handle not available for TMF\n"));
        OSM_ASSERT(0);
    }

    ptmf_handle->flags = 0;

    return ptmf_handle;
}

ACIT_VOID
ap_free_tmf_handle(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TMF_HANDLE            ptmf_handle)
{
    LIST_REMOVE_ELEMENT(&ptmf_handle->controller_link);
    LIST_REMOVE_ELEMENT(&ptmf_handle->target_link);
    LIST_REMOVE_ELEMENT(&ptmf_handle->lun_link);

    LIST_ADDTO_TAIL(&ptmf_handle->controller_link,
                    &pcontroller->tmf_handle_free_list);

    ptmf_handle->pio_handle = ACIT_NULL;
    ptmf_handle->ptmf = ACIT_NULL;
    ptmf_handle->plun = ACIT_NULL;
    ptmf_handle->ptarget = ACIT_NULL;
    pcontroller->counters.tmf_handles_available ++;

    if(pcontroller->counters.queued_tmfs &&
       ap_ready_for_tmf(pcontroller, ACIT_NULL) == ACIT_STATUS_SUCCESS)
    {
        pcontroller->state.flush_queued_tmfs = ACIT_TRUE;
    }
}

PACIT_NOP_HANDLE
ap_alloc_nop_handle(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_DC_LIST               plink;
    PACIT_NOP_HANDLE            pnop_handle = ACIT_NULL;

    if(IS_LIST_NOT_EMPTY(&pcontroller->nop_handle_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                             &pcontroller->nop_handle_free_list);

        pnop_handle = STRUCT_BASE(ACIT_NOP_HANDLE,
                                  controller_link,
                                  plink);
        
        LIST_ADDTO_TAIL(&pnop_handle->controller_link,
                        &pcontroller->nop_handle_busy_list);

        pcontroller->counters.nop_handles_available --;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("NOP Handle not available for TMF"));
        OSM_ASSERT(0);
    }

    return pnop_handle;
}

ACIT_VOID
ap_free_nop_handle(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_NOP_HANDLE            pnop_handle)
{
    LIST_REMOVE_ELEMENT(&pnop_handle->controller_link);

    if(IS_LINK_VALID(&pnop_handle->target_link))
    {
        LIST_REMOVE_ELEMENT(&pnop_handle->target_link);
    }

    LIST_ADDTO_TAIL(&pnop_handle->controller_link,
                    &pcontroller->nop_handle_free_list);

    pcontroller->counters.nop_handles_available ++;
}

PACIT_INVALIDATE_HANDLE
ap_alloc_inv_handle(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_DC_LIST           plink;
    PACIT_INVALIDATE_HANDLE pinv_handle = ACIT_NULL;

    if(IS_LIST_NOT_EMPTY(&pcontroller->inv_handle_free_list))
    {
        LIST_REMOVEFROM_HEAD(&plink, 
                             &pcontroller->inv_handle_free_list);

        pinv_handle = STRUCT_BASE(ACIT_INVALIDATE_HANDLE,
                                  controller_link,
                                  plink);

        LIST_INIT_HEADER(&pinv_handle->list_header);
        pinv_handle->flags = 0;
        pinv_handle->num_invs = 0;
        
        LIST_ADDTO_TAIL(&pinv_handle->controller_link,
                        &pcontroller->inv_handle_busy_list);

        pcontroller->counters.inv_handles_available --;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Invalidate Handle not available"));
        OSM_ASSERT(0);
    }

    return pinv_handle;
}

ACIT_VOID
ap_free_inv_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle)
{
    LIST_REMOVE_ELEMENT(&pinv_handle->controller_link);

    LIST_ADDTO_TAIL(&pinv_handle->controller_link,
                    &pcontroller->inv_handle_free_list);

    pcontroller->counters.inv_handles_available ++;
}



ACIT_VOID
ap_queue_io_ws(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun,
    PACIT_IO_WS                 pio)
{
    LIST_ADDTO_TAIL(&pio->controller_link,
                    &pcontroller->io_queued_list);
    LIST_ADDTO_TAIL(&pio->target_link,
                    &ptarget->io_queued_list);
    LIST_ADDTO_TAIL(&pio->lun_link,
                    &plun->io_queued_list);
    pcontroller->counters.queued_ios ++;
    pio->queued = 1;
}

ACIT_VOID
ap_queue_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle)
{
    PACIT_TMF_WS            ptmf_ws;
    ACIT_UEXACT32           is_queued = 0, targets = 0;
    PACIT_DC_LIST           plink;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s enter CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));

    // validate tmf_link in target
    for(plink = ptarget->tmf_link.next; plink != &ptarget->tmf_link; plink = plink->next)
    {
        targets++;
        OSM_ASSERT(targets <= pcontroller->ptarget_table->occupied_entries);
    }
    targets = 0;

    if(pio_handle)
    {
        // Bug 32855 : Queue internal and external TMF aborts separately. External
        // TMFs were clobbering the queue if an internal TMF was already queued.
        if(pio_handle->ptmf_ws)
        {
            // External TMF. is_queued is not set in this path because the external
            // TMF is usually chained to an internal TMF already queued. Even if
            // the internal TMF is not already queued, it will be queued here and
            // then is_queued will be set.

            if(IS_LINK_INVALID(&pio_handle->ptmf_ws->controller_link))
            {
                //
                // it is an external TMF request with a valid WS. Queue the WS along
                // the controller link
                //
                LIST_ADDTO_TAIL(&pio_handle->ptmf_ws->controller_link, &pcontroller->tmf_queued_list);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("TMFWS 0x%p IOH 0x%p LUN 0x%p Target 0x%p c_link queued\n",
                    pio_handle->ptmf_ws, pio_handle, plun, ptarget));
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("TMFWS 0x%p IOH 0x%p LUN 0x%p Target 0x%p c_link already queued\n",
                    pio_handle->ptmf_ws, pio_handle, plun, ptarget));
            }
        }

        if(ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_TMF_ABORT_QUEUED))
        {
            // Resources were unavailable, queue I/O handle along TMF link
            LIST_ADDTO_TAIL(&pio_handle->tmf_link, &plun->tmf_abort_list);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Queued TMF Abort for IOH 0x%p LUN 0x%p Target 0x%p in internal queue\n",
                pio_handle, plun, ptarget));

            is_queued = 1;

            ACIT_BIT_SET(pio_handle->flags, FLAG_IOH_TMF_ABORT_QUEUED);
        }
        else
        {
            OSM_ASSERT(IS_LINK_VALID(&pio_handle->tmf_link));

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p LUN 0x%p Target 0x%p already queued\n",
                pio_handle, plun, ptarget));
        }
    }

    if(plun)
    {
        if(IS_LINK_INVALID(&plun->tmf_link))
        {
            OSM_ASSERT(!plun->counters.queued_tmfs);
            LIST_ADDTO_TAIL(&plun->tmf_link, &ptarget->tmf_lun_list);
            ptarget->counters.tmf_queued_luns++;
        }
        if(plun->plun_reset_tmf_request)
        {
            plun->counters.lun_reset_queued = ACIT_TRUE;
            ptmf_ws = GET_TMF_WS(plun->plun_reset_tmf_request->ptmf_key);
            if(IS_LINK_INVALID(&ptmf_ws->controller_link))
            {
                //
                // it is an external TMF request with a valid WS. Queue the WS along
                // the controller link
                //
                LIST_ADDTO_TAIL(&ptmf_ws->controller_link, &pcontroller->tmf_queued_list);
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Queued TMF LUN Reset for LUN 0x%p in internal queue\n", plun));
                is_queued = 1;
            }
        }

        if(is_queued)
            plun->counters.queued_tmfs++;
    }

    if(ptarget)
    {
        if(IS_LINK_INVALID(&ptarget->tmf_link))
        {
            OSM_ASSERT(!ptarget->counters.queued_tmfs);
            LIST_ADDTO_TAIL(&ptarget->tmf_link, &pcontroller->tmf_target_list);
            pcontroller->counters.tmf_queued_targets++;
        }
        if(ptarget->ptarget_reset_tmf_request)
        {
            ptarget->target_reset_queued = ACIT_TRUE;
            ptmf_ws = GET_TMF_WS(ptarget->ptarget_reset_tmf_request->ptmf_key);
            if(IS_LINK_INVALID(&ptmf_ws->controller_link))
            {
                //
                // it is an external TMF request with a valid WS. Queue the WS along
                // the controller link
                //
                LIST_ADDTO_TAIL(&ptmf_ws->controller_link, &pcontroller->tmf_queued_list);
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Queued TMF Target Reset for Target 0x%p in internal queue\n", ptarget));
                is_queued = 1;
            }
        }

        if(is_queued)
            ptarget->counters.queued_tmfs++;
    }

    if(is_queued)
        pcontroller->counters.queued_tmfs++;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s exit CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));

    // validate tmf_link in target
    for(plink = ptarget->tmf_link.next; plink != &ptarget->tmf_link; plink = plink->next)
    {
        targets++;
        OSM_ASSERT(targets <= pcontroller->ptarget_table->occupied_entries);
    }
}

ACIT_VOID
ap_dequeue_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle)
{
    PACIT_TMF_WS            ptmf_ws = ACIT_NULL;
    ACIT_UEXACT32           tmf_dequeue = ACIT_FALSE, targets = 0;
    PACIT_DC_LIST           plink;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s enter CWS Queued TMFs %d Target 0x%p Queued TMFs %d\n", 
        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs));

    // validate tmf_link in target
    for(plink = ptarget->tmf_link.next; plink != &ptarget->tmf_link; plink = plink->next)
    {
        targets++;
        OSM_ASSERT(targets <= pcontroller->ptarget_table->occupied_entries);
    }
    targets = 0;

    if(pio_handle)
    {
        OSM_ASSERT(ptarget);
        OSM_ASSERT(plun);

        ptmf_ws = pio_handle->ptmf_ws;

        LIST_REMOVE_ELEMENT(&pio_handle->tmf_link);
        LIST_INIT_ELEMENT(&pio_handle->tmf_link);

        ACIT_BIT_CLR(pio_handle->flags, FLAG_IOH_TMF_ABORT_QUEUED);

        tmf_dequeue = ACIT_TRUE;
        goto dequeue_lun;
    }

    if(plun)
    {
        OSM_ASSERT(plun->plun_reset_tmf_request);
        OSM_ASSERT(ptarget);

        ptmf_ws = GET_TMF_WS(plun->plun_reset_tmf_request->ptmf_key);
        plun->counters.lun_reset_queued = ACIT_FALSE;

        tmf_dequeue = ACIT_TRUE;
        goto dequeue_lun;
    }
    
    if(ptarget)
    {
        OSM_ASSERT(ptarget->ptarget_reset_tmf_request);
        ptmf_ws = GET_TMF_WS(ptarget->ptarget_reset_tmf_request->ptmf_key);
        ptarget->target_reset_queued = ACIT_FALSE;

        goto dequeue_target;
    }

    OSM_ASSERT(tmf_dequeue);

dequeue_lun:
    OSM_ASSERT(plun->counters.queued_tmfs);
    plun->counters.queued_tmfs--;

    if(!plun->counters.queued_tmfs)
    {
        LIST_REMOVE_ELEMENT(&plun->tmf_link);
        LIST_INIT_ELEMENT(&plun->tmf_link);
        LIST_INIT_HEADER(&plun->tmf_abort_list);
        ptarget->counters.tmf_queued_luns--;
    }

dequeue_target:
    OSM_ASSERT(ptarget->counters.queued_tmfs);
    ptarget->counters.queued_tmfs--;
    pcontroller->counters.queued_tmfs--;

    if(!ptarget->counters.queued_tmfs)
    {
        LIST_REMOVE_ELEMENT(&ptarget->tmf_link);
        LIST_INIT_ELEMENT(&ptarget->tmf_link);
        LIST_INIT_HEADER(&ptarget->tmf_lun_list);
        pcontroller->counters.tmf_queued_targets--;
    }

    if(ptmf_ws)
    {
        OSM_ASSERT(IS_LINK_VALID(&ptmf_ws->controller_link));
        LIST_REMOVE_ELEMENT(&ptmf_ws->controller_link);
        LIST_INIT_ELEMENT(&ptmf_ws->controller_link);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s exit CWS Queued TMFs %d Target 0x%p Queued TMFs %d dequeued %d\n", 
        __FUNCTION__, pcontroller->counters.queued_tmfs, ptarget, ptarget->counters.queued_tmfs, tmf_dequeue));

    // validate tmf_link in target
    for(plink = ptarget->tmf_link.next; plink != &ptarget->tmf_link; plink = plink->next)
    {
        targets++;
        OSM_ASSERT(targets <= pcontroller->ptarget_table->occupied_entries);
    }
}

ACIT_UEXACT32
ap_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_LUN_WS            plun;
    PACIT_IO_WS             pio;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_STATUS             status;

    ptarget = GET_TARGET_WS(pio_request->ptarget_key);
    plun = GET_LUN_WS(pio_request->plun_key);
    pio = GET_IO_WS(pio_request->pio_key);

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);
    OSM_ASSERT(plun);
    OSM_ASSERT(pio);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_ap_alloc_io_resources,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    if( (pcontroller->busy) ||
        (ptarget->busy)     ||
        (plun->busy)        )
    {
        /*
         * Can't send this I/O out, will have to wait, put this I/O in queue.
         * We will come back later and check to see if this I/O can be sent.
         */
        pio->pio_handle = ACIT_NULL;

        ap_queue_io_ws(pcontroller, ptarget, plun, pio);

        DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                            FUNC_ap_alloc_io_resources,
                            VASE_MODULE_ACIT,
                            (ACIT_UEXACT64) pcontroller->busy,
                            ptarget->busy,
                            plun->busy);
        return ACIT_STATUS_SUCCESS;
    }

    status = ap_update_io_ticks(pcontroller, ptarget, plun, pio);

    if(status != ACIT_STATUS_SUCCESS)
    {
        pio->pio_handle = ACIT_NULL;
        pio->pio_request->io_status = ACIT_IO_STATUS_IO_TIMEOUT;

        DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                            FUNC_ap_alloc_io_resources,
                            VASE_MODULE_ACIT,
                            (ACIT_UEXACT64) pcontroller->busy,
                            ptarget->busy,
                            plun->busy);

        return ACIT_STATUS_INSUFFICIENT_TIMEOUT;
    }

    pio_handle = ap_alloc_io_handle(pcontroller);

    /*
     * Make sure we got an IO Handle. It is illegal for this function to 
     * not return an IO Handle if controller is not busy.
     */
    OSM_ASSERT(pio_handle);

    pio->pio_handle = pio_handle;

    pio_handle->pio = pio;
    pio_handle->ptarget = ptarget;
    pio_handle->plun = plun;
    pio_handle->exp_xfer_length = pio_request->data_transfer_length;

#if ACIT_CFG_INTERRUPT_REDIRECTION
    pio_handle->message_id = pio_request->message_id;
    pio_handle->cq_id = HWI_GET_CQ_ID(pcontroller, pio_handle->message_id);
#endif

    status = iscsi_alloc_io_resources(pcontroller, ptarget, pio_handle);

    if(status == ACIT_STATUS_SUCCESS)
    {
        hwi_alloc_io_resources(pcontroller, pio_handle);
        iscsi_start_io(pcontroller, ptarget, pio_handle);

        pio_handle->starting_tick_count = CURRENT_TICK(pcontroller);

        /*
         * Resource allocation is complete, link the IO handle along target and lun WS
         */

        LIST_ADDTO_TAIL(&pio_handle->target_link,
                        &ptarget->io_handle_busy_list);
        LIST_ADDTO_TAIL(&pio_handle->lun_link,
                        &plun->io_handle_busy_list);

        /*
        * Increment pending_ios here, since we're all set with allocating resources
        */
        pcontroller->counters.pending_ios ++;
        ptarget->counters.pending_ios ++;
    }
    else
    {
        ap_free_io_handle(pcontroller, pio_handle);
        pio->pio_handle = ACIT_NULL;
        /*
         * Put the IO on the busy queue, we don't have enough credits now
         */
        ap_queue_io_ws(pcontroller, ptarget, plun, pio);
    }

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_ap_alloc_io_resources,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pcontroller->busy,
                        ptarget->busy,
                        plun->busy);
    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_ap_free_io_resources,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
    hwi_free_io_resources(pcontroller, pio_handle);
    iscsi_free_io_resources(pcontroller, pio_handle);
    ap_free_io_handle(pcontroller, pio_handle);

    OSM_ASSERT(pcontroller->counters.pending_ios);
    pcontroller->counters.pending_ios --;
    pio_handle->ptarget->counters.pending_ios --;
    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_ap_free_io_resources,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
}

ACIT_STATUS
ap_check_duplicate_target(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle,
    PACIT_UEXACT32          ptarget_index)
{
    PACIT_TARGET_ENTRY      ptarget_entry;
    ACIT_UEXACT32           index;
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];
    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++) {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
            ACIT_TFLAG_IN_USE) && 
        ptarget_entry[index].session_handle == session_handle){
                if(ptarget_index)
                    *ptarget_index = index;
                return ACIT_STATUS_DUPLICATE_ENTRY;
        }
    }

    if(ptarget_index)
        *ptarget_index = ACIT_U32_MINUS_ONE;
    return ACIT_STATUS_SUCCESS;
}

// No MCS support
#if 0
ACIT_STATUS
ap_init_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           connection_handle)
{
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           is_leading_cxn = 0;
    ACIT_UEXACT32           status;
    ACIT_UEXACT16           cq_id, def_hdr_id, def_data_id;
    PHYS_ADDR               template_address = {{{0,0}}};
    PPHYS_ADDR              ptemplate_address;

    pcxn = iscsi_get_cxn_from_cxn_handle(pcontroller, 
                                         ptarget, 
                                         connection_handle,
                                         &is_leading_cxn);

    ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_INITIALIZING);
    
    ptemplate_address = &template_address;
    ISCSI_GET_PDU_TEMPLATE_ADDRESS(pcontroller, ptemplate_address);
    cq_id = (ACIT_UEXACT16)HWI_GET_CQ_ID(pcontroller, 0);
    def_hdr_id = (ACIT_UEXACT16)HWI_GET_DEF_HDRQ_ID(pcontroller);
    def_data_id = (ACIT_UEXACT16)HWI_GET_DEF_BUFQ_ID(pcontroller);

    if (pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED) 
        status = mgmt_offload_cxn(pcontroller, 
                                    ptarget, 
                                    ptarget->piscsi_ws->session_handle,
                                    pcxn->connection_handle,
                                    pcxn->cid,
                                    cq_id,
                                    def_hdr_id,
                                    def_data_id,
                                    ptemplate_address,
                                    ACIT_FALSE   // synchronous
                                    );
    else
        status = ACIT_STATUS_SUCCESS; 

    return status;
}

ACIT_STATUS
ap_add_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle, 
    ACIT_UEXACT32           connection_handle)
{
    PACIT_TARGET_WS         ptarget;
    PISCSI_TARGET_WS        piscsi_target;    
    PISCSI_CONNECTION       pcxn;

    ptarget = ap_get_target_ws_from_session_handle(pcontroller, session_handle);

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    // Initialize a connection to send out I/Os later
    pcxn = iscsi_alloc_cxn(pcontroller);

    if(pcxn == ACIT_NULL)
    {
        OSM_ASSERT(0);
        return ISCSI_STATUS_NO_CONNECTIONS;
    }

    piscsi_target->num_connections++;
    piscsi_target->pcxn_list[piscsi_target->num_connections] = pcxn;
    pcxn->connection_handle = connection_handle;
    pcxn->piscsi_target_ws = piscsi_target;

    return ACIT_STATUS_SUCCESS;
}
#endif

ACIT_STATUS
ap_add_target(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle, 
    ACIT_UEXACT8            domain,
    ACIT_UEXACT16           error_status,
    ACIT_UEXACT8            is_login_pending)
{
    ACIT_UEXACT32           index;
    PACIT_TARGET_ENTRY      ptarget_entry;

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    pcontroller->ptarget_table->change_count ++;

    /*
     * Find a free slot on the TARGET_TABLE and add the target there.
     */
    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_CLR(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            break;
        }
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("index %d\n", index));

    if(index < pcontroller->params.targets_per_controller)
    {
        /*
         * We found an empty entry that we can use.
         */
        pcontroller->ptarget_table->occupied_entries ++;
        pcontroller->counters.current_num_targets = pcontroller->ptarget_table->occupied_entries;
        ptarget_entry = &ptarget_entry[index];
        ptarget_entry->ptarget_key = &pcontroller->ptarget_key[index];
        ptarget_entry->session_handle = session_handle;
        ptarget_entry->domain = domain;
        ACIT_BIT_SET(ptarget_entry->acit_tflags,
                        (ACIT_TFLAG_IN_USE | ACIT_TFLAG_NEW));

        //
        // Make a note of the source of this target. Whether its from a
        // login IOCTL or from a persistent target
        //
        ptarget_entry->is_login_pending = is_login_pending;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("index %d\n", index));

        /* Notify OSM that a new target is available */
        OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                        ACIT_EVENT_CLASS_TARGET,
                        ACIT_EVENT_TARGET_ADD,
                        index);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_delete_target_entry(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           target_index)
{
    PACIT_TARGET_ENTRY      ptarget_entry;

    //
    // This routine is called to delete the target entry from ACIT
    // target table. This may be called when a target is logged out from 
    // a driver or a firmware owned data session
    //

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];


	// Under certain conditions the logout may be issued while the login process is
	// still incomplete. Check for this here and avoid async notification, BTL manipulation etc.
    if ((ACIT_NULL == pcontroller->ptarget_key[target_index].acit_key) ||
		(ACIT_NULL == ptarget_entry[target_index].ptarget_key ))
    {
       OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Delete target skipping async notification -Index:%d\n", target_index));
       return;
    }


    //
    // Notify OSM that the target has been removed 
    // This has to be done before we delete the entry
    //
    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_TARGET,
                    ACIT_EVENT_TARGET_REMOVE,
                    target_index);

    ptarget_entry[target_index].acit_tflags &=~ ACIT_TFLAG_IN_USE;

    ptarget_entry[target_index].ptarget_key = ACIT_NULL;
    ptarget_entry[target_index].session_handle = 0;
    ptarget_entry[target_index].acit_tflags = 0;
    ptarget_entry[target_index].is_login_pending = 0;

    pcontroller->ptarget_key[target_index].acit_key = ACIT_NULL;

    pcontroller->ptarget_table->change_count ++;
    pcontroller->ptarget_table->occupied_entries --;
    pcontroller->counters.current_num_targets = pcontroller->ptarget_table->occupied_entries;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("deleting target index %d\n", target_index));

    return;
}

ACIT_STATUS
ap_remove_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    ap_free_queued_ws_resources(pcontroller, 
                                ptarget, 
                                ACIT_TRUE);

    // if there are any MGMT IOCTLs pending, then return them now
    mgmt_return_pending_ioctls(pcontroller, ptarget);

    OSM_ASSERT(ptarget->counters.num_io_invs == 0);

    //
    // Logout pending is an unhandled case, it will happen if logout was sent
    // but no responses came from microcode (because the target didn't respond?).
    // However a FIN/RST arrived and the target is being removed
    // ASSERT for now
    //
    //OSM_ASSERT(ACIT_IS_BIT_CLR(ptarget->flags, FLAG_TGT_LOGOUT_RESPONSE_PENDING));
    OSM_ASSERT(ptarget->sub_state != TARGET_SUB_STATE_LOGOUT_PENDING);
    
    // Clear any pending events in the target
    if(IS_LINK_VALID(&ptarget->event_link))
    {
        ap_clear_target_event(pcontroller, ptarget, FLAG_TGT_ALL);
    }
    
    ap_delete_target_entry(pcontroller, ptarget->target_table_index);

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_delete_all_targets(
    PACIT_CONTROLLER_WS         pcontroller)
{
    ACIT_UEXACT32               index = 0;
    PACIT_TARGET_ENTRY          ptarget_entry = ACIT_NULL;
    ACIT_UEXACT32               count = 0;
    volatile ACIT_UEXACT32      poll_count = 0;
    PACIT_TARGET_WS             ptarget;

    //
    // if LDTO has expired, then there are no sessions though
    // the target entries are all valid
    //
    // Bug 14036 - 07/21/09 - do not return here if LDTO has expired
    // Linux OSM requires ACIT to call delete target routine
    // in order to cleanup memory allocated to targets.
    //if(IS_CONTROLLER_LDTO_EXPIRED(pcontroller, 0))
    //{
    //    return ACIT_STATUS_SUCCESS; 
    //}
    
    //
    // Wait for any pending IOCTLs so that they do not interfere
    // with the logout process later. When all IOCTLs are
    // completed, the target states will be clear
    //
    while(pcontroller->counters.pending_ioctls && 
          (pcontroller->eh_stats.ue_detected != 1))
    {
        OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
        acit_process_irq(pcontroller->pcontroller_key);
        OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
        poll_count++;
        if(poll_count == 500)
        {
            // run UE detection since a UE will stall any
            // pending IOCTLs
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Waiting for %d ioctls to complete..\n", pcontroller->counters.pending_ioctls));
            hwi_check_health(pcontroller, ACIT_TRUE);
            poll_count = 0;
        }
    }

    if(pcontroller->eh_stats.ue_detected)
    {
        // Bail out now
        return ACIT_STATUS_SUCCESS; 
    }

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        ((index < pcontroller->params.targets_per_controller) & (pcontroller->eh_stats.ue_detected == 0));
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            count = pcontroller->ptarget_table->occupied_entries;

            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Skipping index %d\n", index));
                continue;
            }

            if(ptarget->initialized == 0)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Remove target index %d target 0x%p busy 0x%x substate 0x%x\n", 
                    index, ptarget, ptarget->busy, ptarget->sub_state));

                mgmt_logout_fw_session(pcontroller,
                                       ptarget);

                //
                // 09/04/2010 - I believe there's a bug here that the connection 
                // structure is not freed. But since we're in shutdown path it
                // should be ok
                //                  
                ap_remove_target(pcontroller, ptarget);
                continue;
            }

            if(!(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY)))
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issue logout for index %d target 0x%p busy 0x%x substate 0x%x\n", 
                    index, ptarget, ptarget->busy, ptarget->sub_state));

                iscsi_logout_session(pcontroller, 
                                     ptarget,
                                     ACIT_NULL);
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Delete wait for index %d target 0x%p busy 0x%x substate 0x%x\n", 
                    index, ptarget, ptarget->busy, ptarget->sub_state));
            }

            //
            // Polled mode, the occupied entries change when a target gets
            // deleted
            //
            poll_count = 0;
            while(count == pcontroller->ptarget_table->occupied_entries && 
                    (pcontroller->eh_stats.ue_detected != 1))
            {
                OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
                acit_process_irq(pcontroller->pcontroller_key);
                OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
                poll_count++;
                if(poll_count == 500)
                {
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Waiting for target# %d to be deleted..\n", index));
                    acit_process_timer_tick(pcontroller->pcontroller_key);
                    poll_count = 0;
                }
            }
        }
    }
    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_return_queued_abort_tmfs(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun,
    ACIT_UEXACT32               tmf_status,
    ACIT_UEXACT32               free_handle)
{
    PACIT_DC_LIST               pabort_link;
    PACIT_IO_HANDLE             pio_handle;
    ACIT_UEXACT32               count = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Enter: Target 0x%p has %d queued TMFs remaining free_handle %d\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs, free_handle));

    if(free_handle)
    {
        if(IS_LIST_NOT_EMPTY(&plun->tmf_abort_list))
        {
            pabort_link = plun->tmf_abort_list.next;

            while(pabort_link != &plun->tmf_abort_list)
            {
                pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                        tmf_link,
                                        pabort_link);

                OSM_ASSERT(pio_handle);

                // double free?
                OSM_ASSERT(ACIT_IS_BIT_CLR(pio_handle->flags, FLAG_IOH_IO_FREE));

                pabort_link = pabort_link->next;

                ap_dequeue_tmf(pcontroller, ptarget, plun, pio_handle);

                // The IOWs may be freed here, but the IO Handle is always valid
                // on return
                ap_decide_io_disposition_in_cleanup(pcontroller, pio_handle);

                if(pio_handle->ptmf_ws)
                {
                    ap_return_tmf(pcontroller, 
                                pio_handle->ptmf_ws->ptmf_request,
                                tmf_status);

                    pio_handle->ptmf_ws = ACIT_NULL;
                }

                ap_free_io_resources(pcontroller, pio_handle);
                count++;
            }
        }

        LIST_INIT_HEADER(&plun->tmf_abort_list);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Returned %d TMFs in LUN 0x%p, remaining %d\n", count, plun, plun->counters.queued_tmfs));
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Exit: Target 0x%p has %d queued TMFs remaining\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs));

    return ACIT_STATUS_SUCCESS;

}

ACIT_UEXACT32
ap_return_queued_lun_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           tmf_status,
    ACIT_UEXACT32           free_handle)
{
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Enter: Target 0x%p has %d queued TMFs remaining free_handle %d\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs, free_handle));

    ap_return_queued_abort_tmfs(pcontroller, 
                                ptarget, 
                                plun, 
                                tmf_status,
                                free_handle);

    if(plun->counters.lun_reset_queued)
    {
        OSM_ASSERT(plun->plun_reset_tmf_request);

        ap_dequeue_tmf(pcontroller, ptarget, plun, ACIT_NULL);

        ap_return_tmf(pcontroller, 
                        plun->plun_reset_tmf_request,
                        tmf_status);

        plun->plun_reset_tmf_request = ACIT_NULL;

        if(IS_LIST_NOT_EMPTY(&plun->io_handle_busy_list))
        {
            plun->eh_stats.io_busy_during_lun_rst_timeout++;
            ptarget->eh_stats.io_busy_during_lun_rst_timeout++;
            pcontroller->eh_stats.io_busy_during_lun_rst_timeout++;
        }

        if(IS_LIST_NOT_EMPTY(&plun->io_queued_list))
        {
            plun->eh_stats.io_queued_during_lun_rst_timeout++;
            ptarget->eh_stats.io_queued_during_lun_rst_timeout++;
            pcontroller->eh_stats.io_queued_during_lun_rst_timeout++;
        }

        if(IS_LIST_NOT_EMPTY(&plun->io_pending_reset_list))
        {
            plun->eh_stats.io_rstpend_during_lun_rst_timeout++;
            ptarget->eh_stats.io_rstpend_during_lun_rst_timeout++;
            pcontroller->eh_stats.io_rstpend_during_lun_rst_timeout++;
        }
    }

    if(free_handle)
    {
        OSM_ASSERT(!plun->counters.queued_tmfs);
        OSM_ASSERT(IS_LINK_INVALID(&plun->tmf_link));
    }

    //
    // mark the LUN as available. It is ok for the LUN state to be free
    // since a pending session recovery or ETO in progress will keep the
    // IOs and TMFs at bay..
    //
    plun->flags = 0;
    ap_flag_ready_lun(pcontroller, plun, LUN_BUSY_RESET_IN_PROGRESS);

    if(!ptarget->counters.queued_tmfs)
    {
        OSM_ASSERT(IS_LINK_INVALID(&ptarget->tmf_link));
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("LUN 0x%p has remaining %d TMFs\n", plun, plun->counters.queued_tmfs));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Exit: Target 0x%p has %d queued TMFs remaining\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_return_queued_target_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           tmf_status,
    ACIT_UEXACT32           free_handle)
{
    PACIT_DC_LIST           plun_link;
    PACIT_LUN_WS            plun;
    ACIT_UEXACT32           count = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Enter: Target 0x%p has %d queued TMFs remaining free_handle %d\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs, free_handle));

    if(IS_LIST_NOT_EMPTY(&ptarget->tmf_lun_list))
    {
        plun_link = ptarget->tmf_lun_list.next;

        /*
         * Loop and travel this linked list.
         */
        while(plun_link != &ptarget->tmf_lun_list)
        {
            plun = STRUCT_BASE(ACIT_LUN_WS,
                               tmf_link,
                               plun_link);

            plun_link = plun_link->next;
            OSM_ASSERT(plun->counters.queued_tmfs);

            ap_return_queued_lun_tmfs(pcontroller, 
                                      ptarget, 
                                      plun, 
                                      tmf_status,
                                      free_handle);

            count++;
        }
    }

    if(ptarget->target_reset_queued)
    {
        OSM_ASSERT(ptarget->ptarget_reset_tmf_request);

        ap_dequeue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);

        ap_return_tmf(pcontroller, 
                        ptarget->ptarget_reset_tmf_request,
                        tmf_status);
        
        ptarget->ptarget_reset_tmf_request = ACIT_NULL;

        if(IS_LIST_NOT_EMPTY(&ptarget->io_handle_busy_list))
        {
            ptarget->eh_stats.io_busy_during_tgt_rst_timeout++;
            pcontroller->eh_stats.io_busy_during_tgt_rst_timeout++;
        }

        if(IS_LIST_NOT_EMPTY(&ptarget->io_queued_list))
        {
            ptarget->eh_stats.io_queued_during_tgt_rst_timeout++;
            pcontroller->eh_stats.io_queued_during_tgt_rst_timeout++;
        }

        if(IS_LIST_NOT_EMPTY(&ptarget->io_pending_reset_list))
        {
            ptarget->eh_stats.io_rstpend_during_tgt_rst_timeout++;
            pcontroller->eh_stats.io_rstpend_during_tgt_rst_timeout++;
        }
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Traversed %d LUNs to cleanup\n", count));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Exit: Target 0x%p has %d queued TMFs remaining\n",
        __FUNCTION__, ptarget, ptarget->counters.queued_tmfs));

    if(free_handle)
    {
        OSM_ASSERT(!ptarget->counters.queued_tmfs);
    }
    
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_return_queued_controller_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           tmf_status,
    ACIT_UEXACT32           free_handle)
{
    PACIT_DC_LIST           ptarget_link;
    PACIT_TARGET_WS         ptarget;
    ACIT_UEXACT32           count = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Enter: Controller has %d queued TMFs remaining free_handle %d\n",
        __FUNCTION__, pcontroller->counters.queued_tmfs, free_handle));

    if(IS_LIST_NOT_EMPTY(&pcontroller->tmf_target_list))
    {
        ptarget_link = pcontroller->tmf_target_list.next;

        /*
         * Loop and travel this linked list.
         */
        while(ptarget_link != &pcontroller->tmf_target_list)
        {
            ptarget = STRUCT_BASE(ACIT_TARGET_WS,
                               tmf_link,
                               ptarget_link);

            ptarget_link = ptarget_link->next;

            ap_return_queued_target_tmfs(pcontroller, 
                                         ptarget, 
                                         tmf_status,
                                         free_handle);

            OSM_ASSERT(count < pcontroller->ptarget_table->occupied_entries);
            count++;
        }
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Traversed %d targets to cleanup\n", count));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("%s Exit: Controller has %d queued TMFs remaining\n",
        __FUNCTION__, pcontroller->counters.queued_tmfs));

    if(free_handle)
    {
        OSM_ASSERT(!pcontroller->counters.queued_tmfs);
        LIST_INIT_HEADER(&pcontroller->tmf_target_list);
    }

    return ACIT_STATUS_SUCCESS;

}

ACIT_UEXACT32
ap_flush_lun_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    PACIT_DC_LIST           pabort_link;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_UEXACT32           count = 0, i;

    if (plun->plun_reset_tmf_request)
    {
        if(ACIT_IS_BIT_CLR(plun->busy, LUN_BUSY_RESET_SUBMITTED))
        {
            ap_dequeue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
            if(ap_check_and_issue_reset(pcontroller, 
                                        ptarget, 
                                        plun) != ACIT_STATUS_SUCCESS)
            {
                ap_queue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
            }
        }
        return ACIT_STATUS_SUCCESS;
    }

    pabort_link = plun->tmf_abort_list.next;
    count = plun->counters.queued_tmfs;

    for(i = 0; i < count; i++)
    {
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE, 
                                 tmf_link, 
                                 pabort_link);
            
        OSM_ASSERT(pio_handle);

        pabort_link = pabort_link->next;
        OSM_ASSERT(pabort_link);

        if(ap_ready_for_tmf(pcontroller, ptarget) == ACIT_STATUS_SUCCESS)
        {
            ap_dequeue_tmf(pcontroller, ptarget, plun, pio_handle);

            if (ap_check_and_issue_abort(pcontroller, 
                                        ptarget, 
                                        plun, 
                                        pio_handle) != ACIT_STATUS_SUCCESS)
            {
                ap_queue_tmf(pcontroller, ptarget, plun, pio_handle);
            }
        }
        else
        {
            break;
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_flush_target_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PACIT_DC_LIST           plun_link;
    PACIT_LUN_WS            plun = ACIT_NULL;
    ACIT_UEXACT32           i;

    if(ptarget->ptarget_reset_tmf_request)
    {
        if(!ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_RESET_START))
        {
            ap_dequeue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
            if(ap_check_and_issue_reset(pcontroller, 
                                        ptarget, 
                                        ACIT_NULL) != ACIT_STATUS_SUCCESS)
            {
                ap_queue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
            }
        }
        return ACIT_STATUS_SUCCESS;
    }

    plun_link = ptarget->tmf_lun_list.next;

    for(i = 0; i < ptarget->counters.tmf_queued_luns; i++)
    {
        if(ap_ready_for_tmf(pcontroller, ptarget) == ACIT_STATUS_SUCCESS)
        {
            plun = STRUCT_BASE(ACIT_LUN_WS,
                               tmf_link,
                               plun_link);

            plun_link = plun_link->next;
            OSM_ASSERT(plun_link);

            OSM_ASSERT(plun->counters.queued_tmfs);
            ap_flush_lun_tmfs(pcontroller, ptarget, plun);
        }
        else
        {
            break;
        }
    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_flush_controller_tmfs(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_DC_LIST           ptarget_link;
    PACIT_TARGET_WS         ptarget = ACIT_NULL;

    ACIT_UEXACT32           targets = 0;
    PACIT_DC_LIST           plink;
    ACIT_UEXACT32           i;

    ptarget_link = pcontroller->tmf_target_list.next;

    for(i = 0; i < pcontroller->counters.tmf_queued_targets; i++)
    {
        ptarget = STRUCT_BASE(ACIT_TARGET_WS, 
                              tmf_link, 
                              ptarget_link);

        ptarget_link = ptarget_link->next;
        OSM_ASSERT(ptarget_link);
        OSM_ASSERT(ptarget->counters.queued_tmfs);

        if(ap_ready_for_tmf(pcontroller, ptarget) == ACIT_STATUS_SUCCESS)
        {
            ap_flush_target_tmfs(pcontroller, ptarget);
        }
    }

    if(IS_LIST_NOT_EMPTY(&pcontroller->tmf_target_list))
    {
        pcontroller->state.flush_queued_tmfs = ACIT_TRUE;
        OSM_ASSERT(pcontroller->counters.queued_tmfs);
    }
    else
    {
        /*
         * All pending TMFs are completed, so clear the flush_queued_tmfs
         * status
         */
        pcontroller->state.flush_queued_tmfs = ACIT_FALSE;
        OSM_ASSERT(!pcontroller->counters.queued_tmfs);
    }


    if(ptarget)
    {
        // validate tmf_link in target
        for(plink = ptarget->tmf_link.next; plink != &ptarget->tmf_link; plink = plink->next)
        {
            targets++;
            OSM_ASSERT(targets <= pcontroller->ptarget_table->occupied_entries);
        }
    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
ap_flush_queued_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         presume_target,
    PACIT_LUN_WS            presume_lun)
{
    ACIT_DC_LIST            temp_list;
    PACIT_DC_LIST           pio_link;
    PACIT_IO_WS             pio;
    PACIT_IO_REQUEST        pio_request;
    PACIT_TARGET_WS         ptarget;
    PACIT_LUN_WS            plun;
    ACIT_UEXACT16           num_flushed = 0;
    ACIT_UEXACT16           num_queued = 0;
    ACIT_UEXACT32           status;

    OSM_ASSERT(pcontroller);

#if ACIT_CFG_DEBUG_BUILD
    ap_validate_lists(pcontroller, ACIT_NULL, ACIT_NULL);
#endif

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Flushing controller IOs..\n"));

    /*
     * This function flushes the pending I/Os from IO WS queue.
     * We could've saved on PCI writes by accumulating all requests
     * and posting them in one go. But since we don't sort the
     * requests based on their CIDs, we don't post multiple requests
     * either
     */

    /* Initialize the linked list and move all the queued IOs to temp list */
    LIST_INIT_HEADER(&temp_list);
    LIST_MOVE(&temp_list, &pcontroller->io_queued_list);

    while( (IS_LIST_NOT_EMPTY(&temp_list)) && (!pcontroller->busy) )
    {
        LIST_REMOVEFROM_HEAD(&pio_link, &temp_list);
        OSM_ASSERT(pio_link);

        pio = STRUCT_BASE(  ACIT_IO_WS,
                            controller_link,
                            pio_link);

        pio_request = (PACIT_IO_REQUEST)pio->pio_request;

        ptarget = GET_TARGET_WS(pio_request->ptarget_key);
        plun = GET_LUN_WS(pio_request->plun_key);

        if( (!ptarget->busy) &&
            (!plun->busy))
        {
            /*
             * When we put IO in busy list we also put it in targets and Luns
             * queued list, since we are restarting this I/O, remove from list
             */
            LIST_REMOVE_ELEMENT(&pio->target_link);
            LIST_REMOVE_ELEMENT(&pio->lun_link);

            /*
             * Call ap_alloc_io_resources to start the IO again.
             */
            status = ap_alloc_io_resources(pcontroller, pio->pio_request);

            /*
            * Fill-up the hardware structs and submit it to H/W if we were able to get
            * all resources for this IO.
            */
            if(pio->pio_handle)
            {
                PACIT_IO_HANDLE pio_handle = pio->pio_handle;

                pio->queued = 0;
                hwi_fill_io_request(pcontroller->phwi_ws, pio_handle);
                hwi_submit_request(pcontroller, pio_handle->pwrb_handle, pio_handle->cri);
                ACIT_BIT_SET(pio_handle->flags, FLAG_IOH_IO_SUBMITTED);
                num_flushed ++;
            }
            else
            {
                if(status == ACIT_STATUS_INSUFFICIENT_TIMEOUT)
                {
                    ap_return_io(pcontroller,
                                 pio,
                                 ACIT_IO_STATUS_IO_TIMEOUT,
                                 ACIT_TRUE);
                }
                else
                {
                    num_queued ++;
                }
            }
        }
        else
        {
            /*
             * Devices are busy and request cannot be restarted right now
             * put this I/O request back on the queue, try it later.
             */
            LIST_ADDTO_TAIL(&pio->controller_link, 
                            &pcontroller->io_queued_list);
        }
    }

    /*
     * We broke out of the loop see if some IOs are still pending in temp_list
     * because the controller is busy. If so put them back on the queued list.
     */
    if(IS_LIST_NOT_EMPTY(&temp_list))
    {
        LIST_MOVETO_TAIL(&pcontroller->io_queued_list, &temp_list);
    }

    if(IS_LIST_NOT_EMPTY(&pcontroller->io_queued_list))
    {
        pcontroller->state.flush_queued_ios = ACIT_TRUE;
    }
    else
    {
        /*
         * All pending IOs are completed, so clear the flush_queued_ios
         * status
         */
        pcontroller->state.flush_queued_ios = ACIT_FALSE;
        OSM_ASSERT(!(pcontroller->counters.queued_ios - (num_flushed+num_queued)));
    }

    pcontroller->counters.queued_ios -=(num_flushed + num_queued);

    ACIT_UPDATE_FLUSH_IO_COUNTERS(pcontroller, num_flushed, num_queued);

#if ACIT_CFG_DEBUG_BUILD
    ap_validate_lists(pcontroller, ACIT_NULL, ACIT_NULL);
#endif

    return ACIT_STATUS_SUCCESS;
}

PACIT_IO_HANDLE
ap_get_io_handle(
    PACIT_DC_LIST           link,
    ACIT_UEXACT32           toremove)
{
    PACIT_IO_HANDLE         pio_handle;

    switch(toremove)
    {
    case LUN_QUEUE:
        
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                            lun_link,
                            link);
        break;
    case TARGET_QUEUE:
        
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                            target_link,
                            link);
        break;
    default:  
        
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                            controller_link,
                            link);
        break;
    }

    return pio_handle;
}

PACIT_TMF_HANDLE
ap_get_tmf_handle(
    PACIT_DC_LIST           link,
    ACIT_UEXACT32           toremove)
{
    PACIT_TMF_HANDLE        ptmf_handle;

    switch(toremove)
    {
    case LUN_QUEUE:
        
        ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                  lun_link,
                                  link);
        break;
    case TARGET_QUEUE:
        
        ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                  target_link,
                                  link);
        break;
    default:  
        
        ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                  controller_link,
                                  link);
        break;
    }
    
    return ptmf_handle;
}

PACIT_NOP_HANDLE
ap_get_nop_handle(
    PACIT_DC_LIST           link,
    ACIT_UEXACT32           toremove)
{
    PACIT_NOP_HANDLE        pnop_handle = ACIT_NULL;

    switch(toremove)
    {
    case LUN_QUEUE:

        OSM_ASSERT(0);
        break;
    case TARGET_QUEUE:
        
        pnop_handle = STRUCT_BASE(ACIT_NOP_HANDLE,
                                  target_link,
                                  link);
        break;
    default:  
        
        pnop_handle = STRUCT_BASE(ACIT_NOP_HANDLE,
                                  controller_link,
                                  link);
        break;
    }
    
    return pnop_handle;
}

PACIT_LOGOUT_HANDLE
ap_get_logout_handle(
    PACIT_DC_LIST           link,
    ACIT_UEXACT32           toremove)
{
    PACIT_LOGOUT_HANDLE     plogout_handle = ACIT_NULL;

    switch(toremove)
    {
    case LUN_QUEUE:

        OSM_ASSERT(0);
        break;

    case TARGET_QUEUE:

        plogout_handle = STRUCT_BASE(ACIT_LOGOUT_HANDLE,
                                     target_link,
                                     link);
        break;

    default:  
        
        plogout_handle = STRUCT_BASE(ACIT_LOGOUT_HANDLE,
                                     controller_link,
                                     link);
        break;
    }
    
    return plogout_handle;
}

PACIT_IO_WS
ap_get_io_ws(
    PACIT_DC_LIST           pio_link,
    ACIT_UEXACT32           toremove)
{
    PACIT_IO_WS             pio = ACIT_NULL;

    switch(toremove)
    {
    case LUN_QUEUE:

        pio = STRUCT_BASE(ACIT_IO_WS, lun_link, pio_link);
        break;

    case TARGET_QUEUE:

        pio = STRUCT_BASE(ACIT_IO_WS, target_link, pio_link);
        break;

    case CONTROLLER_QUEUE:

        pio = STRUCT_BASE(ACIT_IO_WS, controller_link, pio_link);
        break;
    }

    return pio;
}

ACIT_UEXACT32
ap_return_queued_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           io_status,
    ACIT_UEXACT32           port_num)
{
    ACIT_UEXACT32           toremove = 0;
    ACIT_DC_LIST            temp_list = {ACIT_NULL,ACIT_NULL};
    PACIT_DC_LIST           pio_link;
    PACIT_IO_WS             pio = ACIT_NULL;

    if(plun)
    {
        toremove = LUN_QUEUE;
        LIST_MOVE(&temp_list, &plun->io_queued_list);
    }
    else if(ptarget)
    {
        toremove = TARGET_QUEUE;
        LIST_MOVE(&temp_list, &ptarget->io_queued_list);
    }
    else if(pcontroller)
    {
        toremove = CONTROLLER_QUEUE;
        LIST_MOVE(&temp_list, &pcontroller->io_queued_list);
    }

    if(IS_LIST_NOT_EMPTY(&temp_list))
    {
        pio_link = temp_list.next;

        /*
         * Loop and travel this linked list.
         */
        while(pio_link != &temp_list)
        {
            OSM_ASSERT(pio_link);

            pio = ap_get_io_ws(pio_link, toremove);

            /* Move the next pointer before modifying the list */
            pio_link = pio_link->next;

            OSM_ASSERT(pio->pio_request);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Returning IOWS 0x%p queued. target 0x%p lun 0x%p available ticks %d internal ticks %d\n",
                pio, ptarget, plun, pio->available_ticks, pio->internal_available_ticks));

            ap_return_io(pcontroller, pio, io_status, ACIT_TRUE);
        }
    }

#if ACIT_CFG_DEBUG_BUILD
    ap_validate_lists(pcontroller, ACIT_NULL, ACIT_NULL);
#endif

    return ACIT_STATUS_SUCCESS;
}

#if ACIT_CFG_DEBUG_BUILD

ACIT_VOID
ap_validate_iows_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    ACIT_UEXACT32           cws_controller_ios = 0, lws_lun_ios = 0, tws_target_ios = 0;
    PACIT_DC_LIST           piows_link;
    PACIT_IO_WS             pio;

    for(piows_link = pcontroller->io_queued_list.next;
        piows_link != &pcontroller->io_queued_list;
        piows_link = piows_link->next) 
    {
        pio = STRUCT_BASE(ACIT_IO_WS,
                          controller_link,
                          piows_link);

        OSM_ASSERT(pio);

        cws_controller_ios++;
    }

    if(ptarget)
    {
        for(piows_link = ptarget->io_queued_list.next;
            piows_link != &ptarget->io_queued_list;
            piows_link = piows_link->next) 
        {
            pio = STRUCT_BASE(ACIT_IO_WS,
                            target_link,
                            piows_link);

            OSM_ASSERT(pio);
            tws_target_ios++;
        }
    }

    if(plun)
    {
        for(piows_link = plun->io_queued_list.next;
            piows_link != &plun->io_queued_list;
            piows_link = piows_link->next) 
        {
            pio = STRUCT_BASE(ACIT_IO_WS,
                              lun_link,
                              piows_link);

            OSM_ASSERT(pio);

            lws_lun_ios++;
        }
    }

    OSM_ASSERT(pcontroller->counters.queued_ios == cws_controller_ios);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Queued IO count CW 0x%p - %d TWS 0x%p - %d 0x%p - %d\n", 
                                            pcontroller, cws_controller_ios, ptarget, tws_target_ios, plun, lws_lun_ios));
}

ACIT_VOID
ap_validate_lists(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_LUN_WS            plun)
{
    //
    // this routine is specifically to compare the different lists from different
    // WS structures to ensure consistency between all WS strcuts.
    // Right now we check only IO ws lists...
    //

    ap_validate_iows_list(pcontroller, ptarget, plun);
}

ACIT_VOID
ap_print_cmd_ticks(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_LUN_WS            plun)
{
    PACIT_DC_LIST           piows_link, pioh_link;
    PACIT_IO_WS             pio;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_UEXACT32           ios = 0;

    if(plun)
    {
        for(piows_link = plun->io_queued_list.next;
            piows_link != &plun->io_queued_list;
            piows_link = piows_link->next) 
        {
            pio = STRUCT_BASE(ACIT_IO_WS,
                              lun_link,
                              piows_link);

            OSM_ASSERT(pio);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOWS 0x%p Avail %d Internal Avail %d\n",
                pio, pio->available_ticks, pio->internal_available_ticks));

            ios++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("LUN 0x%p has %d IOS in internal list\n",
            plun, ios));
        ios = 0;

        for(pioh_link = plun->io_handle_busy_list.next;
            pioh_link != &plun->io_handle_busy_list;
            pioh_link = pioh_link->next) 
        {
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                     lun_link,
                                     pioh_link);

            OSM_ASSERT(pio_handle);

            if(pio_handle->pio)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p IOWS 0x%p Avail %d Internal Avail %d\n",
                    pio_handle, pio_handle->pio, pio_handle->pio->available_ticks, pio_handle->pio->internal_available_ticks));
            }
            else 
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p time taken %d\n",
                    pio_handle, TIME_SPENT(pcontroller, pio_handle->starting_tick_count)));
            }
            ios++;
        }
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("LUN 0x%p has %d IOS in pending list\n",
            plun, ios));
        ios = 0;

    }
    else if(ptarget)
    {
        for(piows_link = ptarget->io_queued_list.next;
            piows_link != &ptarget->io_queued_list;
            piows_link = piows_link->next) 
        {
            pio = STRUCT_BASE(ACIT_IO_WS,
                            target_link,
                            piows_link);

            OSM_ASSERT(pio);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IO 0x%p Avail %d Internal Avail %d\n",
                pio, pio->available_ticks, pio->internal_available_ticks));

            ios++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p has %d IOS in internal list\n",
            ptarget, ios));
        ios = 0;

        for(pioh_link = ptarget->io_handle_busy_list.next;
            pioh_link != &ptarget->io_handle_busy_list;
            pioh_link = pioh_link->next) 
        {
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                     target_link,
                                     pioh_link);

            OSM_ASSERT(pio_handle);

            if(pio_handle->pio)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p IOWS 0x%p Avail %d Internal Avail %d\n",
                    pio_handle, pio_handle->pio, pio_handle->pio->available_ticks, pio_handle->pio->internal_available_ticks));
            }
            else 
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p time taken %d\n",
                    pio_handle, TIME_SPENT(pcontroller, pio_handle->starting_tick_count)));
            }
            ios++;
        }
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p has %d IOS in pending list\n",
            ptarget, ios));
        ios = 0;
    }
    else
    {
        for(piows_link = pcontroller->io_queued_list.next;
            piows_link != &pcontroller->io_queued_list;
            piows_link = piows_link->next) 
        {
            pio = STRUCT_BASE(ACIT_IO_WS,
                            controller_link,
                            piows_link);

            OSM_ASSERT(pio);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IO 0x%p Avail %d Internal Avail %d\n",
                pio, pio->available_ticks, pio->internal_available_ticks));

            ios++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Controller 0x%p has %d IOS in pending list\n",
            pcontroller, ios));
        ios = 0;

        for(pioh_link = pcontroller->io_handle_busy_list.next;
            pioh_link != &pcontroller->io_handle_busy_list;
            pioh_link = pioh_link->next) 
        {
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                     controller_link,
                                     pioh_link);

            OSM_ASSERT(pio_handle);

            if(pio_handle->pio)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p IOWS 0x%p Avail %d Internal Avail %d\n",
                    pio_handle, pio_handle->pio, pio_handle->pio->available_ticks, pio_handle->pio->internal_available_ticks));
            }
            else 
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("IOH 0x%p time taken %d\n",
                    pio_handle, TIME_SPENT(pcontroller, pio_handle->starting_tick_count)));
            }
            ios++;
        }
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Controller 0x%p has %d IOS in pending list\n",
            pcontroller, ios));
        ios = 0;
    }
}
#endif

ACIT_UEXACT32
ap_free_pending_handles(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           port_num)
{
    ACIT_UEXACT32           toremove = 0;
    ACIT_DC_LIST            temp_ioh_list, temp_tmfh_list;
    ACIT_DC_LIST            temp_noph_list, temp_loh_list;
    ACIT_DC_LIST            temp_ioh_inv_list;
    PACIT_DC_LIST           flink;
    PACIT_IO_HANDLE         pio_handle = ACIT_NULL;
    PACIT_TMF_HANDLE        ptmf_handle = ACIT_NULL;
    PACIT_NOP_HANDLE        pnop_handle = ACIT_NULL;
    PACIT_LOGOUT_HANDLE     plogout_handle = ACIT_NULL;
    ACIT_UEXACT32           count = 0;

    LIST_INIT_HEADER(&temp_ioh_list);
    LIST_INIT_HEADER(&temp_tmfh_list);
    LIST_INIT_HEADER(&temp_noph_list);
    LIST_INIT_HEADER(&temp_loh_list);
    LIST_INIT_HEADER(&temp_ioh_inv_list);

    if(plun)
    {
        toremove = LUN_QUEUE;
        LIST_MOVE(&temp_ioh_list,  &plun->io_handle_busy_list);
        LIST_MOVE(&temp_tmfh_list, &plun->tmf_handle_busy_list);
        LIST_MOVE(&temp_ioh_inv_list, &plun->io_pending_reset_list);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Cleaning up LUN WS 0x%p\n", plun));

    }
    else if(ptarget)
    {
        toremove = TARGET_QUEUE;
        LIST_MOVE(&temp_ioh_list,  &ptarget->io_handle_busy_list);
        LIST_MOVE(&temp_tmfh_list, &ptarget->tmf_handle_busy_list);
        LIST_MOVE(&temp_loh_list,  &ptarget->logout_handle_list);
        LIST_MOVE(&temp_noph_list, &ptarget->nop_handle_list);
        LIST_MOVE(&temp_ioh_inv_list, &ptarget->io_pending_reset_list);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Cleaning up Target WS 0x%p\n", ptarget));

    }
    else if(pcontroller)
    {
        toremove = CONTROLLER_QUEUE;
        LIST_MOVE(&temp_ioh_list,  &pcontroller->io_handle_busy_list);
        LIST_MOVE(&temp_tmfh_list, &pcontroller->tmf_handle_busy_list);
        LIST_MOVE(&temp_noph_list, &pcontroller->nop_handle_busy_list);
        LIST_MOVE(&temp_loh_list, &pcontroller->logout_handle_busy_list);

        LIST_MOVE(&temp_ioh_inv_list, &pcontroller->io_pending_reset_list);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Cleaning up Controller WS 0x%p\n", pcontroller));
    }

    flink = temp_ioh_list.next;

    // Loop and travel the IO handle busy list
    while(flink != &temp_ioh_list)
    {
        OSM_ASSERT(flink);

        pio_handle = ap_get_io_handle(flink, toremove);
        OSM_ASSERT(pio_handle);

        // Move the next pointer before modifying the list
        flink = flink->next;

        // update IO counters before dealing with the IO
        iscsi_update_io_counters(pcontroller, pio_handle);

        // The IOWs may be freed here, but the IO Handle is always valid
        // on return
        ap_decide_io_disposition_in_cleanup(pcontroller, pio_handle);

        if(pio_handle->ptmf_ws)
        {
            //
            // Bug 27997 : The aborts that were issued from OSM could not be issued
            // as a higher level reset was in progress. If ACIT also had started an
            // internal abort on this, then, before starting the abort itself, the
            // IOHandle would have been removed from busy list and we wouldn't have been
            // here.
            //

            ap_return_tmf(pcontroller, 
                        pio_handle->ptmf_ws->ptmf_request,
                        ACIT_TMF_STATUS_SUCCESS);

            pio_handle->ptmf_ws = ACIT_NULL;
        }

        ap_free_io_resources(pcontroller, pio_handle);

        count++;
    }
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reclaiming %d IOS\n", count));

    flink = temp_tmfh_list.next;
    count = 0;

     // Loop and travel TMF handle busy list
    while(flink != &temp_tmfh_list)
    {
        PACIT_TMF_WS pioh_tmf_ws = ACIT_NULL;

        OSM_ASSERT(flink);

        ptmf_handle = ap_get_tmf_handle(flink, toremove);
        OSM_ASSERT(ptmf_handle);

        // Move the next pointer before modifying the list
        flink = flink->next;

        // If there is a IOH associated with this TMFH, deal with it now.
        // The IOWS needs to be either queued or returned back
        if(ptmf_handle->pio_handle)
        {
            ap_decide_io_disposition_in_cleanup(pcontroller, ptmf_handle->pio_handle);

            // Bug 106047 :
            // If an external Abort request came in after a TMF handle was instantiated
            // an abort was already issued to the target, the pio_handle->ptmf_ws is
            // the only reference to the external abort. In that case, ptmf_handle->ptmf
            // can be NULL. So use the pio_handle->ptmf_ws reference to check for
            // external TMF

            pioh_tmf_ws = ptmf_handle->pio_handle->ptmf_ws;
        }

        //
        // if its a external target reset, LUN reset or abort TMF from OSM, make sure 
        // the references are all gone
        //
        if(ptmf_handle->ptmf || pioh_tmf_ws)
        {
            if(ptmf_handle->tmf_function == ACIT_TMF_ABORT_TASK)
            {
                OSM_ASSERT(pioh_tmf_ws);
                OSM_ASSERT(pioh_tmf_ws->ptmf_request);

                ap_return_tmf(pcontroller,
                              pioh_tmf_ws->ptmf_request,
                              ACIT_TMF_STATUS_SUCCESS);

                ptmf_handle->pio_handle->ptmf_ws = ACIT_NULL;
            }
            else if(ptmf_handle->tmf_function == ACIT_TMF_LUN_RESET)
            {
                ap_return_tmf(pcontroller, 
                              ptmf_handle->plun->plun_reset_tmf_request, 
                              ACIT_TMF_STATUS_SUCCESS);
            }
            else if(ACIT_IS_TARGET_RESET_TMF(ptmf_handle->tmf_function))
            {
                ap_return_tmf(pcontroller, 
                              ptmf_handle->ptarget->ptarget_reset_tmf_request, 
                              ACIT_TMF_STATUS_SUCCESS);
            }
            else
            {
                OSM_ASSERT(0);
            }
        }

        ap_free_tmf_resources(pcontroller, ptmf_handle);
        count++;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reclaiming %d TMFs\n", count));

    flink = temp_noph_list.next;
    count = 0;

    // Loop and travel the NOP handle busy list
    while(flink != &temp_noph_list)
    {
        OSM_ASSERT(flink);

        pnop_handle = ap_get_nop_handle(flink, toremove);
        OSM_ASSERT(pnop_handle);

        // Move the next pointer before modifying the list
        flink = flink->next;

        ap_free_nop_resources(pcontroller, pnop_handle);

        count++;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reclaiming %d NOPs\n", count));

    flink = temp_loh_list.next;
    count = 0;

    // Loop and travel logout handle busy list.
    while(flink != &temp_loh_list)
    {
        OSM_ASSERT(flink);

        plogout_handle = ap_get_logout_handle(flink, toremove);
        OSM_ASSERT(plogout_handle);

        // Move the next pointer before modifying the list
        flink = flink->next;

        ap_free_logout_resources(pcontroller, plogout_handle);
        count++;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reclaiming %d logouts\n", count));

    // Process the IO handle invalidated list here 
    flink = temp_ioh_inv_list.next;
    count = 0;

    while(flink != &temp_ioh_inv_list)
    {
        OSM_ASSERT(flink);

        pio_handle = ap_get_io_handle(flink, toremove);
        OSM_ASSERT(pio_handle);

        // Move the next pointer before modifying the list
        flink = flink->next;

        // The IOWs may be freed here, but the IO Handle is always valid
        // on return
        ap_decide_io_disposition_in_cleanup(pcontroller, pio_handle);

        //
        // Bug 27997 : The aborts that were issued from OSM could not be issued
        // right away as it just initiated a command invalidation. Normally,
        // command invalidation would start an abort. But if a session recovery
        // happened to occur before the invalidation completed, the OSM abort
        // was never returned.
        //
        if(pio_handle->ptmf_ws)
        {
            ap_return_tmf(pcontroller, 
                        pio_handle->ptmf_ws->ptmf_request,
                        ACIT_TMF_STATUS_SUCCESS);

            pio_handle->ptmf_ws = ACIT_NULL;
        }

        ap_free_io_resources(pcontroller, pio_handle);
        count++;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Reclaiming %d IO invalidations\n", count));

    if(toremove == LUN_QUEUE)
    {
        ap_return_queued_lun_tmfs(pcontroller, 
                                    ptarget, 
                                    plun, 
                                    ACIT_TMF_STATUS_SUCCESS,
                                    ACIT_TRUE);
    }
    else if(toremove == TARGET_QUEUE)
    {

        ap_return_queued_target_tmfs(pcontroller,
                                        ptarget,
                                        ACIT_TMF_STATUS_SUCCESS,
                                        ACIT_TRUE);
    }
    else if(toremove == CONTROLLER_QUEUE)
    {
        ap_return_queued_controller_tmfs(pcontroller,
                                        ACIT_TMF_STATUS_SUCCESS,
                                        ACIT_FALSE);
    }


#if ACIT_CFG_DEBUG_BUILD
    ap_validate_lists(pcontroller, ptarget, plun);
#endif

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_free_queued_ws_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           free_ios)
{
   
    if(free_ios)
    {
        ap_return_queued_ios(pcontroller, 
                            ptarget, 
                            ACIT_NULL, 
                            ACIT_IO_STATUS_ABORTED, 
                            ACIT_U32_MINUS_ONE);
    }

    if(ptarget)
    {
        ap_return_queued_target_tmfs(pcontroller, 
                                     ptarget, 
                                     ACIT_TMF_STATUS_TARGET_TIMEOUT,
                                     ACIT_TRUE);
    }
    else
    {
        ap_return_queued_controller_tmfs(pcontroller, 
                                         ACIT_TMF_STATUS_TARGET_TIMEOUT,
                                         ACIT_TRUE);
    }
}

ACIT_UEXACT32
ap_get_target_cleanup_flag(
    PACIT_TARGET_WS         ptarget)
{
    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_LOGOUT))
    {
        return TGT_CLEANUP_FLAG_FOR_LOGOUT;
    }

    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET))
    {
        return TGT_CLEANUP_FLAG_FOR_TMF_RESET;
    }

    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_RECOVERY) ||
       (ptarget->sub_state == TARGET_SUB_STATE_TMF_TIMEOUT) ||
       (ptarget->sub_state == TARGET_SUB_STATE_TMF_ERROR))
    {
        return TGT_CLEANUP_FLAG_FOR_RECOVERY;
    }

    return 0;
}

ACIT_VOID
ap_complete_group_invalidations(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    ACIT_UEXACT32           flags = 0;

    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_LDTO_EXPIRED))
    {
        if(!pcontroller->counters.num_io_invs)
        {
            //
            // call the LDTO expiry function. The handler should also make sure
            // that there are no invalidations pending
            //
            iscsi_recover_all_sessions(pcontroller, ACIT_NULL);
        }
        return;
    }

    if(ptarget)
    {
        flags = ap_get_target_cleanup_flag(ptarget);

        if(flags && (!ptarget->counters.num_io_invs))
        {
            if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_LOGOUT))
            {
                iscsi_check_and_close_session(pcontroller, ptarget, ACIT_FALSE);
                return;
            }
            else if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_RECOVERY))
            {
                iscsi_check_and_close_session(pcontroller, ptarget, ACIT_TRUE);
                return;
            }
            else if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET))
            {
                ap_dequeue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
                if(ap_check_and_issue_reset(pcontroller, 
                                            ptarget, 
                                            ACIT_NULL) != ACIT_STATUS_SUCCESS)
                {
                    ap_queue_tmf(pcontroller, ptarget, ACIT_NULL, ACIT_NULL);
                }
                return;
            }
            else
                OSM_ASSERT(0);
        }
    }

    if(plun && (!plun->counters.num_io_invs))
    {
        if (plun->plun_reset_tmf_request)
        {
            ap_dequeue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
            if(ap_check_and_issue_reset(pcontroller, 
                                        ptarget, 
                                        plun) != ACIT_STATUS_SUCCESS)
            {
                ap_queue_tmf(pcontroller, ptarget, plun, ACIT_NULL);
            }
            return;
        }
    }

    return;
}

ACIT_UEXACT32
ap_process_io_disposition_after_timeout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           disposition)
{
    //
    // Called after a timeout, which means that the IO WS
    // was not queued at this instance. So if we're returning
    // the IO back, then make sure to set the unlink to false
    //
    switch(disposition)
    {
        case ACIT_STATUS_SUCCESS:

            pio->retry_count++;
            pio->io_retries++;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Retrying IOWS 0x%p for LUN 0x%p Target 0x%p Ticks available %d\n"),
                pio, plun, ptarget, pio->available_ticks));

            ap_queue_io_ws(pcontroller, ptarget, plun, pio);

            // mark for flush
            pcontroller->state.flush_queued_ios = ACIT_TRUE;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_UCODE,("schedule for flush\n"));

            //
            // If the IO retry count exceeded, then trigger a session
            // level recovery
            //
            if(pio->retry_count >= pcontroller->params.io_retries_b4_erl0)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IO Retries exceeded for IOWS 0x%p, triggering session recovery..\n"),
                                pio));

                ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_IO_RETRIES_EXCEEDED);

                pcontroller->eh_stats.ioretry_exceeded++;
                ptarget->eh_stats.ioretry_exceeded++;
                iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
            }

            pcontroller->eh_stats.io_retry_count++;

            break;

        case ACIT_STATUS_INSUFFICIENT_TIMEOUT:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IO timeout for IOWS 0x%p, returing IO back...\n"),
                            pio));

            pio->retry_count++;
            if(pio->retry_count >= pcontroller->params.io_retries_b4_erl0)
            {
                ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_IO_RETRIES_EXCEEDED);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IO Retries exceeded for IOWS 0x%p, triggering session recovery..\n"),
                                pio));

                pcontroller->eh_stats.ioretry_exceeded++;
                ptarget->eh_stats.ioretry_exceeded++;
                iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
            }

            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_IO_TIMEOUT, 
                         ACIT_FALSE);
            break;
    
        case ACIT_STATUS_LINK_DOWN_TIMEOUT:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("LD timeout for IOWS 0x%p, returing IO back...\n"),
                            pio));

            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_LINK_DOWN_TIMEOUT, 
                         ACIT_FALSE);
            break;

        case ACIT_STATUS_TARGET_TIMEOUT:
        
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target timeout for IOWS 0x%p, returing IO back...\n"),
                            pio));

            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_TARGET_TIMEOUT, 
                         ACIT_FALSE);
            break;

        case ACIT_STATUS_ERROR:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Failing IO 0x%p after max retries\n", pio));

            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_ERROR, 
                         ACIT_FALSE);

            break;

        default:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Unknown error for IO 0x%p\n", pio));
            OSM_ASSERT(0);

            ap_return_io(pcontroller, 
                         pio, 
                         ACIT_IO_STATUS_ERROR, 
                         ACIT_FALSE);

            break;
    }

    // before we return, delink the IO handle from IO WS. ap_return_io would already
    // have done this, but not ap_queue_io_ws
    pio->pio_handle = ACIT_NULL;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_is_reset_pending(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_LUN_WS            plun)
{
    //
    // This function return ACIT_STATUS_PENDING if a reset is pending
    // for the target or LU
    //

    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_LDTO_EXPIRED))
    {
        return ACIT_STATUS_PENDING;
    }

    if(ptarget)
    {
        if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET) ||
           ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_RECOVERY)  ||
           (ptarget->sub_state == TARGET_SUB_STATE_TMF_TIMEOUT)                   ||
           (ptarget->sub_state == TARGET_SUB_STATE_TMF_ERROR))
        {
            return ACIT_STATUS_PENDING;
        }
    }

    if(plun)
    {
        if(plun->plun_reset_tmf_request)
        {
            return ACIT_STATUS_PENDING;
        }
    }

    return ACIT_STATUS_SUCCESS;
}

//
// Bug 31195 : Do not return the IO back until we're sure there's a
// path failover or a UE. Path failover also does not guarantee that
// the IO is really terminated on the target. But there are methods
// via configuration (e.g. periodic target side NOPIn) that can help
// mitigate the risks
//

ACIT_UEXACT32
ap_complete_invalidate_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle)
{
    PACIT_DC_LIST           flink;
    PACIT_IO_HANDLE         pio_handle;
    ACIT_UEXACT32           status, count = 0;

    //
    // When invalidate IO completes, the IO handle cannot be freed. The IO
    // statistics can be updated since IO request may be returned back.
    // The IO is further queued or returned depending on the state of the
    // target and LUN.
    // Correction! The IO is never returned from here nor queued internally
    //
    // Since the IO handle is not freed, it is moved from invalidate handle
    // to a invalidated_list. This list is cleared when either a TMF completes
    // or a invalidate connection returns
    //
    // Invalidate IO completion may also need a TMF abort to be sent. Unless a
    // target or LU level flag indicates that abort should not be sent, a abort
    // TMF will be initiated from here
    //

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&pinv_handle->list_header));

    flink = pinv_handle->list_header.next;

    while(flink != &pinv_handle->list_header)
    {
        pio_handle = ap_get_io_handle(flink, CONTROLLER_QUEUE);
        OSM_ASSERT(pio_handle);
        OSM_ASSERT(pio_handle->pio);
        OSM_ASSERT(pio_handle->cri < pcontroller->params.num_cxn);

        count++;
        flink = flink->next;

        iscsi_update_io_counters(pcontroller, pio_handle);

        LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
        
        LIST_INIT_ELEMENT(&pio_handle->controller_link);
        LIST_INIT_ELEMENT(&pio_handle->target_link);
        LIST_INIT_ELEMENT(&pio_handle->lun_link);

        ap_setup_ioh_for_tmf(pcontroller, pio_handle);

        OSM_ASSERT(pcontroller->counters.num_io_invs);
        OSM_ASSERT(pio_handle->ptarget->counters.num_io_invs);
        OSM_ASSERT(pio_handle->plun->counters.num_io_invs);

        pcontroller->counters.num_io_invs--;
        pio_handle->ptarget->counters.num_io_invs--;
        pio_handle->plun->counters.num_io_invs--;

        if(ap_is_reset_pending(pcontroller, 
                               pio_handle->ptarget, 
                               pio_handle->plun) == ACIT_STATUS_SUCCESS)
        {
            // A command abort is needed
            status = ap_check_and_issue_abort(pcontroller, 
                                              pio_handle->ptarget, 
                                              pio_handle->plun, 
                                              pio_handle);

            OSM_ASSERT(status != ACIT_STATUS_RESET_IN_PROGRESS);

            if(status != ACIT_STATUS_SUCCESS)
            {
                ap_queue_tmf(pcontroller, 
                             pio_handle->ptarget, 
                             pio_handle->plun, 
                             pio_handle);
            }
        }
        else
        {
            //
            // if no abort is sent, then queue the IO handle in io_pending_reset_list.
            // When a target or LUN level cleanup is done, these lists are traversed
            // and cleaned up
            //
            // For aborts alone, the IO handles are not queued along io_pending_reset_list
            // Once queued OR submitted, they are freed either when abort completes or 
            // abort timesout and results in session recovery
            //

            //
            // always queue IO handle for release later if no aborts are issued. Found 
            // this out during MPIO failover testing. Resources were leaked since the 
            // IO handle was removed from the list and never added to any list for cleanup.
            //

//            if(ACIT_IS_BIT_SET(pio_handle->ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET) ||
//                (pio_handle->plun->plun_reset_tmf_request))
            {
                LIST_ADDTO_TAIL(&pio_handle->controller_link, &pcontroller->io_pending_reset_list);
                LIST_ADDTO_TAIL(&pio_handle->target_link, &pio_handle->ptarget->io_pending_reset_list);
                LIST_ADDTO_TAIL(&pio_handle->lun_link, &pio_handle->plun->io_pending_reset_list);
            }
        }

        //
        // Because inv handle may contain IOs from across LUNs or targets, check
        // to see if any group level error handling needs to be triggered
        //
        ap_complete_group_invalidations(pcontroller, 
                                        pio_handle->ptarget,
                                        pio_handle->plun);
    }

    OSM_ASSERT(count == pinv_handle->num_invs);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_get_inv_params(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle,
    PACIT_DC_LIST           element,
    PACIT_UEXACT16          picd,
    PACIT_UEXACT16          pcid)
{
    switch(pinv_handle->flags)
    {
    case ACIT_INVH_TYPE_IO:
    {
        PACIT_IO_HANDLE pio_handle;

        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                 controller_link,
                                 element);

        hwi_mark_wrb_for_invalidate(pio_handle->pwrb_handle);

        *picd = SGLICD_FROM_ITT(pio_handle->itt);
        *pcid = CID_FROM_CRI(pcontroller, pio_handle->cri);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("invalidating IOH 0x%p CID 0x%x ITT 0x%x lun 0x%p time_elapsed %d\n"), 
                                                pio_handle, 
                                                *pcid, 
                                                pio_handle->itt, 
                                                pio_handle->plun,
                                                TIME_SPENT(pcontroller, pio_handle->starting_tick_count)));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IOH 0x%p IOWS 0x%p Original Ticks %d Available Ticks %d Internal Ticks %d\n"),
            pio_handle,
            pio_handle->pio,
            pio_handle->pio->original_ticks,
            pio_handle->pio->available_ticks,
            pio_handle->pio->internal_available_ticks));
    }

    break;

    default:
        OSM_ASSERT(0);
    }
}

ACIT_UEXACT32
ap_invalidate_pending_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle,
    ACIT_UEXACT32           cri)
{
    ACIT_UEXACT32           toremove = CONTROLLER_QUEUE;
    ACIT_STATUS             status = ACIT_STATUS_SUCCESS;

    if(pio_handle) 
    {
        /*
         * For invalidating an I/O alone, create a temp list so that
         * ap_invalidate_io function can still take in a list
         */
        ACIT_DC_LIST    temp_list;
        ACIT_UEXACT32   invs_allowed;

        invs_allowed = (pcontroller->params.ioctls_per_ctrl - pcontroller->counters.pending_ioctls + 1) * 
                            BE_MAX_ITTS_PER_INVALIDATE;

        if(!invs_allowed)
        {
            pcontroller->state.defer_invalidations = ACIT_TRUE;
            pcontroller->eh_stats.invs_deferred++;
            pcontroller->eh_stats.invs_deferred_no_ioctls++;
            status = ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
        else
        {
            LIST_INIT_HEADER(&temp_list);

            LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
            LIST_ADDTO_TAIL(&pio_handle->controller_link, &temp_list);

            status = ap_invalidate_cmd_list(pcontroller, 
                                        &temp_list, 
                                        ACIT_INVH_TYPE_IO,
                                        toremove);
        }
    }
    else if(plun)
    {
        if(IS_LIST_NOT_EMPTY(&plun->io_handle_busy_list))
        {
            toremove = LUN_QUEUE;

            status = ap_invalidate_cmd_list(pcontroller,
                                        &plun->io_handle_busy_list, 
                                        ACIT_INVH_TYPE_IO,
                                        toremove);
        }
        else if(plun->counters.num_io_invs)
            status = ACIT_STATUS_PENDING;
    }
    else if(ptarget)
    {
        if(IS_LIST_NOT_EMPTY(&ptarget->io_handle_busy_list))
        {
            toremove = TARGET_QUEUE;

            status = ap_invalidate_cmd_list(pcontroller,
                                           &ptarget->io_handle_busy_list, 
                                           ACIT_INVH_TYPE_IO,
                                           toremove);
        }
        else if(ptarget->counters.num_io_invs)
            status = ACIT_STATUS_PENDING;
    }
    else if(pcontroller)
    {
        if(IS_LIST_NOT_EMPTY(&pcontroller->io_handle_busy_list))
        {
            status = ap_invalidate_cmd_list(pcontroller,
                                        &pcontroller->io_handle_busy_list, 
                                        ACIT_INVH_TYPE_IO,
                                        toremove);
        }
        else if(pcontroller->counters.num_io_invs)
            status = ACIT_STATUS_PENDING;
    }
    
    return status;
}

ACIT_VOID
ap_unlink_cmd_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_DC_LIST           link,
    PACIT_INVALIDATE_HANDLE pinv_handle,
    ACIT_UEXACT32           toremove,
    PACIT_TARGET_WS         *ptarget,
    PACIT_LUN_WS            *plun)
{
    PACIT_TARGET_WS         plocal_target = ACIT_NULL;
    PACIT_LUN_WS            plocal_lun = ACIT_NULL;

    PACIT_IO_HANDLE         pio_handle;

    switch(pinv_handle->flags)
    {
    case ACIT_INVH_TYPE_IO:

        pio_handle = ap_get_io_handle(link, toremove);
        ACIT_BIT_SET(pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING);
        plocal_target = pio_handle->ptarget;
        plocal_lun = pio_handle->plun;

        /* This removes the IO handle from IO handle busy list */
        LIST_REMOVE_ELEMENT(&pio_handle->controller_link);
        LIST_REMOVE_ELEMENT(&pio_handle->target_link);
        LIST_REMOVE_ELEMENT(&pio_handle->lun_link);

        // Since its taken off the list, count this as the start of invalidation
        pio_handle->inv_start_tick = pcontroller->counters.current_tick_count;

        /* Add the IO handle to the list in invalidation handle */
        LIST_ADDTO_TAIL(&pio_handle->controller_link,
                        &pinv_handle->list_header);

        break;
        
    }

    *ptarget = plocal_target;
    *plun = plocal_lun;
}

ACIT_VOID
ap_track_invalidations(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           cmd_type,
    ACIT_UEXACT16           num_invs)
{
    switch(cmd_type)
    {
    case ACIT_INVH_TYPE_IO:

        //
        // Compilation note!
        // Microsoft C4244 - do not use += on non-integer types
        //
        pcontroller->counters.num_io_invs = pcontroller->counters.num_io_invs + num_invs;

        ptarget->counters.num_io_invs += num_invs;
        plun->counters.num_io_invs += num_invs;

        break;

    }
}

ACIT_UEXACT32
ap_invalidate_cmd_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_DC_LIST           plist,
    ACIT_UEXACT32           cmd_type,
    ACIT_UEXACT32           toremove)
{
    ACIT_UEXACT8            num_invs = 0;
    PACIT_TARGET_WS         ptarget = ACIT_NULL;
    PACIT_LUN_WS            plun = ACIT_NULL;
    PACIT_DC_LIST           flink, save_flink;

    ACIT_STATUS             status = ACIT_STATUS_SUCCESS;
    PACIT_INVALIDATE_HANDLE pinv_handle = ACIT_NULL;

    hwi_check_health(pcontroller, ACIT_TRUE);

#if ACIT_CFG_DEBUG_BUILD
    //hwi_validate_pending_ios(pcontroller, ACIT_NULL, ACIT_NULL);
#endif

    OSM_ASSERT(IS_LIST_NOT_EMPTY(&plist));

    flink = plist->next;

    while(flink != plist)
    {
        ptarget = ACIT_NULL;
        plun = ACIT_NULL;

        if(!pinv_handle)
        {
            pinv_handle = ap_alloc_inv_resources(pcontroller);

            if(pinv_handle == ACIT_NULL)
            {
                // no resource for invalidation.
                pcontroller->state.defer_invalidations = ACIT_TRUE;
                pcontroller->eh_stats.invs_deferred++;
                return ACIT_STATUS_INSUFFICIENT_RESOURCE;
            }
            pinv_handle->flags = (ACIT_UEXACT8)cmd_type;
        }

        /* Increment this before the queue is modified */
        save_flink = flink;
        flink = flink->next;

        ap_unlink_cmd_handle(pcontroller, 
                             save_flink,
                             pinv_handle,
                             toremove,
                             &ptarget,
                             &plun);

        OSM_ASSERT(ptarget);

        num_invs++;
        ap_track_invalidations(pcontroller, ptarget, plun, cmd_type, 1);

        if(num_invs % BE_MAX_ITTS_PER_INVALIDATE == 0 && num_invs > 0)
        {
            //
            // Compilation note!
            // Microsoft C4244 - do not use += on non-integer types
            //
            pinv_handle->num_invs = pinv_handle->num_invs + num_invs;

            status = mgmt_invalidate_icds(pcontroller, pinv_handle);
        
            if(status != MGMT_STATUS_SUCCESS)
            {
                // convert MGMT status to ACIT status
                status = ACIT_STATUS_INSUFFICIENT_RESOURCE;
            }

            OSM_ASSERT(status == ACIT_STATUS_SUCCESS);

            /* Reset values */
            num_invs = 0;
            pinv_handle = ACIT_NULL;
        }
    }

    if(num_invs)
    {
        pinv_handle->num_invs = pinv_handle->num_invs + num_invs;
        status = mgmt_invalidate_icds(pcontroller, pinv_handle);
        if(status != MGMT_STATUS_SUCCESS)
        {
             // convert MGMT status to ACIT status
             status = ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
         
        OSM_ASSERT(status == ACIT_STATUS_SUCCESS);
    }

    if(status == ACIT_STATUS_SUCCESS)
        return ACIT_STATUS_PENDING;

    return status;
}

ACIT_UEXACT32
ap_set_invalidate_icd_status(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle,
    PACIT_DC_LIST           element,
    ACIT_UEXACT32           inv_status,
    ACIT_UEXACT32           byte_count)
{
    ACIT_UEXACT32           h_flag;

    h_flag = ap_cmdh_flag_from_inv_status(inv_status);

    switch(pinv_handle->flags)
    {
    case ACIT_INVH_TYPE_IO:
    {
        PACIT_IO_HANDLE pio_handle;
        ACIT_UEXACT32 tick_diff;

        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                 controller_link,
                                 element);

        ACIT_BIT_CLR(pio_handle->flags, FLAG_IOH_INVALIDATE_PENDING);
        ACIT_BIT_SET(pio_handle->flags, h_flag);

        tick_diff = (pinv_handle->complete_tick - pinv_handle->submit_tick);

        if(pio_handle->pio)
        {
            pio_handle->pio->available_ticks -= tick_diff;
            pio_handle->pio->internal_available_ticks -= tick_diff;

            if(pio_handle->pio->available_ticks < 0)
            {
                pio_handle->pio->available_ticks = 0;
            }

            if(pio_handle->pio->internal_available_ticks < 0)
            {
                pio_handle->pio->internal_available_ticks = 0;
            }
        }

        if(tick_diff > pcontroller->state.invalidation_ticks)
        {
            pcontroller->eh_stats.inv_over_time_limit++;

            if(tick_diff > pcontroller->eh_stats.max_inv_ticks_over_limit)
            {
                pcontroller->eh_stats.max_inv_ticks_over_limit = tick_diff;
            }
        }

        pio_handle->pconnection_handle->total_bytes_xferred += byte_count;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("IOWS 0x%p IOH 0x%p CID 0x%x ITT 0x%x lun 0x%p\n"),
                                pio_handle->pio,
                                pio_handle, 
                                CID_FROM_CRI(pcontroller, pio_handle->cri),
                                pio_handle->itt, pio_handle->plun));
    }

    break;

    default:
        OSM_ASSERT(0);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_complete_invalidate_icds(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle)
{

    //
    // For every invalidation response, check the CQ again to make
    // sure there are no responses that have come in for the commands
    // that are invalidated. This is to close the window of time when
    // CQs could coming in just after we looked at the CQ but before
    // we look into the MCC CQ
    //
    if(ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS) &&
        ((pcontroller->state.init_type == ACIT_INIT_TYPE_POWER_MGMT) ||
        (pcontroller->state.init_type == ACIT_INIT_TYPE_NORMAL)))
    {
        ACIT_UEXACT32 pid;
        for(pid = 0; pid < pcontroller->params.num_cq; pid++)
        {
            hwi_process_cq(pcontroller, pid); 
        }
    }

    switch(pinv_handle->flags)
    {
    case ACIT_INVH_TYPE_IO:

        ap_complete_invalidate_ios(pcontroller, pinv_handle);
        break;

    default:

        OSM_ASSERT(0);
        break;
    }

    ap_free_inv_resources(pcontroller, pinv_handle);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_complete_discovery(
    PACIT_CONTROLLER_WS         pcontroller)
{
    OSM_ASSERT(pcontroller->state.discovery_in_progress == ACIT_TRUE);
    if(pcontroller->state.discovery_type == ACIT_DISCOVER_OPCODE_ASYNCHRONOUS){
        OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_CONTROLLER,
                    ACIT_EVENT_CONTROLLER_DISCOVERY_COMPLETE,
                    0);
    }
    pcontroller->state.discovery_in_progress = ACIT_FALSE;

    return ACIT_STATUS_SUCCESS;
}

PACIT_LOGOUT_HANDLE
ap_alloc_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           reason_code)
{
    PACIT_LOGOUT_HANDLE     plogout_handle = ACIT_NULL;

    if( ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_EH_SGL_HANDLE))
    {
        return ACIT_NULL;
    }

    // Allocate logout handle
    plogout_handle = ap_alloc_logout_handle(pcontroller);

    if(plogout_handle == ACIT_NULL) 
    {
        return ACIT_NULL;
    }

    plogout_handle->cri = (ACIT_UEXACT16)cri;

    if(hwi_alloc_logout_resources(pcontroller, plogout_handle) == ACIT_STATUS_SUCCESS)
    {
        plogout_handle->available_ticks = pcontroller->params.logout_timeout_ticks;
        plogout_handle->starting_tick_count = CURRENT_TICK(pcontroller);

        plogout_handle->ptarget = ptarget;
        plogout_handle->reason_code = reason_code;
        LIST_ADDTO_TAIL(&plogout_handle->target_link, &ptarget->logout_handle_list);
        pcontroller->counters.pending_logouts++;
    }
    else
    {
        ap_free_logout_handle (pcontroller, plogout_handle);        
        return NULL;
    }
    
    return plogout_handle;
}

ACIT_VOID
ap_free_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    hwi_free_logout_resources(pcontroller, plogout_handle);
    ap_free_logout_handle(pcontroller, plogout_handle);
    pcontroller->counters.pending_logouts--;
}

PACIT_TMF_HANDLE
ap_alloc_tmf_resources(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun,
    PACIT_IO_HANDLE             pio_handle,
    PACIT_TMF_WS                ptmf,
    ACIT_UEXACT32               tmf_function)
{
    PACIT_TMF_HANDLE            ptmf_handle = ACIT_NULL;
    ACIT_UEXACT32               status, tmfto = 0;

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);

    /*
     * TMFs can be sent if there are
     * a. Enough resources available
     * b. Can be sent out on the wire
     */

    status = ap_ready_for_tmf(pcontroller, ptarget);

    if(status == ACIT_STATUS_SUCCESS)
    {
        ptmf_handle = ap_alloc_tmf_handle(pcontroller);

        /*
         * Make sure we got an TMF Handle. It is illegal for this function to 
         * not return an TMF Handle since we've checked for it earlier.
         */
        OSM_ASSERT(ptmf_handle);

        ptmf_handle->ptarget = ptarget;
        ptmf_handle->plun = plun;
        ptmf_handle->pio_handle = pio_handle;
        ptmf_handle->tmf_function = tmf_function;
        ptmf_handle->ptmf = ptmf;

        status = iscsi_alloc_tmf_resources(pcontroller, ptarget, ptmf_handle);

        if(status == ACIT_STATUS_SUCCESS)
        {
            hwi_alloc_tmf_resources(pcontroller, ptmf_handle);

            if(status == ACIT_STATUS_SUCCESS)
            {
                //
                // Update! 02/28/2012 - The following is changed again. We were
                // imposing internal timeout value upon OSM given TMFs also.
                // Check the source of TMF and setup available ticks accordingly.
                //

                //
                // 05/04/2009 - Change to hardcode 5 secs to TMF timeout. This is done
                // to complete ETO steps within specified time
                // Earlier we used the same value used last for the command. We took this
                // approach in mangler case in order to provide more time for TMFs as well
                //

                //ptmf_handle->available_ticks = GET_TARGET_EXEC_TICKS(ptarget, 0);
                //ptmf_handle->available_ticks = GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget);

                if(ptmf_handle->ptmf)
                {
                    ACIT_UEXACT32 eto, rto, osmto;

                    // ETO = Extended Timeout
                    // RTO = Recovery Timeout. Time required to detect an ETO
                    // OSMTO = TMF Timeout provided by OSM
                    // TMFTO = TMF Timeout given by ACIT

                    eto = ptarget->eto.adjusted_ticks;
                    rto = MGMT_REOPEN_SESSION_TIMEOUT * TICKS_PER_SEC(pcontroller);
                    osmto = ptmf_handle->ptmf->original_ticks;

                    tmfto = ACIT_MIN(eto, osmto);

                    if(tmfto <= (rto * 2))
                    {
                        // We are bound to miss ETO if the target does not
                        // respond. But we take that chance nonetheless..

                        // Do not subtract anything off the provided tmfto

                        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target 0x%p "
                            "Unable to deduct RTO from OSMTO, ETO reporting "
                            "may be delayed!!\n"),ptarget));
                    }
                    else
                    {
                        tmfto = tmfto - rto;
                    }

                    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target 0x%p "
                                    "TMFH 0x%p OSMTO %d TMFTO %d\n"),
                                                        ptarget,
                                                        ptmf_handle, 
                                                        osmto,
                                                        tmfto));
                }
                else
                {
                    tmfto = GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget);

                    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Target 0x%p "
                                    "TMFH 0x%p TMFTO %d\n"),
                                                        ptarget,
                                                        ptmf_handle, 
                                                        tmfto));
                }

                ptmf_handle->available_ticks = tmfto;
                ptmf_handle->starting_tick_count = CURRENT_TICK(pcontroller);

                LIST_ADDTO_TAIL(&ptmf_handle->target_link,
                                &ptarget->tmf_handle_busy_list);

                if(plun)
                {
                    LIST_ADDTO_TAIL(&ptmf_handle->lun_link,
                                    &plun->tmf_handle_busy_list);
                }

                /*
                * Increment pending_tmfs here, since we're all set with allocating resources
                */
                pcontroller->counters.pending_tmfs ++;
                pcontroller->eh_stats.tmfs_issued ++;
            }
        }
        else
        {
            ap_free_tmf_handle(pcontroller, ptmf_handle);
            ptmf_handle = ACIT_NULL;
        }
    }

    return ptmf_handle;
}

ACIT_UEXACT32
ap_process_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           async)
{
    ACIT_UEXACT32           port_num;

    if(async && ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_WS_INIT_PENDING))
    {
        ACIT_UEXACT32 status;

        status = ap_reinit_target_ws(pcontroller, 
                                     ptarget, 
                                     GET_SESSION_HANDLE(pcontroller, ptarget));

        if(status == ACIT_STATUS_SUCCESS)
        {
            ap_clear_target_event(pcontroller, ptarget, EVENT_TGT_WS_INIT_PENDING);
        }
    }

    if(async && ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_REOPEN_PENDING))
    {
        MGMT_STATUS mgmt_status;
        ACIT_UEXACT32 session_handle = GET_SESSION_HANDLE(pcontroller, ptarget);

        mgmt_status = mgmt_reopen_sessions(pcontroller, 
                                           IOCTL_ISCSI_REOPEN_A_SESSION,
                                           session_handle,
                                           GET_TARGET_DOMAIN(pcontroller, ptarget),
                                           ACIT_FALSE);

        if(mgmt_status == MGMT_STATUS_SUCCESS)
        {
            ap_clear_target_event(pcontroller, ptarget, EVENT_TGT_REOPEN_PENDING);
        }
    }

    if(ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_LOGOUT_PENDING) ||
        ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_CXN_EVENT_PENDING) ||
        ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_RECOVERY_LOGOUT_PENDING))
    {
        iscsi_process_target_event(pcontroller, ptarget, async);
    }

    if(async)
    {
        /* if target is missing AND link is up, countdown the ETO. 
         * This will pause the ETO incase of ldtimeout 
         */
        port_num = ISCSI_GET_TARGET_PORT(ptarget);

        if(IS_TARGET_MISSING(ptarget) && ACIT_IS_CONTROLLER_LINK_UP(pcontroller, port_num))
        {
            if(ptarget->eto.ticks_counter) 
                ptarget->eto.ticks_counter -- ; 
            else
            {
                ap_process_eto_expiry(pcontroller, ptarget);
            }
        }

        //
        // assumes MCS = 0
        // 06/24/09 - add a check to see if there aren't any commands pending
        // There could be a case where a tape target could have only 1 command window
        // and would need lot of time to complete that command. The ping timer would
        // in this case reset the connection not giving the target a chance to respond.
        // Effectively, the ping timer now runs only when there are no outstanding commands
        // from the target.
        //
        if((ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_PING_WAIT)) &&
           (ptarget->piscsi_ws->pcxn_list[0]->num_of_active_cmds == 0))
        {
            if(ptarget->ping_wait_ticks) 
                ptarget->ping_wait_ticks -- ; 
            else
            {
                ACIT_UEXACT32 session_handle;

                ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_PING_WAIT_TIMEOUT);

                session_handle = GET_SESSION_HANDLE(pcontroller, ptarget);

                ap_log_osm_event(pcontroller,
                                 UNI_CFG_LOG_LEVEL_CRITICAL,
                                 UNI_SUBMODULE_ACIT_ISCSI,
                                 UNI_PHASE_RECOVERY,
                                 1,
                                 ACIT_EC_PING_WAIT_TIMEOUT,
                                 0, ACIT_NULL,
                                 &session_handle);

                pcontroller->eh_stats.ping_wait_timeouts++;
                ptarget->eh_stats.ping_wait_timeouts++;

                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("Ping wait timer in target 0x%p timeout, will trigger session recovery...\n"),
                                                        ptarget));

                iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);

                ap_clear_target_event(pcontroller, ptarget, EVENT_TGT_PING_WAIT);
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_log_osm_event(
    PACIT_CONTROLLER_WS     pcontroller,
    UNI_ERROR_LEVEL         level,
    UNI_SUBMODULE           submodule,
    UNI_PHASE               phase,
    ACIT_UEXACT8            addi_len,
    ACIT_UEXACT16           error_code,
    ACIT_UEXACT32           dump_data_size,
    PACIT_VOID              pdump_data,
    PACIT_UEXACT32          paddi_dwords)
{
    ACIT_UEXACT32 status = 0, i;
    ACIT_UEXACT8 module = UNI_MODULE_ACIT;

    status = (
        ((level     << UNI_ERROR_LEVEL_BIT_SHIFT) & UNI_ERROR_LEVEL_BIT_MASK)   |
        ((module    << UNI_MODULE_BIT_SHIFT)      & UNI_MODULE_BIT_MASK)        |
        ((submodule << UNI_SUBMODULE_BIT_SHIFT)   & UNI_SUBMODULE_BIT_MASK)     |
        ((phase     << UNI_PHASE_BIT_SHIFT)       & UNI_PHASE_BIT_MASK)         |
        ((addi_len  << UNI_EXTENSION_BIT_SHIFT)   & UNI_EXTENSION_BIT_MASK)     |
        ((error_code<< UNI_ERROR_CODE_BIT_SHIFT)  & UNI_ERROR_CODE_BIT_MASK));

    if(OSM_LOG_EVENT != NULL)
    {
        ACIT_EVENT_LOG_ENTRY logentry = {0};

        logentry.size_in_bytes = sizeof(logentry);
        logentry.uec = status;
        logentry.dump_data_size = dump_data_size;
        logentry.dump_data_ptr = pdump_data;
        logentry.num_addi_dwords = (ACIT_UEXACT32)addi_len;
        logentry.addi_dwords_ptr = paddi_dwords;

        OSM_LOG_EVENT(pcontroller->pcontroller_key,
                      &logentry);

    }
    else
    {
        if(level >= UNI_CFG_LOG_LEVEL)
        {
            OSM_LOG_ERROR( pcontroller->pcontroller_key,
                           status);

            for(i = 0; i < addi_len; i++)
            {
                OSM_LOG_ERROR( pcontroller->pcontroller_key,
                               *(paddi_dwords + i));
            }
        }
    }

    return ACIT_STATUS_SUCCESS;

}


PACIT_TARGET_WS
ap_get_target_ws_from_session_handle
(
    PACIT_CONTROLLER_WS         pcontroller,
    ACIT_UEXACT32               session_handle)
{
    ACIT_UEXACT32 i;
    PACIT_TARGET_ENTRY ptarget_entry;

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(i = 0; i < pcontroller->params.targets_per_controller; i++)
    {
        if(!(ACIT_IS_BIT_CLR(ptarget_entry[i].acit_tflags,ACIT_TFLAG_IN_USE)))
        {
            if(session_handle == ptarget_entry[i].session_handle)
            {
                return (GET_TARGET_WS_SAFE(ptarget_entry[i].ptarget_key));
            }
        }
    }

    return ACIT_NULL;
}

ACIT_UEXACT32
ap_init_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           eto)
{
    // Initialize default eto value. This can be overridden later  
    ptarget->eto.original_ticks = eto * TICKS_PER_SEC(pcontroller);
    
    // The value in flash is overridden if the value is out of range 

    if((!ACIT_IS_ETO_VALID(eto)))
    {
        ptarget->eto.original_ticks = pcontroller->params.eto * TICKS_PER_SEC(pcontroller);
    }

    ptarget->eto.rto_ticks = GET_TMF_TIMEOUT_TICKS(pcontroller, target) + 
                             (MGMT_REOPEN_SESSION_TIMEOUT * TICKS_PER_SEC(pcontroller)) +
                             pcontroller->params.invalidation_ticks;

    if(ptarget->eto.original_ticks < (MIN_ETO * TICKS_PER_SEC(pcontroller)))
    {
        // Set the adjusted value as per MIN_ETO
        ptarget->eto.adjusted_ticks = MIN_ETO * TICKS_PER_SEC(pcontroller);
    }
    else
    {
        ptarget->eto.adjusted_ticks = ptarget->eto.original_ticks;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, (("Setting target 0x%p ETO to %d secs RTO to %d from offload completion\n"), 
                                            ptarget, 
                                            ptarget->eto.adjusted_ticks / TICKS_PER_SEC(pcontroller),
                                            ptarget->eto.rto_ticks / TICKS_PER_SEC(pcontroller)));

    return ACIT_STATUS_SUCCESS;
}

/*
ACIT_UEXACT32
ap_start_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{

    ACIT_UEXACT32           recovery_ticks = 0;

    ptarget->counters.missing_since = CURRENT_TICK(pcontroller);

    // if target ws says the recovery reason was a TMF timeout, then we will 
    // subtract the internal timeout + TMF timeout + reopen session timeout 
    // value from the overall ETO ticks

    //
    // Bug 19481 : Do not remove 6 secs off ETO timer. We might delay the ETO
    // failure by doing this. But if ARM fw did not take 6 secs for reopen
    // session completion, then we will be giving lesser time to ETO timer.
    //
    //
    //recovery_ticks = (MGMT_REOPEN_SESSION_TIMEOUT * 
    //                    TICKS_PER_SEC(pcontroller));

    //
    // if this target has seen a TMF timeout then subtract some the internal
    // timeout values from overall ETO value. Else give full ETO value
    //
    if(ptarget->recovery_reason.tmf_timeout)
    {
        recovery_ticks += ((ptarget->last_cmd_ticks * ptarget->timeout_factor) + 
            GET_TMF_TIMEOUT_TICKS(pcontroller, ptarget));

        // take the max of recovery ticks calculated and minimum estimated
        // Taking the greater value would cause ETO timer to have lower
        // value and expiry to be faster.
        recovery_ticks = ACIT_MAX(recovery_ticks, MIN_ETO);
    }

    if(recovery_ticks > ptarget->eto.original_ticks)
    {
        ptarget->eto.ticks_counter = 0;
    }
    else
    {
        ptarget->eto.ticks_counter = (ptarget->eto.original_ticks - recovery_ticks);
    }

    /// If failed to reopen session then enter start_eto susbstate
    ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ETO_IN_PROGRESS);
    ap_set_target_event(pcontroller, ptarget, EVENT_TGT_ETO_IN_PROGRESS);

    OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("ETO for target 0x%p started for %d secs\n", 
        ptarget, 
        (ptarget->eto.ticks_counter / TICKS_PER_SEC(pcontroller))));

    return ACIT_STATUS_SUCCESS;
}
*/

ACIT_UEXACT32
ap_start_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    ACIT_UEXACT32           last_update = 0, missing_for = 0;
    ACIT_UEXACT32           tick_now;
    
    tick_now = CURRENT_TICK(pcontroller);

    // if target ws says the recovery reason was a TMF timeout, then we will 
    // subtract the internal timeout + TMF timeout + reopen session timeout 
    // value from the overall ETO ticks

    //
    // Bug 19481 : Do not remove 6 secs off ETO timer. We might delay the ETO
    // failure by doing this. But if ARM fw did not take 6 secs for reopen
    // session completion, then we will be giving lesser time to ETO timer.
    //
    //
    //recovery_ticks = (MGMT_REOPEN_SESSION_TIMEOUT * 
    //                    TICKS_PER_SEC(pcontroller));

    if(ptarget->recovery_reason.tmf_timeout)
    {
        last_update = ptarget->recovery_reason.io_resp_missing_since;
    }
    else
    {
        // The last update this target has seen is the max of last_cmdsn_tick, 
        // last_async_tick and last_cqerr_tick
        last_update = ACIT_MAX(ptarget->stats.last_cmdsn_tick,
                               ptarget->stats.last_async_tick);
        last_update = ACIT_MAX(ptarget->stats.last_cqerr_tick, last_update);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("TMFTimeout %d IOResponseMissingSince %d (Current %d)\n",
                ptarget->recovery_reason.tmf_timeout,
                ptarget->recovery_reason.io_resp_missing_since,
                CURRENT_TICK(pcontroller)));

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("LastCmdSnAt %d LastAsyncAt %d LastCQErrAt %d (Current %d)\n",
                ptarget->stats.last_cmdsn_tick,
                ptarget->stats.last_async_tick,
                ptarget->stats.last_cqerr_tick,
                CURRENT_TICK(pcontroller)));

    missing_for = tick_now - last_update;

    // Update missing_since
    ptarget->counters.missing_since = last_update;

    if(missing_for > ptarget->eto.adjusted_ticks)
    {
        // Already has expired ETO! Probably because of an ARM firmware problem
        // where reopen session has a cascading delay problem, we did not
        // expire ETO on time!!
        ptarget->eto.ticks_counter = 0;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
            ("Target 0x%p ETO Expiry Reporting is Delayed************\n",
            ptarget));
    }
    else
    {
        // Has been missing for missing_for ticks. Give the subtracted value for
        // ETO
        ptarget->eto.ticks_counter = (ptarget->eto.adjusted_ticks - missing_for);
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Target 0x%p Missing for %d secs ETO %d secs\n",
                ptarget,
                missing_for / TICKS_PER_SEC(pcontroller),
                ptarget->eto.adjusted_ticks / TICKS_PER_SEC(pcontroller)));

    /// If failed to reopen session then enter start_eto susbstate
    ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ETO_IN_PROGRESS);
    ap_set_target_event(pcontroller, ptarget, EVENT_TGT_ETO_IN_PROGRESS);

    OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("ETO for target 0x%p started for %d secs\n", 
        ptarget, 
        (ptarget->eto.ticks_counter / TICKS_PER_SEC(pcontroller))));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_stop_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PACIT_DC_LIST           piows_link;
    PACIT_IO_WS             pio;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("ETO for target 0x%p stopped %d secs before expiry\n", 
                            ptarget, (ptarget->eto.ticks_counter / TICKS_PER_SEC(pcontroller))));

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("LastCmdSnAt %d LastAsyncAt %d LastCQErrAt %d (Current %d)\n",
                ptarget->stats.last_cmdsn_tick,
                ptarget->stats.last_async_tick,
                ptarget->stats.last_cqerr_tick,
                CURRENT_TICK(pcontroller)));

    ptarget->counters.missing_since = 0;
    ptarget->eto.ticks_counter = 0;
    ptarget->recovery_reason.tmf_timeout = 0;
    ptarget->recovery_reason.io_resp_missing_since = 0;

    //
    // Targets may have entered session recovery due to other reasons
    // (e.g TMF timeout), so mark the target ready by clearing recovery
    // in progress flag too
    //
    ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_SESSION_RECOVERY);
    ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_ETO_EXPIRED);
 
    // Clear all other flags that may have been set
    ptarget->busy = 0;
    ptarget->event_flag = 0;
    ptarget->cleanup_flag = 0;
    ptarget->flags = 0;
    ptarget->eto_status = 0;
    ptarget->sub_state = 0;

    ap_clear_target_event(pcontroller, ptarget, FLAG_TGT_ALL);

    // walk the target list for queued IOs and reset the IO retry count in its workspace

    for(piows_link = ptarget->io_queued_list.next;
        piows_link != &ptarget->io_queued_list;
        piows_link = piows_link->next) 
    {
        pio = STRUCT_BASE(ACIT_IO_WS,
                          target_link,
                          piows_link);

        pio->retry_count = 0;
    }

    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_TARGET,
                    ACIT_EVENT_TARGET_RECONNECTED,
                    ptarget->target_table_index);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
ap_process_eto_expiry(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    ACIT_UEXACT32           session_handle;

    // Return all the pending commands and queued commands
    ap_cleanup_target(pcontroller, ptarget, CL_REASON_ETO_EXPIRY);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("ETO of %d secs expired on target 0x%p\n", 
        (ptarget->eto.original_ticks / TICKS_PER_SEC(pcontroller)), ptarget));

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("LastCmdSnAt %d LastAsyncAt %d LastCQErrAt %d (Current %d)\n",
                ptarget->stats.last_cmdsn_tick,
                ptarget->stats.last_async_tick,
                ptarget->stats.last_cqerr_tick,
                CURRENT_TICK(pcontroller)));

    session_handle = GET_SESSION_HANDLE(pcontroller, ptarget);

    ap_log_osm_event(pcontroller,
                     UNI_CFG_LOG_LEVEL_ERROR,
                     UNI_SUBMODULE_ACIT_ISCSI,
                     UNI_PHASE_RECOVERY,
                     1,
                     ACIT_EC_TARGET_ETO_EXPIRY,
                     0, ACIT_NULL,
                     &session_handle);

    OSM_ASSERT(IS_LIST_EMPTY(&ptarget->tmf_handle_busy_list));
    OSM_ASSERT(!ptarget->counters.num_io_invs);
    OSM_ASSERT(!ptarget->counters.queued_tmfs);

    ap_clear_target_event(pcontroller, ptarget, EVENT_TGT_ETO_IN_PROGRESS);
    // Target sub state will still be TARGET_SUBSET_ETO_IN_PROGRESS
    ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_ETO_EXPIRED);

    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_TARGET,
                    ACIT_EVENT_TARGET_MISSING,
                    ptarget->target_table_index);

    pcontroller->eh_stats.eto_expiry++;
}

ACIT_VOID
ap_force_eto_expiry(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Forcing ETO expiry on target 0x%p\n"), ptarget));

    ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ETO_IN_PROGRESS);
    ap_process_eto_expiry(pcontroller, ptarget);
}

ACIT_UEXACT32
ap_setup_lun_for_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{
    ACIT_UEXACT32           status = ACIT_STATUS_PENDING;

    //
    // This routine is called from ap_check_and_issue_reset. This function
    // returns ACIT_STATUS_PENDING if reset cannot be issued right away.
    // Otherwise it will return ACIT_STATUS_SUCCESS
    //
    // This routine checks if any IO invalidations are pending on the LUN.
    // Otherwise it will call the invalidate IO routine once and will
    // return ACIT_STATUS_PENDING if any invalidations have been started.
    //
    // Secondly it checks if any TMFs are outstanding on the LUN. If so,
    // it returns ACIT_STATUS_PENDING. Else, it removes any queued aborts
    // on the LUN and returns ACIT_STATUS_SUCCESS to make way for LUN resets
    //

    ap_flag_busy_lun(pcontroller, plun, LUN_BUSY_RESET_WAITING);

    if(!plun->counters.num_io_invs)
    {
        status = ap_invalidate_pending_ios(pcontroller, 
                                           ptarget,
                                           plun,
                                           ACIT_NULL,
                                           ACIT_U32_MINUS_ONE);
    }

    return status;
}

ACIT_UEXACT32
ap_setup_target_for_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           cleanup_flag,
    ACIT_UEXACT32           is_abortive)
{
    ACIT_UEXACT32           status = ACIT_STATUS_SUCCESS;
    PACIT_DC_LIST           ptmf_link;
    PACIT_TMF_HANDLE        ptmf_handle;

    //
    // This routine is called from ap_check_and_issue_reset and iscsi_check_and_close_session. 
    // This function returns ACIT_STATUS_PENDING if any IO invalidations are pending on
    // the target. Otherwise it returns ACIT_STATUS_SUCCESS
    //
    // Since this is the beginning of a cleanup, all queued TMFs that are hierarchically
    // lower than target resets are returned here
    //

    ACIT_BIT_SET(ptarget->cleanup_flag, cleanup_flag);

    switch(cleanup_flag)
    {
    case TGT_CLEANUP_FLAG_FOR_TMF_RESET:

        ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_TMF_RESET_IN_PROGRESS);
        ACIT_SET_TARGET_SUBSTATE(ptarget,TARGET_SUB_STATE_TMF_RESET_WAIT);
        break;

    case TGT_CLEANUP_FLAG_FOR_RECOVERY:

        ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_SESSION_RECOVERY);
        break;

    case TGT_CLEANUP_FLAG_FOR_LOGOUT:

        ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_SESSION_LOGOUT);

        if(ptarget->osm_notification.u.bits.remove_pending == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("** Setting Target 0x%p state to REMOVE_PENDING \n"), ptarget));
            OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                            ACIT_EVENT_CLASS_TARGET,
                            ACIT_EVENT_TARGET_REMOVE_PENDING,
                            ptarget->target_table_index);

            ptarget->osm_notification.u.bits.remove_pending = 1;
        }

        break;

    default:
        OSM_ASSERT(0);
        break;
    }

    // If invalidations are already pending, just return
    if(ptarget->counters.num_io_invs)
    {
        return ACIT_STATUS_PENDING;
    }

    // for recovery cleanups, do not wait for TMF completions
    // proceed with recovery in these cases
    if(cleanup_flag == TGT_CLEANUP_FLAG_FOR_TMF_RESET)
    {
        // This loop checks if any LUN reset are pending at the target. 
        // If one reset is pending then target reset will not be send 
        // to the target.
        if (IS_LIST_NOT_EMPTY(&ptarget->tmf_handle_busy_list))
        {
            ptmf_link = ptarget->tmf_handle_busy_list.next;
            while (ptmf_link != &ptarget->tmf_handle_busy_list)
            {
                ptmf_handle = STRUCT_BASE(ACIT_TMF_HANDLE,
                                        target_link,
                                        ptmf_link);
                ptmf_link = ptmf_link->next;
                if ((ptmf_handle->plun) && 
                    (ACIT_IS_BIT_SET(ptmf_handle->plun->busy, LUN_BUSY_RESET_SUBMITTED)))
                {
                    return ACIT_STATUS_PENDING;
                }
            }
        }
    }

    //
    // If it is not an abortive cleanup, then return status from
    // invalidate routine
    //
    if(!is_abortive)
    {
        status = ap_invalidate_pending_ios(pcontroller, 
                                            ptarget,
                                            ACIT_NULL,
                                            ACIT_NULL,
                                            cri);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target 0x%p Returning status 0x%x from invalidate_pending_ios\n",
            ptarget, status));
    }

    return status;
}

ACIT_UEXACT32
ap_complete_target_recovery(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           status,
    ACIT_UEXACT32           ignore_link)
{
    if((ignore_link) ||
      (ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_LINK_DOWN)))
    {
        if(status != ACIT_STATUS_SUCCESS)
        {
            /*
            * Report this target as missing. This could happen if either
            * the login or offload session has failed
            */
            ap_start_eto(pcontroller, ptarget);
        }
        else
        {
            /* otherwise stop eto */
            ap_stop_eto(pcontroller, ptarget);

    #if ACIT_CFG_DEBUG_BUILD
            ap_validate_lists(pcontroller, ptarget, ACIT_NULL);
    #endif
        }
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Not starting ETO on target 0x%p since link is down\n", ptarget));
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_complete_all_target_recovery(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32               index = 0;
    PACIT_TARGET_ENTRY          ptarget_entry = ACIT_NULL;
    PACIT_TARGET_WS             ptarget;

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
                continue;

            if(ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_LDTO_EXPIRED))
            {
                //
                // Check if the target is not already counting down ETO. Otherwise this will
                // cause the ETO timers to be restarted
                //

                if(!IS_TARGET_MISSING(ptarget))
                {
                    ap_complete_target_recovery(pcontroller, 
                                                ptarget, 
                                                ACIT_STATUS_ERROR,
                                                0); // do not ignore link
                }
                else
                {
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("target 0x%p already in ETO, not restarting ETO timer..\n", ptarget));
                }
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("reporting target 0x%p missing since LDTO expired..\n", ptarget));

                // take out this target since LDTO expired
                ap_cleanup_target(pcontroller, ptarget, CL_REASON_LDTO_EXPIRY);

                OSM_ASSERT(IS_LIST_EMPTY(&ptarget->io_handle_busy_list));
                OSM_ASSERT(IS_LIST_EMPTY(&ptarget->io_queued_list));
                OSM_ASSERT(IS_LIST_EMPTY(&ptarget->io_pending_reset_list));
                OSM_ASSERT(IS_LIST_EMPTY(&ptarget->tmf_handle_busy_list));
                OSM_ASSERT(!ptarget->counters.num_io_invs);
                OSM_ASSERT(!ptarget->counters.queued_tmfs);

                OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                                ACIT_EVENT_CLASS_TARGET,
                                ACIT_EVENT_TARGET_MISSING,
                                ptarget->target_table_index);
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_process_ldto_tick(      
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           port_num, ldto_expired = 0;

    //
    // LDTimeout is handled independent of controller. If a link is down on a 
    // particular port for greater than LDTO, then all I/Os on the port
    // is invalidated and returned back to OSMs
    //
    for(port_num = 0; port_num < MAX_PHYS_PORTS; port_num++)
    {
        if(ACIT_IS_CONTROLLER_LINK_DOWN(pcontroller, port_num))
        {
            if(IS_CONTROLLER_LDTO_EXPIRED(pcontroller, port_num))
            {
                ldto_expired++;
            }
            else
            {
                pcontroller->ld_timeout_ticks[port_num]--;
            }
        }
    }

    if(ldto_expired == MAX_PHYS_PORTS)
    {
        if(ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_LDTO_EXPIRED))
        {
            pcontroller->eh_stats.ld_timeouts++;

            ap_log_osm_event(pcontroller,
                             UNI_CFG_LOG_LEVEL_CRITICAL,
                             UNI_SUBMODULE_ACIT_GENERIC,
                             UNI_PHASE_DONT_CARE,
                             0,
                             ACIT_EC_LDTO_EXPIRY,
                             0, ACIT_NULL,
                             ACIT_NULL);

            // Notify OSMs about the link down timeout expiry
            OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                        ACIT_EVENT_CLASS_CONTROLLER,
                        ACIT_EVENT_CONTROLLER_LINKDOWN_TIMEOUT,
                        0);

            ap_flag_busy_controller(pcontroller, CONTROLLER_BUSY_LDTO_EXPIRED);

            if(!pcontroller->counters.num_io_invs)
            {
                iscsi_recover_all_sessions(pcontroller, ACIT_NULL);
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_process_link_down(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           port_num)
{

#ifdef NO_ARM_PORT_FAILOVER
    ACIT_UEXACT32           index;
    PACIT_TARGET_ENTRY      ptarget_entry;
    PACIT_TARGET_WS         ptarget;
#endif

    pcontroller->ld_timeout_ticks[port_num] = pcontroller->params.ldto * TICKS_PER_SEC(pcontroller); 
    pcontroller->link_down[port_num] = ACIT_TRUE;

    ap_flag_busy_controller(pcontroller, CONTROLLER_BUSY_LINK_DOWN);

    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_CONTROLLER,
                    ACIT_EVENT_CONTROLLER_LINK_DOWN,
                    port_num);

    ap_log_osm_event(pcontroller,
                     UNI_CFG_LOG_LEVEL_CRITICAL,
                     UNI_SUBMODULE_ACIT_GENERIC,
                     UNI_PHASE_DONT_CARE,
                     0,
                     ACIT_EC_LINK_DOWN,
                     0, ACIT_NULL,
                     ACIT_NULL);

#ifdef NO_ARM_PORT_FAILOVER
    /*
     * Check to see what targets may be marked busy since this port
     * went down
     */
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
                continue;

            iscsi_process_linkdown_event(pcontroller, 
                                         GET_TARGET_WS(ptarget_entry[index].ptarget_key),
                                         port_num);
        }
    }
#endif

    return ACIT_STATUS_SUCCESS;
}
 
ACIT_UEXACT32
ap_process_link_up(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           port_num)
{

#ifdef NO_ARM_PORT_FAILOVER
    ACIT_UEXACT32           index;
    PACIT_TARGET_ENTRY      ptarget_entry;
    PACIT_TARGET_WS         ptarget;
#endif

    pcontroller->ld_timeout_ticks[port_num] = 0; 
    pcontroller->link_down[port_num] = ACIT_FALSE;

    OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                    ACIT_EVENT_CLASS_CONTROLLER,
                    ACIT_EVENT_CONTROLLER_LINK_UP,
                    port_num);

    ap_flag_ready_controller(pcontroller, CONTROLLER_BUSY_LINK_DOWN);
    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_LDTO_EXPIRED))
    {
        ap_flag_ready_controller(pcontroller, CONTROLLER_BUSY_LDTO_EXPIRED);

        // Bug 24354: Check if there are any targets that need ETO timers to 
        // be started. Or else, the logouts will be failed indefinitely because
        // link is back, but ETO timers aren't running
        ap_complete_all_target_recovery(pcontroller);
    }

    ap_log_osm_event(pcontroller,
                    UNI_CFG_LOG_LEVEL_INFO,
                    UNI_SUBMODULE_ACIT_GENERIC,
                    UNI_PHASE_DONT_CARE,
                    0,
                    ACIT_EC_LINK_UP,
                    0, ACIT_NULL,
                    ACIT_NULL);

#ifdef NO_ARM_PORT_FAILOVER
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(ptarget)
            {
                iscsi_process_linkup_event(pcontroller, 
                                            GET_TARGET_WS(ptarget_entry[index].ptarget_key),
                                            port_num);
            }
        }
    }
#endif

    return ACIT_STATUS_SUCCESS;
}

/*
static void ap_compile_time_asserts(void)
{
    ACIT_COMPILETIME_ASSERT(!(sizeof(ACIT_CONTROLLER_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(ACIT_TARGET_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(ACIT_LUN_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(ACIT_IO_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(ISCSI_CONTROLLER_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(ISCSI_TARGET_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(MGMT_CONTROLLER_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ACIT_COMPILETIME_ASSERT(!(sizeof(MGMT_TARGET_WS) & ACIT_MEM_ALIGNMENT_MASK));
    ap_compile_time_asserts();
}
*/

#if ACIT_CFG_SAVE_CMD_STATE

ACIT_VOID
ap_save_cmd_state(
    PACIT_CONTROLLER_WS         pcontroller,
    ACIT_UEXACT32               itt,
    ACIT_UEXACT32               cmd_sn)
{

    if(pcontroller->num_cmds)
        pcontroller->last_cmd++;

    pcontroller->num_cmds++;

    if(pcontroller->num_cmds >= ACIT_MAX_CMDS)
        pcontroller->num_cmds = ACIT_MAX_CMDS;
        
    if(pcontroller->last_cmd >= ACIT_MAX_CMDS)
        pcontroller->last_cmd = 0;

    pcontroller->cmd_state[pcontroller->last_cmd].itt = itt;
    pcontroller->cmd_state[pcontroller->last_cmd].cmd_sn = cmd_sn;

}

ACIT_VOID
ap_print_cmd_state(
    PACIT_CONTROLLER_WS         pcontroller)
{
    ACIT_U32                    i;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_FATAL,(("Num cmds - %d Last cmd %d\n"),
                        pcontroller->last_cmd,
                        pcontroller->num_cmds));

    for(i = 0; i < ACIT_MAX_CMDS; i++)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_FATAL,(("itt %d cmd_sn %d\n"),
                        pcontroller->cmd_state[i].itt, 
                        pcontroller->cmd_state[i].cmd_sn));
    }
}

/* Taken from iscsipdu.h */
#define SERIAL_BITS 32
#define SMALLEST_INVALID_DELTA  (ACIT_UEXACT32) (1 << (SERIAL_BITS - 1))

#define FORMER_IS_LESSER(Former, Latter)    (   \
    ((Former) < (Latter) && (Latter) - (Former) < SMALLEST_INVALID_DELTA) ||    \
    ((Former) > (Latter) && (Former) - (Latter) > SMALLEST_INVALID_DELTA)   \
)


#endif

#if ACIT_CFG_DEBUG_BUILD

ACIT_VOID
ap_process_dbg_flags(
    PACIT_CONTROLLER_WS         pcontroller)
{
#if ACIT_CFG_DEBUG_BUILD
    if(pcontroller->params.dbg_flags & ACIT_DBG_FLAGS_DUMP_PENDING_CMDS)
    {
        ap_dbg_dump_pending_cmds(pcontroller);
    }
    else if(pcontroller->params.dbg_flags & ACIT_DBG_FLAGS_VALIDATE_PENDING_IOS)
    {
#if 0
        // STORPORT_DUPLEX not implemented
        hwi_validate_pending_ios(pcontroller, ACIT_NULL, ACIT_NULL);
#endif
    }
#endif
}

ACIT_VOID
ap_dbg_dump_pending_cmds(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_IO_HANDLE             pio_handle;
    PACIT_DC_LIST               flink, pio_link;

    flink = pcontroller->io_handle_busy_list.next;
    pio_link = pcontroller->io_handle_busy_list.next;

    if(!pcontroller->counters.pending_ios)
        return;

    while(flink != &pcontroller->io_handle_busy_list)
    {
        pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                 controller_link,
                                 pio_link);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("IO Handle 0x%p CmdSN 0x%x Tgt %d\n"), 
                                                        (PACIT_UEXACT32)pio_handle, 
                                                        pio_handle->cmd_sn,
                                                        pio_handle->ptarget->target_table_index));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("\n"));

        flink = flink->next;
        pio_link = pio_link->next;
    }
}
#endif

ACIT_STATUS
ap_get_param_num (
    PACIT_CONTROLLER_KEY        pcontroller_key, 
    const ACIT_PARAM_PROPERTIES *pp, 
    ACIT_UEXACT32               *mask,
    ACIT_UEXACT32               operating_mem_level) 
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_CONFIG_PARAM config_param;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT (pcontroller);
    OSM_ASSERT (pp);
    OSM_ASSERT (pp->name);
    OSM_ASSERT (pp->min_value <= pp->max_value);
    

    OSM_ASSERT (mask);
    OSM_ASSERT ((*mask & pp->default_mask) == 0);
    
    OSM_MEMSET (&config_param, 0, sizeof(config_param));
    OSM_MEMCPY (config_param.osm_config_param, pp->name, strlen (pp->name) + 1);

    config_param.min_value      = pp->min_value;
    config_param.max_value      = pp->max_value;

    switch(operating_mem_level)
    {
    case ACIT_MEM_LEVEL_MINIMUM:
        config_param.numeric_value  = pp->min_value;
        break;

    case ACIT_MEM_LEVEL_1:
        config_param.numeric_value = pp->mem_level1_value;
        break;

    case ACIT_MEM_LEVEL_DEFAULT:
        config_param.numeric_value  = pp->mem_level2_value;
        break;

    case ACIT_MEM_LEVEL_3:
        config_param.numeric_value  = pp->mem_level3_value;
        break;

    case ACIT_MEM_LEVEL_MAXIMUM:
        config_param.numeric_value  = pp->max_value;
        break;
    }

    config_param.acit_pflag     = ACIT_PFLAG_PARAMETER_NUMERIC;

    if (operating_mem_level > ACIT_MEM_LEVEL_MINIMUM)
        if (OSM_GET_CONFIG_VALUE (pcontroller_key, &config_param)){
            OSM_DEBUG_PRINT (ACIT_DBGMSG_FATAL, ("Failed to query parameter %s.", pp->name));
            return ACIT_STATUS_ERROR;
        }

    if((config_param.osm_pflag == ACIT_OSM_PFLAG_DEFAULT_VALUE) ||
       (config_param.numeric_value < pp->min_value) ||
       (config_param.numeric_value > pp->max_value) )
    {
        switch (operating_mem_level) {
            case ACIT_MEM_LEVEL_MINIMUM:
                *pp->value = pp->min_value;
                break;
            case ACIT_MEM_LEVEL_1:
                *pp->value = pp->mem_level1_value;
                break;
            case ACIT_MEM_LEVEL_2:
                *pp->value = pp->mem_level2_value;
                break;
            case ACIT_MEM_LEVEL_3:
                *pp->value = pp->mem_level3_value;
                break;
            case ACIT_MEM_LEVEL_MAXIMUM:
                *pp->value = pp->max_value;
                break;

            default:
                OSM_DEBUG_PRINT (ACIT_DBGMSG_FATAL, ("Illegal memory level %d.\n", operating_mem_level));
                OSM_ASSERT (0);
        }

        *mask |= pp->default_mask;

    } 
    else 
    {
        *pp->value = config_param.numeric_value;
    }

    OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Parameter %s = 0x%x (%d)\n", pp->name, *pp->value, *pp->value));

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
ap_init_fw_config(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR membar_ctrl_int_ctrl;
    ACIT_UEXACT32 status = ACIT_STATUS_SUCCESS;
    PACIT_PARAMETERS p = &pcontroller->params;
    ACIT_UEXACT32 num_frag_sz;

    membar_ctrl_int_ctrl.dw = ACIT_PCICFG0_READ (pcontroller, membar_ctrl_intr_ctrl);
    pcontroller->fw_config.pci_function = membar_ctrl_int_ctrl.PFNUM;

    switch(pcontroller->fw_config.pci_function)
    {
    case 0:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr0);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x0;
        pcontroller->fw_config.iscsi_cid_start = 0x0;
#endif
        break;

    case 1:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr1);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x200;
        pcontroller->fw_config.iscsi_cid_start = 0x40;
#endif
        break;

    case 2:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr2);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x400;
        pcontroller->fw_config.iscsi_cid_start = 0x80;
#endif
        break;

    case 3:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr3);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x600;
        pcontroller->fw_config.iscsi_cid_start = 0xc0;
#endif
        break;

    case 4:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr4);
#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x0;
        pcontroller->fw_config.iscsi_cid_start = 0x0;
#endif
        break;

    case 5:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr5);
#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x200;
        pcontroller->fw_config.iscsi_cid_start = 0x40;
#endif
        break;

    case 6:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr6);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x400;
        pcontroller->fw_config.iscsi_cid_start = 0x80;
#endif
        break;

    case 7:
        pcontroller->fw_config.isr_offset = \
                STRUCT_FIELD_OFFSET (BLADE_ENGINE_CSRMAP, cev.isr7);

#ifdef FPGA_ARM
        pcontroller->fw_config.iscsi_icd_start = 0x600;
        pcontroller->fw_config.iscsi_cid_start = 0xc0;
#endif
        break;

    default:
        OSM_ASSERT(0);
        status = ACIT_STATUS_ERROR;
        break;
    }

 //   OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Overwrite cid start and icd start for function %d as %d and %d\n",
 //                                       pcontroller->fw_config.pci_function,
 //                                       pcontroller->fw_config.iscsi_cid_start,
 //                                       pcontroller->fw_config.iscsi_icd_start));

    if(pcontroller->fw_config.iscsi_cid_count < p->num_cxn)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Override number of connections from %d to %d \n",
                                            p->num_cxn,
                                            pcontroller->fw_config.iscsi_cid_count));

        p->num_cxn = pcontroller->fw_config.iscsi_cid_count;
        p->targets_per_controller = p->num_cxn;
    }

    if(pcontroller->fw_config.iscsi_icd_count < p->icds_per_controller)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Override number of ICDs from %d to %d \n",
                                            p->icds_per_controller,
                                            pcontroller->fw_config.iscsi_icd_count));

        p->icds_per_controller = pcontroller->fw_config.iscsi_icd_count;

        if(p->icds_per_controller > p->num_cxn)
        {
            p->ios_per_controller = (p->icds_per_controller - p->num_cxn);
        }
        else
        {
            if(p->icds_per_controller > BE_MAX_ICDS/8)
            {
                // Atleast 64 ICDs are present. Set IOs to 64
                p->ios_per_controller = BE_MAX_ICDS/8;
            }
            else
            {
                // Not even 64 ICDs
                if(p->icds_per_controller > 8)
                {
                    // Atleast 2 for EH since NOP Out need a SGL
                    // to be reserved
                    p->ios_per_controller = (8-1);
                }
                else
                {
                    p->ios_per_controller = 0;
                    p->icds_per_controller = 0;
                }
            }
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("New IO Count is set to %d\n",
                                            p->ios_per_controller));

        num_frag_sz = sizeof(SGE)* p->num_sge * p->icds_per_controller;
        pcontroller->acit_var_desc_index.run_time.frag = 
                (ACIT_UEXACT8)sa_ceiling(num_frag_sz, ACIT_CFG_MAX_CONTIGUOUS_MEM);
    }

    if((p->icds_per_controller == 0) ||
       (p->targets_per_controller == 0))
    {
        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_INIT,
                         0,
                         ACIT_EC_FW_CFG_BAD_ISCSI_IPL,
                         0, ACIT_NULL,
                         ACIT_NULL);

        // we don't return an error only because we want the driver load
        // to succeed. Though the driver is pretty much useless at this
        // point it will allow management applications to update IPL file
        // or change personality
    }

    // Note! Note!
    // Although the operating parameters are overridden above, the memory will go waste!
    // This is because by now, the memory is already allocated. Ideally, ACIT has to
    // determine the firmware config _before_ providing memory requirement to OSM. But
    // for this, ACIT has to implement the get firmware config IOCTL in mailbox mode and
    // all OSMs should have a means to allocate memory dynamically.
    //
                                        
    return status;
}

ACIT_UEXACT32
ap_determine_load(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_UEXACT32          pio_load,
    PACIT_UEXACT32          pload_factor,
    PACIT_UEXACT32          ptarget_load,
    ACIT_UEXACT32           pid)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_TARGET_ENTRY      ptarget_entry;
    ACIT_UEXACT32           index, io_targets = 0;

    ACIT_UEXACT32           overload = 0, target_load = 0;

    if(pcontroller->partition[pid].dic.ios_per_sec > pcontroller->params.dic.io_threshold_for_dic)
    {
        overload = (pcontroller->partition[pid].dic.ios_per_sec / pcontroller->params.dic.io_threshold_for_dic);

        ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

        for(index = 0; 
            index < pcontroller->params.targets_per_controller;
            index ++)
        {
            if(&pcontroller->ptarget_key[index] &&
               pcontroller->ptarget_key[index].acit_key)
            {
                ptarget = pcontroller->ptarget_key[index].acit_key;

                /*
                if((ptarget->partition == pid) && 
                   (ptarget->stats.last_completed_tick == (CURRENT_TICK(pcontroller) - 1)))
                {
                    io_targets++;
                }
                */

                // check to see if there were completions on this partition for this target
                if(ptarget->stats.io_completion[pid] - ptarget->stats.last_completion[pid])
                {
                    io_targets++;
                }

                ptarget->stats.last_completion[pid] = ptarget->stats.io_completion[pid];
            }
        }

        if(io_targets >= pcontroller->params.dic.tgt_threshold_for_load)
        {
            ACIT_UEXACT32 target_factor;
            target_factor = (io_targets / pcontroller->params.dic.tgt_threshold_for_load) - 1;
            overload += target_factor;

            target_load = io_targets / pcontroller->params.dic.tgt_threshold_for_load + 1;
        }
        else if((io_targets > pcontroller->params.dic.tgt_threshold_for_load / 2) &&
               (io_targets <= pcontroller->params.dic.tgt_threshold_for_load))
        {
            target_load = 1;
        }

        if(overload > pcontroller->partition[pid].dic.max_overload)
        {
            pcontroller->partition[pid].dic.max_overload = overload;
        }

        *pio_load = 1;
        *pload_factor = overload + 1;
        *ptarget_load = target_load;
    }
    else
    {
        *pio_load = 0;
        *pload_factor = 0;
        *ptarget_load = 0;
    }

    return ACIT_STATUS_SUCCESS;
}

#if 0

ACIT_UEXACT32
ap_determine_load(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_UEXACT32          pio_load,
    PACIT_UEXACT32          pload_factor,
    PACIT_UEXACT32          ptarget_load)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_TARGET_ENTRY      ptarget_entry;
    ACIT_UEXACT32           index, io_targets = 0;

    ACIT_UEXACT32           overload = 0, target_load = 0;

    if(pcontroller->dic.ios_per_sec > pcontroller->params.dic.io_threshold_for_dic)
    {
        overload = (pcontroller->dic.ios_per_sec / pcontroller->params.dic.io_threshold_for_dic);

        ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

        for(index = 0; 
            index < pcontroller->params.targets_per_controller;
            index ++)
        {
            if(&pcontroller->ptarget_key[index] &&
               pcontroller->ptarget_key[index].acit_key)
            {
                ptarget = pcontroller->ptarget_key[index].acit_key;

                if(ptarget->stats.last_completed_tick == (CURRENT_TICK(pcontroller) - 1))
                {
                    io_targets++;
                }
            }
        }

        if(io_targets >= pcontroller->params.dic.tgt_threshold_for_load)
        {
            ACIT_UEXACT32 target_factor;
            target_factor = (io_targets / pcontroller->params.dic.tgt_threshold_for_load) - 1;
            overload += target_factor;

            target_load = io_targets / pcontroller->params.dic.tgt_threshold_for_load + 1;
        }
        else if((io_targets > pcontroller->params.dic.tgt_threshold_for_load / 2) &&
               (io_targets <= pcontroller->params.dic.tgt_threshold_for_load))
        {
            target_load = 1;
        }

        if(overload > pcontroller->dic.max_overload)
        {
            pcontroller->dic.max_overload = overload;
        }

        *pio_load = 1;
        *pload_factor = overload + 1;
        *ptarget_load = target_load;
    }
    else
    {
        *pio_load = 0;
        *pload_factor = 0;
        *ptarget_load = 0;
    }

    return ACIT_STATUS_SUCCESS;
}

#endif

ACIT_UEXACT32
ap_process_scsi_event(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_VOID              psense_buffer, 
    ACIT_UEXACT16           sense_len)
{
    PACIT_SCSI_SENSE_DATA   psense_data;

    psense_data = (PACIT_SCSI_SENSE_DATA)psense_buffer;

    if(((psense_data->response_code != SENSE_DATA_CURRENT_ERRORS) &&
       (psense_data->response_code != SENSE_DATA_DEFERRED_ERRORS)) ||
       (psense_data->valid == 0))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Invalid SCSI response or valid bit 0x%x 0x%x\n",
            psense_data->response_code, psense_data->valid));

        return ACIT_STATUS_ERROR;
    }

    if(sense_len < sizeof(ACIT_SCSI_SENSE_DATA))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Sense length too small expected %d received %d\n",
            (ACIT_UEXACT32)sizeof(ACIT_SCSI_SENSE_DATA), sense_len));

        return ACIT_STATUS_ERROR;
    }

    switch (psense_data->sense_key)
    {
    case SENSE_KEY_UNIT_ATTENTION:

        if((psense_data->asc == ASC_REPORT_LUN_DATA_HAS_CHANGED) &&
           (psense_data->ascq == ASCQ_REPORT_LUN_DATA_HAS_CHANGED))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Target Report LUN data changed, notifying OSM\n"));

            OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                            ACIT_EVENT_CLASS_TARGET,
                            ACIT_EVENT_TARGET_RESCAN_REQUIRED,
                            ptarget->target_table_index);
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Not handling sense key 0x%x addi_sns_len %d asc 0x%x ascq 0x%x\n",
                                psense_data->sense_key,
                                psense_data->addi_sense_len,
                                psense_data->asc,
                                psense_data->ascq));
        }

        break;

    default:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Not handling sense key 0x%x addi_sns_len %d asc 0x%x ascq 0x%x\n",
                            psense_data->sense_key,
                            psense_data->addi_sense_len,
                            psense_data->asc,
                            psense_data->ascq));

        return ACIT_STATUS_ERROR;

        break;
    }

    return ACIT_STATUS_SUCCESS;

}

ACIT_VOID
ap_update_lun_retry_param(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun)
{
    //
    // Updates retry_io parameter in the LUN based on device type and
    // global parameters. dev_type must be valid in the LUN workspace
    // when this function is called.
    //

    if(pcontroller->params.io_retries == 0)
    {
        plun->retry_io = ACIT_FALSE;
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Turning off retries on Lun 0x%p based on global param\n", plun));
    }
    else
    {
        if(plun->dev_type == ACIT_SCSI_DEVTYPE_DISK)
        {
            plun->retry_io = ACIT_TRUE;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Turning on retries on Lun 0x%p based on dev_type DISK\n", plun));
        }
        else
        {
            plun->retry_io = ACIT_FALSE;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Turning off retries on Lun 0x%p based on dev_type 0x%x\n", plun, plun->dev_type));
        }
    }
}

#if 0
ACIT_UEXACT32
ap_call_all_events(PACIT_CONTROLLER_WS pcontroller)
{
    ACIT_UEXACT32 ec[3], loop;

    ec[0] = 0x5fcd09ac;
    ec[1] = 0x1d9c2232;
    ec[2] = 0x1a32f323;

    for(loop = 0; loop < 32; loop++)
    {
        ap_log_osm_event(pcontroller,
                        UNI_CFG_LOG_LEVEL_CRITICAL,
                        UNI_SUBMODULE_ACIT_GENERIC,
                        UNI_PHASE_RECOVERY,
                        1,
                        ACIT_EC_UNHANDLED_TMF_REQUEST,
                        0, ACIT_NULL,
                        (PACIT_UEXACT32)(&ec[0]));

        ap_log_osm_event(pcontroller,
                        UNI_CFG_LOG_LEVEL_CRITICAL,
                        UNI_SUBMODULE_ACIT_GENERIC,
                        UNI_PHASE_TMF,
                        0,
                        ACIT_EC_TASK_INVALID_FOR_OSM_ABORT,
                        0, ACIT_NULL,
                        ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_ISCSI,
                         UNI_PHASE_RECOVERY,
                         1,
                         ACIT_EC_PING_WAIT_TIMEOUT,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_ERROR,
                         UNI_SUBMODULE_ACIT_ISCSI,
                         UNI_PHASE_RECOVERY,
                         1,
                         ACIT_EC_TARGET_ETO_EXPIRY,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_DONT_CARE,
                         0,
                         ACIT_EC_LDTO_EXPIRY,
                         0, ACIT_NULL,
                         ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_DONT_CARE,
                         0,
                         ACIT_EC_LINK_DOWN,
                         0, ACIT_NULL,
                         ACIT_NULL);

        ap_log_osm_event(pcontroller,
                        UNI_CFG_LOG_LEVEL_INFO,
                        UNI_SUBMODULE_ACIT_GENERIC,
                        UNI_PHASE_DONT_CARE,
                        0,
                        ACIT_EC_LINK_UP,
                        0, ACIT_NULL,
                        ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         3,
                         ACIT_EC_HWI_UE,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_INIT_FAILED,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller, 
                         UNI_CFG_LOG_LEVEL_ERROR,
                         UNI_SUBMODULE_ACIT_ISCSI,
                         UNI_PHASE_RECOVERY,
                         3,
                         ACIT_EC_ISCSI_CMDSN_ERROR,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_INIT,
                         0,
                         ACIT_EC_FIRMWARE_MISMATCH,
                         0, ACIT_NULL,
                         ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_GENERIC,
                         UNI_PHASE_INIT,
                         0,
                         ACIT_EC_FW_CONFIG_FAILED,
                         0, ACIT_NULL,
                         ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_MGMT,
                         UNI_PHASE_IOCTL,
                         1,
                         ACIT_EC_OFFLOAD_ERROR,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_MGMT,
                         UNI_PHASE_IOCTL,
                         2,
                         ACIT_EC_MCC_SGE_EXCEEDED,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_ERROR,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         0,
                         ACIT_EC_HWI_IN_UE_TOLERANCE,
                         0, ACIT_NULL,
                         ACIT_NULL);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_INFO,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_UE_RECOVERY_SUCCESS,
                         0, ACIT_NULL,
                         &ec[0]);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_UE_RECOVERY_FAILED,
                         0, ACIT_NULL,
                         &ec[0]);
    }

    return 1;
}
#endif

ACIT_VOID
ap_cleanup_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cleanup_type)
{
    ACIT_UEXACT32           io_status = 0;
    ACIT_UEXACT32           tmf_status = 0;

    switch(cleanup_type)
    {
    case CL_REASON_ETO_EXPIRY:
        io_status = ACIT_IO_STATUS_TARGET_TIMEOUT;
        tmf_status = ACIT_TMF_STATUS_TARGET_TIMEOUT;

        break;

    case CL_REASON_LDTO_EXPIRY:
    case CL_REASON_UE_TOLERANCE:
    default:
        io_status = ACIT_IO_STATUS_TARGET_TIMEOUT;
        tmf_status = ACIT_TMF_STATUS_TARGET_TIMEOUT;
        break;
    }

    ap_return_queued_ios(pcontroller, 
                            ptarget, 
                            ACIT_NULL, 
                            io_status,
                            ACIT_U32_MINUS_ONE);

    ap_return_queued_target_tmfs(pcontroller, 
                                 ptarget, 
                                 tmf_status,
                                 ACIT_TRUE);

    ap_free_pending_handles(pcontroller,
                            ptarget,
                            ACIT_NULL,
                            ACIT_U32_MINUS_ONE);

    if(cleanup_type == CL_REASON_UE_TOLERANCE)
    {
        //
        // Clear all event processing if there is a UE. Events will get lost,
        // but it is ok since a UE recovery will anyway re-init the state
        //
        if(IS_LINK_VALID(&ptarget->event_link))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Clear all target 0x%p events because of UE\n", ptarget));
            ap_clear_target_event(pcontroller, ptarget, FLAG_TGT_ALL);
        }
    }
}

ACIT_VOID
ap_process_ue_occurrance(
    PACIT_CONTROLLER_WS     pcontroller)
{
    // This routine is called when a UE is detected.

    ap_flag_busy_controller(pcontroller, CONTROLLER_BUSY_UE);

process_begin:

    if(ACIT_IS_UE_RECOVERY_ENABLED(pcontroller))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Starting UE Recovery...\n"));

        // Initiate recovery from UE. This routine will only initiate
        // a recovery. The polling for completion is done from the
        // health check routine
        if(hwi_start_ue_recovery(pcontroller) != ACIT_STATUS_SUCCESS)
        {
            pcontroller->eh_stats.ue_recoveries++;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Could not start UE Recovery!\n"));
            OSM_ASSERT(0);
            goto process_begin;
        }
    }
    else
    {
        // Operate in UE tolerance mode

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Operating in UE Tolerance mode...\n"));

        // This has to be called explicitly to let beclib complete
        // any pending IOCTLs
        hwi_cleanup_after_ue(pcontroller);

        iscsi_cleanup_all_sessions(pcontroller);

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_ERROR,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         0,
                         ACIT_EC_HWI_IN_UE_TOLERANCE,
                         0, ACIT_NULL,
                         ACIT_NULL);
    }
}

ACIT_VOID
ap_flag_ue_recovery_complete(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ap_flag_ready_controller(pcontroller, CONTROLLER_BUSY_UE);
    pcontroller->eh_stats.ue_detected = 0;

    ap_log_osm_event(pcontroller,
                     UNI_CFG_LOG_LEVEL_INFO,
                     UNI_SUBMODULE_ACIT_HWI,
                     UNI_PHASE_DONT_CARE,
                     1,
                     ACIT_EC_HWI_UE_RECOVERY_SUCCESS,
                     0, ACIT_NULL,
                     &pcontroller->eh_stats.ue_recoveries);
}

ACIT_UEXACT32
ap_complete_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           recovery_status)
{
    ACIT_UEXACT32           status;

    // This function is called when the UE recovery has completed. The
    // recovery_status indicates whether the recovery was a success or
    // not.

    // Count the number of ue_recoveries attemps so far, 
    // count even the unsuccessful ones
    pcontroller->eh_stats.ue_recoveries++;

    if(recovery_status == ACIT_STATUS_SUCCESS)
    {
        // recovered successfully from UE.
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Sucessfully recovered from UE!\n"));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Starting Re-initialization!!\n"));

        pcontroller->ue_detection.reinit_after_ue = 1;

        //
        // Re-init the controller. Note that the osm_handle parameter can be
        // NULL when doing a re-init
        //
        status = acit_init_controller(pcontroller->pcontroller_key,
                             pcontroller->state.init_type,
                             ACIT_NULL);

        if(status == ACIT_STATUS_SUCCESS)
        {
            // Turn off UE status
            ap_flag_ue_recovery_complete(pcontroller);
        }
    }
    else
    {
        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_UE_RECOVERY_FAILED,
                         0, ACIT_NULL,
                         &pcontroller->eh_stats.ue_recoveries);

        // A UE recovery action has failed. Check if we can attempt it again
        ap_process_ue_occurrance(pcontroller);
    }

    return ACIT_STATUS_SUCCESS;
}

#if 1
INLINE
ACIT_UEXACT32
ap_update_io_ticks(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_WS             pio)
{
    //PACIT_IO_REQUEST        pio_request;
    ACIT_UEXACT32           ito, eto, rto, cto, invto, tmfto;
    ACIT_UEXACT32           min_ticks;

    //
    // Goal of this routine is to calculate the time given for an IO based on
    // various factors. The earlier algorithm always put a hard limit of 8 secs
    // for every IO. Do away with that and try to maximize the amount of time
    // given for a command and yet honor all the ETO requirements that OEMs
    // require.
    // Note!
    // One thing to watch out for is if the OS timers aren't accurate or if
    // command invalidations take more time than expected, we run the risk
    // of TMFs from OS since ito is now very close to cto. Only when cto
    // is > (eto - rto) is when there's enough buffer between cto and ito to
    // account for these unexpected delays.
    //

    // invto = Max amount of time estimated for invalidation
    // cto   = command timeout given by the OSM with some buffer for invalidation
    // eto   = Extended Timeout (ETO). Adjusted to MIN_ETO if < MIN_ETO
    // rto   = Recovery Timeout. Time required for detecting ETO (static)
    // ito   = Internal Timeout ultimately given to a command

    invto = pcontroller->state.invalidation_ticks;
    cto = pio->available_ticks;
    eto = ptarget->eto.adjusted_ticks;
    rto = ptarget->eto.rto_ticks;
    tmfto = GET_TMF_TIMEOUT_TICKS(pcontroller, target);

    //
    // Bug 32886 : Some Linux versions issue IO with very less timeout
    // initially. Hence we cannot expect to have a min_tick that includes
    // TMF timeout also. So require large remaining time only during
    // a retry.
    //
    // If the IO is queued, then this is being attempted on a retry.
    // So require a higher internal timeout.
    if(pio->queued)
    {
        min_ticks = tmfto + invto;
    }
    else
    {
        min_ticks = invto * 2;
    }

    // Check if we are given enough timeout to invalidate the command 
    if(cto <= min_ticks)
    {
        return ACIT_STATUS_INSUFFICIENT_TIMEOUT;
    }

    // if retry has been turned off, then give full timeout
    if(plun->retry_io == 0)
    {
        ito = cto - min_ticks;
    }
    else
    {
        // We want rto ticks to determine if we are going to enter ETO. So rto must
        // be subtracted from the eto and that should be the ito. But if cto is
        // lesser than this, then ito should be cto - min_ticks.
        ito = ACIT_MIN((eto - rto), (cto - min_ticks));

        // In this new scheme, the following workaround is hopefully not required.

        // Linux specific workaround
        //pio_request = pio->pio_request;
        //if(pio_request->cdb[0] == ACIT_SCSI_CMD_FORMAT)
        //{
        //    pio->internal_available_ticks = pio->available_ticks - 
        //        	                            pcontroller->state.invalidation_ticks;
        //}
    }

    pio->internal_available_ticks = ito;
    
    // When dynamic timer is implemented this needs to be enabled
    // pio->last_update = CURRENT_TICK(pcontroller);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("PIO 0x%p CTO %d ITO %d Current %d\n",
        pio, cto, ito, CURRENT_TICK(pcontroller)));

    OSM_ASSERT(pio->internal_available_ticks > pcontroller->state.invalidation_ticks);
    OSM_ASSERT(pio->internal_available_ticks < pio->available_ticks);
    OSM_ASSERT(pio->internal_available_ticks < (ACIT_SEXACT32)ptarget->eto.adjusted_ticks);

    return ACIT_STATUS_SUCCESS;
}

#else

ACIT_UEXACT32
ap_update_io_ticks(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_WS             pio)
{
    PACIT_IO_REQUEST        pio_request;

    pio_request = pio->pio_request;

    if(plun->retry_io)
    {
        // Bug 25401 : Let ETO timers take precedence even for untagged
        // commands. Else the untagged commands in the mix will delay
        // failover beyond ETO limits
        if(pio_request->task_attribute == ACIT_TASK_ATTRIBUTE_UNTAGGED)
        {
            pio->internal_available_ticks = pio->available_ticks - 
                	                            pcontroller->state.invalidation_ticks;

            if(ACIT_IS_ETO_ENABLED(ptarget))
            {
                // Set internal_available_ticks to command timeout and ETO timeout
                pio->internal_available_ticks = ACIT_MIN(pio->internal_available_ticks,
                                                         (ACIT_SEXACT32)ptarget->eto.original_ticks);
            }
            else
            {
                // Set internal_available_ticks to command timeout and ETO timeout
                // Since ETO is 0, use MIN_ETO in this case
                pio->internal_available_ticks = ACIT_MIN(pio->internal_available_ticks,
                                                         (ACIT_SEXACT32)MIN_ETO);
            }
        }

        // Linux specific workaround
        if(pio_request->cdb[0] == ACIT_SCSI_CMD_FORMAT)
        {
            pio->internal_available_ticks = pio->available_ticks - 
                	                            pcontroller->state.invalidation_ticks;
        } else  
            pio->internal_available_ticks = GET_TARGET_EXEC_TICKS(ptarget, pio->retry_count);
    }
    else
    {
        // if no retry is required, set internal timeout to slightly
        // less than what the OSMs gave, do not set default internal
        // timeouts
        pio->internal_available_ticks = pio->available_ticks - 
                                            pcontroller->state.invalidation_ticks;
    }

    OSM_ASSERT(pio->available_ticks);
    OSM_ASSERT(pio->internal_available_ticks);

    /*
     *  Check if we are given enough timeout to invalidate the command 
     */
    if(pio->available_ticks <= pcontroller->state.invalidation_ticks)
    {
        DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                            FUNC_ap_alloc_io_resources,
                            VASE_MODULE_ACIT,
                            (ACIT_UEXACT64) pcontroller->busy,
                            ptarget->busy,
                            plun->busy);
        return ACIT_STATUS_INSUFFICIENT_TIMEOUT;
    }

    /*
     * Adjust the internal available ticks if needed. This is done since
     * the internal available ticks must always be lower than the ticks
     * OSM gave to an IO
     */
    if(pio->available_ticks <= pio->internal_available_ticks)
    {
        pio->internal_available_ticks = pio->available_ticks - 
                                            pcontroller->state.invalidation_ticks;
    }

    return ACIT_STATUS_SUCCESS;
}

#endif

ACIT_VOID
ap_decide_io_disposition_in_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PACIT_TARGET_WS         ptarget;
    PACIT_LUN_WS            plun;
    ACIT_UEXACT32           disposition;

    ptarget = pio_handle->ptarget;
    plun = pio_handle->plun;

    OSM_ASSERT(pio_handle->pio);

    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_LOGOUT))
    {
        // Logging out, so return the IO
        ap_return_io(pcontroller, 
                    pio_handle->pio, 
                    ACIT_IO_STATUS_TARGET_DELETED, 
                    ACIT_FALSE);
        return;
    }

    //
    // Add the check for external TMF. If we get to ap_free_pending_handles
    // because of a connection reset _while_ there were OSM TMFs, then we
    // should return the IOs back and not attemtpt a retry
    //
    if(ACIT_IS_BIT_SET(ptarget->cleanup_flag, TGT_CLEANUP_FLAG_FOR_TMF_RESET) ||
        (plun->plun_reset_tmf_request) ||
        (pio_handle->ptmf_ws))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("Return IO 0x%p IOH 0x%p LUN 0x%p Target 0x%p back because of external TMF\n"),
            pio_handle->pio, pio_handle, plun, ptarget));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("IOWS 0x%p Ticks - Original %d Available %d Internal %d\n"),
            pio_handle->pio, pio_handle->pio->original_ticks, pio_handle->pio->available_ticks,
            pio_handle->pio->internal_available_ticks));

        // External OS TMF, so return the IO
        ap_return_io(pcontroller, 
                    pio_handle->pio, 
                    ACIT_IO_STATUS_ABORTED, 
                    ACIT_FALSE);

        return;
    }

    if((plun->retry_io == ACIT_FALSE) ||
       (IS_TARGET_ETO_EXPIRED(ptarget))||
       (IS_CONTROLLER_LDTO_EXPIRED(pcontroller, 0)) ||
       (IS_CONTROLLER_UE(pcontroller)))
    {
        ACIT_UEXACT32 io_status = ACIT_IO_STATUS_IO_TIMEOUT;

        // The IO handle may have additional status about the reason for timeout. 
        // CQ errors and Rejected commands will be timed-out with some
        // additional state. If that is the case, then return the IO with an error and
        // not with timeout status
        if((pio_handle->flags & FLAG_IOH_CQ_ERROR) ||
            (pio_handle->flags & FLAG_IOH_REJECTED_BY_TARGET))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("IOH 0x%p rejected/CQ error 0x%x, return error\n",
                pio_handle, pio_handle->flags));
            io_status = ACIT_IO_STATUS_ERROR;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Not retrying IOWS 0x%p for LUN 0x%p Target 0x%p\n"),
            pio_handle->pio, plun, ptarget));

        ap_return_io(pcontroller, 
                    pio_handle->pio, 
                    io_status, 
                    ACIT_FALSE);

        return;
    }
    
    // Since there's a possiblity that the IOWS is queued again, recalculate the
    // available tick to the IOWS
    ap_recalc_io_tick(pcontroller, pio_handle, pio_handle->pio);

    // Now we can check the timeouts alone. The following routine has some
    // redundant checks for this path, but it's required in the acit_init_io_ws
    // path. Do it here too anyway
    disposition = ap_decide_io_disposition(pcontroller, 
                                           pio_handle->ptarget, 
                                           pio_handle->pio);

    ap_process_io_disposition_after_timeout(pcontroller, 
                                            pio_handle->ptarget, 
                                            pio_handle->plun, 
                                            pio_handle->pio, 
                                            disposition);
}

ACIT_VOID
ap_recalc_io_tick(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_IO_WS             pio)
{
    ACIT_SEXACT32           ticks_spent;

    // The goal of this routine is to setup the available_ticks for the IO so
    // that the IO disposition routine can make the right decision about
    // queueing it or returning it back

    //
    // As long as the IOHandle is in the busy list or the IOWS is in the queued
    // list, the ticks are accounted for. When the IOH is removed from busy list,
    // the inv_start_tick is initialized. Count the number of ticks since then
    // and subtract from the IOWS avaialable ticks
    //
    if(pio_handle->inv_start_tick)
    {
        ticks_spent = TICKS_SPENT(pcontroller, pio_handle->inv_start_tick);

        if(pio->available_ticks >= ticks_spent)
        {
            pio->available_ticks -= ticks_spent;
        }
        else
        {
            // This IO is already timed out! Likely we're seeing TMFs from OS..
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("!!!IOH 0x%p IOWS 0x%p on Target 0x%p Lun 0x%p already exceeded time quota for exec!!!\n",
                pio_handle, pio, pio_handle->ptarget, pio_handle->plun));
            pio->available_ticks = 0;
        }
    }
}
